package jfrace;

import jfrace.device.RAM;

import java.util.Vector;

/**
 *  The address space is the backbone of the framework.
 *  CPU's and devices know from each other due to the
 *  address space. It is similar to the bus in a computer system.
 *  An address can be any non-negative integer below 2<sup>31</sup>.
 *  <p>
 *  The address space is organized in non-overlapping address
 *  banks. An address bank is defined when a {@link Device}
 *  is attached (by the method {@link addDevice})
 *  to a certain address.
 *  The size of the <code>Device</code> (obtained by the method
 *  {@link Device#getNumberOfUnits}) defines the size of the
 *  address bank. For Reading and writing the invokation of
 *  the methods {@link #getValue}/{@link #getByte} and
 *  {@link #setValue}/{@link #setByte}, resp.,
 *  are redirected to the corresponding methods of
 *  <code>Device</code>. Usually {@link CPU CPUs} will call
 *  these methods. Reading from or writing to an address
 *  outside any address bank may lead to a
 *  {@link BusErrorException}.
 *  <p>
 *  Also a <code>CPU</code> can be added to an address
 *  (by the method {@link #addCPU}).
 *  CPU's are stored in a list. Devices interested in invoking
 *  an interrupt on a CPU can get the corresponding instance
 *  with the method {@link #getCPU}.
 *  <p>
 *  Devices and CPU's (i.e., any {@link Hardware}) can be
 *  removed from the <code>AddressSpace</code>.
 *  If a <code>Hardware</code> is added or removed all subscribed
 *  {@link AddressSpaceListener} will be notified.
 *
 *  @author Franz-Josef Elmer
 */
public class AddressSpace {
    private Vector listeners; // List of AddressSpaceListeners

    // Address banks and devices
    private int[] firstAddresses = new int[0];
    private int[] lastAddresses = new int[0];
    private int[] modeMasks = new int[0];
    private Device[] devices = new Device[0];
    // To increase performance RAMs are handled differently
    private boolean[] isRAM = new boolean[0];
    private byte[][] memories = new byte[0][];
    private boolean[] isWritable = new boolean[0];


    private CPU[] processors = new CPU[0];
    private int[] addressOffsets = new int[0];

    private int defaultValue;
    private byte defaultByteValue;

    /**
     *  Default constructor. Identical with
     *  <code>AddressSpace(0)</code>.
     */
    public AddressSpace() {}

    /**
     *  Constructor defining the default value which is returned
     *  when no address bank could be found
     *  in an access due to the {@link #getValue} or
     *  {@link #getByte} method.
     *  @param aDefaultValue Default value for nonsuccessful access.
     */
    public AddressSpace(int aDefaultValue) {
        defaultValue = aDefaultValue;
        defaultByteValue = (byte) defaultValue;
    }

    /**
     *  Add an <code>AddressSpaceListener</code>. Does nothing if already
     *  in the list.
     *  @param asl A new <code>AddressSpaceListener</code>.
     */
    public synchronized void addAddressSpaceListener(AddressSpaceListener asl) {
        if (listeners == null) {
            listeners = new Vector();
        }
        if (!listeners.contains(asl)) {
            listeners.addElement(asl);
        }
    }

    /**
     *  Remove an <code>AddressSpaceListener</code>.
     *  @param asl <code>AddressSpaceListener</code> to be removed.
     */
    public void removeAddressSpaceListener(AddressSpaceListener asl) {
        listeners.removeElement(asl);
    }

    /**
     *  Notify registrated <code>AddressSpaceListeners</code>.
     *  @param anEvent Event to be sent.
     */
    private void notifyListeners(AddressSpaceEvent anEvent) {
        if (listeners != null) {
            synchronized (listeners) {
                for (int i = 0; i < listeners.size(); i++) {
                    ((AddressSpaceListener) listeners.elementAt(i))
                                            .hardwareMoved(anEvent);
                }
            }
        }
    }

    /** Return the number of address banks. */
    public int getNumberOfAddressBanks() {
        return firstAddresses.length;
    }

    /**
     *  Return the first address of a given address bank.
     *  @param index Index of an address bank.
     */
    public int getFirstAddress(int index) {
        return firstAddresses[index];
    }

    /**
     *  Return the address just following the last address
     *  of a given address bank.
     *  @param index Index of address bank.
     */
    public int getLastAddress(int index) {
        return lastAddresses[index];
    }

    /**
     *  Return the mode mask of a given address bank.
     *  @param index Index of address bank.
     */
    public int getModeMask(int index) {
        return modeMasks[index];
    }

    /**
     *  Return the device of a given address bank.
     *  @param index Index of address bank.
     */
    public Device getDevice(int index) {
        return devices[index];
    }

    /**
     *  Add a new device and thereby define a new address bank.
     *  Inform interested {@link AddressSpaceListener}s.
     *  @param firstAddress First address of the added address bank.
     *  @param modeMask Mask for checking the access mode.
     *  @param device Device to be added.
     *  @exception AddressBankOverlapException if an overlap with an
     *             already existing address bank has been detected.
     *  @exception IllegalArgumentException if <code>firstAddress &lt; 0</code>
     *             or <code>device.numberOfUnits() &lt;= 0</code>.
     */
    public synchronized void addDevice(int firstAddress, int modeMask,
                                       Device device)
           throws AddressBankOverlapException, IllegalArgumentException  {
        // Test on illegal arguments
        if (firstAddress < 0) {
            throw new IllegalArgumentException("first address is negative: "
                                               + firstAddress);
        }
        int numberOfUnits = device.getNumberOfUnits();
        if (numberOfUnits <= 0) {
            throw new IllegalArgumentException("Number of units is <= 0: "
                                               + numberOfUnits);
        }

        // Test on overlap and find insert position
        int lastAddress = firstAddress + numberOfUnits;
        int index = 0; // index of insert position
        boolean overlap;
        for (int i = 0; i < firstAddresses.length; i++) {
            overlap = firstAddress < lastAddresses[i]
                      && lastAddress > firstAddresses[i];
            if (overlap) {
                throw new AddressBankOverlapException("["
                    + Integer.toHexString(firstAddress) + ", "
                    + Integer.toHexString(firstAddress
                                          + device.getNumberOfUnits() - 1)
                    + "] overlaps with ["
                    + Integer.toHexString(firstAddresses[i]) + ", "
                    + Integer.toHexString(lastAddresses[i]) + "].");
            }
            if (firstAddresses[i] < firstAddress) {
                index = i + 1;
            }
        }

        // Make new arrays and copy old arrays into the new ones
        int[] newFirstAddresses = new int[firstAddresses.length + 1];
        int[] newLastAddresses = new int[newFirstAddresses.length];
        int[] newModeMasks = new int[newFirstAddresses.length];
        Device[] newDevices = new Device[newFirstAddresses.length];
        boolean[] newIsRAM = new boolean[newFirstAddresses.length];
        boolean[] newIsWritable = new boolean[newFirstAddresses.length];
        byte[][] newMemories = new byte[newFirstAddresses.length][];
        if (index > 0) {
            System.arraycopy(firstAddresses, 0, newFirstAddresses, 0, index);
            System.arraycopy(lastAddresses, 0, newLastAddresses, 0, index);
            System.arraycopy(modeMasks, 0, newModeMasks, 0, index);
            System.arraycopy(devices, 0, newDevices, 0, index);
            System.arraycopy(isRAM, 0, newIsRAM, 0, index);
            System.arraycopy(isWritable, 0, newIsWritable, 0, index);
            System.arraycopy(memories, 0, newMemories, 0, index);
        }
        int len = firstAddresses.length - index;
        if (len > 0) {
            int i1 = index + 1;
            System.arraycopy(firstAddresses, index,
                             newFirstAddresses, i1, len);
            System.arraycopy(lastAddresses, index,
                             newLastAddresses, i1, len);
            System.arraycopy(modeMasks, index, newModeMasks, i1, len);
            System.arraycopy(devices, index, newDevices, i1, len);
            System.arraycopy(isRAM, index, newIsRAM, i1, len);
            System.arraycopy(isWritable, index, newIsWritable, i1, len);
            System.arraycopy(memories, index, newMemories, i1, len);
        }
        firstAddresses = newFirstAddresses;
        lastAddresses = newLastAddresses;
        modeMasks = newModeMasks;
        devices = newDevices;
        isRAM = newIsRAM;
        isWritable = newIsWritable;
        memories = newMemories;

        // Store address space, mode mask, and device. Notify listeners.
        firstAddresses[index] = firstAddress;
        lastAddresses[index] = lastAddress;
        modeMasks[index] = modeMask;
        devices[index] = device;
        isRAM[index] = device instanceof RAM;
        if (isRAM[index]) {
            memories[index] = ((RAM) device).getMemory();
            isWritable[index] = !((RAM) device).isReadOnly();
        }
        notifyListeners(new AddressSpaceEvent(this, AddressSpaceEvent.ADD,
                                              firstAddress, modeMask, device));
    }

    /**
     *  Remove a device and the associated address bank.
     *  Does nothing if given device is not attached to this
     *  <code>AddressSpace</code>. Interested
     *  {@link AddressSpaceListener}s will be informed.
     *  @param device Device to be removed.
     */
    public synchronized void removeDevice(Device device) {
        // search for device
        int index = -1;
        for (int i = 0; i < devices.length; i++) {
            if (devices[i] == device) {
                index = i;
                break;
            }
        }

        if (index < 0) { // not found
            return;
        }
        int firstAddress = firstAddresses[index];
        int modeMask = modeMasks[index];

        // remove device and address bank
        Device[] newDevices = new Device[devices.length - 1];
        int[] newFirstAddresses = new int[newDevices.length];
        int[] newLastAddresses = new int[newDevices.length];
        int[] newModeMasks = new int[newDevices.length];
        boolean[] newIsRAM = new boolean[newDevices.length];
        boolean[] newIsWritable = new boolean[newDevices.length];
        byte[][] newMemories = new byte[newDevices.length][];
        System.arraycopy(devices, 0, newDevices, 0, index);
        System.arraycopy(firstAddresses, 0, newFirstAddresses, 0, index);
        System.arraycopy(lastAddresses, 0, newLastAddresses, 0, index);
        System.arraycopy(modeMasks, 0, newModeMasks, 0, index);
        System.arraycopy(isRAM, 0, newIsRAM, 0, index);
        System.arraycopy(isWritable, 0, newIsWritable, 0, index);
        System.arraycopy(memories, 0, newMemories, 0, index);
        System.arraycopy(devices, index + 1, newDevices, index,
                         newDevices.length - index);
        System.arraycopy(firstAddresses, index + 1, newFirstAddresses, index,
                         newFirstAddresses.length - index);
        System.arraycopy(lastAddresses, index + 1, newLastAddresses, index,
                         newLastAddresses.length - index);
        System.arraycopy(modeMasks, index + 1, newModeMasks, index,
                         newModeMasks.length - index);
        System.arraycopy(isRAM, index + 1, newIsRAM, index,
                         newIsRAM.length - index);
        System.arraycopy(isWritable, index + 1, newIsWritable, index,
                         newIsWritable.length - index);
        System.arraycopy(memories, index + 1, newMemories, index,
                         newMemories.length - index);
        firstAddresses = newFirstAddresses;
        lastAddresses = newLastAddresses;
        modeMasks = newModeMasks;
        devices = newDevices;
        isRAM = newIsRAM;
        isWritable = newIsWritable;
        memories = newMemories;

        // Notify registrated AddressSpaceListeners
        notifyListeners(new AddressSpaceEvent(this, AddressSpaceEvent.REMOVE,
                                              firstAddress, modeMask, device));
    }

    /**
     *  Return the number of attached CPUs.
     */
    public int getNumberOfCPUs() {
        return processors.length;
    }

    /**
     *  Return the CPU for the given index.
     */
    public CPU getCPU(int index) {
        return processors[index];
    }

    /**
     *  Return the address offset for the CPU of given index.
     */
    public int getCPUAddressOffset(int index) {
        return addressOffsets[index];
    }

    /**
     *  Add a CPU. Interested {@link AddressSpaceListener}s will
     *  be informed.
     *  @param anAddressOffset Address offset.
     *  @param aCpu CPU which will be added.
     */
    public synchronized void addCPU(int anAddressOffset, CPU aCpu) {
        CPU[] newProcessors = new CPU[processors.length + 1];
        int[] newAddressOffsets = new int[newProcessors.length];
        System.arraycopy(processors, 0, newProcessors, 0, processors.length);
        System.arraycopy(addressOffsets, 0, newAddressOffsets, 0,
                         addressOffsets.length);
        newProcessors[processors.length] = aCpu;
        newAddressOffsets[addressOffsets.length] = anAddressOffset;
        processors = newProcessors;
        addressOffsets = newAddressOffsets;
        notifyListeners(new AddressSpaceEvent(this, AddressSpaceEvent.ADD,
                                              anAddressOffset, 0, aCpu));
    }

    /**
     *  Remove a CPU. Does nothing if given CPU is not attached to this
     *  <code>AddressSpace</code>. Interested
     *  {@link AddressSpaceListener}s will be informed.
     *  @param aCpu CPU to be removed.
     */
    public synchronized void removeCPU(CPU aCpu) {
        // search for cpu
        int index = -1;
        for (int i = 0; i < processors.length; i++) {
            if (processors[i] == aCpu) {
                index = i;
                break;
            }
        }

        if (index < 0) { // not found
            return;
        }
        int addressOffset = addressOffsets[index];

        // remove cpu
        CPU[] newProcessors = new CPU[processors.length - 1];
        int[] newAddressOffsets = new int[newProcessors.length];
        System.arraycopy(processors, 0, newProcessors, 0, index);
        System.arraycopy(processors, index + 1, newProcessors, index,
                         newProcessors.length - index);
        System.arraycopy(addressOffsets, 0, newAddressOffsets, 0, index);
        System.arraycopy(addressOffsets, index + 1, newAddressOffsets, index,
                         newAddressOffsets.length - index);
        processors = newProcessors;
        addressOffsets = newAddressOffsets;

        // Notify registrated AddressSpaceListeners
        notifyListeners(new AddressSpaceEvent(this, AddressSpaceEvent.REMOVE,
                                              addressOffset, 0, aCpu));
    }

    /**
     *  Return the index of the address bank for the given address.
     *  @param anAddress An address.
     *  @return -1 if no address bank found.
     */
    private int getAddressBank(int anAddress) {
        for (int i = 0; i < firstAddresses.length; i++) {
            if (firstAddresses[i] <= anAddress && anAddress < lastAddresses[i]) {
                return i;
            }
        }
        return -1;
    }

    /**
     *  Get the value of the addressable unit at the given address.
     *  If the addressable unit is a byte {@link #getByte} should be
     *  be used.
     *  @param anAddress Address in the address space.
     *  @param mode Access mode.
     *  @return the value.
     *  @exception BusErrorException will be thrown if
     *             <code>mode > 0</code> <i>and</i>
     *             <code>mode</code> masked with the
     *             <code>modeMask</code> attribute of the address
     *             bank yields zero.
     */
    public final int getValue(int anAddress, int mode) throws BusErrorException {
        int index = getAddressBank(anAddress);
        if (index < 0) {
            return defaultValue;
        }
        if (mode == 0 || (mode & modeMasks[index]) != 0) {
            return devices[index].getValue(anAddress - firstAddresses[index]);
        } else {
            throw new BusErrorException("address = "
                + Integer.toHexString(anAddress) + ", mode = "
                + Integer.toHexString(mode));
        }
    }

    /**
     *  Set the value of the addressable unit at the given address.
     *  If the addressable unit is a byte {@link #setByte} should be
     *  be used.
     *  @param anAddress Address in the address space.
     *  @param mode Access mode.
     *  @param value Value.
     *  @exception BusErrorException will be thrown if
     *             <code>mode > 0</code> <i>and</i>
     *             <code>mode</code> masked with the
     *             <code>modeMask</code> attribute of the address
     *             bank yields zero.
     */
    public final void setValue(int anAddress, int mode, int value)
           throws BusErrorException {
        int index = getAddressBank(anAddress);
        if (index >= 0) {
            if (mode == 0 || (mode & modeMasks[index]) != 0) {
                devices[index].setValue(anAddress - firstAddresses[index],
                                        value);
            } else {
                throw new BusErrorException("address = "
                    + Integer.toHexString(anAddress) + ", mode = "
                    + Integer.toHexString(mode));
            }
        }
    }

    /**
     *  Get the byte at the given address.
     *  This method is similar {@link #getValue}
     *  but has a better performance.
     *  <br>WARNING: This method should
     *  be used only when the addressable unit is a byte.
     *  @param anAddress Address in the address space.
     *  @param mode Access mode.
     *  @return the value.
     *  @exception BusErrorException will be thrown if
     *             <code>mode > 0</code> <i>and</i>
     *             <code>mode</code> masked with the
     *             <code>modeMask</code> attribute of the address
     *             bank yields zero.
     */
    public final byte getByte(int anAddress, int mode) throws BusErrorException {
        int index = getAddressBank(anAddress);
        if (index < 0) {
            return defaultByteValue;
        }
        if (mode == 0 || (mode & modeMasks[index]) != 0) {
            int adr = anAddress - firstAddresses[index];
            if (isRAM[index]) {
                return memories[index][adr];
            } else {
                return devices[index].getByte(adr);
            }
        } else {
            throw new BusErrorException("address = "
                + Integer.toHexString(anAddress) + ", mode = "
                + Integer.toHexString(mode));
        }
    }

    /**
     *  Set the byte at the given address.
     *  This method is similar to {@link #setValue} but has a
     *  better performance.
     *  <br>WARNING: This method should
     *  be used only when the addressable unit is a byte.
     *  @param anAddress Address in the address space.
     *  @param mode Access mode.
     *  @param value Value.
     *  @exception BusErrorException will be thrown if
     *             <code>mode > 0</code> <i>and</i>
     *             <code>mode</code> masked with the
     *             <code>modeMask</code> attribute of the address
     *             bank yields zero.
     */
    public final void setByte(int anAddress, int mode, byte value)
           throws BusErrorException {
        int index = getAddressBank(anAddress);
        if (index >= 0) {
            if (mode == 0 || (mode & modeMasks[index]) != 0) {
                int adr = anAddress - firstAddresses[index];
                if (isRAM[index]) {
                    if (isWritable[index]) {
                        memories[index][adr] = value;
                    }
                } else {
                    devices[index].setByte(adr, value);
                }
            } else {
                throw new BusErrorException("address = "
                    + Integer.toHexString(anAddress) + ", mode = "
                    + Integer.toHexString(mode));
            }
        }
    }
}
