package jfrace;

/**
 *  Event which happens when a piece of hardware (either a
 *  {@link Device} or a {@link CPU})
 *  is added or removed from an {@link AddressSpace}.
 *
 *  @author Franz-Josef Elmer
 */
public class AddressSpaceEvent {
    /** Event type: add a piece of hardware. */
    public static final int ADD = 0;
    /** Event type: remove a piece of hardware. */
    public static final int REMOVE = 1;

    private AddressSpace addressSpace;
    private int type;
    private int addressOffset;
    private int modeMask;
    private Hardware hardware;

    /**
     *  Create an event.
     *  @param anAddressSpace Address space from which (if
     *                        <code>aType == REMOVE</code>) or to
     *                        which (if <code>aType == ADD</code>)
     *                        <code>aPieceOfHardware</code> has been
     *                        removed or added, resp.
     *  @param aType Type of action. Either <code>ADD</code>
     *               or <code>REMOVE</code>.
     *  @param anAddressOffset Address offset.
     *  @param aModeMask Mode mask. Only needed for devices.
     *  @param aPieceOfHardware Hardware which is either added or removed.
     */
    public AddressSpaceEvent(AddressSpace anAddressSpace, int aType,
                             int anAddressOffset, int aModeMask,
                             Hardware aPieceOfHardware) {
        addressSpace = anAddressSpace;
        type = aType;
        addressOffset = anAddressOffset;
        modeMask = aModeMask;
        hardware = aPieceOfHardware;
    }

    /**
     *  Return the address space from which a piece of hardware has been
     *  added or removed.
     */
    public AddressSpace getAddressSpace() {
        return addressSpace;
    }

    /**
     *  Return the type of action. Possible return values are
     *  <code>ADD</code> or <code>REMOVE</code>.
     */
    public int getType() {
        return type;
    }

    /**
     *  Return the address offset of the added or removed hardware.
     */
    public int getAddressOffset() {
        return addressOffset;
    }

    /**
     *  Return the mode mask of the added or removed device.
     *  @return 0 in the case CPUs.
     */
    public int getModeMask() {
        return modeMask;
    }

    /**
     *  Return the added or removed hardware.
     */
    public Hardware getHardware() {
        return hardware;
    }

    /** Compose a string with all basic informations. */
    public String toString() {
        StringBuffer sb = new StringBuffer("AddressSpaceEvent: ");
        sb.append(Integer.toHexString(addressOffset));
        switch (type) {
        case ADD:
            sb.append(", ADD ");
            break;
        case REMOVE:
            sb.append(", REMOVE ");
            break;
        default:
            sb.append(", unknown type ");
        }
        sb.append(hardware.getClass().getName());
        sb.append('@');
        sb.append(Integer.toHexString(hardware.hashCode()));
        sb.append(", ");
        sb.append(addressSpace);
        if (hardware instanceof Device) {
            sb.append(", ");
            sb.append(Integer.toHexString(modeMask));
        }

        return new String(sb);
    }
}
