package jfrace;

/**
 *  Interface for a central processing unit (CPU).
 *
 *  @author Franz-Josef Elmer
 */
public interface CPU extends Hardware {
    /**
     *  Return the name of the CPU which is emulated.
     */
    public String getName();

    /**
     *  Return the number of registers.
     */
    public int getNumberOfRegisters();

    /**
     *  Return a register.
     *  @param index Index of the register.
     *  @return <code>null</code> if no register exists for the
     *          given index.
     */
    public Register getRegister(int index);

    /**
     *  Return the program counter. The program counter is a
     *  special register.
     */
    public Register getProgramCounter();

    /**
     *  Set the value of a register.
     *  @param index Index of the register.
     *  @param value New value.
     */
    public void setRegisterValue(int index, int value);

    /**
     *  Set the address space.
     */
    public void setAddressSpace(AddressSpace addressSpace);

    /**
     *  Get the address space.
     */
    public AddressSpace getAddressSpace();

    /**
     *  Interrupt the CPU at the next call of {@link #next}.
     *  @param interrupt Bit pattern of the interrupt. The meaning of
     *                   the bits depend on the implementation.
     */
    public void interrupt(long interrupt);

    /**
     *  Return error description. The meaning of the error
     *  code depends on the implementation. This method should be
     *  called when {@link #next} returns <code>-1</code>.
     */
    public String getErrorDescription();

    /**
     *  Proceed one opcode further. Return the number of clock cycles.
     *  @return <code>-1</code> if an error occured.
     */
    public int next();

    /**
     *  Return <code>true</code> if disassembling is supported.
     */
    public boolean supportsDisassembling();

    /**
     *  Disassemble the next op-code at the given address and append
     *  it to the given string buffer. If disassembling isn't
     *  supported -1 will be returned.
     *  @param sb String buffer.
     *  @param address Address of op-code.
     *  @return number of addressable units of the op-code at
     *          <code>address</code>.
     */
    public int disassemble(StringBuffer sb, int address);
}
