package jfrace;

import java.lang.reflect.Method;
import java.util.StringTokenizer;
import java.util.Vector;

/**
 *  Create a {@link CPU} and add it to an {@link AddressSpace}.
 *  Creation and configuration is determined by a description string
 *  of the following syntax:
 *  <p>
 *  <code><nobr><i>&lt;class name&gt;</i> [<i>&lt;offset&gt;</i>]
 *  [<i>&lt;method call 1&gt; &lt;method call 2&gt;</i> ...]</nobr></code>.
 *  </p>
 *  where <code><i>&lt;class name&gt;</i></code> is the class name
 *  of the CPU without the common subpackage name,
 *  <code><i>&lt;offset&gt;</i></code> is the optional offset in
 *  the <code>AddressSpace</code> (default value is 0), and
 *  <code><i>&lt;method call 1&gt; &lt;method call 2&gt;</i> ...</code>
 *  are optional method calls of the form
 *  <p>
 *  <code><nobr><i>&lt;method name&gt;</i>(<i>&lt;1. argument&gt;</i>,<i>&lt;2. argument&gt;</i>,...)</nobr></code>
 *  </p>
 *  Arguments starting with a digit will be treated as <code>int</code>.
 *  Otherwise <code>String</code> will be assumed.
 *  NOTE: Because white space is used as the delimiter between
 *  method calls there should be no white space in a method call.
 *
 *  @author Franz-Josef Elmer
 */
public class CPUConfigurator {
    /**
     *  Create a {@link CPU} from the given description and add
     *  it to the given <code>AddressSpace</code>.
     *  @param as address space.
     *  @param description CPU description.
     *  @param packageCPUPrefix Class name prefix of all classes
     *                          implementing <code>CPU</code>.
     *  @throws ConfigurationException in the case of an error
     *          (including syntax errors in <code>description</code>).
     */
    public static void configureCPU(AddressSpace as, String description,
                                    String packageCPUPrefix)
                  throws ConfigurationException {
        StringTokenizer st = new StringTokenizer(description);
        int tokens = st.countTokens();

        // obtain CPU class name
        if (tokens == 0) {
            throw new ConfigurationException(
                    "class name missing in CPU description.");
        }
        String className = packageCPUPrefix + st.nextToken();
        tokens--;

        // obtain offset in address space
        int offset = 0;
        String token = null;
        if (tokens > 0) {
            token = st.nextToken();
            tokens--;
            if (Character.isDigit(token.charAt(0))) {
                offset = Util.parseNumber(token);
                if (tokens > 0) {
                    token = st.nextToken();
                    tokens--;
                } else {
                    token = null;
                }
            }
        }

        // Create CPU
        CPU cpu = null;
        try {
            cpu = (CPU) Class.forName(className).newInstance();
        } catch (Exception e) {
            throw new ConfigurationException("Couldn't create CPU from '"
                                             + className
                                             + "' because of " + e);
        }

        // Connect CPU with Address Space
        if (cpu instanceof AddressSpaceListener) {
            as.addAddressSpaceListener((AddressSpaceListener) cpu);
        }
        as.addCPU(offset, cpu);
        cpu.setAddressSpace(as);

        // Reset and initialize CPU
        cpu.reset();
        while (token != null) {
            processCPUMethod(cpu, token);
            if (tokens-- > 0) {
                token = st.nextToken();
            } else {
                token = null;
            }
        }
    }

    /**
     *  Execute a method call for the given CPU.
     *  @param cpu CPU.
     *  @param methodCall Method name and arguments.
     *  @throws ConfigurationException in the case of an error
     *          (including syntax errors in <code>methodCall</code>).
     */
    private static void processCPUMethod(CPU cpu, String methodCall)
                            throws ConfigurationException {
        int index = methodCall.indexOf('(');
        if (index < 0) {
            throw new ConfigurationException("'(' missing in "
                                             + methodCall);
        }
        String methodName = methodCall.substring(0, index);
        int index2 = methodCall.indexOf(')');
        if (index2 < 0) {
            throw new ConfigurationException("')' missing in "
                                             + methodCall);
        }
        StringTokenizer st = new StringTokenizer(
                                methodCall.substring(index + 1, index2), ",");
        Vector arguments = new Vector();
        String token;
        while (st.hasMoreTokens()) {
            token = st.nextToken();
            if (Character.isDigit(token.charAt(0))) {
                arguments.addElement(new Integer(Util.parseNumber(token)));
            } else {
                arguments.addElement(token);
            }
        }

        Object[] args = new Object[arguments.size()];
        Class[] classes = new Class[args.length];
        for (int i = 0; i < args.length; i++) {
            args[i] = arguments.elementAt(i);
            if (args[i] instanceof Integer) {
                classes[i] = Integer.TYPE;
            } else {
                classes[i] = args[i].getClass();
            }
        }
        try {
            Method method = cpu.getClass().getMethod(methodName, classes);
            method.invoke(cpu, args);
        } catch (Exception e) {
            throw new ConfigurationException("Error for method '"
                                             + methodCall + "': " + e);
        }
    }
}
