package jfrace;

import jfrace.device.RAM;

import java.io.BufferedReader;
import java.io.FileReader;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Vector;

/**
 *  Test engine for testing implementations of the {@link CPU}
 *  interface. A test is performed by invoking the method
 *  <code>makeTest</code> with a test script. A test script contains
 *  a header, comments (starting with '<code>#</code>')
 *  and many test description lines.
 *  <p>
 *  The header defines the CPU used and configures the RAM. It contains
 *  two lines with the following syntax:
 *  <p>
 *  <nobr><code>CPU = <i>&lt;class name> &lt;offset in address space></i></code></nobr>
 *  <p>
 *  and
 *  <p>
 *  <nobr><code>RAM = <i>&lt;first address> &lt;size></i>
 *        [<i>&lt;mode mask></i>]</code></nobr>
 *  <p>
 *  The meaning of the parameters are:
 *  <table border=0 cellpadding=0>
 *  <tr><td valign=top><code><i>&lt;class name></i></code>&nbsp;</td>
 *      <td valign=top>Complete name of the CPU class</td>
 *  </tr>
 *  <tr><td valign=top><code><i>&lt;offset&nbsp;in&nbsp;address&nbsp;space></i></code>&nbsp;&nbsp;</td>
 *      <td valign=top>Mapping of the First cell the CPU can address onto
 *                     an address in the <code>AddressSpace</code></td>
 *  </tr>
 *  <tr><td valign=top><code><i>&lt;first address></i></code>&nbsp;</td>
 *      <td valign=top>Address of the first cell of the RAM in
 *                     the <code>AddressSpace</code></td>
 *  </tr>
 *  <tr><td valign=top><code><i>&lt;size></i></code>&nbsp;</td>
 *      <td valign=top>Number of cells of the RAM</td>
 *  </tr>
 *  <tr><td valign=top><code><i>&lt;mode mask></i></code>&nbsp;</td>
 *      <td valign=top>Optional mode mask</td>
 *  </tr>
 *  </table>
 *  <p>
 *  With the test description lines the behavior of the CPU
 *  will be tested on specified op-codes. A test description line
 *  contains information in order to prepare the CPU state and
 *  the RAM and exceptions about the state of CPU and RAM after
 *  execution of the <code>next</code> method. If there is
 *  a difference between expection and the true state an
 *  error message will be printed.
 *  <p>
 *  Here is the syntax (in EBNF) of a test description line:
 *  <pre>
 *  testDescriptionLine := initialState ">" numberOfOpCodeCycles expectations
 *  initialState := statement | statement ";" initialState
 *  numberOfOpCodeCycles := number
 *  expectations := statement | statement ";" expectations
 *  statement := registerName {"="|"=="} number | address {"="|"=="} numbers
 *  address := "(" {number | registerName} ")"
 *  numbers := number | number "," numbers
 *  number := binaryNumber | octalNumber | decimalNumber | hexNumber
 *  binaryNumber := {"0b"|"0B"} binaryDigit<sup>*</sup>
 *  binaryDigit := "0" | "1"
 *  octalNumber := "0" octalDigit<sup>*</sup>
 *  octalDigit := "0" | "1" | "2" | "3" | "4" | "5" | "6" | "7"
 *  decimalNumber := nonZeroDigit digit<sup>*</sup>
 *  nonZeroDigit := "1" | "2" | "3" | "4" | "5" | "6" | "7" | "8" | "9"
 *  digit := "0" | nonZeroDigit
 *  hexNumber := {"0x"|"0X"} hexDigit<sup>*</sup>
 *  hexDigit := digit | "a" | "A" | "b" | "B" | "c" | "C" | "d" | "D" | "e" | "E" | "f" | "F"
 *  </pre>
 *  <p>
 *  Here is an example of a test script:
 *  <pre>
 *  # Header
 *  CPU = jfrace.cpu.FJE5 0
 *  RAM = 0 0x400
 *
 *  # test description lines
 *  SP==1000; PC=0x123; (0x123)==0 > 1 PC=0x124
 *  SP==1000; PC=123; (123)==1; A=0b10111; I=0b01100 > 1 PC=124; A=0b01100; I=0b10111
 *  SP==1000; PC=0x333; (0x333)==2,23; A=0b10111; I==0b01100 > 2 PC=0x335; A=23
 *  SP==1000; PC=12; (12)==3,30; A==22; I=11 > 2 PC=14; I=30
 *  </pre>
 *
 *  A <code>registerName</code> is either the name of a single
 *  register of the CPU (as given by
 *  <code>CPU.getRegister(index).getName()</code>) or concatenation
 *  of possible register names.
 *  A <code>statement</code> contains either a <code>=</code> or a
 *  <code>==</code>. In the later case
 *  such a <code>statement</code> of the <code>initialState</code>
 *  will be duplicated to the <code>statement</code> list of the
 *  <code>expectations</code>.
 *  <p>
 *  Before preparation RAM will be reseted by invoking the method
 *  <code>reset</code>).
 *  In addition all CPU registers will be set to zero
 *  (invoking <code>reset</code> would not do the job in CPUs like
 *  6502).
 *  The returned value of <code>next</code> will be checked with
 *  <code>numberOfOpCodeCycles</code> in the test description line.
 *  Not only the exceptions describe in the test decription line
 *  are tested but also the unmentioned registers and memory cells
 *  are tested against their default values.
 *
 *  @author Franz-Josef Elmer
 */
public class CPUTestEngine {
    /**
     *  Run the test engine for one or several test scripts.
     *  @param scripts File names of the test scripts.
     */
    public static void main(String[] scripts) {
        if (scripts.length == 0) {
            System.out.println("Usage: java jfrace.CPUTestEngine <test script 1> <test script 2> ...");
            System.exit(0);
        }
        CPUTestEngine engine = new CPUTestEngine(true);
        for (int i = 0; i < scripts.length; i++) {
            System.out.println(" ================= " + (i + 1)
                               + ": Run script " + scripts[i]);
            if (!engine.makeTest(scripts[i])) {
                System.out.println("======== ERROR =========");
                break;
            }
        }
    }

    private int count = 0; // line counter for the script
    private CPU cpu;
    private AddressSpace as;
    private Hashtable regHashTable;
    private boolean totalResult;
    private boolean onSystemOut;
    private StringBuffer protocol = new StringBuffer();

    /**
     *  Create a new instance.
     *  @param onSystemOutFlag If <code>true</code> output goes onto
     *                         <code>System.out</code> otherwise
     *                         it will be stored in a <code>StringBuffer</code>
     *                         which can be obtained by the method
     *                         <code>getProtocol</code>.
     */
    public CPUTestEngine(boolean onSystemOutFlag) {
        onSystemOut = onSystemOutFlag;
    }

    /** Return the protocol. */
    public String getProtocol() {
        return new String(protocol);
    }

    private void print(Object object) {
        if (onSystemOut) {
            System.out.print(object);
        } else {
            protocol.append(object);
        }
    }

    private void println(Object object) {
        if (onSystemOut) {
            System.out.println(object);
        } else {
            protocol.append(object);
            protocol.append('\n');
        }
    }

    /**
     *  Run the test with the given test script.
     *  @param testScript Script with the test statements.
     *  @return <code>true</code> for a successful test.
     */
    public boolean makeTest(String testScript) {
        totalResult = true;

        as = new AddressSpace();
        try {
            BufferedReader br = new BufferedReader(new FileReader(testScript));
            String line;
            String errorMessage;
            long time = System.currentTimeMillis();
            int numberOfTests = 0;
            int numberOfErrors = 0;
            while ((line = br.readLine()) != null) {
                count++;
                if (line.length() > 0 && !line.startsWith("#")) {
                    if (line.startsWith("CPU")) {
                        processCPULine(line, as);
                    } else if (line.startsWith("RAM")) {
                        if (cpu != null) {
                            processRAMLine(line, as);
                        } else {
                            print("ERR RAM specification should ");
                            println("appear after the CPU specification.");
                            return false;
                        }
                    } else {
                        numberOfTests++;
                        errorMessage = test(line);
                        if (errorMessage != null) {
                            totalResult = false;
                            numberOfErrors++;
                            println(errorMessage);
                            println(count + ": " + line);
                        }
                    }
                }
            }
            println("Test finished after "
                               + (System.currentTimeMillis() - time)
                               + " msec.");
            println("Number of tests: " + numberOfTests);
            println("Number of errors: " + numberOfErrors);
            br.close();
        } catch (Exception e) {
            totalResult = false;
            println("ERR in line " + count);
            println(e);
        }
        return totalResult;
    }

    /**
     *  Initialize the hashtable for the registers. The register
     *  name is the key of <code>regHashTable</code> where
     *  the index and the mask are the corresponding value.
     */
    private void initRegHashTable() {
        regHashTable = new Hashtable();
        int n = cpu.getNumberOfRegisters();
        Register r;
        int mask;
        int[] indexBitsMask;
        for (int i = 0; i < n; i++) {
            r = cpu.getRegister(i);
            indexBitsMask = new int[3];
            indexBitsMask[0] = i;
            indexBitsMask[1] = r.getNumberOfBits();
            mask = 0;
            for (int j = 0; j < indexBitsMask[1]; j++) {
                mask = (mask << 1) | 1;
            }
            indexBitsMask[2] = mask;
            regHashTable.put(r.getName(), indexBitsMask);
        }
    }

    /**
     *  Test the CPU in occordance to the given description.
     *  The test steps are:
     *  <ol><li>Reset the CPU and all devices attached at the address space
     *          by invoking their <code>reset</code> methods.
     *      <li>Set registers and memory in accordance to the description.
     *      <li>Execute a single step by invoking the method
     *          <code>next</code>
     *      <li>Compare the register values with the expected
     *          results as given in the description string.
     *  </ol>
     *  @param testDescription describes intial state and expected
     *                         result. The syntax of the description
     *                         is explained above.
     *  @return <code>null</code> if the test was successful. Otherwise
     *          an error message explaining the failure is returned.
     */
    private String test(String testDescription) {
        StringTokenizer st;
        int index = testDescription.indexOf('>');
        if (index < 0) {
            return "Error: '>' missing.";
        }

        // 1. Reset devices (like RAM)
        int n = as.getNumberOfAddressBanks();
        for (int i = 0; i < n; i++) {
            as.getDevice(i).reset();
        }

        // 2. Reset CPU and set the CPU registers to zero
        cpu.reset();
        for (int i = 0; i < cpu.getNumberOfRegisters(); i++) {
            cpu.setRegisterValue(i, 0);
        }

        // 3. Set registers and memory
        Vector list = new Vector();
        st = new StringTokenizer(testDescription.substring(0, index), ";");
        Statement statement;
        String errorMessage;
        while (st.hasMoreTokens()) {
            statement = new Statement(st.nextToken(), regHashTable);
            if (statement.isDoubleEqualSign()) {
                list.addElement(statement);
            }
            errorMessage = statement.setState(cpu, as);
            if (errorMessage != null) {
                return errorMessage;
            }
        }

        // 4. Execute a single step
//        println(cpu);
        int opCodeCycles = cpu.next();
//        println(cpu);

        // 5. Compare with expectations
        boolean[] testedRegisters = new boolean[cpu.getNumberOfRegisters()];
        // 5.1 First with '==' statements from the initial state part
        StringBuffer sb = new StringBuffer();
        for (int i = 0; i < list.size(); i++) {
            statement = (Statement) list.elementAt(i);
            statement.test(sb, cpu, as);
            statement.markTestedRegisters(testedRegisters);
        }
        // 5.2 statements from the exception part
        testDescription = testDescription.substring(index + 1).trim();
        index = testDescription.indexOf(' ');
        int cycles = Integer.parseInt(testDescription.substring(0, index));
        if (cycles != opCodeCycles) {
            return "Wrong number of opcode cycles: " + opCodeCycles;
        }
        st = new StringTokenizer(testDescription.substring(index + 1), ";");
        while (st.hasMoreTokens()) {
            statement = new Statement(st.nextToken(), regHashTable);
            statement.test(sb, cpu, as);
            statement.markTestedRegisters(testedRegisters);
        }
        // 5.3 test all default registers
        int value;
        Register r;
        for (int i = 0; i < testedRegisters.length; i++) {
            if (!testedRegisters[i] && cpu.getRegister(i).getValue() != 0) {
                r = cpu.getRegister(i);
                sb.append(' ');
                sb.append(r.getName());
                sb.append('=');
                Util.appendHex(sb, (r.getNumberOfBits() + 3) >> 2, r.getValue());
                sb.append("!=0");
            }
        }
        // 5.4 test memory
        int lastAddress;
        int count=0;
        try {
            for (int i = 0; i < as.getNumberOfAddressBanks(); i++) {
                lastAddress = as.getLastAddress(i);
                for (int j = as.getFirstAddress(i); j < lastAddress; j++) {
                    if (as.getByte(j, 0) != 0) {
                        sb.append(" (");
                        Util.appendHex(sb, j);
                        sb.append(")=");
                        Util.appendHex(sb, as.getByte(j, 0));
                        sb.append("!=0");
                    }
                }
            }
        } catch (Exception e) {
            println("Oops! " + e);
        }

        if (sb.length() == 0) {
            return null;
        } else {
            sb.insert(0, "Error:");
            return new String(sb);
        }
    }

    private void processCPULine(String line, AddressSpace as) throws Exception {
        int index = line.indexOf('=');
        if (index < 0) {
            totalResult = false;
            println("ERR CPU missing: " + count + ": " + line);
        } else {
            try {
                CPUConfigurator.configureCPU(as,
                                             line.substring(index + 1).trim(),
                                             "");
                cpu = as.getCPU(0);
            } catch (Exception e) {
                System.out.println(e);
            }
            initRegHashTable();
        }
    }

    private void processRAMLine(String line, AddressSpace as) {
        int index = line.indexOf('=');
        if (index < 0) {
            totalResult = false;
            println("ERR RAM specification missing: " + count
                               + ": " + line);
        } else {
            try {
                RAM.createRAMAndAddToAS(as, line.substring(index + 1).trim(),
                                        null);
            } catch (Exception e) {
                System.out.println(e);
            }
        }
    }

}
