package jfrace;

/**
 *  Interface for devices (e.g.&nbsp;RAM, graphic card,
 *  printer, etc). A device is a block of consecutive
 *  addressable units which will be attached to an
 *  {@link AddressSpace}. Other <code>Hardware</code>
 *  objects (mainly {@link CPU CPUs}) can read from or
 *  write to the <code>Device</code> via the <code>AddressSpace</code>.
 *  The addresses of these calls will be translated by
 *  the <code>AddressSpace</code> to the relative address of
 *  the <code>Device</code>.
 *
 *  @author Franz-Josef Elmer
 */
public interface Device extends Hardware {
    /** Return the number of addressable units. */
    public int getNumberOfUnits();

    /**
     *  Get the value of the addressable unit at the given address.
     *  If the addressable unit is a byte <code>getByte</code> should be
     *  supported.
     *  @param address Memory address.
     *  @return the value.
     */
    public int getValue(int address);

    /**
     *  Set the value of the addressable unit at the given address.
     *  If the addressable unit is a byte <code>setByte</code> should be
     *  supported.
     *  @param address Memory address.
     *  @param value the new value.
     */
    public void setValue(int address, int value);

    /**
     *  Get the byte at the given address.
     *  This method is similar
     *  to <code>getValue</code> but has a better performance.
     *  <br>WARNING: This method should
     *  be used only when the addressable unit is a byte.
     *  @param address Memory address.
     *  @return the value of the byte at <code>address</code>.
     */
    public byte getByte(int address);

    /**
     *  Set the byte at the given address.
     *  This method is similar
     *  to <code>setValue</code> but has a better performance.
     *  <br>WARNING: This method should
     *  be used only when the addressable unit is a byte.
     *  @param address Memory address.
     *  @param value new value of the byte at <code>address</code>.
     */
    public void setByte(int address, byte value);
}
