package jfrace;

import jfrace.device.RAM;
import jfrace.view.EmulatorGUI;

import java.applet.Applet;
import java.lang.reflect.Method;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.net.URL;
import java.util.StringTokenizer;
import java.util.Vector;

/**
 *  The <code>Emulator</code> configurates a system and emulates it.
 *  It is a singleton and can be used as an application or inside an
 *  applet.
 *  <p>
 *  The configuration is done with the help of a configuration files.
 *  They are similar to Java properties files. For more details
 *  see {@link SystemConfigurator}.
 *
 *  @author Franz-Josef Elmer
 */
public class Emulator {
    static final int YIELD_INTERVAL = 500;
    private static final Emulator emulator = new Emulator();

    /**
     *  Return the single instance of <code>Emulator</code>.
     */
    public static Emulator getEmulator() {
        return emulator;
    }

    /**
     *  Configure a system and run it.
     *  The usage is:
     *  <pre>
     *  java ce.Emulator [&lt;options>] &lt;configSource> [&lt;additional options>]
     *  </pre>
     *  where
     *  <dl><dt><code>&lt;options></code></dt>
     *      <dd>Command line options
     *      <dt><code>&lt;configSource></code></dt>
     *      <dd>Source of the configuration file either an URL or a
     *          (local) file
     *      <dt><code>&lt;additional options></code></dt>
     *      <dd>These additional options will be delivered to
     *          configurators inside the configuration files
     *  </dl>
     *  <p>
     *  List of the valid command line options:
     *  <dl><dt><code>-d<br>-d<i>start</i><br>-d<i>start</i>.<i>end</i><br>
     *          -d<i>start</i>.<i>end</i>.<i>step</i></code></dt>
     *      <dd>Debug mode. It starts after the <code><i>start</i></code>-th
     *          instruction has been executed and ends just before the
     *          <code><i>end</i></code>-th instruction.
     *          Every <code><i>step</i></code>-th step the CPUs will be
     *          printed on the console. The default values of
     *          <code><i>start</i></code>, <code><i>end</i></code>,
     *          and <code><i>step</i></code> are 0, infinity, and 1 resp.
     *      <dt><code>-p</code></dt>
     *      <dd>Print effective clock frequency in regular time steps.
     *          The time step is given (in msec) by the property
     *          <code>adaptationTime</code> of the configuration file.
     *      <dt><code>-g</code></dt>
     *      <dd>Show GUI which visualize CPU state and memory.
     *      <dt><code>-n</code></dt>
     *      <dd>Does not start emulation.
     *      <dt><code>-t</code></dt>
     *      <dd>In the case of more than one processor each processor will
     *          be emulated in its own thread.
     *  </dl>
     *
     */
    public static void main(String[] args) {
        try {
            URL documentBase = new URL("file:///"
                                       + System.getProperty("user.dir")
                                       + '/');
            emulator.setDocumentBase(documentBase);
        } catch (Exception e) {}
        String configFile = null;
        String[] additionalArgs = null;
        int offset = 0;
        for (int i = 0; i < args.length; i++) {
            if (configFile == null) {
                if (args[i].charAt(0) == '-') {
                    if (!emulator.analyseOption(args[i])) {
                        System.out.println("Invalid option: " + args[i]);
                    }
                } else {
                    configFile = args[i];
                    offset = i + 1;
                    additionalArgs = new String[args.length - offset];
                }
            } else {
                additionalArgs[i - offset] = args[i];
            }
        }
        if (configFile == null) {
            System.out.println("Usage: java Emulator [<options>] "
                               + "<config file> [<additional options>]");
            System.exit(0);
        } else {
            try {
                emulator.configure(configFile, additionalArgs);
                if (emulator.isNoEmulation()) {
                    System.out.println("Configuration finished but no emulation.");
                } else {
                    emulator.runEmulation();
                }
            } catch (ConfigurationException e) {
                e.printStackTrace();
            }
        }
        if (emulator.isWithGUI()) {
            new EmulatorGUI().show();
        }
    }

    private SystemConfigurator systemConfigurator;
    private AddressSpace as;
    private long yieldInterval;
    private double clockFrequency;
    private long adaptationTime;
    private boolean printEffectiveClockFreq;
    private URL documentBase;
    private Applet applet;
    private boolean runsInsideAnApplet;
    private boolean debuging;
    private int debugStart;
    private int debugEnd;
    private int debugStep = 1;
    private boolean noEmulation;
    private boolean withGUI;
    private boolean ownThread;
    private EmulatorThread[] emuThreads;

    /** Default constructor is private because this is a singlton. */
    private Emulator() {}

    /** Reset emulator. Needed only for testing purpose. */
    void reset() {
        as = null;
        debuging = false;
        debugStart = 0;
        debugEnd = 0;
        debugStep = 1;
        ownThread = false;
        noEmulation = false;
        withGUI = false;
        emuThreads = null;
    }

    boolean getDebuging() {
        return debuging;
    }

    int getDebugStart() {
        return debugStart;
    }

    int getDebugEnd() {
        return debugEnd;
    }

    int getDebugStep() {
        return debugStep;
    }

    boolean getOwnThread() {
        return ownThread;
    }

    /** Set the applet. */
    void setApplet(Applet theApplet) {
        applet = theApplet;
    }

    /** Return applet. */
    public Applet getApplet() {
        return applet;
    }

    /** Return <code>true</code> if the emulator runs inside an Applet. */
    public boolean insideAnApplet() {
        return applet != null;
    }

    /**
     *  Return <code>true</code> if the no-emulation flag is set.
     *  Is happens if a command line option was "-n".
     */
    public boolean isNoEmulation() {
        return noEmulation;
    }

    /**
     *  Return <code>true</code> if GUI flag is set. Is happens
     *  if a command line option was "-g".
     */
    public boolean isWithGUI() {
        return withGUI;
    }

    /**
     *  Return the address space.
     */
    public AddressSpace getAddressSpace() {
        return as;
    }

    /** Set the document base for relative URL sources. */
    public void setDocumentBase(URL anURL) {
        documentBase = anURL;
        systemConfigurator = new SystemConfigurator(documentBase);
    }

    /**
     *  Return the absolute URL built from
     *  <code>relativeOrAbsoluteURL</code>. If
     *  <code>relativeOrAbsoluteURL</code> starts with "http://"
     *  it is an absolute URL otherwise a relative one is assumed
     *  and an URL will be returned which is built by the document base
     *  (as it is set by the method <code>setDocumentBase()</code>)
     *  and <code>relativeOrAbsoluteURL</code>.
     */
    public URL getAbsoluteURL(String relativeOrAbsoluteURL) {
        try {
            if (relativeOrAbsoluteURL.startsWith("http://")) {
                return new URL(relativeOrAbsoluteURL);
            } else {
                return new URL(documentBase, relativeOrAbsoluteURL);
            }
        } catch (Exception e) {
            return null;
        }
    }

    /**
     *  Analyse a command line option.
     *  @param option Command line option including '-'.
     *  @return <code>true</code> if the option was a valid one.
     */
    public boolean analyseOption(String option) {
        if (option.startsWith("-d")) { // debug option
            debuging = true;
            StringTokenizer st = new StringTokenizer(option.substring(2), ".");
            if (st.hasMoreTokens()) {
                debugStart = Util.parseNumber(st.nextToken());
            }
            if (st.hasMoreTokens()) {
                debugEnd = Util.parseNumber(st.nextToken());
            }
            if (st.hasMoreTokens()) {
                debugStep = Util.parseNumber(st.nextToken());
            }
            return true;
        } else if (option.equals("-g")) {
            noEmulation = true;
            withGUI = true;
            return true;
        } else if (option.equals("-p")) {
            printEffectiveClockFreq = true;
            return true;
        } else if (option.equals("-n")) { // no-emulation option
            noEmulation = true;
            return true;
        } else if (option.equals("-t")) { // thread option
            ownThread = true;
            return true;
        }
        return false;
    }

    /**
     *  Build and configurate the system in accordance with
     *  the configuration file.
     *  @param configSource Source of the configuration file.
     *  @param args Additional arguments.
     *  @throws ConfigurationException in the case of an error
     *          (including syntax errors in the configuration file).
     */
    public void configure(String configSource, String[] args)
                throws ConfigurationException {
        systemConfigurator.configure(configSource, args);
        as = systemConfigurator.getAddressSpace();
        yieldInterval = systemConfigurator.getYieldInterval();
        clockFrequency = systemConfigurator.getClockFrequency();
        adaptationTime = systemConfigurator.getAdaptationTime();
    }

    /**
     *  Run emulation.
     */
    public void runEmulation() {
        if (as == null) {
            System.out.println("Can't run because no address space defined.");
            return;
        }
        if (as.getNumberOfCPUs() == 0) {
            System.out.println("Can't run because no cpu is defined.");
            return;
        }
        if (ownThread) {
            CPU[] cpus;
            emuThreads = new EmulatorThread[as.getNumberOfCPUs()];
            for (int i = 0; i < emuThreads.length; i++) {
                cpus = new CPU[1];
                cpus[0] = as.getCPU(i);
                emuThreads[i] = new EmulatorThread(cpus, yieldInterval,
                                                   clockFrequency,
                                                   adaptationTime,
                                                   printEffectiveClockFreq,
                                                   debuging, debugStart,
                                                   debugEnd, debugStep);
                emuThreads[i].start();
            }
        } else {
            CPU[] cpus = new CPU[as.getNumberOfCPUs()];
            for (int i = 0; i < cpus.length; i++) {
                cpus[i] = as.getCPU(i);
            }
            emuThreads = new EmulatorThread[1];
            emuThreads[0] = new EmulatorThread(cpus, yieldInterval,
                                               clockFrequency, adaptationTime,
                                               printEffectiveClockFreq,
                                               debuging, debugStart, debugEnd,
                                               debugStep);
            emuThreads[0].start();
        }
    }

    /**
     *  Stop emulation.
     */
    public void stopEmulation() {
        if (emuThreads != null) {
            for (int i = 0; i < emuThreads.length; i++) {
                emuThreads[i] = null;
            }
        }
    }

}


