package jfrace;

import jfrace.view.EmulatorGUI;

import java.applet.Applet;
import java.awt.Label;
import java.util.StringTokenizer;

/**
 *  Running the emulator inside an applet.
 *  The possible parameters are.
 *  <table border=1 cellpadding=4>
 *      <tr><th>Parameter</th><th>Meaning</th></tr>
 *      <tr><td><code>options</code></td>
 *          <td>Correspond to the (optional) command line options
 *              in the application (for more details see
 *              <a href="Emulator.html#main(java.lang.String[])"><code>main</code></a>
 *              method of the <code>Emulator</code> class).
 *              The options are separated by white spaces.
 *      </tr>
 *      <tr><td><code>configFile</code></td>
 *          <td>Configuration file.</td>
 *      </tr>
 *      <tr><td><code>additionalOptions</code></td>
 *          <td>Additional options. Used by configurators inside
 *              the configuration file.
 *              The options are separated by white spaces.
 *      </tr>
 *  </table>
 *
 *  @author Franz-Josef Elmer
 */
public class EmulatorApplet extends Applet implements Runnable {
    private Thread runningThread;
    private Emulator emu = Emulator.getEmulator();
    private boolean emulatorReady = false;

    /** Initialize emulator based on the applet parameters. */
    public void init() {
        emu.reset();
        emu.setApplet(this);
        emu.setDocumentBase(getCodeBase());

        String options = getParameter("options");
        if (options != null) {
            StringTokenizer st = new StringTokenizer(options);
            String token;
            while (st.hasMoreTokens()) {
                token = st.nextToken();
                if (!emu.analyseOption(token)) {
                    add(new Label("Invalid option " + token));
                    return;
                }
            }
        }
        options = getParameter("additionalOptions");
        String[] args = new String[0];
        if (options != null) {
            StringTokenizer st = new StringTokenizer(options);
            args = new String[st.countTokens()];
            for (int i = 0; i < args.length; i++) {
                args[i] = st.nextToken();
            }
        }
        String configFile = getParameter("configFile");
        if (configFile == null) {
            add(new Label("Parameter 'configFile' missing"));
            return;
        }

        try {
            emu.configure(configFile, args);
            if (emu.isNoEmulation()) {
                add(new Label("Configuration finished but no emulation."));
            } else {
                add(new Label("Emulator is running."));
            }
            emulatorReady = true;
        } catch (ConfigurationException e) {
            add(new Label(e.getMessage()));
        }

        if (emu.isWithGUI()) {
            new EmulatorGUI().show();
        }
    }

    /**
     *  Start emulation thread if emulator is ready and neither
     *  "-n" nor the "-g" is set in the options.
     */
    public void start() {
        System.out.println("start");
        if (emulatorReady && !emu.isNoEmulation()) {
            runningThread = new Thread(this);
            runningThread.start();
        }
    }

    /** Stop emulation if running. */
    public void stop() {
        runningThread = null;
    }

//
//  Implementation of the Runnable interface
//

    public void run() {
        emu.runEmulation();
    }
}
