package jfrace;

import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.Locale;

/**
 *  Thread running a group of CPUs asynchronously from
 *  the other ones.
 *
 *  @author Franze-Josef Elmer
 */
public class EmulatorThread extends Thread {
    private CPU[] cpus;
    private long[] numberOfCycles;
    private double clockFrequency;
    private long yieldInterval;
    private long adaptationTime;
    private boolean printEffectiveClockFrequency;
    private DecimalFormat df;
    private boolean debuging;
    private int debugStart;
    private int debugEnd;
    private int debugStep;
    private double timeUnit;

    /**
     *  Create a new thread for the given group of CPUs.
     *  @param cpus CPUs to be run.
     *  @param yieldInterval Number of clock cycles between calls of
     *                       the method <code>Thread.yield()</code>.
     *  @param clockFrequency The clock frequency in MHz.
     *  @param adaptationTime Time interval in milliseconds between
     *                        resetings of the time origin for the
     *                        clock frequency algorithm.
     *  @param printEffectiveClockFrequency Print the effective clock
     *                                      frequency if <code>true</code>.
     *  @param debuging <code>true</code> in the case of debuging.
     *  @param debugStart Number of instruction when debuging starts.
     *  @param debugEnd Number of instruction when debuging ends.
     *                  0 if no end.
     *  @param debugStep Number of steps between two debug prints
     */
    public EmulatorThread(CPU[] cpus, long yieldInterval,
                          double clockFrequency, long adaptationTime,
                          boolean printEffectiveClockFrequency,
                          boolean debuging, int debugStart,
                          int debugEnd, int debugStep) {
        this.cpus = cpus;
        this.clockFrequency = clockFrequency;
        this.printEffectiveClockFrequency = printEffectiveClockFrequency;
        this.yieldInterval = yieldInterval;
        this.adaptationTime = adaptationTime;
        this.debuging = debuging;
        this.debugStart = debugStart;
        this.debugEnd = debugEnd;
        this.debugStep = debugStep;

        if (clockFrequency > 0) {
            timeUnit = 0.001 / clockFrequency;
        }
        numberOfCycles = new long[cpus.length];
        df = (DecimalFormat) NumberFormat.getInstance(Locale.US);
        df.applyPattern("0.000");
    }

    /** Run emulation. */
    public void run() {
        long time0 = System.currentTimeMillis();
        if (debuging) {
            runInDebugMode();
        } else if (cpus.length != 1) { // no debugging but more than one cpu
            runMultipleCPUs();
        } else { // no debugging and one cpu
            // this method has a better performance than runMultipleCPUs()
            runSingleCPU();
        }
        long time = System.currentTimeMillis() - time0;

        System.out.println(time + " msec.");
        for (int i = 0; i < numberOfCycles.length; i++) {
            System.out.println((i + 1) + ". CPU (" + cpus[i].getName() + "): "
                               + numberOfCycles[i]
                               + " cycles. Corresponding clock frequency = "
                               + df.format(numberOfCycles[i] / (1000.0 * time))
                               + " MHz");
        }
    }

    /** Run in debugging mode. */
    private void runInDebugMode() {
        boolean stopped = false;
        boolean[] cpuStopped = new boolean[cpus.length];
        int cycles = 0;
        long[] numberOfOperations = new long[cpus.length];
        long countCycles = 0;
        while (!stopped) {
            if ((countCycles++ % yieldInterval) == 0) {
                Thread.yield();
            }
            stopped = true;
            for (int i = 0; i < cpus.length; i++) {
                if (numberOfCycles[i] < countCycles) {
                    numberOfOperations[i]++;
                    cycles = cpus[i].next();
                    numberOfCycles[i] += cycles;
                    if (cycles < 0) {
                        cpuStopped[i] = true;
                    }
                    if (numberOfOperations[i] > debugStart
                            && (debugEnd == 0
                                || numberOfOperations[i] <= debugEnd)
                            && ((numberOfOperations[i] - debugStart - 1)
                                 % debugStep == 0)) {
                        System.out.println(cpus[i].toString());
                    }

                }
                stopped &= cpuStopped[i];
            }
        }
    }

    /** Run more than one single CPU in non-debugging mode. */
    private void runMultipleCPUs() {
        boolean stopped = false;
        boolean[] cpuStopped = new boolean[cpus.length];
        int cycles = 0;
        long nc = 0;
        long nc0 = 0;
        long time1 = System.currentTimeMillis();
        long time, deltaT;
        while (!stopped) {
            if ((nc++ % yieldInterval) == 0) {
                Thread.yield();
                time = System.currentTimeMillis() - time1;
                deltaT = (long) ((nc - nc0) * timeUnit) - time;
                if (deltaT > 0) {
                    try {
                        Thread.sleep(deltaT);
                    } catch (Exception e) {}
                }
                if (time > adaptationTime) {
                    if (printEffectiveClockFrequency) {
                        System.out.println(
                                "effective clock frequency (in MHz): "
                                + df.format((nc - nc0) / (1000.0 * time)));
                    }
                    time1 = System.currentTimeMillis();
                    nc0 = nc;
                }
            }
            stopped = true;
            for (int i = 0; i < cpus.length; i++) {
                if (numberOfCycles[i] < nc && !cpuStopped[i]) {
                    cycles = cpus[i].next();
                    numberOfCycles[i] += cycles;
                    if (cycles < 0) {
                        cpuStopped[i] = true;
                    }
                }
                stopped &= cpuStopped[i];
            }
        }
    }

    /** Run a single CPU in non-debugging mode. */
    private void runSingleCPU() {
        boolean stopped = false;
        int cycles = 0;
        CPU cpu = cpus[0];
        long count = 0;
        long nc = 0;
        long nc0 = 0;
        long ncNextYield = 0;
        int delay1 = 0;
        int delay = 0;
        int delayAv;
        int index = 0;
        long time1 = System.currentTimeMillis();
        long time, deltaT;
        while (cycles >= 0) {
//            if ((count++ % yieldInterval) == 0) {
            if (nc >= ncNextYield) {
                ncNextYield += yieldInterval;
                Thread.yield(); // allow other threads running
                // Delay to adapt to clock frequency
    //                    for (long l = 0; l < delay; l++);
                // Calculate new delay
                time = System.currentTimeMillis() - time1;
                deltaT = (long) ((nc - nc0) * timeUnit) - time;
                if (deltaT > 0) {
                    try {
                        Thread.sleep(deltaT);
                    } catch (Exception e) {}
                }
                /*
                delayAv = (delay + delay1) / 2;
                delay1 = delay;
                delay = delayAv + (int) (1000.0
                          * ((nc - nc0) * timeUnit - time));
                if (delay < 0) {
                    delay = 0;
                }
                */
                // Reset origin of time and cycle count
                if (time > adaptationTime) {
                    if (printEffectiveClockFrequency) {
                        System.out.println(
                                "effective clock frequency (in MHz): "
                                + df.format((nc - nc0) / (1000.0 * time)));
                    }
                    time1 = System.currentTimeMillis();
                    nc0 = nc;
                }
            }
            cycles = cpu.next();
            nc += cycles;
        }
        numberOfCycles[0] = nc;
    }
}
