package jfrace;

/**
 *  This is a helper class which holds information about a
 *  register of the {@link CPU}. These informations are
 *  <ul><li>number of bits
 *      <li>type
 *      <li>name
 *      <li>value
 *  </ul>
 *  The register type can be any combination of the 6 basic types
 *  <code>PROGRAM_COUNTER</code>,
 *  <code>DATA_REGISTER</code>,
 *  <code>ADDRESS_REGISTER</code>,
 *  <code>STACK_POINTER</code>,
 *  <code>ACCUMULATOR</code>, and
 *  <code>STATUS_REGISTER</code>.
 *  Note that the number of bits has to be less than or equal 32.
 *
 *  @author Franz-Josef Elmer
 */
public class Register {
    /** Register type mask. */
    public static final int PROGRAM_COUNTER = 1;
    /** Register type mask. */
    public static final int DATA_REGISTER = 2;
    /** Register type mask. */
    public static final int ADDRESS_REGISTER = 4;
    /** Register type mask. */
    public static final int STACK_POINTER = 8;
    /** Register type mask. */
    public static final int ACCUMULATOR = 16;
    /** Register type mask. */
    public static final int STATUS_REGISTER = 32;

    private static final String[] REG_NAMES
        = {"program counter", "data register", "address register",
           "stack pointer", "accumulator", "status register"};
    private int numberOfBits;
    private int type;
    private String name;
    private int value;

    /**
     *  Create a new instance.
     *  @param numberOfBits the number of bits.
     *  @param type the type of the register
     *              (e.g. <code>PROGRAM_COUNTER</code>).
     *  @param name the name of the register.
     *  @param value the value of the bits of the register. Only
     *               lowest <code>numberOfBits</code> are meaningful.
     */
    public Register(int numberOfBits, int type, String name, int value) {
        this.numberOfBits = numberOfBits;
        this.type = type;
        this.name = (name == null) ? "" : name;
        this.value = value;
    }

    /** Return the number of bits. */
    public int getNumberOfBits() {
        return numberOfBits;
    }

    /** Return the type of the register. */
    public int getType() {
        return type;
    }

    /**
     *  Return the name of the register. Return an empty string if
     *  the corresponding argument in the constructor was <code>null</code>.
     */
    public String getName() {
        return name;
    }

    /** Return the value of the register. */
    public int getValue() {
        return value;
    }

    public String toString() {
        StringBuffer sb = new StringBuffer(name);
        sb.append(": ");
        sb.append(numberOfBits);
        sb.append(" bit register, type");
        int mask = 1;
        boolean first = true;
        for (int i = 0; i < REG_NAMES.length; i++) {
            if ((type & mask) != 0) {
                sb.append(first ? '=' : '+');
                sb.append(REG_NAMES[i]);
            }
            mask <<= 1;
        }
        sb.append(", value=");
        Util.appendHex(sb, numberOfBits >> 2, value);

        return new String(sb);
    }
}
