package jfrace;

import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Vector;

/**
 *  Test statement. Helper class for {@link CPUTestSuite}.
 *  A test statement contains some registers or an address.
 *  These registers or address can be set to numerical values.
 *
 *  @author Franz-Josef Elmer.
 */
public class Statement {
    private String errorMessage;
    private boolean memory;
    private int[] regIndices;
    private int[] shifts;
    private int[] masks;
    private int address;
    private int[] numbers;
    private boolean doubleEqualSign;

    /**
     *  Create a statement in accordance of the given statement
     *  string.
     *  @param statement description string.
     *  @param regNames register names. Needed for obtaining the
     *                  register index from the name.
     */
    public Statement(String statement, Hashtable regNames) {
        int index = statement.indexOf('=');
        if (index < 0) {
            errorMessage = "'=' is missing in '" + statement + "'";
        } else {
            extractAddresses(statement.substring(0, index++), regNames);
            if (statement.charAt(index) == '=') {
                doubleEqualSign = true;
                index++;
            }
            extractNumbers(statement.substring(index));
        }
/*
        System.out.println("Statement: memory="+memory+" address="+address);
        if (regIndices != null) {
            for (int i = 0; i < regIndices.length; i++) {
                System.out.println(i+":"+regIndices[i]+" "+shifts[i]
                                   +" "+Integer.toHexString(masks[i]));
            }
        }
        if (numbers != null) {
            System.out.print("  numbers =");
            for (int i = 0; i < numbers.length; i++) {
                System.out.print(" "+numbers[i]);
            }
            System.out.println();
        }
*/
    }

    /** Return <code>true</code> if '==' appeared instead of '='. */
    public boolean isDoubleEqualSign() {
        return doubleEqualSign;
    }

    /**
     *  Set the state of the given CPU or memory in accordance
     *  to this statement.
     *  @param cpu CPU
     *  @param as address space of the memory
     *  @return <code>null</code> if no error. Otherwise an
     *          error message is returned.
     */
    public String setState(CPU cpu, AddressSpace as) {
        if (errorMessage != null) {
            return errorMessage;
        }

        if (memory) {
            int memAdr = address;
            if (regIndices != null) { // memory address defined by registers
                memAdr = 0;
                int value;
                for (int i = 0; i < regIndices.length; i++) {
                    value = (int) cpu.getRegister(regIndices[i]).getValue();
                    memAdr |= ((value & masks[i]) << shifts[i]);
                }
            }
            try {
                for (int i = 0; i < numbers.length; i++) {
                    as.setByte(memAdr++, 0, (byte) numbers[i]);
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        } else {
            int value;
            for (int i = 0; i < regIndices.length; i++) {
                value = (int) ((numbers[0] >> shifts[i]) & masks[i]);
                cpu.setRegisterValue(regIndices[i], value);
            }
        }
        return null;
    }

    /**
     *  Test the state of CPU and memory in accordance to this
     *  statement.
     *  @param sb an error message will be apend to this string buffer.
     *  @param cpu CPU
     *  @param as address space of the memory.
     */
    public void test(StringBuffer sb, CPU cpu, AddressSpace as) {
        if (errorMessage != null) {
            sb.append(' ');
            sb.append(errorMessage);
        }

        if (memory) {
            int memAdr = address;
            if (regIndices != null) { // memory address defined by registers
                memAdr = 0;
                int value;
                for (int i = 0; i < regIndices.length; i++) {
                    value = (int) cpu.getRegister(regIndices[i]).getValue();
                    memAdr |= ((value & masks[i]) << shifts[i]);
                }
            }
            try {
                byte b;
                for (int i = 0; i < numbers.length; i++) {
                    b = (byte) (numbers[i] ^ as.getByte(memAdr, 0));
                    as.setByte(memAdr++, 0, b); // if correct zero is stored
                    if (b != 0) {
                        sb.append(" (0x");
                        Util.appendHex(sb, 4, (long) (memAdr - 1));
                        sb.append(")=0x");
                        Util.appendHex(sb, as.getByte(memAdr - 1, 0));
                        sb.append("!=");
                        Util.appendHex(sb, (byte) numbers[i]);
                    }
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        } else {
            int v1, v2;
            Register r;
            for (int i = 0; i < regIndices.length; i++) {
                v1 = (numbers[0] >> shifts[i]) & masks[i];
                r = cpu.getRegister(regIndices[i]);
                v2 = r.getValue();
                if (v1 != v2) {
                    sb.append(' ');
                    sb.append(r.getName());
                    sb.append("=0x");
                    Util.appendHex(sb, r.getNumberOfBits() / 4, v2);
                    sb.append("!=0x");
                    Util.appendHex(sb, r.getNumberOfBits() / 4, v1);
                }
            }
        }
    }

    /**
     *  Mark tested registers. Set the corresponding bit of
     *  <code>testedRegisters</code> for those registers which
     *  will be tested by the <code>test</code> method in
     *  accordance of the statement description given in the constructor.
     *  @param testedRegisters the length of this array has to be equal
     *                         to the number of registers of the CPU.
     */
    public void markTestedRegisters(boolean[] testedRegisters) {
        if (!memory) {
            for (int i = 0; i < regIndices.length; i++) {
                testedRegisters[regIndices[i]] = true;
            }
        }
    }

    private void extractAddresses(String descriptor, Hashtable regNames) {
        descriptor = descriptor.trim();
        if (descriptor.length() > 0 && descriptor.charAt(0) == '(') {
            memory = true;
            int len = descriptor.length();
            if (descriptor.charAt(len - 1) != ')') {
                errorMessage = "')' missing in '" + descriptor + "'";
            } else {
                descriptor = descriptor.substring(1, len - 1).trim();
                if (Character.isDigit(descriptor.charAt(0))) {
                    address = Util.parseNumber(descriptor);
                } else {
                    extractRegisters(descriptor, regNames);
                }
            }
        } else {
            extractRegisters(descriptor, regNames);
        }
    }

    private void extractRegisters(String descriptor, Hashtable regNames) {
        int len = descriptor.length();
        if (len == 0) {
            errorMessage = "missing register name in '" + descriptor + "'";
        }
        int i0 = 0;
        int i1 = 0;
        int[] indexBitsMask;
        String regName;
        Vector regList = new Vector();
        while (i0 < len) {
            indexBitsMask = null;
            for (i1 = len; i1 > i0 && indexBitsMask == null; i1--) {
                regName = descriptor.substring(i0, i1);
                indexBitsMask = (int[]) regNames.get(regName);
            }
            if (indexBitsMask != null) {
                regList.addElement(indexBitsMask);
                i1++;
            } else {
                i1 = len;
            }
            i0 = i1;
        }
        regIndices = new int[regList.size()];
        shifts = new int[regIndices.length];
        masks = new int[regIndices.length];
        int shift = 0;
        for (int i = regIndices.length - 1; i >= 0; i--) {
            indexBitsMask = (int[]) regList.elementAt(i);
            regIndices[i] = indexBitsMask[0];
            shifts[i] = shift;
            masks[i] = indexBitsMask[2];
            shift += indexBitsMask[1];
        }
    }

    private void extractNumbers(String numberList) {
        StringTokenizer st = new StringTokenizer(numberList.trim(), ",");
        numbers = new int[st.countTokens()];
        for (int i = 0; i < numbers.length; i++) {
            numbers[i] = Util.parseNumber(st.nextToken().trim());
        }
    }
}

