package jfrace;

import jfrace.device.RAM;

import java.io.*;
import java.net.*;
import java.util.*;

/**
 *  Configure a computer system based on a configuration file.
 *  The configuration file is like a Java Properties file with
 *  key-value pairs (where '=' is the delimiter between the key
 *  and the value). A key-value pair has to be in one line.
 *  Empty lines and lines starting with '#' will be ignored.
 *  Contrary to properties file there can be several
 *  key-value pairs for the same key.
 *  <p>
 *  Numerical values can be denoted either decimal, binary
 *  (with a leading '0b'), octal (with a leading '0'), or
 *  hexadecimal (with a leading 0x).
 *  </p>
 *  <p>
 *  The following table explains all keys and the meaning of their
 *  values. Note, that <code>packagePrefix</code> and
 *  <code>defaultValue</code> have to be defined before the first
 *  <code>CPU, RAM</code>, or <code>device</code> definition.
 *  </p>
 *  <p>
 *  <table border="1" cellpadding="5">
 *  <tr><th>Key</th><th>Default Value</th>
 *      <th width="10%">Multiple Definitions?</th>
 *      <th>Meaning</th></tr>
 *  <tr><td valign="top"><code>packagePrefix</code></td><td valign="top"><i>mandatory</i></td>
 *      <td valign="top">no</td>
 *      <td valign="top">Package prefix for all dynamically loaded classes.
 *          The subpackages <code>cpu, device</code>, and
 *          <code>terminal</code> are derived from the main
 *          package.
 *      </td></tr>
 *  <tr><td valign="top"><code>defaultValue</code></td><td valign="top"><i>mandatory</i></td>
 *      <td valign="top">no</td>
 *      <td valign="top">Default return value of the {@link AddressSpace}
 *          when no address bank is defined for a given address
 *          in a <code>getValue()</code> or <code>getByte()</code>
 *          call.
 *      </td></tr>
 *  <tr><td valign="top"><code>yieldInterval</code></td><td valign="top"><code>1000</code></td>
 *      <td valign="top">no</td>
 *      <td valign="top">Number of clock cycles between two calls of
 *          <code>Thread.yield()</code>. After such a call the
 *          <code>Thread.sleep()</code> will be invoked if
 *          a delay is necessary to adapt the effective
 *          clock frequency to the frequency defined in the
 *          property <code>clockFrequency</code>.
 *      </td></tr>
 *  <tr><td valign="top"><code>clockFrequency</code></td><td valign="top"><i>the maximum</i></td>
 *      <td valign="top">no</td>
 *      <td valign="top">The frequency of the system clock (in MHz). After
 *          <code>yieldInterval</code> numbers of clock cycles a
 *          delay will be made to adapt the emulation speed to this
 *          frequency. If <code>clockFrequency</code> is not defined.
 *          no delay will be performed an the emualtion runs at the
 *          maximum speed possible.
 *      </td></tr>
 *  <tr><td valign="top"><code>adaptationTime</code></td><td valign="top"><code>2000</code></td>
 *      <td valign="top">no</td>
 *      <td valign="top">The time (in milliseconds) between two recalibration
 *          of the time origin. In addition the effective clock
 *          frequency will be printed onto the console if the
 *          command line option <code>-p</code> is set.
 *      </td></tr>
 *  <tr><td valign="top"><code>CPU</code></td><td valign="top">&nbsp;</td>
 *      <td valign="top">yes</td>
 *      <td valign="top">Definition of a CPU. The value has to be
 *          <p>
 *          <code><i>&lt;class name&gt;</i> [<i>&lt;offset&gt;</i>]
 *          [<i>&lt;method call 1&gt; &lt;method call 2&gt;</i> ...]</code>
 *          </p>
 *          For more details see {@link CPUConfigurator}.
 *      </td></tr>
 *  <tr><td valign="top"><code>RAM</code></td><td valign="top">&nbsp;</td>
 *      <td valign="top">yes</td>
 *      <td valign="top">Definition of a memory bank (RAM or ROM).
 *          The value has to be
 *          <p>
 *          <code><i>&lt;number of bits&gt; &lt;first address&gt;</i>
 *          {<i>&lt;size&gt;</i>|<i>&lt;raw data&gt;</i>}
 *          [<i>&lt;mode mask&gt;</i>] [readOnly]</code>
 *          </p>
 *          For more details see {@link jfrace.device.RAM#createRAMAndAddToAS
 *          RAM.createRAMAndAddToAS()}.
 *      </td></tr>
 *  <tr><td valign="top"><code>device</code></td><td valign="top">&nbsp;</td>
 *      <td valign="top">yes</td>
 *      <td valign="top">Definition of a device. The value has to be
 *          <p>
 *          <code><i>&lt;class name&gt; &lt;first address&gt;</i>
 *          [<i>&lt;mode mask&gt;</i>]</code>
 *          </p>
 *          or
 *          <p>
 *          <code><i>&lt;class name&gt; &lt;configuration resources&gt;</i>
 *          </code>
 *          </p>
 *          where
 *          <dl><dt><code><i>&lt;class name&gt;</i></code></dt>
 *              <dd>is the class name of the device (without
 *                  package name)</dd>
 *              <dt><code><i>&lt;first address&gt;</i></code></dt>
 *              <dd>is the first address in the {@link AddressSpace}.</dd>
 *              <dt><code><i>&lt;mode mask&gt;</i></code></dt>
 *              <dd>is the optional mode mask.</dd>
 *              <dt><code><i>&lt;configuration resources&gt;</i></code></dt>
 *              <dd>is a string which does not start with a digit.</dd>
 *          </dl>
 *          In the first form the device will be added to
 *          the address space without doing some configuration.
 *          In the second case the device will be created and
 *          configurated by a special {@link Configurator} for
 *          this device. The interpretation of the string
 *          <code><i>&lt;configuration resources&gt;</i></code>
 *          depends on the <code>Configurator</code>. In most
 *          cases it will be the name of a configuration file.
 *      </td></tr>
 *  </table>
 *  </p>
 *  <p>Example of a configuration file:
 *  <pre>
 *      packagePrefix = jfrace
 *      defaultValue = 0
 *      CPU = MC6502
 *      CPU = Intel8080 0x20000 setPortBiasAddress(0x30000)
 *      RAM = 8 0 0xc000
 *      RAM = 8 0xc000 ROM.dat 3
 *      RAM = 8 0x20000 65536
 *      device = GI_AY_3_8910 0x300b6
 *      device = CPMEmulator CPMConfig1.cfg
 *  </pre>
 *  </p>
 *
 *  @author Franz-Josef Elmer
 */
public class SystemConfigurator {
    private URL documentBase;
    private String packagePrefix;
    private String packageCPUPrefix;
    private String packageDevicePrefix;
    private AddressSpace as;
    private long yieldInterval = 1000;
    private double clockFrequency;
    private long adaptationTime = 2000;

    /**
     *  Create an instance for the given document base.
     *  The document base is need for reading configuration file
     *  with no absolute URL.
     *  @param theDocumentBase
     */
    public SystemConfigurator(URL theDocumentBase) {
        documentBase = theDocumentBase;
    }

    /**
     *  Return the address space.
     *  Call this method after {@link #configure}.
     */
    public AddressSpace getAddressSpace() {
        return as;
    }

    /**
     *  Return the clock frequency. Call this method after
     *  {@link #configure}.
     */
    public double getClockFrequency() {
        return clockFrequency;
    }

    /**
     *  Return the property <code>yieldInterval</code> read from
     *  the configuration data. Default value is 1000.
     *  Call this method after {@link #configure}.
     */
    public long getYieldInterval() {
        return yieldInterval;
    }

    /**
     *  Return the property <code>adaptationTime</code> read from
     *  the configuration data. Default value is 2000.
     *  Call this method after {@link #configure}.
     */
    public long getAdaptationTime() {
        return adaptationTime;
    }

    /**
     *  Configure a computer system based on
     *  the given configuration source.
     *  @param configSource Absolute or relative URL of the
     *                      configuration file.
     *  @param args Arguments needed by {@link Configurator
     *              Configurators} of {@link Device Devices}
     *              declared in the configuration file.
     *  @throws ConfigurationException in the case of an error
     *          (including syntax errors in the configuration file).
     */
    public void configure(String configSource, String[] args)
                            throws ConfigurationException {
        BufferedReader br = null;
        String line = null;
        int count = 0;
        try {
            br = new BufferedReader(new InputStreamReader(
                                        getAbsoluteURL(configSource)
                                            .openStream()));
            int index;
            while ((line = br.readLine()) != null) {
                count++;
                if (line.length() > 0 && line.charAt(0) != '#') {
                    index = line.indexOf('=');
                    if (index < 0) {
                        throw new ConfigurationException("'=' missing.");
                    } else {
                        process(line.substring(0, index).trim(),
                                line.substring(index + 1).trim(), args);
                    }
                }
            }
        } catch (ConfigurationException e) {
            throw new ConfigurationException(
                        "System configuration error in line\n" + count
                        + ": " + line + "\n" + e.getMessage());
        } catch (Exception e) {
            throw new ConfigurationException(
                        "System configuration error in line\n" + count
                        + ": " + line + "\n" + e);
        } finally {
            try {
                if (br != null) {
                    br.close();
                }
            } catch (IOException e) {
                throw new ConfigurationException(
                            "Couldn't close '" + configSource + "'.");
            }
        }
    }

    /**
     *  Process a line from the configuration file.
     *  @param type String before '='
     *  @param type String after '='
     *  @param args[] Additional command line arguments after the
     *                configuration file name.
     *  @throws ConfigurationException in the case of an error
     *          (including syntax errors in the configuration file).
     */
    private void process(String type, String description, String[] args)
                throws ConfigurationException {
        if (type.equals("packagePrefix")) {
            packagePrefix = description;
            packageCPUPrefix = description + ".cpu.";
            packageDevicePrefix = description + ".device.";
        } else if (type.equals("defaultValue")) {
            as = new AddressSpace(Util.parseNumber(description));
        } else if (type.equals("yieldInterval")) {
            try {
                yieldInterval = Long.parseLong(description);
            } catch (NumberFormatException e) {
                throw new ConfigurationException(
                        "Value of 'yieldInterval' is not a number: "
                        + description);
            }
        } else if (type.equals("clockFrequency")) {
            try {
                clockFrequency = new Double(description).doubleValue();
            } catch (NumberFormatException e) {
                throw new ConfigurationException(
                        "Value of 'clockFrequency' is not a number: "
                        + description);
            }
        } else if (type.equals("adaptationTime")) {
            try {
                adaptationTime = Long.parseLong(description);
            } catch (NumberFormatException e) {
                throw new ConfigurationException(
                        "Value of 'adaptationTime' is not a number: "
                        + description);
            }
        } else {
            if (packagePrefix == null) {
                throw new ConfigurationException(
                        "key 'packagePrefix' undefined or in a wrong order.");
            }
            if (as == null) {
                throw new ConfigurationException(
                        "key 'defaultValue' undefined or in a wrong order.");
            }
            if (type.equals("CPU")) {
                processCPU(description);
            } else if (type.equals("RAM")) {
                processRAM(description);
            } else if (type.equals("device")) {
                processDevice(description, args);
            } else {
                throw new ConfigurationException("unkown type >" + type + "<");
            }
        }
    }

    /**
     *  Process a CPU line from the configuration file.
     *  @param description Details after the '=' character.
     *  @return <code>true</code> if successfully processed.
     */
    private boolean processCPU(String description) {
        try {
            CPUConfigurator.configureCPU(as, description, packageCPUPrefix);
            return true;
        } catch (Exception e) {
            System.out.println(e);
            return false;
        }
    }

    /**
     *  Process a RAM line from the configuration file.
     *  @param description Details after the '=' character.
     *  @return <code>true</code> if successfully processed.
     */
    private boolean processRAM(String description) {
        try {
            RAM.createRAMAndAddToAS(as, description, documentBase);
            return true;
        } catch (Exception e) {
            System.out.println(e);
            return false;
        }
    }

    /**
     *  Process a device line from the configuration file.
     *  @param description Details after the '=' character.
     *  @param args[] Additional command line arguments after the
     *                configuration file name.
     *  @return <code>true</code> if successfully processed.
     */
    private boolean processDevice(String description, String[] args) {
        StringTokenizer st = new StringTokenizer(description);
        int numberOfTokens = st.countTokens();
        if (numberOfTokens < 2) {
            System.out.println("two items expected.");
            return false;
        }
        String className = packageDevicePrefix + st.nextToken();
        String secondToken = st.nextToken();
        if (Character.isDigit(secondToken.charAt(0))) {
            // simple case without Configurator
            int offset = Util.parseNumber(secondToken);
            int modeMask = 0;
            if (numberOfTokens > 2) {
                modeMask = Util.parseNumber(st.nextToken());
            }
            try {
                Device device = (Device) Class.forName(className).newInstance();
                if (device instanceof AddressSpaceListener) {
                    as.addAddressSpaceListener((AddressSpaceListener) device);
                }
                as.addDevice(offset, modeMask, device);
                return true;
            } catch (Exception e) {
                e.printStackTrace(System.out);
                return false;
            }
        } else {
            // case where configuration is delegated
            StringBuffer configResources = new StringBuffer(secondToken);
            while (st.hasMoreTokens()) {
                configResources.append(' ');
                configResources.append(st.nextToken());
            }
            className += "Configurator";
            try {
                Configurator configurator
                    = (Configurator) Class.forName(className).newInstance();
                return configurator.configurate(as,
                                                new String(configResources),
                                                packagePrefix,
                                                args);
            } catch (Exception e) {
                e.printStackTrace(System.out);
                return false;
            }
        }
    }



    private URL getAbsoluteURL(String relativeOrAbsoluteURL) {
        try {
            if (relativeOrAbsoluteURL.startsWith("http://")) {
                return new URL(relativeOrAbsoluteURL);
            } else {
                return new URL(documentBase, relativeOrAbsoluteURL);
            }
        } catch (Exception e) {
            return null;
        }
    }

}
