package jfrace;

import java.awt.Image;
import java.awt.Toolkit;
import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.Reader;
import java.net.URL;
import java.net.URLConnection;

/**
 *  Collection of utility functions.
 *
 *  @author Franz-Josef Elmer
 */
public class Util {
    /**
     *  Parse a string as a number. The number is given either as a
     *  binary one if it starts with '0b', or a
     *  decimal one if it starts with a nonzero digit, or an octal
     *  one if it starts with zero, or a hexdecimal one if it starts
     *  with '0x'. Return zero if the argument does not start with
     *  a digit (including '-' and white spaces).
     */
    public static int parseNumber(String numberString) {
        int result = 0;
        if (numberString != null && numberString.length() > 0) {
            int len = numberString.length();
            numberString = numberString.toLowerCase();
            int base = 10;
            int index = 0;
            if (numberString.startsWith("0b")) {
                base = 2;
                index = 2;
            } else if (numberString.startsWith("0x")) {
                base = 16;
                index = 2;
            } else if (numberString.charAt(0) == '0') {
                base = 8;
                index = 1;
            }
            char c;
            int digit;
            while (index < len) {
                c = numberString.charAt(index++);
                digit = Character.digit(c, base);
                if (digit < 0) {
                    break;
                }
                result = result * base + digit;
            }
        }
        return result;
    }

    /**
     *  Append a two-digit hexadecimal number (upper case) to the given string buffer.
     *  @param sb string buffer.
     *  @param number 8-bit number.
     */
    public static void appendHex(StringBuffer sb, byte number) {
        int index = sb.length();
        char digit;
        for (int i = 0; i < 2; i++) {
            digit = Character.forDigit((int) number & 0xf, 16);
            sb.insert(index, Character.toUpperCase(digit));
            number >>= 4;
        }
    }

    /**
     *  Append an eight-digit hexadecimal number (upper case)
     *  to the given string buffer.
     *  @param sb string buffer.
     *  @param number 32-bit number.
     */
    public static void appendHex(StringBuffer sb, int number) {
        int index = sb.length();
        char digit;
        for (int i = 0; i < 8; i++) {
            digit = Character.forDigit((int) number & 0xf, 16);
            sb.insert(index, Character.toUpperCase(digit));
            number >>= 4;
        }
    }

    /**
     *  Append a n-digit hexadecimal number (upper case) to the given string buffer.
     *  @param sb string buffer.
     *  @param n number of digits.
     *  @param number number.
     */
    public static void appendHex(StringBuffer sb, int n, long number) {
        int index = sb.length();
        char digit;
        for (int i = 0; i < n; i++) {
            digit = Character.forDigit((int) number & 0xf, 16);
            sb.insert(index, Character.toUpperCase(digit));
            number >>= 4;
        }
    }

    /**
     *  Read some binary data from an URL.
     *  @param url URL of the source of the data.
     *  @return data in a byte array.
     *  @Exception if anything went wrong.
     */
    public static byte[] readBinaryData(URL url) throws Exception {
        URLConnection connection = url.openConnection();
        connection.connect();
        BufferedInputStream bis;
        bis = new BufferedInputStream(connection.getInputStream(), 2048);
        long length = connection.getContentLength();
        if (length < 0) { // sometimes the wrong number is returned
            length = bis.available(); // try this method
        }

        // Read in the data
        if (length == -1) {
            bis.close();
            throw new IOException("Data length = -1");
        } else {
            byte[] data = new byte[(int) length];
            int soFar = 0;
            int remaining;
            int incoming;
            while (soFar < length) {
                remaining = (int) length - soFar;
                incoming = bis.read(data, soFar, Math.min(remaining, 2048));
                soFar += incoming;
            }
            bis.close();
            return data;
        }
    }

    /**
     *  Read a binary file. If file type is <code>hex</code>
     *  Intel-Hex-Format is assumed.
     */
    public static byte[] readBinaryData(String fileName) {
        try {
            if (fileName.endsWith(".hex") || fileName.endsWith(".HEX")) {
                // Intel-Hex-Format
                Reader reader = new FileReader(fileName);
                BufferedReader br = new BufferedReader(reader);
                String line;
                StringBuffer sb = new StringBuffer();
                while ((line = br.readLine()) != null) {
                    sb.append(line);
                }
                br.close();
                return generateBinaryData(new String(sb));
            } else { // Binary
                FileInputStream fis = new FileInputStream(fileName);
                byte[] data = new byte[fis.available()];
                fis.read(data);
                fis.close();
                return data;
            }
        } catch (IOException e) {
            System.out.println(e);
            return null;
        }
    }

    /**
     *  Generate binary data from Intel Hex Format.
     *  Assumes correct Intel Hex Format and does not check checksums.
     *  @param intelHexData data in Intel Hex Format
     *  @return binary data.
     */
    public static byte[] generateBinaryData(String intelHexData) {
        // Determine the start address
        int index = intelHexData.lastIndexOf(':');
        String hexWord = intelHexData.substring(index + 3, index + 7);
        int startAddress = 0;

        // Determine highest address
        index = intelHexData.lastIndexOf(':', index - 1);
        String hexByte = intelHexData.substring(index + 1, index + 3);
        hexWord = intelHexData.substring(index + 3, index + 7);
        int endAddress = Integer.parseInt(hexWord, 16)
                         + Integer.parseInt(hexByte, 16);

        // extract data
        byte[] data = new byte[endAddress - startAddress];
        int address;
        int count = -1;
        index = 0;
        while (count != 0) {
            index = intelHexData.indexOf(':', index) + 1;
            hexByte = intelHexData.substring(index, index + 2);
            count = Integer.parseInt(hexByte, 16);
            hexWord = intelHexData.substring(index + 2, index + 6);
            address = Integer.parseInt(hexWord, 16) - startAddress;
            index += 6;
            for (int i = 0; i < count; i++) {
                index += 2;
                hexByte = intelHexData.substring(index, index + 2);
                data[address++] = (byte) Integer.parseInt(hexByte, 16);
            }
        }
        return data;
    }

    /**
     *  Load an image. Use the <code>Emulator</code> singleton
     *  in order to decide whether this method is invoked inside
     *  an application or an applet.
     *  @param imgName file name of the image. Path is relative
     *                 to <code>codeBase</code>.
     *  @return <code>null</code> if image not found.
     */
    public static Image loadImage(String imgName) {
        Emulator emu = Emulator.getEmulator();
        URL url = emu.getAbsoluteURL(imgName);
        try {
            return emu.insideAnApplet() ?
                            emu.getApplet().getImage(url) :
                            Toolkit.getDefaultToolkit().getImage(url);
        } catch (Exception e) {
            System.out.println("Can not load image " + url);
            return null;
        }
    }
}
