package jfrace.cpu;

import jfrace.AddressSpace;
import jfrace.CPU;
import jfrace.Register;
import jfrace.Util;

/**
 *  A simple 5-bit CPU. It has the following registers:
 *  <table border=0 cellpadding=0>
 *  <tr><td valign=top><code><b>A</b></code>&nbsp;</td>
 *      <td>5-bit accumulator</td>
 *  </tr>
 *  <tr><td valign=top><code><b>I</b></code>&nbsp;</td>
 *      <td>5-bit index register. It defines the higher order bits
 *          of a 10-bit address in instructions with a bias.
 *      </td>
 *  </tr>
 *  <tr><td valign=top><code><b>SP</b></code>&nbsp;</td>
 *      <td>10-bit stack pointer</td>
 *  </tr>
 *  <tr><td valign=top><code><b>PC</b></code>&nbsp;</td>
 *      <td>10-bit program counter</td>
 *  </tr>
 *  </table>
 *  <p>
 *  Here is the list of operations:
 *  <table border=1 cellpadding=5>
 *     <tr><th>Assembler-Operation</th><th>Machine-code</th>
 *         <th>Meaning</th>
 *     </tr>
 *     <tr><td valign=top><code>NOP</code></td><td valign=top><code>00000</code></td>
 *         <td valign=top>No operation.<br><code>PC &lt;- PC + 1</code></td>
 *     </tr>
 *     <tr><td valign=top><code>XAI</code></td><td valign=top><code>00001</code></td>
 *         <td valign=top>Exchange accumulator with index register.<br>
 *             <code>A &lt;-> I</code><br>
 *             <code>PC &lt;- PC + 1</code></td>
 *     </tr>
 *     <tr><td valign=top><code>LDA</code> <i>value</i></td>
 *         <td valign=top><code>00010</code> <i>value</i></td>
 *         <td valign=top>Load accumulator with <i>value</i><br>
 *             <code>A &lt;- </code><i>value</i><br>
 *             <code>PC &lt;- PC + 2</code></td>
 *     </tr>
 *     <tr><td valign=top><code>LDI</code> <i>value</i></td>
 *         <td valign=top><code>00011</code> <i>value</i></td>
 *         <td valign=top>Load index register with <i>value</i><br>
 *             <code>I &lt;- </code><i>value</i><br>
 *             <code>PC &lt;- PC + 2</code></td>
 *     </tr>
 *     <tr><td valign=top><code>LDAI</code> <i>bias</i></td>
 *         <td valign=top><code>00100</code> <i>bias</i></td>
 *         <td valign=top>Load accumulator indirectly from the location build by
 *             the index register and the <i>bias</i>.
 *             <br>
 *             <code>A &lt;- (32*I+</code><i>bias</i><code>)</code><br>
 *             <code>PC &lt;- PC + 2</code></td>
 *     </tr>
 *     <tr><td valign=top><code>LDAD</code> <i>bias</i></td>
 *         <td valign=top><code>00101</code> <i>bias</i></td>
 *         <td valign=top>Load accumulator double indirectly.
 *             <br>
 *             <code>A &lt;- (32*(32*I+</code><i>bias</i><code>)+(32</code>*I+</code><i>bias</i><code>+1))</code><br>
 *             <code>PC &lt;- PC + 2</code></td>
 *     </tr>
 *     <tr><td valign=top><code>STAI</code> <i>bias</i></td>
 *         <td valign=top><code>00110</code> <i>bias</i></td>
 *         <td valign=top>Store accumulator indirectly to a location build by
 *             the index register and the <i>bias</i>.
 *             <br>
 *             <code>(32*I+</code><i>bias</i><code>) &lt;- A</code><br>
 *             <code>PC &lt;- PC + 2</code></td>
 *     </tr>
 *     <tr><td valign=top><code>STAD</code> <i>bias</i></td>
 *         <td valign=top><code>00111</code> <i>bias</i></td>
 *         <td valign=top>Store accumulator double indirectly.
 *             <br>
 *             <code>(32*(32*I+</code><i>bias</i><code>)+(32</code>*I+</code><i>bias</i><code>+1)) &lt;- A</code><br>
 *             <code>PC &lt;- PC + 2</code></td>
 *     </tr>
 *     <tr><td valign=top><code>PUSH</code></td><td valign=top><code>01000</code></td>
 *         <td valign=top>Push accumulator and index register on stack<br>
 *             <code>(SP-1) &lt;- A<br>(SP-2) &lt;- I<br>SP &lt;- SP - 2</code><br>
 *             <code>PC &lt;- PC + 1</code></td>
 *     </tr>
 *     <tr><td valign=top><code>POP</code></td><td valign=top><code>01001</code></td>
 *         <td valign=top>Pop accumulator and index register from stack<br>
 *             <code>I &lt;- (SP)<br>A &lt;- (SP+1)<br>SP &lt;- SP + 2</code><br>
 *             <code>PC &lt;- PC + 1</code></td>
 *     </tr>
 *     <tr><td valign=top><code>CALL</code> <i>address</i></td>
 *         <td valign=top><code>01010</code> <i>address-high</i> <i>address-low</i></td>
 *         <td valign=top>Call a subroutine. Push the program counter on stack.<br>
 *             <code>(SP-1) &lt;- (PC + 3)</code><sub>low</sub><br>
 *             <code>(SP-2) &lt;- (PC + 3)</code><sub>high</sub><br>
 *             <code>SP &lt;- SP - 2<br>
 *             PC &lt;- 32 * </code><i>address-high</i> <code>+</code> <i>address-low</i></td>
 *     </tr>
 *     <tr><td valign=top><code>RET</code></td><td valign=top><code>01011</code></td>
 *         <td valign=top>Return from subroutine. Pop the program counter from the stack<br>
 *             <code>PC &lt;- 32 * (SP) + (SP+1)<br>
 *             SP &lt;- SP + 2</code></td>
 *     </tr>
 *     <tr><td valign=top><code>LDS</code> <i>value</i></td>
 *         <td valign=top><code>01100</code> <i>value-high</i> <i>value-low</i></td>
 *         <td valign=top>Load stack pointer<br>
 *             <code>SP &lt;- 32 * </code><i>value-high</i> <code>+</code> <i>value-low</i><br>
 *             <code>PC &lt;- PC + 3</code></td>
 *     <tr><td valign=top><code>JP</code> <i>address</i></td>
 *         <td valign=top><code>01101</code> <i>address-high</i> <i>address-low</i></td>
 *         <td valign=top>Jump to <i>address</i><br>
 *             <code>PC &lt;- 32 * </code><i>address-high</i> <code>+</code> <i>address-low</i><br></td>
 *     </tr>
 *     <tr><td valign=top><code>JPZ</code> <i>address</i></td>
 *         <td valign=top><code>01110</code> <i>address-high</i> <i>address-low</i></td>
 *         <td valign=top>Jump to <i>address</i> if accumulator equals zero<br>
 *             if <code>A == 0: PC &lt;- 32 * </code><i>address-high</i> <code>+</code> <i>address-low</i><br>
 *             if <code>A != 0: PC &lt;- PC + 3</code></td>
 *     </tr>
 *     <tr><td valign=top><code>JPNZ</code> <i>address</i></td>
 *         <td valign=top><code>01111</code> <i>address-high</i> <i>address-low</i></td>
 *         <td valign=top>Jump to <i>address</i> if accumulator not equals zero<br>
 *             if <code>A != 0: PC &lt;- 32 * </code><i>address-high</i> <code>+</code> <i>address-low</i><br>
 *             if <code>A == 0: PC &lt;- PC + 3</code></td>
 *     </tr>
 *     <tr><td valign=top><code>ADD</code> <i>bias</i></td>
 *         <td valign=top><code>10000</code> <i>bias</i></td>
 *         <td valign=top>Add to accumulator a value from the location build by
 *             the index register and the <i>bias</i>.
 *             <br>
 *             <code>A &lt;- A + (32*I+</code><i>bias</i><code>)</code><br>
 *             <code>PC &lt;- PC + 2</code></td>
 *     </tr>
 *     <tr><td valign=top><code>SUB</code> <i>bias</i></td>
 *         <td valign=top><code>10010</code> <i>bias</i></td>
 *         <td valign=top>Subtract from accumulator a value from the location build by
 *             the index register and the <i>bias</i>.
 *             <br>
 *             <code>A &lt;- A - (32*I+</code><i>bias</i><code>)</code><br>
 *             <code>PC &lt;- PC + 2</code></td>
 *     </tr>
 *     <tr><td valign=top><code>AND</code> <i>bias</i></td>
 *         <td valign=top><code>10100</code> <i>bias</i></td>
 *         <td valign=top>Logically AND to accumulator a value from the location build by
 *             the index register and the <i>bias</i>.
 *             <br>
 *             <code>A &lt;- A AND (32*I+</code><i>bias</i><code>)</code><br>
 *             <code>PC &lt;- PC + 2</code></td>
 *     </tr>
 *     <tr><td valign=top><code>OR</code> <i>bias</i></td>
 *         <td valign=top><code>10110</code> <i>bias</i></td>
 *         <td valign=top>Logically OR to accumulator a value from the location build by
 *             the index register and the <i>bias</i>.
 *             <br>
 *             <code>A &lt;- A OR (32*I+</code><i>bias</i><code>)</code><br>
 *             <code>PC &lt;- PC + 2</code></td>
 *     </tr>
 *     <tr><td valign=top><code>XOR</code> <i>bias</i></td>
 *         <td valign=top><code>11000</code> <i>bias</i></td>
 *         <td valign=top>Logicall XOR to accumulator a value from the location build by
 *             the index register and the <i>bias</i>.
 *             <br>
 *             <code>A &lt;- A XOR (32*I+</code><i>bias</i><code>)</code><br>
 *             <code>PC &lt;- PC + 2</code></td>
 *     </tr>
 *     <tr><td valign=top><code>SARA</code></td><td valign=top><code>11010</code></td>
 *         <td valign=top>Shift arithmetically right accumulator. The uppermost bit
 *             will be shifted from the left into the accumulator.
 *             Corresponds to a division by two of a signed number.<br>
 *             <code>A &lt;- A >> 1</code></td>
 *     </tr>
 *     <tr><td valign=top><code>SLRA</code></td><td valign=top><code>11100</code></td>
 *         <td valign=top>Shift logically right accumulator. A zero bit
 *             will be shifted from the left into the accumulator.<br>
 *             Corresponds to a division by two of an unsigned number.<br>
 *             <code>A &lt;- A >>> 1</code></td>
 *     </tr>
 *     <tr><td valign=top><code>SLA</code></td><td valign=top><code>11110</code></td>
 *         <td valign=top>Shift left accumulator. A zero bit
 *             will be shifted from the right into the accumulator.
 *             Corresponds to a multiplication with 2.<br>
 *             <code>A &lt;- A &lt;&lt; 1</code></td>
 *     </tr>
 *     <tr><td valign=top><code>HALT</code></td><td valign=top><code>11111</code></td>
 *         <td valign=top>Stops exceution. Program counter will not be incremented.</td>
 *     </tr>
 *  </table>
 *
 *  @author Franz-Josef Elmer
 */
public class FJE5 implements CPU {
    /** Activity mask: Program read */
    public static final int READ_PRG = 0x10000;
    /** Activity mask: Data read */
    public static final int READ_DATA = 0x20000;
    /** Activity mask: Data write */
    public static final int WRITE_DATA = 0x40000;

    /** Index of register <code>A</code> (accumulator). */
    public static final int A = 0;
    /** Index of register <code>I</code> (index register). */
    public static final int I = 1;
    /** Index of register <code>SP</code> (stack pointer). */
    public static final int SP = 2;
    /** Index of register <code>PC</code> (pogram counter). */
    public static final int PC = 3;

    private static final int NOP = 0;
    private static final int XAI = 1;
    private static final int LDA = 2;
    private static final int LDI = 3;
    private static final int LDAI = 4;
    private static final int LDAD = 5;
    private static final int STAI = 6;
    private static final int STAD = 7;
    private static final int PUSH = 010;
    private static final int POP = 011;
    private static final int CALL = 012;
    private static final int RET = 013;
    private static final int LDS = 014;
    private static final int JP = 015;
    private static final int JPZ = 016;
    private static final int JPNZ = 017;
    private static final int ADD = 020;
    private static final int SUB = 022;
    private static final int AND = 024;
    private static final int OR = 026;
    private static final int XOR = 030;
    private static final int SARA = 032;
    private static final int SLRA = 034;
    private static final int SLA = 036;
    private static final int HALT = 037;

    private int pc;
    private int sp;
    private int ireg;
    private int accu;
    private AddressSpace as;
    private String errorDescription = "ERROR";

    private int[] activityAddresses = new int[5];
    private int activityIndex;

    /**
     *  Return the activity address for a given index.
     *  An activity address is an address together with kind
     *  of reading or writing activity (i.e. <code>READ_PRG</code>,
     *  <code>READ_DATA</code>, or <code>WRITE_DATA</code>).
     *  There are up to five activity address stored during the
     *  execution of the <code>next()</code> method.
     */
    public int getActivityAddress(int index) {
        return activityAddresses[index];
    }

    private void setValue(int address, int mode, int value) {
        activityAddresses[activityIndex++] = address + mode;
        as.setValue(address, 0, value);
    }

    private int getValue(int address, int mode) {
        activityAddresses[activityIndex++] = address + mode;
        return as.getValue(address, 0);
    }

    /** Return the state of the CPU as a well-formed string. */
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("I=");
        appendBinary(sb, 5, ireg >> 5);
        sb.append(" A=");
        appendBinary(sb, 5, accu);
        sb.append(" SP=");
        appendBinary(sb, 10, sp);
        sb.append("=0x");
        Util.appendHex(sb, 3, (long) sp);
        sb.append(" PC=");
        appendBinary(sb, 10, pc);
        sb.append("=0x");
        Util.appendHex(sb, 3, (long) pc);
        sb.append(' ');
        disassemble(sb, pc);
        return new String(sb);
    }

    private void appendBinary(StringBuffer sb, int numberOfBits, int value) {
        int mask = 1 << numberOfBits;
        for (int i = 0; i < numberOfBits; i++) {
            value <<= 1;
            sb.append((value & mask) == 0 ? '0' : '1');
        }
    }

    private int disassembleOneUnit(StringBuffer sb, String opCode,
                                   int address) {
        sb.append(opCode);
        Util.appendHex(sb, (byte) as.getValue(address, 0));
        return 2;
    }

    private int disassembleTwoUnits(StringBuffer sb, String opCode,
                                    int address) {
        sb.append(opCode);
        Util.appendHex(sb, 3, (long) ((as.getValue(address, 0) << 5)
                       + as.getValue(++address, 0)));
        return 3;
    }

//
//  Implementation of the CPU interface
//

    /** Return "FJE5". */
    public String getName() {
        return "FJE5";
    }

    /** Return 5. */
    public int getNumberOfBits() {
        return 5;
    }

    /** Return 4. */
    public int getNumberOfRegisters() {
        return 4;
    }

    public Register getRegister(int index) {
        Register register = null;
        switch (index) {
        case A:
            register = new Register(5, Register.ACCUMULATOR, "A",
                                    accu & 0x1f);
            break;
        case I:
            register = new Register(5, Register.ADDRESS_REGISTER, "I",
                                    (ireg >> 5) & 0x1f);
            break;
        case SP:
            register = new Register(10, Register.STACK_POINTER, "SP",
                                    sp & 0x3ff);
            break;
        case PC:
            register = new Register(10, Register.PROGRAM_COUNTER, "PC",
                                    pc & 0x3ff);
            break;
        }
        return register;
    }

    public Register getProgramCounter() {
        return new Register(10, Register.PROGRAM_COUNTER, "PC", pc & 0x3ff);
    }

    public void setRegisterValue(int index, int value) {
        switch (index) {
        case A: accu = value & 0x1f; break;
        case I: ireg = (value & 0x1f) << 5; break;
        case SP: sp = value & 0x3ff; break;
        case PC: pc = value & 0x3ff; break;
        }
    }

    public void setAddressSpace(AddressSpace addressSpace) {
        as = addressSpace;
    }

    public AddressSpace getAddressSpace() {
        return as;
    }

    public void reset() {
        accu = 0;
        ireg = 0;
        sp = 0;
        pc = 0;
    }

    /** Not implemented. */
    public void interrupt(long interrupt) {}

    public String getErrorDescription() {
        return errorDescription;
    }

    public int next() {
        activityIndex = 0;
        int adr;
        int numberOfCycles = 1;
        int opcode = getValue(pc++, READ_PRG) & 0x1f;
        switch (opcode) {
        case NOP:
            break;
        case XAI:
            int dummy = (ireg >> 5) & 0x1f;
            ireg = accu << 5;
            accu = dummy;
            break;
        case LDA:
            accu = getValue(pc++, READ_PRG) & 0x1f;
            numberOfCycles = 2;
            break;
        case LDI:
            ireg = getValue(pc++, READ_PRG) << 5;
            numberOfCycles = 2;
            break;
        case LDAI:
            accu = getValue(ireg + getValue(pc++, READ_PRG), READ_DATA);
            numberOfCycles = 3;
            break;
        case LDAD:
            adr = ireg + getValue(pc++, READ_PRG);
            adr = (getValue(adr, READ_DATA) << 5) + getValue(adr + 1, READ_DATA);
            accu = getValue(adr, READ_DATA);
            numberOfCycles = 5;
            break;
        case STAI:
            setValue(ireg + getValue(pc++, READ_PRG), WRITE_DATA, accu);
            numberOfCycles = 3;
            break;
        case STAD:
            adr = ireg + getValue(pc++, READ_PRG);
            adr = (getValue(adr, READ_DATA) << 5) + getValue(adr + 1, READ_DATA);
            setValue(adr, WRITE_DATA, accu);
            numberOfCycles = 5;
            break;
        case PUSH:
            setValue(--sp, WRITE_DATA, accu);
            setValue(--sp, WRITE_DATA, ireg >> 5);
            numberOfCycles = 3;
            break;
        case POP:
            ireg = getValue(sp++, READ_DATA) << 5;
            accu = getValue(sp++, READ_DATA) ;
            numberOfCycles = 3;
            break;
        case CALL:
            adr = (getValue(pc++, READ_PRG) << 5)
                  + getValue(pc++, READ_PRG);
            setValue(--sp, WRITE_DATA, pc & 0x1f);
            setValue(--sp, WRITE_DATA, (pc >> 5) & 0x1f);
            pc = adr;
            numberOfCycles = 5;
            break;
        case RET:
            pc = (getValue(sp++, READ_DATA) << 5) + getValue(sp++, READ_DATA);
            numberOfCycles = 3;
            break;
        case LDS:
            sp = (getValue(pc++, READ_PRG) << 5) + getValue(pc++, READ_PRG);
            numberOfCycles = 3;
            break;
        case JP:
            pc = (getValue(pc++, READ_PRG) << 5) + getValue(pc, READ_PRG);
            numberOfCycles = 3;
            break;
        case JPZ:
            if (accu == 0) {
                pc = (getValue(pc++, READ_PRG) << 5) + getValue(pc, READ_PRG);
                numberOfCycles = 3;
            } else {
                pc += 2;
            }
            break;
        case JPNZ:
            if (accu != 0) {
                pc = (getValue(pc++, READ_PRG) << 5) + getValue(pc, READ_PRG);
                numberOfCycles = 3;
            } else {
                pc += 2;
            }
            break;
        case ADD:
            accu = (accu + getValue(ireg + getValue(pc++, READ_PRG),
                                    READ_DATA))
                   & 0x1f;
            numberOfCycles = 3;
            break;
        case SUB:
            accu = (accu - getValue(ireg + getValue(pc++, READ_PRG),
                                    READ_DATA))
                   & 0x1f;
            numberOfCycles = 3;
            break;
        case AND:
            accu = (accu & getValue(ireg + getValue(pc++, READ_PRG),
                                    READ_DATA))
                   & 0x1f;
            numberOfCycles = 3;
            break;
        case OR:
            accu = (accu | getValue(ireg + getValue(pc++, READ_PRG),
                                    READ_DATA))
                   & 0x1f;
            numberOfCycles = 3;
            break;
        case XOR:
            accu = (accu ^ getValue(ireg + getValue(pc++, READ_PRG),
                                    READ_DATA))
                   & 0x1f;
            numberOfCycles = 3;
            break;
        case SARA:
            accu = (((accu & 0x10) << 1) + accu) >> 1;
            break;
        case SLRA:
            accu >>= 1;
            break;
        case SLA:
            accu = (accu << 1) & 0x1f;
            break;
        case HALT:
            pc--;
            break;
        default:
            errorDescription = "Undefined opcode: "
                               + Integer.toHexString(opcode);
            numberOfCycles = -1;
        }
        pc &= 0x3ff;
        return numberOfCycles;
    }

    /** Disassembling is supported. */
    public boolean supportsDisassembling() {
        return true;
    }

    /**
     */
    public int disassemble(StringBuffer sb, int address) {
        if (as != null) {
            int opcode = as.getValue(address++, 0);
            switch (opcode) {
            case NOP:  sb.append("NOP"); return 1;
            case XAI:  sb.append("XAI"); return 1;
            case LDA:  return disassembleOneUnit(sb, "LDA  ", address);
            case LDI:  return disassembleOneUnit(sb, "LDI  ", address);
            case LDAI: return disassembleOneUnit(sb, "LDAI ", address);
            case STAI: return disassembleOneUnit(sb, "STAI ", address);
            case LDAD: return disassembleOneUnit(sb, "LDAD ", address);
            case STAD: return disassembleOneUnit(sb, "STAD ", address);
            case PUSH: sb.append("PUSH"); return 1;
            case POP:  sb.append("POP"); return 1;
            case CALL: return disassembleTwoUnits(sb, "CALL ", address);
            case RET:  sb.append("RET"); return 1;
            case LDS:  return disassembleTwoUnits(sb, "LDS  ", address);
            case JP:   return disassembleTwoUnits(sb, "JP   ", address);
            case JPZ:  return disassembleTwoUnits(sb, "JPZ  ", address);
            case JPNZ: return disassembleTwoUnits(sb, "JPNZ ", address);
            case ADD:  return disassembleOneUnit(sb, "ADD  ", address);
            case SUB:  return disassembleOneUnit(sb, "SUB  ", address);
            case AND:  return disassembleOneUnit(sb, "AND  ", address);
            case OR:   return disassembleOneUnit(sb, "OR   ", address);
            case XOR:  return disassembleOneUnit(sb, "XOR  ", address);
            case SARA: sb.append("SARA"); return 1;
            case SLRA: sb.append("SLRA"); return 1;
            case SLA:  sb.append("SLA"); return 1;
            case HALT: sb.append("HALT"); return 1;
            default:
                sb.append("? ");
                Util.appendHex(sb, (byte) opcode);
                return 1;
            }
        }
        return 0;
    }

}
