package jfrace.cpu;

import jfrace.*;
import jfrace.device.RAM;

/**
 *  Emulation of the 8-bit microprocessor Intel 8080.
 *  The 256 port addresses are mapped as a block to the
 *  <a href="../AddressSpace.html"><code>AddressSpace</code></a>.
 *  The first address (i.e. the address of port 0)
 *  is defined by the method
 *  <a href="#setPortBiasAddress(int)"><code>
 *  setPortBiasAddress</code></a>.
 *  <p>
 *  In order to increase the performance of memory reading and
 *  writing the 65536 bytes of memory will not be accessed via
 *  the <code>AddressSpace</code> but directly on the
 *  <code>byte</code> array. This memory is obtained from
 *  the first <a href="../device/RAM.html">RAM</code></a> added
 *  to the <code>AddressSpace</code>.
 *  <p>
 *  Note: Currently interrupts are not supported. The method
 *  <code>next</code> returns always 1 and not the correct number
 *  of clock cycles.
 *
 *  @author Franz-Josef Elmer
 */
public class Intel8080 implements AddressSpaceListener, CPU {
    /** Index of register <code>A</code> (accumulator). */
    public static final int A = 7;
    /** Index of register <code>F</code> (flags). */
    public static final int F = 6;
    /** Index of register <code>B</code>. */
    public static final int B = 0;
    /** Index of register <code>C</code>. */
    public static final int C = 1;
    /** Index of register <code>D</code>. */
    public static final int D = 2;
    /** Index of register <code>E</code>. */
    public static final int E = 3;
    /** Index of register <code>H</code>. */
    public static final int H = 4;
    /** Index of register <code>L</code>. */
    public static final int L = 5;
    /** Index of register <code>SP</code> (stack pointer). */
    public static final int SP = 8;
    /** Index of register <code>PC</code> (program counter). */
    public static final int PC = 9;

    /** Bit position of the sign flag. */
    public static final int SIGN_POS = 7;
    /** Bit position of the zero flag. */
    public static final int ZERO_POS = 6;
    /** Bit position of the auxiliary carry flag. */
    public static final int AUX_CARRY_POS = 4;
    /** Bit position of the parity flag. */
    public static final int PARITY_POS = 2;
    /** Bit position of the carry flag. */
    public static final int CARRY_POS = 0;

    /** Mask for the sign flag. */
    public static final int SIGN_MASK = 1 << SIGN_POS;
    /** Mask for the zero flag. */
    public static final int ZERO_MASK = 1 << ZERO_POS;
    /** Mask for the auxiliary carry flag. */
    public static final int AUX_CARRY_MASK = 1 << AUX_CARRY_POS;
    /** Mask for the parity flag. */
    public static final int PARITY_MASK = 1 << PARITY_POS;
    /** Mask for the carry flag. */
    public static final int CARRY_MASK = 1 << CARRY_POS;

    /** Combined flag mask for ADD/SUB operations. */
    protected static final int ADD_MASK = (AUX_CARRY_MASK + CARRY_MASK
                                         + SIGN_MASK + ZERO_MASK
                                         + PARITY_MASK) ^ 0xff;
    /** Combined flag mask for INR/DCR operations. */
    protected static final int INR_MASK = (AUX_CARRY_MASK + SIGN_MASK
                                         + ZERO_MASK + PARITY_MASK) ^ 0xff;
    /** Combined flag mask for ROT operations. */
    protected static final int ROT_MASK = (AUX_CARRY_MASK + CARRY_MASK) ^ 0xff;

    protected static final String[] DATA_REGISTER_NAMES
                         = {"B", "C", "D", "E", "H", "L", "M", "A"};
    protected static final String[] REGISTER_PAIR_NAMES
                         = {"B", "D", "H", "SP"};
    protected static final String[] REGISTER_PAIR_NAMES2
                         = {"B", "D", "H", "PSW"};
    protected static final String[] CONDITION_NAMES
                         = {"NZ", "Z ", "NC", "C ", "PO", "PE", "P ", "M "};

    protected static final short UDEF =  0; // undefined
    protected static final short ACI  =  1;
    protected static final short ADC  =  2;
    protected static final short ADD  =  3;
    protected static final short ADI  =  4;
    protected static final short ANA  =  5;
    protected static final short ANI  =  6;
    protected static final short CALL =  7;
    protected static final short Cc   =  8;// conditional call
    protected static final short CMA  =  9;
    protected static final short CMC  = 10;
    protected static final short CMP  = 11;
    protected static final short CPI  = 12;
    protected static final short DAA  = 13;
    protected static final short DAD  = 14;
    protected static final short DCR  = 15;
    protected static final short DCX  = 16;
    protected static final short DI   = 17;
    protected static final short EI   = 18;
    protected static final short HLT  = 19;
    protected static final short IN   = 20;
    protected static final short INR  = 21;
    protected static final short INX  = 22;
    protected static final short Jc   = 23; // conditional jump
    protected static final short JMP  = 24;
    protected static final short LDA  = 25;
    protected static final short LDAX = 26;
    protected static final short LHLD = 27;
    protected static final short LXI  = 28;
    protected static final short MOV  = 29;
    protected static final short MVI  = 30;
    protected static final short NOP  = 31;
    protected static final short ORA  = 32;
    protected static final short ORI  = 33;
    protected static final short OUT  = 34;
    protected static final short PCHL = 35;
    protected static final short POP  = 36;
    protected static final short PUSH = 37;
    protected static final short RAL  = 38;
    protected static final short RAR  = 39;
    protected static final short Rc   = 40; // conditional return
    protected static final short RET  = 41;
    protected static final short RLC  = 42;
    protected static final short RRC  = 43;
    protected static final short RST  = 44;
    protected static final short SBB  = 45;
    protected static final short SBI  = 46;
    protected static final short SHLD = 47;
    protected static final short SPHL = 48;
    protected static final short STA  = 49;
    protected static final short STAX = 50;
    protected static final short STC  = 51;
    protected static final short SUB  = 52;
    protected static final short SUI  = 53;
    protected static final short XCHG = 54;
    protected static final short XRA  = 55;
    protected static final short XRI  = 56;
    protected static final short XTHL = 57;
    /** Mapping op-code to instruction type. */
    protected static final short[] OP_CODE_TYPE
            = {NOP,  LXI,  STAX, INX,  INR,  DCR,  MVI,  RLC,
               UDEF, DAD,  LDAX, DCX,  INR,  DCR,  MVI,  RRC,
               UDEF, LXI,  STAX, INX,  INR,  DCR,  MVI,  RAL,
               UDEF, DAD,  LDAX, DCX,  INR,  DCR,  MVI,  RAR,
               UDEF, LXI,  SHLD, INX,  INR,  DCR,  MVI,  DAA,
               UDEF, DAD,  LHLD, DCX,  INR,  DCR,  MVI,  CMA,
               UDEF, LXI,  STA,  INX,  INR,  DCR,  MVI,  STC,
               UDEF, DAD,  LDA,  DCX,  INR,  DCR,  MVI,  CMC, // 0x40
               MOV,  MOV,  MOV,  MOV,  MOV,  MOV,  MOV,  MOV,
               MOV,  MOV,  MOV,  MOV,  MOV,  MOV,  MOV,  MOV,
               MOV,  MOV,  MOV,  MOV,  MOV,  MOV,  MOV,  MOV,
               MOV,  MOV,  MOV,  MOV,  MOV,  MOV,  MOV,  MOV,
               MOV,  MOV,  MOV,  MOV,  MOV,  MOV,  MOV,  MOV,
               MOV,  MOV,  MOV,  MOV,  MOV,  MOV,  MOV,  MOV,
               MOV,  MOV,  MOV,  MOV,  MOV,  MOV,  HLT,  MOV,
               MOV,  MOV,  MOV,  MOV,  MOV,  MOV,  MOV,  MOV, // 0x80
               ADD,  ADD,  ADD,  ADD,  ADD,  ADD,  ADD,  ADD,
               ADC,  ADC,  ADC,  ADC,  ADC,  ADC,  ADC,  ADC,
               SUB,  SUB,  SUB,  SUB,  SUB,  SUB,  SUB,  SUB,
               SBB,  SBB,  SBB,  SBB,  SBB,  SBB,  SBB,  SBB,
               ANA,  ANA,  ANA,  ANA,  ANA,  ANA,  ANA,  ANA,
               XRA,  XRA,  XRA,  XRA,  XRA,  XRA,  XRA,  XRA,
               ORA,  ORA,  ORA,  ORA,  ORA,  ORA,  ORA,  ORA,
               CMP,  CMP,  CMP,  CMP,  CMP,  CMP,  CMP,  CMP, // 0xC0
               Rc,   POP,  Jc,   JMP,  Cc,   PUSH, ADI,  RST,
               Rc,   RET,  Jc,   UDEF, Cc,   CALL, ACI,  RST,
               Rc,   POP,  Jc,   OUT,  Cc,   PUSH, SUI,  RST,
               Rc,   UDEF, Jc,   IN,   Cc,   UDEF, SBI,  RST,
               Rc,   POP,  Jc,   XTHL, Cc,   PUSH, ANI,  RST,
               Rc,   PCHL, Jc,   XCHG, Cc,   UDEF, XRI,  RST,
               Rc,   POP,  Jc,   DI,   Cc,   PUSH, ORI,  RST,
               Rc,   SPHL, Jc,   EI,   Cc,   UDEF, CPI,  RST};

    protected static final byte NULL_BYTE = (byte) 0;

    protected String errorDescription = "";

    protected byte[] registers = new byte[8];
    protected int stackPointer;
    protected int programCounter;
    protected boolean interruptEnabled = true;
    protected AddressSpace as;
    protected byte[] memory;
    protected int portBiasAddress = 0x10000;

    /** Default constructor. */
    public Intel8080() {
        reset();
    }

    /**
     *  Return the bias address for ports.
     *  A 8080 port is mapped onto the address space where
     *  its address is the bias address plus the port number.
     */
    public int getPortBiasAddress() {
        return portBiasAddress;
    }

    /**
     *  Set the bias address for ports.
     *  A 8080 port is mapped onto the address space where
     *  its address is the bias address plus the port number.
     *  @param address bias address.
     */
    public void setPortBiasAddress(int address) {
        portBiasAddress = address;
    }


    /** Show register states like CP/M DDT. */
    public String toString() {
        StringBuffer sb = new StringBuffer(getName());
        int flags = (int) registers[F];
        sb.append(": C");
        sb.append((flags & CARRY_MASK) >> CARRY_POS);
        sb.append('Z');
        sb.append((flags & ZERO_MASK) >> ZERO_POS);
        sb.append('M');
        sb.append((flags & SIGN_MASK) >> SIGN_POS);
        sb.append('E');
        sb.append((flags & PARITY_MASK) >> PARITY_POS);
        sb.append('I');
        sb.append((flags & AUX_CARRY_MASK) >> AUX_CARRY_POS);
        sb.append(" A=");
        Util.appendHex(sb, registers[A]);
        sb.append(" B=");
        Util.appendHex(sb, registers[B]);
        Util.appendHex(sb, registers[C]);
        sb.append(" D=");
        Util.appendHex(sb, registers[D]);
        Util.appendHex(sb, registers[E]);
        sb.append(" H=");
        Util.appendHex(sb, registers[H]);
        Util.appendHex(sb, registers[L]);
        sb.append(" S=");
        Util.appendHex(sb, 4, stackPointer);
        sb.append(" P=");
        Util.appendHex(sb, 4, programCounter);
        sb.append(' ');
        disassemble(sb, programCounter);
        return new String(sb);
    }

    /**
     *  Process an undefined op-code. Returns <code>false</code>.
     *  May be overwritten in subclasses.
     */
    protected boolean processUDEF(byte opCode) {
        errorDescription = "undefined op-code: "
                           + Integer.toHexString(((int) opCode) & 0xff);
        return false;
    }

    /**
     *  Process <code>ADD</code> and <code>ADC</code> commands.
     *  @param opCode op-code.
     *  @param carryBit a bit which is zero for an <code>ADD</code>
     *                  command and the carry bit for an
     *                  <code>ADC</code> command.
     */
    private boolean processADDorADC(byte opCode, int carryBit) {
        addToAccu(readByte(opCode & 7), carryBit, ADD_MASK);
        ++programCounter ;
        return true;
    }

    /** Process <code>INR</code> and <code>DCR</code> commands. */
    protected boolean processINRorDCR(byte opCode, int increment) {
        int index = (opCode & 070) >> 3;
        byte b = readByte(index);
        int accu = (b & 0xf) + increment;
        int flags = (accu & 0x10) << (AUX_CARRY_POS - 4);
        accu = (b & 0xff) + increment;
        flags += (accu & 0x80) << (SIGN_POS - 7);
        flags += ((accu & 0xff) == 0) ? ZERO_MASK : 0;
        accu = b + increment;
        boolean signFlag = (accu & 0x80) != 0;
        flags += (signFlag ? accu > 127 : accu < -128) ? PARITY_MASK : 0;
        storeByte(index, (byte) accu);
        registers[F] = (byte) ((registers[F] & INR_MASK) + flags);
        ++programCounter ;
        return true;
    }

    /**
     *  Process <code>ADI</code> and <code>ACI</code> commands.
     *  @param carryBit a bit which is zero for an <code>ADI</code>
     *                  command and the carry bit for an
     *                  <code>ACI</code> command.
     */
    private boolean processADIorACI(int carryBit) {
        addToAccu(memory[++programCounter], carryBit, ADD_MASK);
        ++programCounter ;
        return true;
    }

    /** Process <code>RAR</code> command. */
    private boolean processRAR() {
        int accu = registers[A] & 0xff;
        int flags = (accu & 1) << CARRY_POS;
        accu >>= 1;
        if ((registers[F] & CARRY_MASK) != 0) {
            accu += 0x80;
        }
        registers[A] = (byte) accu;
        registers[F] = (byte) ((registers[F] & ROT_MASK) + flags);
        ++programCounter ;
        return true;
    }

    /** Process <code>RAL</code> command. */
    private boolean processRAL() {
        int accu = (registers[A] & 0xff) << 1;
        if ((registers[F] & CARRY_MASK) != 0) {
            accu++;
        }
        int flags = (accu & 0x100) >> (8 - CARRY_POS);
        registers[A] = (byte) accu;
        registers[F] = (byte) ((registers[F] & ROT_MASK) + flags);
        ++programCounter ;
        return true;
    }

    /** Process <code>RLC</code> command. */
    private boolean processRLC() {
        int accu = (registers[A] & 0xff) << 1;
        int flags = (accu & 0x100) >> 8;
        accu += flags;
        flags <<= CARRY_POS;
        registers[A] = (byte) accu;
        registers[F] = (byte) ((registers[F] & ROT_MASK) + flags);
        ++programCounter ;
        return true;
    }

    /** Process <code>RRC</code> command. */
    private boolean processRRC() {
        int accu = registers[A] & 0xff;
        int flags = accu & 1;
        accu >>= 1;
        accu += flags << 7;
        flags <<= CARRY_POS;
        registers[A] = (byte) accu;
        registers[F] = (byte) ((registers[F] & ROT_MASK) + flags);
        ++programCounter ;
        return true;
    }

    /**
     *  Process <code>SUB</code> and <code>SBB</code> commands.
     *  @param opCode op-code.
     *  @param carryBit a bit which is zero for an <code>SUB</code>
     *                  command and the carry bit for an
     *                  <code>SBB</code> command.
     */
    private boolean processSUBorSBB(byte opCode, int carryBit) {
        subtractFromAccu(readByte(opCode & 7), carryBit, ADD_MASK);
        ++programCounter ;
        return true;
    }

    /**
     *  Process <code>SUI</code> and <code>SBI</code> commands.
     *  @param carryBit a bit which is zero for an <code>SUI</code>
     *                  command and the carry bit for an
     *                  <code>SBI</code> command.
     */
    private boolean processSUIorSBI(int carryBit) {
        subtractFromAccu(memory[++programCounter], carryBit, ADD_MASK);
        ++programCounter ;
        return true;
    }

    /** Process <code>CMP</code> code. */
    private boolean processCMP(byte opCode) {
        byte accu = registers[A];
        subtractFromAccu(readByte(opCode & 7), 0, ADD_MASK);
        registers[A] = accu;
        ++programCounter ;
        return true;
    }

    /** Process <code>CPI</code> code. */
    private boolean processCPI() {
        byte accu = registers[A];
        subtractFromAccu(memory[++programCounter], 0, ADD_MASK);
        registers[A] = accu;
        ++programCounter ;
        return true;
    }

    /** Process <code>DAD</code> code. */
    private boolean processDAD(byte opCode) {
        int accu = ((registers[H] & 0xff) << 8) + (registers[L] & 0xff);
        accu += getRegisterPair(opCode, false);
        registers[H] = (byte) (accu >> 8);
        registers[L] = (byte) accu;
        int flags = (accu & 0x10000) >> (16 - CARRY_POS);
        registers[F] = (byte) ((registers[F] & ROT_MASK) + flags);

        ++programCounter ;
        return true;
    }

    /** Process <code>INX</code> and <code>DCX</code> commands. */
    private boolean processINXorDCX(byte opCode, int increment) {
        setRegisterPair(opCode, getRegisterPair(opCode, false) + increment,
                        false);
        ++programCounter ;
        return true;
    }

    /** Process <code>ANA</code> code. */
    private boolean processANA(byte opCode) {
        setAccuAndFlags(readByte(opCode & 7) & registers[A], AUX_CARRY_MASK);
        ++programCounter ;
        return true;
    }

    /** Process <code>ANI</code> code. */
    private boolean processANI() {
        setAccuAndFlags(memory[++programCounter] & registers[A], AUX_CARRY_MASK);
        ++programCounter ;
        return true;
    }

    /** Process <code>ORA</code> code. */
    private boolean processORA(byte opCode) {
        setAccuAndFlags(readByte(opCode & 7) | registers[A], 0);
        ++programCounter ;
        return true;
    }

    /** Process <code>ORI</code> code. */
    private boolean processORI() {
        setAccuAndFlags(memory[++programCounter] | registers[A], 0);
        ++programCounter ;
        return true;
    }

    /** Process <code>XRA</code> code. */
    private boolean processXRA(byte opCode) {
        setAccuAndFlags(readByte(opCode & 7) ^ registers[A], 0);
        ++programCounter ;
        return true;
    }

    /** Process <code>XRI</code> code. */
    private boolean processXRI() {
        setAccuAndFlags(memory[++programCounter] ^ registers[A], 0);
        ++programCounter ;
        return true;
    }

    /** Process <code>CALL</code> code. */
    private boolean processCALL() {
        pushOnStack(programCounter + 3);
        programCounter = getWord(programCounter + 1);
        return true;
    }

    /**
     *  Process conditional calls. Handles <code>CNZ, CZ, CNC, CC, CPO, CPE,
     *  CP,</code> and <code>CM</code>.
     */
    private boolean processCc(byte opCode) {
        if (checkCondition(opCode)) {
            return processCALL();
        } else {
            programCounter += 3;
            return true;
        }
    }

    /** Process <code>CMA</code> code. */
    private boolean processCMA() {
        registers[A] = (byte) ~registers[A];
        registers[F] = (byte) (registers[F] | AUX_CARRY_MASK);
        ++programCounter ;
        return true;
    }

    /** Process <code>CMC</code> code. */
    private boolean processCMC() {
        registers[F] = (byte) (registers[F] ^ CARRY_MASK);
        ++programCounter ;
        return true;
    }

    /** Process <code>DAA</code> code. */
    private boolean processDAA(byte opCode) {
        int digit0 = registers[A] & 0xf;
        int digit1 = ((registers[A] >> 4) & 0xf)
                     + (((registers[F] & CARRY_MASK) != 0) ? 16 : 0);
        int flags = 0;
        if (digit0 > 9) {
            flags += AUX_CARRY_MASK;
            digit0 -= 10;
            digit1++;
        }
        if (digit1 > 9) {
            flags += CARRY_MASK;
            digit1 -= 10;
        }
        digit0 += digit1 << 4;
        flags += (digit0 == 0) ? ZERO_MASK : 0;
        flags += (digit0 & 0x80) << (SIGN_POS - 7);
        flags += calculateParityFlag((int) registers[A]);
        registers[A] = (byte) digit0;
        registers[F] = (byte) ((registers[F] & ADD_MASK) + flags);

        ++programCounter ;
        return true;
    }

    protected boolean processHLT() {
        errorDescription = "ErrorHLT";
        return false;
    }

    /** Process <code>IN</code> command. */
    private boolean processIN() {
        int port = memory[++programCounter] & 0xff;
        try {
            registers[A] = as.getByte(portBiasAddress + port, 0);
        } catch (Exception e) {
            e.printStackTrace(System.out);
        }
        ++programCounter ;
        return true;
    }

    /** Process <code>OUT</code> command. */
    private boolean processOUT() {
        int port = memory[++programCounter] & 0xff;
        try {
            as.setByte(portBiasAddress + port, 0, registers[A]);
        } catch (Exception e) {
            e.printStackTrace(System.out);
        }
        ++programCounter ;
        return true;
    }

    /**
     *  Process conditional jumps. Handles <code>JNZ, JZ, JNC, JC, JPO, JPE,
     *  JP,</code> and <code>JM</code>.
     */
    private boolean processJc(byte opCode) {
        programCounter = checkCondition(opCode) ? getWord(programCounter + 1)
                                                : programCounter + 3;
        return true;
    }

    /** Process <code>LDA</code> command. */
    private boolean processLDA() {
        registers[A] = memory[getWord(programCounter + 1)];
        programCounter += 3;
        return true;
    }

    /** Process <code>STA</code> command. */
    private boolean processSTA() {
        memory[getWord(programCounter + 1)] = registers[A];
        programCounter += 3;
        return true;
    }

    /** Process <code>LDAX</code> commands. */
    private boolean processLDAX(byte opCode) {
        registers[A] = memory[getRegisterPair(opCode, false)];
        ++programCounter ;
        return true;
    }

    /** Process <code>STAX</code> commands. */
    private boolean processSTAX(byte opCode) {
        memory[getRegisterPair(opCode, false)] = registers[A];
        ++programCounter ;
        return true;
    }

    /** Process <code>LHLD</code> command. */
    private boolean processLHLD() {
        int address = getWord(programCounter + 1);
        registers[L] = memory[address];
        registers[H] = memory[address + 1];
        programCounter += 3;
        return true;
    }

    /** Process <code>SHLD</code> command. */
    private boolean processSHLD() {
        int address = getWord(programCounter + 1);
        memory[address] = registers[L];
        memory[address + 1] = registers[H];
        programCounter += 3;
        return true;
    }

    /** Process <code>LXI</code> commands. */
    private boolean processLXI(byte opCode) {
        setRegisterPair(opCode, getWord(programCounter + 1), false);
        programCounter += 3;
        return true;
    }

    /** Process <code>MOV</code> commands. */
    private boolean processMOV(byte opCode) {
        storeByte((opCode & 0070) >> 3, readByte(opCode & 7));
        ++programCounter ;
        return true;
    }

    /** Process <code>MVI</code> commands. */
    private boolean processMVI(byte opCode) {
        storeByte((opCode & 070) >> 3, memory[++programCounter]);
        ++programCounter ;
        return true;
    }

    /** Process <code>PCHL</code> commands. */
    private boolean processPCHL() {
        programCounter = ((registers[H] & 0xff) << 8) + (registers[L] & 0xff);
        return true;
    }

    /** Process <code>POP</code> commands. */
    private boolean processPOP(byte opCode) {
        setRegisterPair(opCode, popFromStack(), true);
        ++programCounter ;
        return true;
    }

    /** Process <code>PUSH</code> commands. */
    private boolean processPUSH(byte opCode) {
        pushOnStack(getRegisterPair(opCode, true));
        ++programCounter ;
        return true;
    }

    /**
     *  Process conditional return. Handles <code>RNZ, RZ, RNC, RC, RPO, RPE,
     *  RP,</code> and <code>RM</code>.
     */
    private boolean processRc(byte opCode) {
        programCounter = checkCondition(opCode) ? popFromStack()
                                                : programCounter + 1;
        return true;
    }

    /** Process <code>RST</code> commands. */
    private boolean processRST(byte opCode) {
        pushOnStack(programCounter + 1);
        programCounter = opCode & 070;
        return true;
    }

    /** Process <code>SPHL</code> command. */
    private boolean processSPHL() {
        stackPointer = ((registers[H] & 0xff) << 8) + (registers[L] & 0xff);
        ++programCounter ;
        return true;
    }

    /** Process <code>STC</code> command. */
    private boolean processSTC() {
        registers[F] = (byte) (registers[F] | CARRY_MASK);
        ++programCounter ;
        return true;
    }

    /** Process <code>XCHG</code> command. */
    private boolean processXCHG() {
        byte z = registers[D];
        registers[D] = registers[H];
        registers[H] = z;
        z = registers[E];
        registers[E] = registers[L];
        registers[L] = z;
        ++programCounter ;
        return true;
    }

    /** Process <code>XTHL</code> command. */
    private boolean processXTHL() {
        byte b = registers[H];
        registers[H] = memory[stackPointer + 1];
        memory[stackPointer + 1] = b;
        b = registers[L];
        registers[L] = memory[stackPointer];
        memory[stackPointer] = b;
        ++programCounter ;
        return true;
    }

    protected int disassembleUDEF(byte opCode, StringBuffer sb, int address) {
        sb.append(Integer.toHexString(opCode & 0xff) + " ??");
        return 0;
    }

    protected int disassembleByte(StringBuffer sb, int address) {
        Util.appendHex(sb, memory[address + 1]);
        return 2;
    }

    protected int disassembleWord(StringBuffer sb, int address) {
        Util.appendHex(sb, 4, (long) getWord(address + 1));
        return 3;
    }

    /**
     *  Read the byte from a register or from memory. The list shows
     *  the mapping between registers and address code:
     *  <table border=1>
     *      <tr><th>Register</th><th>Address code</th></tr>
     *      <tr><td><code>B</code></td><td>0</td></tr>
     *      <tr><td><code>C</code></td><td>1</td></tr>
     *      <tr><td><code>D</code></td><td>2</td></tr>
     *      <tr><td><code>E</code></td><td>3</td></tr>
     *      <tr><td><code>H</code></td><td>4</td></tr>
     *      <tr><td><code>L</code></td><td>5</td></tr>
     *      <tr><td><code>Memory</code></td><td>6</td></tr>
     *      <tr><td><code>A</code></td><td>7</td></tr>
     *  </table>
     *  The memory address is given by the register pair <code>HL</code>.
     *  @param address address code.
     *  @return the read byte
     */
    protected byte readByte(int address) {
        byte result = registers[address];
        if (address == 6) {
            result = memory[((registers[H] << 8) & 0xff00)
                                + (registers[L] & 0xff)];
        }
        return result;
    }

    /**
     *  Store a byte in a register or in memory. The list shows
     *  the mapping between registers and address code:
     *  <table border=1>
     *      <tr><th>Register</th><th>Address code</th></tr>
     *      <tr><td><code>B</code></td><td>0</td></tr>
     *      <tr><td><code>C</code></td><td>1</td></tr>
     *      <tr><td><code>D</code></td><td>2</td></tr>
     *      <tr><td><code>E</code></td><td>3</td></tr>
     *      <tr><td><code>H</code></td><td>4</td></tr>
     *      <tr><td><code>L</code></td><td>5</td></tr>
     *      <tr><td><code>Memory</code></td><td>6</td></tr>
     *      <tr><td><code>A</code></td><td>7</td></tr>
     *  </table>
     *  The memory address is given by the register pair <code>HL</code>.
     *  @param address address code.
     *  @param value value to be stored.
     */
    protected void storeByte(int address, byte value) {
        if (address == 6) { // destination memory
            memory[((registers[H] << 8) & 0xff00)
                       + (registers[L] & 0xff)] =  value;
        } else {
            registers[address] = value;
        }
    }

    /**
     *  Get the value of a register pair.
     *  <table border=1>
     *      <tr><th>Op-code pattern</th><th>Register pair</th></tr>
     *      <tr><td>xx00xxxx</td><td><code>BC</code></td></tr>
     *      <tr><td>xx01xxxx</td><td><code>DE</code></td></tr>
     *      <tr><td>xx10xxxx</td><td><code>HL</code></td></tr>
     *      <tr><td>xx11xxxx</td>
     *          <td><code>AF</code> if <code>isPopPushPair == true</code><br>
     *              <code>SP</code> if <code>isPopPushPair == false</code><br>
     *      </td></tr>
     *  </table>
     *  @param opCode op-code.
     *  @param isPopPushPair <code>true</code> if a register pair is
     *                       meant for a <code>POP</code> or
     *                       <code>PUSH</code> command.
     *  @return a word.
     */
    protected int getRegisterPair(byte opCode, boolean isPopPushPair) {
        int addressCode = (opCode & 0x30) >> 3;
        int value = 0;
        if (addressCode < 6) {
            value = (registers[addressCode] & 0xff) << 8;
            value += registers[addressCode + 1] & 0xff;
        } else {
            if (isPopPushPair) {
                value = (registers[A] & 0xff) << 8;
                value += registers[F] & 0xff;
            } else {
                value = stackPointer;
            }
        }
        return value;
    }

    /**
     *  Set the value of a register pair.
     *  <table border=1>
     *      <tr><th>Op-code pattern</th><th>Register pair</th></tr>
     *      <tr><td>xx00xxxx</td><td><code>BC</td></tr>
     *      <tr><td>xx01xxxx</td><td><code>DE</td></tr>
     *      <tr><td>xx10xxxx</td><td><code>HL</td></tr>
     *      <tr><td>xx11xxxx</td>
     *          <td><code>AF</code> if <code>isPopPushPair == true</code><br>
     *              <code>SP</code> if <code>isPopPushPair == false</code><br>
     *      </td></tr>
     *  </table>
     *  @param opCode op-code.
     *  @param value value of the pair..
     *  @param isPopPushPair <code>true</code> if a register pair is
     *                       meant for a <code>POP</code> or
     *                       <code>PUSH</code> command.
     */
    protected void setRegisterPair(byte opCode, int value,
                                   boolean isPopPushPair) {
        int addressCode = (opCode & 0x30) >> 3;
        if (addressCode < 6) {
            registers[addressCode + 1] = (byte) value;
            registers[addressCode] = (byte) (value >> 8);
        } else {
            if (isPopPushPair) {
                registers[F] = (byte) value;
                registers[A] = (byte) (value >> 8);
            } else {
                stackPointer = value & 0xffff;
            }
        }
    }

    /**
     *  Obtain a 16-bit word from the memory. Assume lower order bit first.
     *  @param address memory address.
     *  @return word.
     */
    protected int getWord(int address) {
        int word = memory[address] & 0xff;
        word += (memory[address + 1] & 0xff) << 8;
        return word;
    }

    /**
     *  Push a 16-bit word onto the stack.
     *  The high byte will be stored in <code>stackPointer - 1</code>
     *  and the low byte in <code>stackPointer - 2</code>.
     *  The stack pointer will be decreased by two.
     *  @param word word to be push onto the stack.
     */
    protected void pushOnStack(int word) {
        memory[--stackPointer] = (byte) (word >> 8);
        memory[--stackPointer] = (byte) word;
    }

    /**
     *  Pop a 16-bit word from the stack.
     *  The low byte will be read from <code>stackPointer</code>
     *  and the high byte in <code>stackPointer + 1</code>.
     *  The stack pointer will be increased by two.
     *  @return poped word.
     */
    protected int popFromStack() {
        int word = memory[stackPointer++] & 0xff;
        word += (memory[stackPointer++] & 0xff) << 8;
        return word;
    }

    /**
     *  Check condition coded in the bits 3-5 of the op-code.
     *  <table border=1>
     *      <tr><th>Op-code pattern</th><th>Condition</th></tr>
     *      <tr><td>xx000xxx <code>NZ</code></td>
     *          <td><code>zeroFlag == false</code></td></tr>
     *      <tr><td>xx001xxx <code>Z</code></td>
     *          <td><code>zeroFlag == true</code></td></tr>
     *      <tr><td>xx010xxx <code>NC</code></td>
     *          <td><code>carryFlag == false</code></td></tr>
     *      <tr><td>xx011xxx <code>C</code></td>
     *          <td><code>carryFlag == true</code></td></tr>
     *      <tr><td>xx100xxx <code>PO</code></td>
     *          <td><code>parityFlag == false</code></td></tr>
     *      <tr><td>xx101xxx <code>PE</code></td>
     *          <td><code>parityFlag == true</code></td></tr>
     *      <tr><td>xx110xxx <code>P</code></td>
     *          <td><code>signFlag == false</code></td></tr>
     *      <tr><td>xx111xxx <code>M</code></td>
     *          <td><code>signFlag == true</code></td></tr>
     *  </table>
     *  @param opCode op-code.
     *  @return <code>true</code> if the condition is fulfilled.
     */
    protected boolean checkCondition(byte opCode) {
        switch ((opCode & 070) >> 3) {
        case 0: return (registers[F] & ZERO_MASK) == 0;
        case 1: return (registers[F] & ZERO_MASK) != 0;
        case 2: return (registers[F] & CARRY_MASK) == 0;
        case 3: return (registers[F] & CARRY_MASK) != 0;
        case 4: return (registers[F] & PARITY_MASK) == 0;
        case 5: return (registers[F] & PARITY_MASK) != 0;
        case 6: return (registers[F] & SIGN_MASK) == 0;
        case 7: return (registers[F] & SIGN_MASK) != 0;
        }
        return false;
    }

    /** Add operand and carry bit to accumulator and set the flags. */
    protected void addToAccu(byte operand, int carry, int flagMask) {
        int flags = 0;
        int accu = (registers[A] & 0xf) + (operand & 0xf) + carry;
        flags += (accu & 0x10) << (AUX_CARRY_POS - 4);
        accu = (registers[A] & 0xff) + (operand & 0xff) + carry;
        flags += (accu & 0x100) >> (8 - CARRY_POS);
        accu = registers[A] + operand + carry;
        flags += (accu & 0x80) << (SIGN_POS - 7);
        flags += ((accu & 0xff) == 0) ? ZERO_MASK : 0;
        boolean signFlag = (accu & 0x80) != 0;
        flags += (signFlag ? accu > 127 : accu < -128) ? PARITY_MASK : 0;
        registers[A] = (byte) accu;
        registers[F] = (byte) ((registers[F] & flagMask) + flags);
    }

    /** Subtract operand and carry from accumulator and set the flags. */
    protected void subtractFromAccu(byte operand, int carry, int flagMask) {
        int flags = 0;
        int accu = (registers[A] & 0xf) - (operand & 0xf) - carry;
        flags += (accu & 0x10) << (AUX_CARRY_POS - 4);
        accu = (registers[A] & 0xff) - (operand & 0xff) - carry;
        flags += (accu & 0x100) >> (8 - CARRY_POS);
        accu = registers[A] - operand - carry;
        flags += (accu & 0x80) << (SIGN_POS - 7);
        flags += ((accu & 0xff) == 0) ? ZERO_MASK : 0;
        boolean signFlag = (accu & 0x80) != 0;
        flags += (signFlag ? accu > 127 : accu < -128) ? PARITY_MASK : 0;
        registers[A] = (byte) accu;
        registers[F] = (byte) ((registers[F] & flagMask) + flags);
    }

    /**
     *  Set accumulator and flags after a bit wise.
     *  @param accu result of the bit wise operation.
     *  @param flags initial flags set.
     */
    protected void setAccuAndFlags(int accu, int flags) {
        registers[A] = (byte) accu;
        flags += (accu & 0x80) << (SIGN_POS - 7);
        flags += ((accu & 0xff) == 0) ? ZERO_MASK : 0;
        flags += calculateParityFlag((int) registers[A]);
        registers[F] = (byte) ((registers[F] & ADD_MASK) + flags);
    }

    /**
     *  Calculate the parity flag. Return <code>PARITY_MASK</code>
     *  if there is an even number of bits set in the lowest order
     *  byte of <code>accu</code>.
     *  @return <code>PARITY_MASK</code> if an even number bits
     *          of the accumulator have the value 1. Otherwise 0 is
     *          returned.
     */
    protected int calculateParityFlag(int accu) {
        int count = 0;
        for (int i = 0; i < 8; i++) {
            count += accu & 1;
            accu >>= 1;
        }
        return ((count % 2) == 0) ? PARITY_MASK : 0;
    }

//
//  Implementation of the AddressSpaceListener interface
//

    /**
     *  When a <code>RAM</code> is added to the
     *  <code>AddressSpace</code> of the CPU the
     *  memory byte array is taken from it in order to
     *  improve reading/writing performance to the memory.
     */
    public void hardwareMoved(AddressSpaceEvent event) {
        if (event.getType() == AddressSpaceEvent.ADD
            && event.getHardware() instanceof RAM
            && event.getAddressOffset() == 0) {
            memory = ((RAM) event.getHardware()).getMemory();
        }
    }

//
//  Implementation of the CPU interface
//

    /**
     *  Return "Intel 8080".
     */
    public String getName() {
        return "Intel 8080";
    }

    /**
     *  Return 8.
     */
    public int getNumberOfBits() {
        return 8;
    }

    /**
     *  Return 10.
     */
    public int getNumberOfRegisters() {
        return 10;
    }

    public Register getRegister(int index) {
        Register register = null;
        switch (index) {
        case A:
            register = new Register(8, Register.ACCUMULATOR, "A",
                                    registers[A] & 0xff);
            break;
        case F:
            register = new Register(8, Register.STATUS_REGISTER, "F",
                                    registers[F] & 0xff);
            break;
        case SP:
            register = new Register(16, Register.STACK_POINTER, "SP",
                                    stackPointer);
            break;
        case PC:
            register = new Register(16, Register.PROGRAM_COUNTER, "PC",
                                    programCounter);
            break;
        default:
            if (index >= 0 && index < 6) {
                register = new Register(8, Register.DATA_REGISTER,
                                        DATA_REGISTER_NAMES[index],
                                        registers[index] & 0xff);
            }
        }
        return register;
    }

    public Register getProgramCounter() {
        return getRegister(PC);
    }

    public void setRegisterValue(int index, int value) {
        switch (index) {
        case F:
        case A: registers[index] = (byte) value; break;
        case SP: stackPointer = value & 0xffff; break;
        case PC: programCounter = value & 0xffff; break;
        default:
            if (index >= 0 && index < 6) {
                registers[index] = (byte) value;
            }
        }
    }

    /**
     *  Set the address space. Looks for already
     *  existing <code>RAM</code>. If one found it takes
     *  the corresponding memory array.
     */
    public void setAddressSpace(AddressSpace addressSpace) {
        as = addressSpace;
        as.addAddressSpaceListener(this);
        int n = as.getNumberOfAddressBanks();
        Device d;
        for (int i = 0; i < n; i++) {
            d = as.getDevice(i);
            if (d instanceof RAM && memory == null) {
                memory = ((RAM) d).getMemory();
                break;
            }
        }
    }

    public AddressSpace getAddressSpace() {
        return as;
    }

    /**
     *  Reset the CPU. The internal state will be set into a
     *  well-defined initial state.
     */
    public void reset() {
        for (int i = 0; i < 8; i++) {
            registers[i] = 0;
        }
        stackPointer = 0;
        programCounter = 0;
        interruptEnabled = true;
    }

    /**
     *  Interrupts are not supported.
     */
    public void interrupt(long interrupt) {}

    /**
     *  Return error description. The meaning of the error
     *  code depends on the implementation. This method should be
     *  called when <code>next()</code> return <code>false</code>.
     *  @return an empty string if no error occured.
     */
    public String getErrorDescription() {
        return errorDescription;
    }

    /**
     *  Proceed one opcode.
     *  @return always 1 excepted in the case of an error.
     */
    public int next() {
        return oldNext() ? 1 : -1;
    }

    private boolean oldNext() {
        errorDescription = "";
        try {
//            byte opCode = as.getByte(programCounter);
            byte opCode = memory[programCounter];
            switch (OP_CODE_TYPE[opCode & 0xff]) {
            case UDEF: return processUDEF(opCode);
            case ACI:
                return processADIorACI(((registers[F] & CARRY_MASK) != 0)
                                       ? 1 : 0);
            case ADC:
                return processADDorADC(opCode,
                                       ((registers[F] & CARRY_MASK) != 0)
                                       ? 1 : 0);
            case ADD:  return processADDorADC(opCode, 0);
            case ADI:  return processADIorACI(0);
            case ANA:  return processANA(opCode);
            case ANI:  return processANI();
            case CALL: return processCALL();
            case Cc:   return processCc(opCode);
            case CMA:  return processCMA();
            case CMC:  return processCMC();
            case CMP:  return processCMP(opCode);
            case CPI:  return processCPI();
            case DAA:  return processDAA(opCode);
            case DAD:  return processDAD(opCode);
            case DCR:  return processINRorDCR(opCode, -1);
            case DCX:  return processINXorDCX(opCode, -1);
            case DI:
                interruptEnabled = false;
                ++programCounter;
                return true;
            case EI:
                interruptEnabled = true;
                ++programCounter;
                return true;
            case HLT:  return processHLT();
            case IN:   return processIN();
            case INR:  return processINRorDCR(opCode, 1);
            case INX:  return processINXorDCX(opCode, 1);
            case Jc:   return processJc(opCode);
            case JMP:
                programCounter = getWord(programCounter + 1);
                return true;
            case LDA:  return processLDA();
            case LDAX: return processLDAX(opCode);
            case LHLD: return processLHLD();
            case LXI:  return processLXI(opCode);
            case MOV:  return processMOV(opCode);
            case MVI:  return processMVI(opCode);
            case NOP:  ++programCounter; return true;
            case ORA:  return processORA(opCode);
            case ORI:  return processORI();
            case OUT:  return processOUT();
            case PCHL: return processPCHL();
            case POP:  return processPOP(opCode);
            case PUSH: return processPUSH(opCode);
            case RAL:  return processRAL();
            case RAR:  return processRAR();
            case Rc:   return processRc(opCode);
            case RET:  programCounter = popFromStack(); return true;
            case RLC:  return processRLC();
            case RRC:  return processRRC();
            case RST:  return processRST(opCode);
            case SBB:
                return processSUBorSBB(opCode,
                                       ((registers[F] & CARRY_MASK) != 0)
                                       ? 1 : 0);
            case SBI:
                return processSUIorSBI(((registers[F] & CARRY_MASK) != 0)
                                       ? 1 : 0);
            case SHLD: return processSHLD();
            case SPHL: return processSPHL();
            case STA:  return processSTA();
            case STAX: return processSTAX(opCode);
            case STC:  return processSTC();
            case SUB:  return processSUBorSBB(opCode, 0);
            case SUI:  return processSUIorSBI(0);
            case XCHG: return processXCHG();
            case XRA:  return processXRA(opCode);
            case XRI:  return processXRI();
            case XTHL: return processXTHL();
            }
            return true;
        } catch (Exception e) {
            errorDescription = e.toString();
            e.printStackTrace();
            return false;
        }
    }

    /** Return <code>true</code>. */
    public boolean supportsDisassembling() {
        return true;
    }

    /**
     *  Return disassemble instruction in standard Intel 8080
     *  assembler syntax.
     */
    public int disassemble(StringBuffer sb, int address) {
        try {
            byte opCode = memory[address];
            switch (OP_CODE_TYPE[((int) opCode) & 0xff]) {
            case UDEF: return disassembleUDEF(opCode, sb, address);
            case ACI:
                sb.append("ACI  ");
                return disassembleByte(sb, address);
            case ADC:
                sb.append("ADC  ");
                sb.append(DATA_REGISTER_NAMES[opCode & 7]);
                return 1;
            case ADD:
                sb.append("ADD  ");
                sb.append(DATA_REGISTER_NAMES[opCode & 7]);
                return 1;
            case ADI:
                sb.append("ADI  ");
                return disassembleByte(sb, address);
            case ANA:
                sb.append("ANA  ");
                sb.append(DATA_REGISTER_NAMES[opCode & 7]);
                return 1;
            case ANI:
                sb.append("ANI  ");
                return disassembleByte(sb, address);
            case CALL:
                sb.append("CALL ");
                return disassembleWord(sb, address);
            case Cc:
                sb.append('C');
                sb.append(CONDITION_NAMES[(opCode & 070) >> 3]);
                sb.append("  ");
                return disassembleWord(sb, address);
            case CMA:  sb.append("CMA"); return 1;
            case CMC:  sb.append("CMC"); return 1;
            case CMP:
                sb.append("CMP  ");
                sb.append(DATA_REGISTER_NAMES[opCode & 7]);
                return 1;
            case CPI:
                sb.append("CPI  ");
                return disassembleByte(sb, address);
            case DAA:   sb.append("DAA"); return 1;
            case DAD:
                sb.append("DAD  ");
                sb.append(REGISTER_PAIR_NAMES[(opCode & 0x30) >> 4]);
                return 1;
            case DCR:
                sb.append("DCR  ");
                sb.append(DATA_REGISTER_NAMES[(opCode & 070) >> 3]);
                return 1;
            case DCX:
                sb.append("DCX  ");
                sb.append(REGISTER_PAIR_NAMES[(opCode & 0x30) >> 4]);
                return 1;
            case DI:   sb.append("DI"); return 1;
            case EI:   sb.append("EI"); return 1;
            case HLT:  sb.append("HLT"); return 0;
            case IN:
                sb.append("IN   ");
                return disassembleByte(sb, address);
            case INR:
                sb.append("INR  ");
                sb.append(DATA_REGISTER_NAMES[(opCode & 070) >> 3]);
                return 1;
            case INX:
                sb.append("INX  ");
                sb.append(REGISTER_PAIR_NAMES[(opCode & 0x30) >> 4]);
                return 1;
            case Jc:
                sb.append('J');
                sb.append(CONDITION_NAMES[(opCode & 070) >> 3]);
                sb.append("  ");
                return disassembleWord(sb, address);
            case JMP:
                sb.append("JMP  ");
                return disassembleWord(sb, address);
            case LDA:
                sb.append("LDA  ");
                return disassembleWord(sb, address);
            case LDAX:
                sb.append("LDAX ");
                sb.append(REGISTER_PAIR_NAMES2[(opCode & 0x30) >> 4]);
                return 1;
            case LHLD:
                sb.append("LHLD ");
                return disassembleWord(sb, address);
            case LXI:
                sb.append("LXI  ");
                sb.append(REGISTER_PAIR_NAMES[(opCode & 0x30) >> 4]);
                sb.append(',');
                Util.appendHex(sb, 4, (long) getWord(address + 1));
                return 3;
            case MOV:
                sb.append("MOV  ");
                sb.append(DATA_REGISTER_NAMES[(opCode & 070) >> 3]);
                sb.append(',');
                sb.append(DATA_REGISTER_NAMES[opCode & 7]);
                return 1;
            case MVI:
                sb.append("MVI  ");
                sb.append(DATA_REGISTER_NAMES[(opCode & 070) >> 3]);
                sb.append(',');
                Util.appendHex(sb, memory[address + 1]);
                return 2;
            case NOP:  sb.append("NOP"); return 1;
            case ORA:
                sb.append("ORA  ");
                sb.append(DATA_REGISTER_NAMES[opCode & 7]);
                return 1;
            case ORI:
                sb.append("ORI  ");
                return disassembleByte(sb, address);
            case OUT:
                sb.append("OUT  ");
                return disassembleByte(sb, address);
            case PCHL: sb.append("PCHL"); return 1;
            case POP:
                sb.append("POP  ");
                sb.append(REGISTER_PAIR_NAMES2[(opCode & 0x30) >> 4]);
                return 1;
            case PUSH:
                sb.append("PUSH ");
                sb.append(REGISTER_PAIR_NAMES2[(opCode & 0x30) >> 4]);
                return 1;
            case RAL:  sb.append("RAL"); return 1;
            case RAR:  sb.append("RAR"); return 1;
            case Rc:
                sb.append('R');
                sb.append(CONDITION_NAMES[(opCode & 070) >> 3]);
                return 1;
            case RET:  sb.append("RET"); return 1;
            case RLC:  sb.append("RLC"); return 1;
            case RRC:  sb.append("RRC"); return 1;
            case RST:
                sb.append("RST  ");
                sb.append((opCode & 070) >> 3);
                return 1;
            case SBB:
                sb.append("SBB  ");
                sb.append(DATA_REGISTER_NAMES[opCode & 7]);
                return 1;
            case SBI:
                sb.append("SBI  ");
                return disassembleByte(sb, address);
            case SHLD:
                sb.append("SHLD ");
                return disassembleWord(sb, address);
            case SPHL: sb.append("SPHL"); return 1;
            case STA:
                sb.append("STA  ");
                return disassembleWord(sb, address);
            case STAX:
                sb.append("STAX ");
                sb.append(REGISTER_PAIR_NAMES2[(opCode & 0x30) >> 4]);
                return 1;
            case STC:  sb.append("STC"); return 1;
            case SUB:
                sb.append("SUB  ");
                sb.append(DATA_REGISTER_NAMES[opCode & 7]);
                return 1;
            case SUI:
                sb.append("SUI  ");
                return disassembleByte(sb, address);
            case XCHG: sb.append("XCHG"); return 1;
            case XRA:
                sb.append("XRA  ");
                sb.append(DATA_REGISTER_NAMES[opCode & 7]);
                return 1;
            case XRI:
                sb.append("XRI  ");
                return disassembleByte(sb, address);
            case XTHL: sb.append("XTHL"); return 1;
            }
            return 0;
        } catch (Exception e) {
            errorDescription = e.toString();
            return 0;
        }
    }

}
