package jfrace.cpu;

import jfrace.AddressSpace;
import jfrace.CPU;
import jfrace.Register;
import jfrace.Util;

/**
 *  The 6502 microprocessor.
 *
 *  @author Franz-Josef Elmer
 */
public class MC6502 implements CPU {
    /** Index of register <code>A</code> (accumulator). */
    public static final int A = 0;
    /** Index of register <code>X</code> (index register). */
    public static final int X = 1;
    /** Index of register <code>X</code> (index register). */
    public static final int Y = 2;
    /** Index of register <code>S</code> (stack pointer). */
    public static final int S = 3;
    /** Index of register <code>P</code> (processor status). */
    public static final int P = 4;
    /** Index of register <code>PC</code> (pogram counter). */
    public static final int PC = 5;

    /** Bit position of the negative flag. */
    public static final int N_POS = 7;
    /** Bit position of the overflow flag. */
    public static final int V_POS = 6;
    /** Bit position of the break flag. */
    public static final int B_POS = 4;
    /** Bit position of the decimal mode flag. */
    public static final int D_POS = 3;
    /** Bit position of the interrupt disable flag. */
    public static final int I_POS = 2;
    /** Bit position of the zero flag. */
    public static final int Z_POS = 1;
    /** Bit position of the carry flag. */
    public static final int C_POS = 0;

    private static final int MISC_GROUP = 0;
    private static final int _U_ = 0;
    private static final int BRK = 2;
    private static final int JMP = 3;
    private static final int JSR = 4;
    private static final int NOP = 5;
    private static final int STA = 6;

    private static final int ARITHMETIC_GROUP = 0x10;
    private static final int ADC = 0x10;
    private static final int AND = 0x11;
    private static final int CMP = 0x12;
    private static final int EOR = 0x13;
    private static final int LDA = 0x14;
    private static final int ORA = 0x15;
    private static final int SBC = 0x16;
    private static final int BIT = 0x17;


    private static final int SHIFT_GROUP = 0x20;
    private static final int ASL = 0x20;
    private static final int LSR = 0x22;
    private static final int ROL = 0x24;
    private static final int ROR = 0x26;

    private static final int BRANCH_GROUP = 0x30;
    private static final int BCC = 0x30;
    private static final int BCS = 0x31;
    private static final int BEQ = 0x33;
    private static final int BMI = 0x3f;
    private static final int BNE = 0x32;
    private static final int BPL = 0x3e;
    private static final int BVC = 0x3c;
    private static final int BVS = 0x3d;

    private static final int CLEAR_SET_GROUP = 0x40;
    private static final int CLC = 0x40;
    private static final int CLD = 0x46;
    private static final int CLI = 0x44;
    private static final int CLV = 0x4c;
    private static final int SEC = 0x41;
    private static final int SED = 0x47;
    private static final int SEI = 0x45;

    private static final int CPX_CPY_GROUP = 0x50;
    private static final int CPX = 0x50;
    private static final int CPY = 0x51;

    private static final int INC_DEC_GROUP = 0x60;
    private static final int DEC = 0x60;
    private static final int DEX = 0x62;
    private static final int DEY = 0x64;
    private static final int INC = 0x61;
    private static final int INX = 0x63;
    private static final int INY = 0x65;

    private static final int LOAD_STORE_XY_GROUP = 0x70;
    private static final int LDX = 0x70;
    private static final int LDY = 0x72;
    private static final int STX = 0x71;
    private static final int STY = 0x73;

    private static final int PUSH_PULL_GROUP = 0x80;
    private static final int PHA = 0x80;
    private static final int PHP = 0x81;
    private static final int PLA = 0x82;
    private static final int PLP = 0x83;

    private static final int RETURN_GROUP = 0x90;
    private static final int RTI = 0x90;
    private static final int RTS = 0x91;

    private static final int TRANSFER_GROUP = 0xa0;
    private static final int TAX = 0xa0;
    private static final int TAY = 0xa2;
    private static final int TSX = 0xa4;
    private static final int TXA = 0xa1;
    private static final int TXS = 0xa5;
    private static final int TYA = 0xa3;

    private static final short[] OP_CODE_TYPE
            = {BRK, ORA, _U_, _U_, _U_, ORA, ASL, _U_,
                PHP, ORA, ASL, _U_, _U_, ORA, ASL, _U_,
                 BPL, ORA, _U_, _U_, _U_, ORA, ASL, _U_,
                  CLC, ORA, _U_, _U_, _U_, ORA, ASL, _U_,

               JSR, AND, _U_, _U_, BIT, AND, ROL, _U_,
                PLP, AND, ROL, _U_, BIT, AND, ROL, _U_,
                 BMI, AND, _U_, _U_, _U_, AND, ROL, _U_,
                  SEC, AND, _U_, _U_, _U_, AND, ROL, _U_,

               RTI, EOR, _U_, _U_, _U_, EOR, LSR, _U_,
                PHA, EOR, LSR, _U_, JMP, EOR, LSR, _U_,
                 BVC, EOR, _U_, _U_, _U_, EOR, LSR, _U_,
                  CLI, EOR, _U_, _U_, _U_, EOR, LSR, _U_,

               RTS, ADC, _U_, _U_, _U_, ADC, ROR, _U_,
                PLA, ADC, ROR, _U_, JMP, ADC, ROR, _U_,
                 BVS, ADC, _U_, _U_, _U_, ADC, ROR, _U_,
                  SEI, ADC, _U_, _U_, _U_, ADC, ROR, _U_,

               _U_, STA, _U_, _U_, STY, STA, STX, _U_,
                DEY, _U_, TXA, _U_, STY, STA, STX, _U_,
                 BCC, STA, _U_, _U_, STY, STA, STX, _U_,
                  TYA, STA, TXS, _U_, _U_, STA, _U_, _U_,

               LDY, LDA, LDX, _U_, LDY, LDA, LDX, _U_,
                TAY, LDA, TAX, _U_, LDY, LDA, LDX, _U_,
                 BCS, LDA, _U_, _U_, LDY, LDA, LDX, _U_,
                  CLV, LDA, TSX, _U_, LDY, LDA, LDX, _U_,

               CPY, CMP, _U_, _U_, CPY, CMP, DEC, _U_,
                INY, CMP, DEX, _U_, CPY, CMP, DEC, _U_,
                 BNE, CMP, _U_, _U_, _U_, CMP, DEC, _U_,
                  CLD, CMP, _U_, _U_, _U_, CMP, DEC, _U_,

               CPX, SBC, _U_, _U_, CPX, SBC, INC, _U_,
                INX, SBC, NOP, _U_, CPX, SBC, INC, _U_,
                 BEQ, SBC, _U_, _U_, _U_, SBC, INC, _U_,
                  SED, SBC, _U_, _U_, _U_, SBC, INC, _U_};

    private static final int NMI_VECTOR = 0xfffa;
    private static final int RESET_VECTOR = 0xfffc;
    private static final int IRQ_VECTOR = 0xfffe;

    private static final int[] LDA_CYCLES = {2, -1, -2, 0, 1, 0, 0, 0};
    private static final int[] STA_CYCLES = {2, -1, -2, 0, 2, 0, 1, 1};
    private static final int[] SHIFT_CYCLES = {0, 5, 2, 6, 0, 6, 0, 7};


    private int pc;
    private byte s;
    private byte x;
    private byte y;
    private byte a;
    private byte p;
    private AddressSpace as;
    private String errorDescription = "ERROR";

    /** Return the state of the CPU as a well-formed string. */
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("P=");
        Util.appendHex(sb, p);
        sb.append(" A=");
        Util.appendHex(sb, a);
        sb.append(" X=");
        Util.appendHex(sb, x);
        sb.append(" Y=");
        Util.appendHex(sb, y);
        sb.append(" S=");
        Util.appendHex(sb, s);
        sb.append(" PC=");
        Util.appendHex(sb, 4, pc);
        sb.append(' ');
        disassemble(sb, pc);
        return new String(sb);
    }

    private int processMiscGroup(int opCode, int opCodeType) {
        int numberOfCycles = -1;
        switch (opCodeType) {
        case BRK:
            numberOfCycles = 7;
            ++pc;
            pushPC();
            p |= 0x10;
            push(p);
            pc = getWord(IRQ_VECTOR);
            break;
        case JMP:
            int adr = getWord(pc);
            if ((opCode & 0x20) == 0) { // absolute
                numberOfCycles = 3;
                pc = adr & 0xffff;
            } else { // indirect
                numberOfCycles = 5;
                if ((adr & 0xff) == 0xff) { // page boundary bug
                    pc = ((as.getByte(adr - 0xff, 0) & 0xff) << 8)
                         + (as.getByte(adr, 0) & 0xff);
                } else {
                    pc = getWord(adr);
                }
            }
            break;
        case JSR:
            numberOfCycles = 6;
            adr = getWord(pc++);
            pushPC();
            pc = adr & 0xffff;
            break;
        case NOP:
            numberOfCycles = 2;
            break;
        case STA:
            int addressMode = (opCode & 0x1c) >> 2;
            numberOfCycles = 4 + STA_CYCLES[addressMode];
            setByte(addressMode, x, a);
            break;
        default:
            errorDescription = "Undefined opcode: "
                               + Integer.toHexString(opCode);
        }
        return numberOfCycles;
    }

    private int processArithmeticGroup(int opCode, int opCodeType) {
        int addressMode = (opCode & 0x1c) >> 2;
        int numberOfCycles = 4 + LDA_CYCLES[addressMode];
        int value = getByte(addressMode, x);
        switch (opCodeType) {
        case ADC:
            if ((p & 8) == 0) { // binary arithmetics
//                value += a + (p & 1);
                int op1 = value & 0xff;
                int op2 = (a & 0xff) + (p & 1);
                value = op1 + op2;
                setStatus(value, 0x7c);
                int overflow = (~(op1 ^ op2) & (op1 ^ value) & 0x80) >> 1;
                p = (byte) ((p & 0xbf) | overflow);
            } else { // BCD arithmetics
                int carry = p & 1;
                int a0 = a & 0xf;
                int a1 = a & 0xf0;
                int v0 = value & 0xf;
                int v1 = value & 0xf0;
                int s0 = a0 + v0 + carry;
                carry = 0;
                if (s0 > 9) {
                    s0 -= 10;
                    carry = 16;
                }
                int s1 = a1 + v1 + carry;
                carry = 0;
                if (s1 > 160) {
                    s1 -= 160;
                    carry = 1;
                }
                value = s1 + s0;
                p = (byte) ((p & 0xfe) + carry);
            }
            a = (byte) value;
            break;
        case AND:
            a = (byte) (value & a);
            setStatus((int) a, 0x7d);
            break;
        case BIT:
            setStatus(a & value, 0x7d);
            break;
        case CMP:
            value = a - value;
            setStatus(value, 0x7c);
            break;
        case CPX:
            value = x - value;
            setStatus(value, 0x7c);
            break;
        case CPY:
            value = y - value;
            setStatus(value, 0x7c);
            break;
        case EOR:
            a = (byte) (value ^ a);
            setStatus((int) a, 0x7d);
            break;
        case LDA:
            setStatus(value, 0x7d);
            a = (byte) value;
            break;
        case ORA:
            a = (byte) (value | a);
            setStatus((int) a, 0x7d);
            break;
        case SBC:
            if ((p & 8) == 0) {
                value = a - value - (p & 1);
                setStatus(value, 0x3c);
            } else { // BCD arithmetics
                int carry = p & 1;
                int a0 = a & 0xf;
                int a1 = a & 0xf0;
                int v0 = value & 0xf;
                int v1 = value & 0xf0;
                int s0 = a0 - v0 - carry;
                carry = 0;
                if (s0 < 0) {
                    s0 += 10;
                    carry = 16;
                }
                int s1 = a1 - v1 - carry;
                carry = 0;
                if (s1 < 0) {
                    s1 += 160;
                    carry = 1;
                }
                value = s1 + s0;
                p = (byte) ((p & 0xfe) + carry);
            }
            a = (byte) value;
            break;
        default:
            errorDescription = "Undefined opcode: "
                               + Integer.toHexString(opCode);
            numberOfCycles = -1;
        }
        return numberOfCycles;
    }

    private int processShiftGroup(int opCode, int opCodeType) {
        int addressMode = (opCode & 0x1c) >> 2;
        int numberOfCycles = SHIFT_CYCLES[addressMode];
        boolean isAccu = (addressMode & 1) == 0;
        int adr = 0;
        int value = a & 0xff;
        if (!isAccu) {
            adr = as.getByte(pc++, 0) & 0xff;
            if ((addressMode & 2) == 2) { // absolute mode
                adr += (as.getByte(pc++, 0) & 0xff) << 8;
            }
            if ((addressMode & 4) == 4) { // with index x
                adr += x & 0xff;
            }
            value = as.getByte(adr, 0) & 0xff;
        }
        switch (opCodeType) {
        case ASL:
            value <<= 1;
            break;
        case LSR:
            value = (value >> 1) | ((value & 1) << 8);
            break;
        case ROL:
            value = (value << 1) | (p & 1);
            break;
        case ROR:
            value = (value >> 1) | ((p & 1) << 7) | ((value & 1) << 8);
            break;
        default:
            errorDescription = "Undefined opcode: "
                               + Integer.toHexString(opCode);
            return -1;
        }
        setStatus(value, 0x7c);
        if (isAccu) {
            a = (byte) value;
        } else {
            as.setByte(adr, 0, (byte) value);
        }

        return numberOfCycles;
    }


    private int processBranchGroup(int opCode, int opCodeType) {
        int numberOfCycles = 2;
        int mask = 1 << ((opCodeType >> 1) & 7);
        if (((p >> ((opCodeType >> 1) & 7)) & 1) == (opCodeType & 1)) {
            pc = (pc + as.getByte(pc, 0) + 1) & 0xffff;
            ++numberOfCycles;
        } else {
            ++pc;
        }
        return numberOfCycles;
    }

    private int processClearSetGroup(int opCode, int opCodeType) {
        int numberOfCycles = 2;
        int mask = 1 << ((opCodeType >> 1) & 7);
        if ((opCodeType & 1) == 0) {
            p &= ~mask;
        } else {
            p |= mask;
        }
        return numberOfCycles;
    }

    private int processCPXCPYGroup(int opCode, int opCodeType) {
        int addressMode = (opCode & 0x1c) >> 2;
        if (addressMode == 0) {
            addressMode = 2;
        }
        int numberOfCycles = 4 + LDA_CYCLES[addressMode];
        int value = getByte(addressMode, x);
        switch (opCodeType) {
        case CPX:
            value = x - value;
            setStatus(value, 0x7c);
            break;
        case CPY:
            value = y - value;
            setStatus(value, 0x7c);
            break;
        default:
            errorDescription = "Undefined opcode: "
                               + Integer.toHexString(opCode);
            numberOfCycles = -1;
        }
        return numberOfCycles;
    }

    private int processIncDecGroup(int opCode, int opCodeType) {
        int addressMode = (opCode & 0x1c) >> 2;
        int numberOfCycles = SHIFT_CYCLES[addressMode];
        boolean isReg = (addressMode & 1) == 0;
        boolean isY = (opCodeType & 2) == 0;
        int adr = 0;
        int value = 0;
        if (isReg) {
            value = (isY ? y : x) & 0xff;
        } else {
            adr = as.getByte(pc++, 0) & 0xff;
            if ((addressMode & 2) == 2) { // absolute mode
                adr += (as.getByte(pc++, 0) & 0xff) << 8;
            }
            if ((addressMode & 4) == 4) { // with index x
                adr += x & 0xff;
            }
            value = as.getByte(adr, 0) & 0xff;
        }
        value += ((opCodeType & 1) << 1) - 1;
        setStatus(value, 0x7d);
        if (isReg) {
            if (isY) {
                y = (byte) value;
            } else {
                x = (byte) value;
            }
        } else {
            as.setByte(adr, 0, (byte) value);
        }

        return numberOfCycles;
    }


    private int processLoadStoreXYGroup(int opCode, int opCodeType) {
        int addressMode = (opCode & 0x1c) >> 2;
        if (addressMode == 0) {
            addressMode = 2;
        }
        int numberOfCycles = 4 + LDA_CYCLES[addressMode];
        boolean isX = (opCodeType & 2) == 0;
        if ((opCodeType & 1) == 0) { // load
            int value = 0;
            if (isX) {
                value = getByte(addressMode, y);
                x = (byte) value;
            } else {
                value = getByte(addressMode, x);
                y = (byte) value;
            }
            setStatus(value, 0x7d);
        } else { // store
            if (isX) {
                setByte(addressMode, y, x);
            } else {
                setByte(addressMode, x, y);
            }
        }
        return numberOfCycles;
    }

    private int processPushPullGroup(int opCode, int opCodeType) {
        int numberOfCycles = -1;
        switch (opCodeType) {
        case PHA:
            numberOfCycles = 3;
            push(a);
            break;
        case PHP:
            numberOfCycles = 3;
            push(p);
            break;
        case PLA:
            numberOfCycles = 4;
            a = pull();
            break;
        case PLP:
            numberOfCycles = 4;
            p = pull();
            break;
        default:
            errorDescription = "Undefined opcode: "
                               + Integer.toHexString(opCode);
        }
        return numberOfCycles;
    }

    private int processReturnGroup(int opCode, int opCodeType) {
        int numberOfCycles = 6;
        switch (opCodeType) {
        case RTI:
            p = pull();
            pullPC();
            break;
        case RTS:
            pullPC();
            pc = (pc + 1) & 0xffff;
            break;
        default:
            numberOfCycles = -1;
            errorDescription = "Undefined opcode: "
                               + Integer.toHexString(opCode);
        }
        return numberOfCycles;
    }

    private int processTransferGroup(int opCode, int opCodeType) {
        int numberOfCycles = 2;
        switch (opCodeType) {
        case TAX:
            x = a;
            setStatus(x & 0xff, 0x7d);
            break;
        case TAY:
            y = a;
            setStatus(y & 0xff, 0x7d);
            break;
        case TSX:
            x = s;
            setStatus(x & 0xff, 0x7d);
            break;
        case TXA:
            a = x;
            setStatus(a & 0xff, 0x7d);
            break;
        case TXS:
            s = x;
            break;
        case TYA:
            a = y;
            setStatus(a & 0xff, 0x7d);
            break;
        default:
            numberOfCycles = -1;
            errorDescription = "Undefined opcode: "
                               + Integer.toHexString(opCode);
        }
        return numberOfCycles;
    }

    /** Get byte for the standard address modes. */
    private int getByte(int addressMode, byte indexRegister) {
        int value = (int) as.getByte(pc++, 0);
        int adr = value & 0xff;
        if ((addressMode & 1) == 0) {
            switch (addressMode) {
            case 2: // immediate
                return value;
            case 0: // indexed indirect x
                adr = getWord(adr + (x & 0xff));
                break;
            case 4: // indirect indexed y
                adr = getWord(adr) + (y & 0xff);
                break;
            case 6: // absolute indexed y
                adr += ((as.getByte(pc++, 0) & 0xff) << 8)
                       + (y & 0xff);
            }
        } else { // zero-page and absolute modes w/wo x
            if ((addressMode & 2) == 2) { // absolute mode
                adr += (as.getByte(pc++, 0) & 0xff) << 8;
            }
            if ((addressMode & 4) == 4) { // with index x
                adr += indexRegister & 0xff;
            }
        }
        return (int) as.getByte(adr & 0xffff, 0);
    }

    /** Store byte for the given standard address mode. */
    private void setByte(int addressMode, byte indexRegister, byte b) {
        int value = (int) as.getByte(pc++, 0);
        int adr = value & 0xff;
        if ((addressMode & 1) == 0) {
            switch (addressMode) {
            case 0: // indexed indirect x
                adr = getWord(adr + (x & 0xff));
                break;
            case 4: // indirect indexed y
                adr = getWord(adr) + (y & 0xff);
                break;
            case 6: // absolute indexed y
                adr += ((as.getByte(pc++, 0) & 0xff) << 8)
                       + (y & 0xff);
            }
        } else { // zero-page and absolute modes w/wo x
            if ((addressMode & 2) == 2) { // absolute mode
                adr += (as.getByte(pc++, 0) & 0xff) << 8;
            }
            if ((addressMode & 4) == 4) { // with index x
                adr += indexRegister & 0xff;
            }
        }
        as.setByte(adr & 0xffff, 0, b);
    }

    private void setStatus(int value, int maskOfUnaffectedFlags) {
        int status = value & 0x80; // sign
        status |= (value >> 8) & 1; // carry
        if (value > 127 || value < -128) {
            status |= 0x40;
        }
        if ((value & 0xff) == 0) {
            status |= 2;
        }
        p = (byte) ((status & ~maskOfUnaffectedFlags)
                    | (p & maskOfUnaffectedFlags));
    }

    /**
     *  Push a byte onto the 6502 stack.
     *  Decrement <code>S</code> by one.
     */
    private void push(byte b) {
        as.setByte(0x100 + (s & 0xff), 0, b);
        s = (byte) (s - 1);
    }

    /**
     *  Pull a byte from the 6502 stack. Increment <code>S</code>
     *  by one before poping.
     */
    private byte pull() {
        int adr = (s + 1) & 0xff;
        s = (byte) adr;
        return as.getByte(0x100 + adr, 0);
    }

    /**
     *  Push <code>PC</code> onto the 6502 stack.
     *  Decrement <code>S</code> by 2.
     */
    private void pushPC() {
        push((byte) (pc >> 8));
        push((byte) pc);
    }

    /**
     *  Pop <code>PC</code> from the 6502 stack.
     *  Increment <code>S</code> by 2.
     */
    private void pullPC() {
        pc = (pull() & 0xff) + ((pull() & 0xff) << 8);
    }

    /**
     *  Read a 16 bit value from the address space.
     *  @param adr Address.
     *  @return <code>(adr) + 256 * (adr + 1)</code>.
     */
    private int getWord(int adr) {
        return (as.getByte(adr, 0) & 0xff)
               + ((as.getByte(adr + 1, 0) & 0xff) << 8);
    }

//
//  Implementation of the CPU interface
//

    /** Return "6502". */
    public String getName() {
        return "6502";
    }

    /** Return 8. */
    public int getNumberOfBits() {
        return 8;
    }

    /** Return 6. */
    public int getNumberOfRegisters() {
        return 6;
    }

    public Register getRegister(int index) {
        Register register = null;
        switch (index) {
        case A:
            register = new Register(8, Register.ACCUMULATOR, "A",
                                    a & 0xff);
            break;
        case X:
            register = new Register(8, Register.ADDRESS_REGISTER, "X",
                                    x & 0xff);
            break;
        case Y:
            register = new Register(8, Register.ADDRESS_REGISTER, "Y",
                                    y & 0xff);
            break;
        case S:
            register = new Register(8, Register.STATUS_REGISTER, "S",
                                    s & 0xff);
            break;
        case P:
            register = new Register(8, Register.STACK_POINTER, "P",
                                    p & 0xff);
            break;
        case PC:
            register = new Register(16, Register.PROGRAM_COUNTER, "PC",
                                    pc & 0xffff);
            break;
        }
        return register;
    }

    public Register getProgramCounter() {
        return new Register(16, Register.PROGRAM_COUNTER, "PC", pc & 0xffff);
    }

    public void setRegisterValue(int index, int value) {
        switch (index) {
        case A: a = (byte) value; break;
        case X: x = (byte) value; break;
        case Y: y = (byte) value; break;
        case P: p = (byte) value; break;
        case S: s = (byte) value; break;
        case PC: pc = value & 0xffff; break;
        }
    }

    public void setAddressSpace(AddressSpace addressSpace) {
        as = addressSpace;
    }

    public AddressSpace getAddressSpace() {
        return as;
    }

    public void reset() {
        pc = getWord(RESET_VECTOR);
    }

    /** Not implemented. */
    public void interrupt(long interrupt) {}

    public String getErrorDescription() {
        return errorDescription;
    }

    public int next() {
        int numberOfCycles = -1;
        int opCode = as.getByte(pc++, 0) & 0xff;
        int opCodeType = OP_CODE_TYPE[opCode];
        switch (opCodeType & 0xf0) {
        case MISC_GROUP:
            numberOfCycles = processMiscGroup(opCode, opCodeType);
            break;
        case ARITHMETIC_GROUP:
            numberOfCycles = processArithmeticGroup(opCode, opCodeType);
            break;
        case SHIFT_GROUP:
            numberOfCycles = processShiftGroup(opCode, opCodeType);
            break;
        case BRANCH_GROUP:
            numberOfCycles = processBranchGroup(opCode, opCodeType);
            break;
        case CLEAR_SET_GROUP:
            numberOfCycles = processClearSetGroup(opCode, opCodeType);
            break;
        case INC_DEC_GROUP:
            numberOfCycles = processIncDecGroup(opCode, opCodeType);
            break;
        case CPX_CPY_GROUP:
            numberOfCycles = processCPXCPYGroup(opCode, opCodeType);
            break;
        case LOAD_STORE_XY_GROUP:
            numberOfCycles = processLoadStoreXYGroup(opCode, opCodeType);
            break;
        case PUSH_PULL_GROUP:
            numberOfCycles = processPushPullGroup(opCode, opCodeType);
            break;
        case RETURN_GROUP:
            numberOfCycles = processReturnGroup(opCode, opCodeType);
            break;
        case TRANSFER_GROUP:
            numberOfCycles = processTransferGroup(opCode, opCodeType);
            break;
        default:
            errorDescription = "Undefined opcode: "
                               + Integer.toHexString(opCode);
        }
        return numberOfCycles;
    }

    /** Disassembling is supported. */
    public boolean supportsDisassembling() {
        return true;
    }

    /**
     */
    public int disassemble(StringBuffer sb, int address) {
        if (as != null) {
            int opCode = as.getByte(address++, 0) & 0xff;
            int mi = MNEMONIC_INDEX[opCode];
            if (mi >= 0) {
                sb.append(MNEMONICS.charAt(mi++));
                sb.append(MNEMONICS.charAt(mi++));
                sb.append(MNEMONICS.charAt(mi++));
                switch (ADDRESS_MODE[opCode]) {
                case IXYM:
                    sb.append(" ($");
                    Util.appendHex(sb, as.getByte(address, 0));
                    sb.append("),Y");
                    break;
                case ZPGM:
                    sb.append(" $");
                    Util.appendHex(sb, as.getByte(address, 0));
                    break;
                case IMMM:
                    sb.append(" #$");
                    Util.appendHex(sb, as.getByte(address, 0));
                    break;
                case ACUM:
                    sb.append("A");
                    break;
                case ABSM:
                    sb.append(" $");
                    Util.appendHex(sb, as.getByte(address + 1, 0));
                    Util.appendHex(sb, as.getByte(address, 0));
                    break;
                case RELM:
                    sb.append(" $");
                    Util.appendHex(sb, 4, address + as.getByte(address, 0) + 1);
                    break;
                case INXM:
                    sb.append(" ($");
                    Util.appendHex(sb, as.getByte(address, 0));
                    sb.append(",X)");
                    break;
                case ZPYM:
                    sb.append(" $");
                    Util.appendHex(sb, as.getByte(address, 0));
                    sb.append(",Y");
                    break;
                case ZPXM:
                    sb.append(" $");
                    Util.appendHex(sb, as.getByte(address, 0));
                    sb.append(",X");
                    break;
                case ABYM:
                    sb.append(" $");
                    Util.appendHex(sb, as.getByte(address + 1, 0));
                    Util.appendHex(sb, as.getByte(address, 0));
                    sb.append(",Y");
                    break;
                case ABXM:
                    sb.append(" $");
                    Util.appendHex(sb, as.getByte(address + 1, 0));
                    Util.appendHex(sb, as.getByte(address, 0));
                    sb.append(",X");
                    break;
                }
            } else {
                Util.appendHex(sb, 2, opCode);
                sb.append(" ??");
            }
        }
        return 0;
    }

    private static final String MNEMONICS =
        "ADCANDASLBCCBCSBEQBITBMIBNEBPLBRKBVCBVSCLCCLDCLICLVCMPCPXCPYDECDEX"
        + "DEYEORINCINXINYJMPJSRLDALDXLDYLSRNOPORAPHAPHPPLAPLPROLRORRTIRTS"
        + "SBCSECSEDSEISTASTXSTYTAXTAYTSXTXATXSTYA";
    private static final int[] MNEMONIC_INDEX =
              {10 * 3, 34 * 3, -1, -1, -1, 34 * 3, 2 * 3, -1,
                36 * 3, 34 * 3, 2 * 3, -1, -1, 34 * 3, 2 * 3, -1,
                 9 * 3, 34 * 3, -1, -1, -1, 34 * 3, 2 * 3, -1,
                  13 * 3, 34 * 3, -1, -1, -1, 34 * 3, 2 * 3, -1,

               28 * 3, 1 * 3, -1, -1, 6 * 3, 1 * 3, 39 * 3, -1,
                38 * 3, 1 * 3, 39 * 3, -1, 6 * 3, 1 * 3, 39 * 3, -1,
                 7 * 3, 1 * 3, -1, -1, -1, 1 * 3, 39 * 3, -1,
                  44 * 3, 1 * 3, -1, -1, -1, 1 * 3, 39 * 3, -1,

               41 * 3, 23 * 3, -1, -1, -1, 23 * 3, 32 * 3, -1,
                35 * 3, 23 * 3, 32 * 3, -1, 27 * 3, 23 * 3, 32 * 3, -1,
                 11 * 3, 23 * 3, -1, -1, -1, 23 * 3, 32 * 3, -1,
                  15 * 3, 23 * 3, -1, -1, -1, 23 * 3, 32 * 3, -1,

               42 * 3, 0 * 3, -1, -1, -1, 0 * 3, 40 * 3, -1,
                37 * 3, 0 * 3, 40 * 3, -1, 27 * 3, 0 * 3, 40 * 3, -1,
                 12 * 3, 0 * 3, -1, -1, -1, 0 * 3, 40 * 3, -1,
                  46 * 3, 0 * 3, -1, -1, -1, 0 * 3, 40 * 3, -1,

               -1, 47 * 3, -1, -1, 49 * 3, 47 * 3, 48 * 3, -1,
                22 * 3, -1, 53 * 3, -1, 49 * 3, 47 * 3, 48 * 3, -1,
                 3 * 3, 47 * 3, -1, -1, 49 * 3, 47 * 3, 48 * 3, -1,
                  55 * 3, 47 * 3, 54 * 3, -1, -1, 47 * 3, -1, -1,

               31 * 3, 29 * 3, 30 * 3, -1, 31 * 3, 29 * 3, 30 * 3, -1,
                51 * 3, 29 * 3, 50 * 3, -1, 31 * 3, 29 * 3, 30 * 3, -1,
                 4 * 3, 29 * 3, -1, -1, 31 * 3, 29 * 3, 30 * 3, -1,
                  16 * 3, 29 * 3, 52 * 3, -1, 31 * 3, 29 * 3, 30 * 3, -1,

               19 * 3, 17 * 3, -1, -1, 19 * 3, 17 * 3, 20 * 3, -1,
                26 * 3, 17 * 3, 21 * 3, -1, 19 * 3, 17 * 3, 20 * 3, -1,
                 8 * 3, 17 * 3, -1, -1, -1, 17 * 3, 20 * 3, -1,
                  14 * 3, 17 * 3, -1, -1, -1, 17 * 3, 20 * 3, -1,

               18 * 3, 43 * 3, -1, -1, 18 * 3, 43 * 3, 24 * 3, -1,
                25 * 3, 43 * 3, 33 * 3, -1, 18 * 3, 43 * 3, 24 * 3, -1,
                 5 * 3, 43 * 3, -1, -1, -1, 43 * 3, 24 * 3, -1,
                  45 * 3, 43 * 3, -1, -1, -1, 43 * 3, 24 * 3, -1};

    private static final int NOAM = 0;
    private static final int IXYM = 1;
    private static final int ZPGM = 2;
    private static final int IMMM = 3;
    private static final int ACUM = 4;
    private static final int ABSM = 5;
    private static final int RELM = 6;
    private static final int INXM = 7;
    private static final int ZPXM = 8;
    private static final int ZPYM = 9;
    private static final int ABYM = 10;
    private static final int ABXM = 11;
    private static final int[] ADDRESS_MODE =
            {NOAM, INXM, NOAM, NOAM, NOAM, ZPGM, ZPGM, NOAM,
              NOAM, IMMM, ACUM, NOAM, NOAM, ABSM, ABSM, NOAM,
               RELM, IXYM, NOAM, NOAM, NOAM, ZPXM, ZPXM, NOAM,
                NOAM, ABYM, NOAM, NOAM, NOAM, ABXM, ABXM, NOAM,

             ABSM, IXYM, NOAM, NOAM, ZPGM, ZPGM, ZPGM, NOAM,
              NOAM, IMMM, ACUM, NOAM, ABSM, ABSM, ABSM, NOAM,
               RELM, IXYM, NOAM, NOAM, NOAM, ZPXM, ZPXM, NOAM,
                NOAM, ABYM, NOAM, NOAM, NOAM, ABXM, ABXM, NOAM,

             NOAM, INXM, NOAM, NOAM, NOAM, ZPGM, ZPGM, NOAM,
              NOAM, IMMM, ACUM, NOAM, ABSM, ABSM, ABSM, NOAM,
               RELM, IXYM, NOAM, NOAM, NOAM, ZPXM, ZPXM, NOAM,
                NOAM, ABYM, NOAM, NOAM, NOAM, ABXM, ABXM, NOAM,

             NOAM, INXM, NOAM, NOAM, NOAM, ZPGM, ZPGM, NOAM,
              NOAM, IMMM, ACUM, NOAM, ABSM, ABSM, ABSM, NOAM,
               RELM, IXYM, NOAM, NOAM, NOAM, ZPXM, ZPXM, NOAM,
                NOAM, ABYM, NOAM, NOAM, NOAM, ABXM, ABXM, NOAM,

             NOAM, INXM, NOAM, NOAM, ZPGM, ZPGM, ZPGM, NOAM,
              NOAM, NOAM, ACUM, NOAM, ABSM, ABSM, ABSM, NOAM,
               RELM, IXYM, NOAM, NOAM, ZPXM, ZPXM, ZPYM, NOAM,
                NOAM, ABYM, NOAM, NOAM, NOAM, ABXM, NOAM, NOAM,

             IMMM, INXM, IMMM, NOAM, ZPGM, ZPGM, ZPGM, NOAM,
              NOAM, IMMM, ACUM, NOAM, ABSM, ABSM, ABSM, NOAM,
               RELM, IXYM, NOAM, NOAM, ZPXM, ZPXM, ZPYM, NOAM,
                NOAM, ABYM, NOAM, NOAM, ABXM, ABXM, ABYM, NOAM,

             IMMM, INXM, NOAM, NOAM, ZPGM, ZPGM, ZPGM, NOAM,
              NOAM, IMMM, ACUM, NOAM, ABSM, ABSM, ABSM, NOAM,
               RELM, IXYM, NOAM, NOAM, NOAM, ZPXM, ZPXM, NOAM,
                NOAM, ABYM, NOAM, NOAM, NOAM, ABXM, ABXM, NOAM,

             IMMM, INXM, NOAM, NOAM, ZPGM, ZPGM, ZPGM, NOAM,
              NOAM, IMMM, NOAM, NOAM, ABSM, ABSM, ABSM, NOAM,
               RELM, IXYM, NOAM, NOAM, NOAM, ZPXM, ZPXM, NOAM,
                NOAM, ABYM, NOAM, NOAM, NOAM, ABXM, ABXM, NOAM};


}
