package jfrace.cpu;

import jfrace.CPU;
import jfrace.Register;
import jfrace.Util;

/**
 *  Implements an emulation of Zilog's Z80 8-bit processor.
 *  All the restrictions of <code>Intel8080</code> hold also
 *  for this Z80 implementation.
 *
 *  @author Franz-Josef Elmer
 */
public class Z80 extends Intel8080 {
    /** Index of register <code>A'</code> (accumulator). */
    public static final int As = 10;
    /** Index of register <code>F'</code> (flags). */
    public static final int Fs = 11;
    /** Index of register <code>B'</code>. */
    public static final int Bs = 12;
    /** Index of register <code>C'</code>. */
    public static final int Cs = 13;
    /** Index of register <code>D'</code>. */
    public static final int Ds = 14;
    /** Index of register <code>E'</code>. */
    public static final int Es = 15;
    /** Index of register <code>H'</code>. */
    public static final int Hs = 16;
    /** Index of register <code>L'</code>. */
    public static final int Ls = 17;
    /** Index of register <code>IX</code>. */
    public static final int IX = 18;
    /** Index of register <code>IY</code>. */
    public static final int IY = 19;
    /** Index of register <code>I</code>. */
    public static final int I = 20;
    /** Index of register <code>R</code>. */
    public static final int R = 21;

    /** Bit position of the substraction flag. */
    public static final int SUB_POS = 1;

    /** Mask for the substraction flag. */
    public static final int SUB_MASK = 1 << SUB_POS;
    private static final int RESET_SUB_MASK = ~SUB_MASK;

    private static final String[] AUX_DATA_REGISTER_NAMES
                         = {"A'", "F'", "B'", "C'", "D'", "E'", "H'", "L'"};
    private static final String[] DATA_REGISTER_NAMES2
                         = {"B", "C", "D", "E", "H", "L", "(HL)", "A"};
    private static final String[] REGISTER_PAIR_NAMES3
                         = {"BC", "DE", "HL", "SP"};
    private static final String[] REGISTER_PAIR_NAMES4
                         = {"BC", "DE", "HL", "AF"};

    private byte[] auxRegisters ;
    private byte registerI;
    private byte registerR;
    private int[] iReg;         // IX=iReg[0], IY=iReg[1]
    private int interruptMode;

    /** Show register states like CP/M DDTZ. */
    public String toString() {
        StringBuffer sb = new StringBuffer();
        int flags = (int) registers[F];
        sb.append(((flags & SIGN_MASK) == 0) ? ' ' : 'S');
        sb.append(((flags & ZERO_MASK) == 0) ? ' ' : 'Z');
        sb.append(((flags & AUX_CARRY_MASK) == 0) ? ' ' : 'H');
        sb.append(((flags & PARITY_MASK) == 0) ? ' ' : 'V');
        sb.append(((flags & SUB_MASK) == 0) ? ' ' : 'N');
        sb.append(((flags & CARRY_MASK) == 0) ? ' ' : 'C');
        sb.append("   A =");
        Util.appendHex(sb, registers[A]);
        sb.append(" BC =");
        Util.appendHex(sb, registers[B]);
        Util.appendHex(sb, registers[C]);
        sb.append(" DE =");
        Util.appendHex(sb, registers[D]);
        Util.appendHex(sb, registers[E]);
        sb.append(" HL =");
        Util.appendHex(sb, registers[H]);
        Util.appendHex(sb, registers[L]);
        sb.append(" SP=");
        Util.appendHex(sb, 4, stackPointer);
        sb.append(" PC=");
        Util.appendHex(sb, 4, programCounter);
        sb.append("       ");
        disassemble(sb, programCounter);
        sb.append('\n');
        flags = (int) auxRegisters[Fs];
        sb.append(((flags & SIGN_MASK) == 0) ? ' ' : 'S');
        sb.append(((flags & ZERO_MASK) == 0) ? ' ' : 'Z');
        sb.append(((flags & AUX_CARRY_MASK) == 0) ? ' ' : 'H');
        sb.append(((flags & PARITY_MASK) == 0) ? ' ' : 'V');
        sb.append(((flags & SUB_MASK) == 0) ? ' ' : 'N');
        sb.append(((flags & CARRY_MASK) == 0) ? ' ' : 'C');
        sb.append("   A'=");
        Util.appendHex(sb, auxRegisters[As]);
        sb.append(" BC'=");
        Util.appendHex(sb, auxRegisters[Bs]);
        Util.appendHex(sb, auxRegisters[Cs]);
        sb.append(" DE'=");
        Util.appendHex(sb, auxRegisters[Ds]);
        Util.appendHex(sb, auxRegisters[Es]);
        sb.append(" HL'=");
        Util.appendHex(sb, auxRegisters[Hs]);
        Util.appendHex(sb, auxRegisters[Ls]);
        sb.append(" IX=");
        Util.appendHex(sb, 4, iReg[0]);
        sb.append(" IY=");
        Util.appendHex(sb, 4, iReg[1]);
        sb.append(" I=");
        Util.appendHex(sb, registerI);
        return new String(sb);
    }

    /** Return "Zilog Z80". */
    public String getName() {
        return "Zilog Z80";
    }

    /** Return 22. */
    public int getNumberOfRegisters() {
        return 22;
    }

    public Register getRegister(int index) {
        if (index < 10) {
            return super.getRegister(index);
        }
        Register register = null;
        switch (index) {
        case I:
            register = new Register(8, Register.STATUS_REGISTER, "I",
                                    registerI & 0xff);
            break;
        case R:
            register = new Register(8, Register.STATUS_REGISTER, "R",
                                    registerR & 0xff);
            break;
        case IX:
            register = new Register(16, Register.ADDRESS_REGISTER,
                                    "IX", iReg[0]);
            break;
        case IY:
            register = new Register(16, Register.ADDRESS_REGISTER,
                                    "IY", iReg[1]);
            break;
        default:
            if (index >= 10 && index < 18) {
                register = new Register(8, Register.DATA_REGISTER,
                                        AUX_DATA_REGISTER_NAMES[index - 10],
                                        auxRegisters[index] & 0xff);
            }
        }
        return register;
    }

    public void setRegisterValue(int index, int value) {
        if (index < 10) {
            super.setRegisterValue(index, value);
        }
        switch (index) {
        case I: registerI = (byte) value; break;
        case R: registerR = (byte) value; break;
        case IX: iReg[0] = value & 0xffff; break;
        case IY: iReg[1] = value & 0xffff; break;
        default:
            if (index >= 10 && index < 18) {
                auxRegisters[index] = (byte) value;
            }
        }
    }

    public void reset() {
        super.reset();

        if (auxRegisters == null) {
            auxRegisters = new byte[18];
        }
        for (int i = 0; i < 18; i++) {
            auxRegisters[i] = 0;
        }
        if (iReg == null) {
            iReg = new int[2];
        }
        iReg[0] = 0;
        iReg[1] = 0;
        registerI = 0;
        registerR = 0;
        interruptMode = 0;
    }

    public int next() {
        int result = super.next();
        registerR = (byte) ((registerR + 1) & 0x7f);
        return result;
    }

    protected boolean processINRorDCR(byte opCode, int increment) {
        if (increment > 0) {
            registers[F] &= RESET_SUB_MASK;
        } else {
            registers[F] |= SUB_MASK;
        }
        return super.processINRorDCR(opCode, increment);
    }

    /**
     *  Increment a certain byte in the memory by the given amount.
     *  Set the flags except carry bit.
     *  @param adr address of the byte
     *  @param increment increment. May be negative.
     *  @retrun always <code>true</code>.
     */
    private boolean increment(int adr, int increment) {
        int flags = registers[F] & 0x29;
        flags += (increment < 0) ? SUB_MASK : 0;

        byte b = memory[adr];
        int accu = (b & 0xf) + increment;
        flags += accu & 0x10;
        accu = (b & 0xff) + increment;
        flags += accu & 0x80;
        flags += ((accu & 0xff) == 0) ? ZERO_MASK : 0;
        accu = b + increment;
        boolean signFlag = (accu & 0x80) != 0;
        flags += (signFlag ? accu > 127 : accu < -128) ? PARITY_MASK : 0;

        registers[F] = (byte) flags;
        memory[adr] = (byte) accu;

        return true;
    }


    protected void addToAccu(byte operand, int carry, int flagMask) {
        registers[F] &= RESET_SUB_MASK;
        super.addToAccu(operand, carry, flagMask);
    }

    protected void subtractFromAccu(byte operand, int carry, int flagMask) {
        registers[F] |= SUB_MASK;
        super.subtractFromAccu(operand, carry, flagMask);
    }

    private int getRegisterPair(byte opCode, int r) {
        switch ((opCode & 0x30) >> 3) {
        case 0:
            return ((registers[B] & 0xff) << 8) + (registers[C] & 0xff);
        case 2:
            return ((registers[D] & 0xff) << 8) + (registers[E] & 0xff);
        case 4:
            return iReg[r];
        case 6:
            return stackPointer;
        }
        return 0;
    }

    /**
     *  Roll cyclic left with carry bit. Set flags.
     *  @param b byte to roll.
     *  @return result.
     */
    private byte rollLeftWithCarry(int b) {
        int result = (b << 1) + ((registers[F] & CARRY_MASK) >> CARRY_POS);
        return setFlags(result);
    }

    /**
     *  Roll cyclic left without carry bit. Set flags.
     *  @param b byte to roll.
     *  @return result.
     */
    private byte rollLeft(int b) {
        int result = (b << 1) + ((b & 0x80) >> 7);
        return setFlags(result);
    }

    /**
     *  Roll cyclic right with carry bit. Set flags.
     *  @param b byte to roll.
     *  @return result.
     */
    private byte rollRightWithCarry(int b) {
        b = (b & 0xff) + ((registers[F] & CARRY_MASK) << (8 - CARRY_POS));
        int result = (b >> 1) + ((b & 1) << 8);
        return setFlags(result);
    }

    /**
     *  Roll cyclic Right without carry bit. Set flags.
     *  @param b byte to roll.
     *  @return result.
     */
    private byte rollRight(int b) {
        int result = ((b & 0xff) >> 1) + (((b & 1) * 3) << 7);
        return setFlags(result);
    }

    /**
     *  Shift right arithmetic. Set flags.
     *  @param b byte to shift.
     *  @return result.
     */
    private byte shiftRightArithmetic(int b) {
        int result = ((b >> 1) & 0xff) + ((b & 1) << 8);
        return setFlags(result);
    }

    /**
     *  Shift right logic. Set flags.
     *  @param b byte to shift.
     *  @return result.
     */
    private byte shiftRightLogic(int b) {
        int result = ((b & 0xff) >> 1) + ((b & 1) << 8);
        return setFlags(result);
    }

    /**
     *  Set carry, sign, parity, and zero flag according to
     *  <code>result</code>. Reset auxiliary carry
     *  flag and substraction flag.
     *  @return <code>result</code> casted as <code>byte</code>.
     */
    private byte setFlags(int result) {
        int flags = registers[F] & 0x28;
        flags += (result & 0x100) >> (8 - CARRY_POS);
        flags += (result & 0x80) << (SIGN_POS - 7);
        flags += ((result & 0xff) == 0) ? ZERO_MASK : 0;
        flags += calculateParityFlag(result);
        registers[F] = (byte) flags;
        return (byte) result;
    }

    private void testBit(int bitPos, int b) {
        int flags = (registers[F] & 0x29) + 0x50;
        flags ^= ((b >> bitPos) & 1) << ZERO_POS;
        registers[F] = (byte) flags;
    }

    private boolean processLDIorLDD(int increment) {
        int bc = ((registers[B] & 0xff) << 8) + (registers[C] & 0xff);
        int de = ((registers[D] & 0xff) << 8) + (registers[E] & 0xff);
        int hl = ((registers[H] & 0xff) << 8) + (registers[L] & 0xff);
        memory[de] = memory[hl];
        de = (de + increment) & 0xffff;
        hl = (hl + increment) & 0xffff;
        bc = (bc - 1) & 0xffff;
        int flags = registers[F] & 0xe9;
        flags += (bc == 0) ? 0 : PARITY_MASK;
        registers[F] = (byte) flags;
        return saveBCDEHL(bc, de, hl);
    }

    private boolean processLDIRorLDDR(int increment) {
        int bc = ((registers[B] & 0xff) << 8) + (registers[C] & 0xff);
        int de = ((registers[D] & 0xff) << 8) + (registers[E] & 0xff);
        int hl = ((registers[H] & 0xff) << 8) + (registers[L] & 0xff);
        do {
            memory[de] = memory[hl];
            de = (de + increment) & 0xffff;
            hl = (hl + increment) & 0xffff;
            bc = (bc - 1) & 0xffff;
        } while (bc != 0);
        registers[F] &= 0xe9;
        return saveBCDEHL(bc, de, hl);
    }

    private boolean saveBCDEHL(int bc, int de, int hl) {
        registers[B] = (byte) (bc >> 8);
        registers[C] = (byte) bc;
        registers[D] = (byte) (de >> 8);
        registers[E] = (byte) de;
        registers[H] = (byte) (hl >> 8);
        registers[L] = (byte) hl;
        return true;
    }

    private boolean processCPIorCPD(int increment) {
        int bc = ((registers[B] & 0xff) << 8) + (registers[C] & 0xff);
        int hl = ((registers[H] & 0xff) << 8) + (registers[L] & 0xff);
        byte a, m;
        a = registers[A];
        m = memory[hl];
        hl = (hl + increment) & 0xffff;
        bc = (bc - 1) & 0xffff;
        return setFlagsAndSaveBCHL(bc, hl, a, m);
    }

    private boolean processCPIRorCPDR(int increment) {
        int bc = ((registers[B] & 0xff) << 8) + (registers[C] & 0xff);
        int hl = ((registers[H] & 0xff) << 8) + (registers[L] & 0xff);
        byte a, m;
        do {
            a = registers[A];
            m = memory[hl];
            hl = (hl + increment) & 0xffff;
            bc = (bc - 1) & 0xffff;
        } while (a != m && bc != 0);
        return setFlagsAndSaveBCHL(bc, hl, a, m);
    }

    private boolean setFlagsAndSaveBCHL(int bc, int hl, byte a, byte m) {
        int flags = (registers[F] & 0x29) + SUB_MASK;
        int accu = (a & 0xf) - (m & 0xf);
        flags += accu & 0x10;
        accu = a - m;
        flags += accu & 0x80;
        flags += (bc == 0) ? 0 : PARITY_MASK;
        flags += (registers[A] == memory[hl]) ? ZERO_MASK : 0;
        flags += ((accu & 0xff) == 0) ? ZERO_MASK : 0;
        registers[F] = (byte) flags;

        registers[B] = (byte) (bc >> 8);
        registers[C] = (byte) bc;
        registers[H] = (byte) (hl >> 8);
        registers[L] = (byte) hl;

        return true;
    }

    private boolean processRRD() {
        int hl = ((registers[H] & 0xff) << 8) + (registers[L] & 0xff);
        byte a = registers[A];
        byte m = memory[hl];
        memory[hl] = (byte) (((m >> 4) & 0xf) + ((a << 4) & 0xf0));
        registers[A] = (byte) ((a & 0xf0) + (m & 0xf));
        return setFlagsAfterRollDigits(registers[A]);
    }

    private boolean processRLD() {
        int hl = ((registers[H] & 0xff) << 8) + (registers[L] & 0xff);
        byte a = registers[A];
        byte m = memory[hl];
        memory[hl] = (byte) (((m << 4) & 0xf0) + (a & 0xf));
        registers[A] = (byte) ((a & 0xf0) + ((m >> 4) & 0xf));
        return setFlagsAfterRollDigits(registers[A]);
    }

    private boolean setFlagsAfterRollDigits(byte accu) {
        int flags = registers[F] & 0x29;
        flags += accu & 0x80;
        flags += (accu == 0) ? ZERO_MASK : 0;
        flags += calculateParityFlag((int) accu);
        registers[F] = (byte) flags;
        return true;
    }

    private boolean setFlagsAndHL(int hl) {
        registers[H] = (byte) (hl >> 8);
        registers[L] = (byte) hl;
        int flags = registers[F] & 0x2a;
        flags += (hl & 0x10000) >> 16;
        flags += ((hl & 0xffff) == 0) ? ZERO_MASK : 0;
        boolean signFlag = (hl & 0x8000) != 0;
        flags += signFlag ? SIGN_MASK : 0;
        flags += (signFlag ? hl > Short.MAX_VALUE
                           : hl < Short.MIN_VALUE) ? PARITY_MASK : 0;
        registers[F] = (byte) flags;
        return true;
    }

    /**
     *  Process an op-code which is undefined for Intel 8080.
     */
    protected boolean processUDEF(byte opCode) {
        switch (opCode & 0xff) {
        case 0x08:              // EX AF,AF'
            byte b = registers[A];
            registers[A] = auxRegisters[As];
            auxRegisters[As] = b;
            b = registers[F];
            registers[F] = auxRegisters[Fs];
            auxRegisters[Fs] = b;
            ++programCounter;
            return true;
        case 0x10:              // DJNZ
            programCounter += 2;
            if (--registers[B] != 0) {
                programCounter += memory[programCounter - 1];
            }
            return true;
        case 0x18:              // JR
            programCounter += memory[programCounter + 1] + 2;
            return true;
        case 0x20:              // JR NZ
            programCounter += 2;
            if ((registers[F] & ZERO_MASK) == 0) {
                programCounter += memory[programCounter - 1];
            }
            return true;
        case 0x28:              // JR Z
            programCounter += 2;
            if ((registers[F] & ZERO_MASK) != 0) {
                programCounter += memory[programCounter - 1];
            }
            return true;
        case 0x30:              // JR NC
            programCounter += 2;
            if ((registers[F] & CARRY_MASK) == 0) {
                programCounter += memory[programCounter - 1];
            }
            return true;
        case 0x38:              // JR C
            programCounter += 2;
            if ((registers[F] & CARRY_MASK) != 0) {
                programCounter += memory[programCounter - 1];
            }
            return true;
        case 0xcb:              // shift and bit commands
            return processBitCommands(memory[++programCounter], 0, false);
        case 0xd9:              // EXX
            System.arraycopy(registers, B, auxRegisters, 0, 6);
            System.arraycopy(auxRegisters, Bs, registers, B, 6);
            System.arraycopy(auxRegisters, 0, auxRegisters, Bs, 6);
            ++programCounter;
            return true;
        case 0xdd:              // IX commands
            return processIndexRegCommands(memory[++programCounter], 0);
        case 0xfd:              // IY commands
            return processIndexRegCommands(memory[++programCounter], 1);
        case 0xed:
            return processMiscCommand(memory[++programCounter]);
        default:
            errorDescription = "undefined op-code: "
                               + Integer.toHexString(((int) opCode) & 0xff);
        }
        return false;
    }

    private boolean processMiscCommand(byte opCode) {
        ++programCounter;
        switch (opCode & 0xff) {
        case 0104:      // NEG
            byte a = registers[A];
            registers[A] = 0;
            subtractFromAccu(a, 0, ADD_MASK);
            return true;
        case 0105:      // RETN
        case 0115:      // RETI
            return false;
        case 0106:      // IM 0
            interruptMode = 0;
            return true;
        case 0107:      // LD I,A
            registerI = registers[A];
            return true;
        case 0117:      // LD R,A
            registerR = (byte) (registers[A] & 0x7f);
            return true;
        case 0126:      // IM 1
            interruptMode = 1;
            return true;
        case 0136:      // IM 2
            interruptMode = 2;
            return true;
        case 0127:      // LD A,I
        case 0137:      // LD A,R
            registers[A] = (opCode == 0127) ? registerI : registerR;
            int flags = registers[F] & 1;
            flags += interruptEnabled ? 4 : 0;
            flags += registers[A] & 0x80;
            flags += (registers[A] == 0) ? ZERO_MASK : 0;
            return true;
        case 0147:      // RRD
            return processRRD();
        case 0157:      // RLD
            return processRLD();
        case 0240:      // LDI
            return processLDIorLDD(1);
        case 0250:      // LDD
            return processLDIorLDD(-1);
        case 0260:      // LDIR
            return processLDIRorLDDR(1);
        case 0270:      // LDDR
            return processLDIRorLDDR(-1);
        case 0241:      // CPI
            return processCPIorCPD(1);
        case 0251:      // CPD
            return processCPIorCPD(-1);
        case 0261:      // CPIR
            return processCPIRorCPDR(1);
        case 0271:      // CPDR
            return processCPIRorCPDR(-1);
        default:
            switch (opCode & 0307) {
            case 0100:      // IN r,(C)
                int port = registers[C] & 0xff;
                a = registers[A];
                try {
                    setAccuAndFlags((byte) as.getValue(portBiasAddress + port,
                                                       0),
                                    0);
                    if ((opCode & 070) != 060) {
                        storeByte((opCode & 070) >> 3, registers[A]);
                    }
                    registers[A] = a;
                    return true;
                } catch (Exception e) {
                    errorDescription = e.toString();
                    return false;
                }
            case 0101:      // OUT (C),r
                int word = (readByte((opCode & 070) >> 3) & 0xff)
                           + ((registers[B] & 0xff) << 8);
                try {
                    as.setValue(portBiasAddress + (registers[C] & 0xff),
                                0, word);
                    return true;
                } catch (Exception e) {
                    errorDescription = e.toString();
                    return false;
                }
            }
            switch (opCode & 0317) {
            case 0103:      // LD (nn),
                int adr = (memory[programCounter++] & 0xff)
                          + ((memory[programCounter++] & 0xff) << 8);
                int rp = getRegisterPair(opCode, false);
                memory[adr++] = (byte) rp;
                memory[adr] = (byte) (rp >> 8);
                return true;
            case 0113:      // LD ,(nn)
                adr = (memory[programCounter++] & 0xff)
                      + ((memory[programCounter++] & 0xff) << 8);
                int value = (memory[adr++] & 0xff)
                            + ((memory[adr] & 0xff) << 8);
                setRegisterPair(opCode, value, false);
                return true;
            case 0112:      // ADC HL
                int accu = ((registers[H] & 0xff) << 8) + (registers[L] & 0xff);
                return setFlagsAndHL(accu + getRegisterPair(opCode, false)
                                     + (registers[F] & 1));
            case 0102:      // SBC HL
                accu = ((registers[H] & 0xff) << 8) + (registers[L] & 0xff);
                return setFlagsAndHL(accu - getRegisterPair(opCode, false)
                                     - (registers[F] & 1));
            }
        }
        return false;
    }

    /**
     *  Process rotate, shift, and bit commands of the <code>0xCB</code>
     *  block.
     *  @param opCode second op-code. Program pointer points on it.
     *  @param adr absolut memory address if <code>isIReg == true</code>.
     *             Otherwise not used.
     */
    private boolean processBitCommands(byte opCode, int adr, boolean isIReg) {
        ++programCounter;
        int b = 0;
        if (isIReg) {
            b = memory[adr];
        } else {
            adr = opCode & 7;
            b = readByte(adr);
        }
        byte result = 0;
        switch (opCode & 0370) {
        case 0000: result = rollLeft(b); break;
        case 0010: result = rollRight(b); break;
        case 0020: result = rollLeftWithCarry(b); break;
        case 0030: result = rollRightWithCarry(b); break;
        case 0040: result = setFlags(b << 1); break;
        case 0050: result = shiftRightArithmetic(b); break;
        case 0070: result = shiftRightLogic(b); break;
        default:
            int bitPos = (opCode & 070) >> 3;
            switch (opCode & 0300) {
            case 0100: testBit(bitPos, b); return true;
            case 0200: result = (byte) (b & (~(1 << bitPos))); break;
            case 0300: result = (byte) (b | (1 << bitPos)); break;
            default: return false;
            }
        }
        if (isIReg) {
            memory[adr] = result;
        } else {
            storeByte(adr, result);
        }
        return true;
    }

    /**
     *  Process commands involving one of the index registers.
     *  @param opCode second op-code. The program counter is pointing
     *                onto it.
     *  @param r index register. 0 for <code>IX</code> and 1 for
     *           <code>IY</code>.
     *  @return <code>true</code> if op-code exists.
     */
    private boolean processIndexRegCommands(byte opCode, int r) {
        ++programCounter;
        switch (opCode & 0300) {
        case 0000:          // 16-bit load and arithmetic operations
            switch (opCode & 0xff) {
            case 0x21:      // Load index register immediately
                iReg[r] = (memory[programCounter++] & 0xff)
                          + ((memory[programCounter++] & 0xff) << 8);
                return true;
            case 0x22:      // Store index
                int adr = (memory[programCounter++] & 0xff)
                           + ((memory[programCounter++] & 0xff) << 8);
                memory[adr] = (byte) iReg[r];
                memory[adr + 1] = (byte) (iReg[r] >> 8);
                return true;
            case 0x23:      // increment index register
                iReg[r] = (iReg[r] + 1) & 0xffff;
                return true;
            case 0x2a:      // Load index register indirect
                iReg[r] = getWord((memory[programCounter++] & 0xff)
                                  + ((memory[programCounter++] & 0xff) << 8));
                return true;
            case 0x2b:      // decrement index register
                iReg[r] = (iReg[r] - 1) & 0xffff;
                return true;
            case 0x34:      // 8-bit increment indirect
                return increment(iReg[r] + memory[programCounter++], 1);
            case 0x35:      // 8-bit decrement indirect
                return increment(iReg[r] + memory[programCounter++], -1);
            case 0x36:      // 8-bit store indirect
                adr = iReg[r] + memory[programCounter++];
                memory[adr] = memory[programCounter++];
                return true;
            default:
                switch (opCode & 0xcf) {
                case 0x09:      // ADD
                    iReg[r] += getRegisterPair(opCode, r);
                    int flags = registers[F] & 0xec;
                    flags += (iReg[r] & 0x10000) >> 16;
                    registers[F] = (byte) flags;
                    iReg[r] &= 0xffff;
                    return true;
                }

            }
            return true;
        case 0100:          // 8-bit LD commands
            int adr = iReg[r] + memory[programCounter++];
            int a1 = opCode & 7;
            int a2 = (opCode & 070) >> 3;
            if (a1 == 6) { // load register
                if (a2 == 6) { // HALT
                    return processHLT();
                }
                storeByte(a2, memory[adr]);
            } else {                 // store register
                memory[adr] = readByte(a1);
            }
            return true;
        case 0200:          // 8-bit arithmetic and logic operations
            adr = iReg[r] + memory[programCounter++];
            switch (opCode & 0xff) {
            case 0206:      // ADD
                addToAccu(memory[adr], 0, ADD_MASK);
                return true;
            case 0216:      // ADC
                addToAccu(memory[adr], registers[F] & 1, ADD_MASK);
                return true;
            case 0226:      // SUB
                subtractFromAccu(memory[adr], 0, ADD_MASK);
                return true;
            case 0236:      // SBC
                subtractFromAccu(memory[adr], registers[F] & 1, ADD_MASK);
                return true;
            case 0246:      // AND
                setAccuAndFlags(memory[adr] & registers[A], AUX_CARRY_MASK);
                return true;
            case 0256:      // XOR
                setAccuAndFlags(memory[adr] ^ registers[A], 0);
                return true;
            case 0266:      // OR
                setAccuAndFlags(memory[adr] | registers[A], 0);
                return true;
            case 0276:      // CP
                byte accu = registers[A];
                subtractFromAccu(memory[adr], 0, ADD_MASK);
                registers[A] = accu;
                return true;
            default: return false;
            }
        case 0300:          // misc. and bit commands
            switch (opCode & 0xff) {
            case 0xe1:      // POP
                iReg[r] = popFromStack();
                return true;
            case 0xe3:      // EX (SP)
                int memSP = ((memory[stackPointer + 1] & 0xff) << 8)
                            + (memory[stackPointer] & 0xff);
                memory[stackPointer + 1] = (byte) (iReg[r] >> 8);
                memory[stackPointer] = (byte) iReg[r];
                iReg[r] = memSP;
                return true;
            case 0xe5:      // PUSH
                pushOnStack(iReg[r]);
                return true;
            case 0xe9:      // JP
                programCounter = iReg[r];
                return true;
            case 0xf9:      // LD SP
                stackPointer = iReg[r];
                return true;
            case 0xcb:      // Rotate, shift, and bit operations
                adr = iReg[r] + memory[programCounter++];
                return processBitCommands(memory[programCounter],
                                          adr, true);
            }
        }
        return false;
    }

    protected int disassembleUDEF(byte opCode, StringBuffer sb, int address) {
        switch (opCode & 0xff) {
        case 0x08:
            sb.append("EX   AF,AF'");
            return 1;
        case 0x10:
            sb.append("DJNZ ");
            Util.appendHex(sb, 4, address + memory[address + 1] + 2);
            return 2;
        case 0x18:
            sb.append("JR   ");
            Util.appendHex(sb, 4, address + memory[address + 1] + 2);
            return 2;
        case 0x20:
        case 0x28:
        case 0x30:
        case 0x38:
            sb.append("JR   ");
            sb.append(CONDITION_NAMES[(opCode & 030) >> 3]);
            sb.append(',');
            Util.appendHex(sb, 4, address + memory[address + 1] + 2);
            return 2;
        case 0xcb:
            opCode = memory[address + 1];
            switch (opCode & 0370) {
            case 0000: sb.append("RLC  "); break;
            case 0010: sb.append("RRC  "); break;
            case 0020: sb.append("RL   "); break;
            case 0030: sb.append("RR   "); break;
            case 0040: sb.append("SLA  "); break;
            case 0050: sb.append("SRA  "); break;
            case 0070: sb.append("SRL  "); break;
            default:
                switch (opCode & 0300) {
                case 0100: sb.append("BIT  "); break;
                case 0200: sb.append("RES  "); break;
                case 0300: sb.append("SET  "); break;
                default: return super.disassembleUDEF(opCode, sb, address);
                }
                sb.append((opCode & 070) >> 3);
                sb.append(',');
            }
            sb.append(DATA_REGISTER_NAMES2[opCode & 7]);
            return 2;
        case 0xd9:              // EXX
            sb.append("EXX");
            return 1;
        case 0xdd:
        case 0xed:
        case 0xfd:
        default:
            sb.append(Integer.toHexString(opCode & 0xff) + " ??");
        }
        return 0;
    }

    /**
     *  Return disassemble instruction in standard Zilog Z80
     *  assembler syntax.
     */
    public int disassemble(StringBuffer sb, int address) {
        try {
            byte opCode = memory[address];
            switch (OP_CODE_TYPE[((int) opCode) & 0xff]) {
            case UDEF: return disassembleUDEF(opCode, sb, address);
            case ACI:
                sb.append("ADC  A,");
                return disassembleByte(sb, address);
            case ADC:
                sb.append("ADC  A,");
                sb.append(DATA_REGISTER_NAMES2[opCode & 7]);
                return 1;
            case ADD:
                sb.append("ADD  A,");
                sb.append(DATA_REGISTER_NAMES2[opCode & 7]);
                return 1;
            case ADI:
                sb.append("ADD  A,");
                return disassembleByte(sb, address);
            case ANA:
                sb.append("AND  ");
                sb.append(DATA_REGISTER_NAMES2[opCode & 7]);
                return 1;
            case ANI:
                sb.append("AND  ");
                return disassembleByte(sb, address);
            case CALL:
                sb.append("CALL ");
                return disassembleWord(sb, address);
            case Cc:
                sb.append("CALL ");
                sb.append(CONDITION_NAMES[(opCode & 070) >> 3]);
                sb.append(',');
                return disassembleWord(sb, address);
            case CMA:  sb.append("CPL"); return 1;
            case CMC:  sb.append("CCF"); return 1;
            case CMP:
                sb.append("CP   ");
                sb.append(DATA_REGISTER_NAMES2[opCode & 7]);
                return 1;
            case CPI:
                sb.append("CP   ");
                return disassembleByte(sb, address);
            case DAA:   sb.append("DAA"); return 1;
            case DAD:
                sb.append("ADD  HL,");
                sb.append(REGISTER_PAIR_NAMES3[(opCode & 0x30) >> 4]);
                return 1;
            case DCR:
                sb.append("DEC  ");
                sb.append(DATA_REGISTER_NAMES2[(opCode & 070) >> 3]);
                return 1;
            case DCX:
                sb.append("DEC  ");
                sb.append(REGISTER_PAIR_NAMES3[(opCode & 0x30) >> 4]);
                return 1;
            case DI:   sb.append("DI"); return 1;
            case EI:   sb.append("EI"); return 1;
            case HLT:  sb.append("HALT"); return 0;
            case IN:
                sb.append("IN   A,(");
                disassembleByte(sb, address);
                sb.append(')');
                return 2;
            case INR:
                sb.append("INC  ");
                sb.append(DATA_REGISTER_NAMES2[(opCode & 070) >> 3]);
                return 1;
            case INX:
                sb.append("INC  ");
                sb.append(REGISTER_PAIR_NAMES3[(opCode & 0x30) >> 4]);
                return 1;
            case Jc:
                sb.append("JP   ");
                sb.append(CONDITION_NAMES[(opCode & 070) >> 3]);
                sb.append(',');
                return disassembleWord(sb, address);
            case JMP:
                sb.append("JP   ");
                return disassembleWord(sb, address);
            case LDA:
                sb.append("LD   A,(");
                disassembleWord(sb, address);
                sb.append(')');
                return 3;
            case LDAX:
                sb.append("LD   A,(");
                sb.append(REGISTER_PAIR_NAMES4[(opCode & 0x30) >> 4]);
                sb.append(')');
                return 1;
            case LHLD:
                sb.append("LD   HL,(");
                disassembleWord(sb, address);
                sb.append(')');
                return 3;
            case LXI:
                sb.append("LD   ");
                sb.append(REGISTER_PAIR_NAMES3[(opCode & 0x30) >> 4]);
                sb.append(',');
                Util.appendHex(sb, 4, (long) getWord(address + 1));
                return 3;
            case MOV:
                sb.append("LD   ");
                sb.append(DATA_REGISTER_NAMES2[(opCode & 070) >> 3]);
                sb.append(',');
                sb.append(DATA_REGISTER_NAMES2[opCode & 7]);
                return 1;
            case MVI:
                sb.append("LD   ");
                sb.append(DATA_REGISTER_NAMES2[(opCode & 070) >> 3]);
                sb.append(',');
                Util.appendHex(sb, memory[address + 1]);
                return 2;
            case NOP:  sb.append("NOP"); return 1;
            case ORA:
                sb.append("OR   ");
                sb.append(DATA_REGISTER_NAMES2[opCode & 7]);
                return 1;
            case ORI:
                sb.append("OR   ");
                return disassembleByte(sb, address);
            case OUT:
                sb.append("OUT  (");
                disassembleByte(sb, address);
                sb.append("),A");
                return 2;
            case PCHL: sb.append("JP   (HL)"); return 1;
            case POP:
                sb.append("POP  ");
                sb.append(REGISTER_PAIR_NAMES4[(opCode & 0x30) >> 4]);
                return 1;
            case PUSH:
                sb.append("PUSH ");
                sb.append(REGISTER_PAIR_NAMES4[(opCode & 0x30) >> 4]);
                return 1;
            case RAL:  sb.append("RLA"); return 1;
            case RAR:  sb.append("RRA"); return 1;
            case Rc:
                sb.append("RET   ");
                sb.append(CONDITION_NAMES[(opCode & 070) >> 3]);
                return 1;
            case RET:  sb.append("RET"); return 1;
            case RLC:  sb.append("RLCA"); return 1;
            case RRC:  sb.append("RRCA"); return 1;
            case RST:
                sb.append("RST  ");
                sb.append(opCode & 070);
                return 1;
            case SBB:
                sb.append("SBC  ");
                sb.append(DATA_REGISTER_NAMES2[opCode & 7]);
                return 1;
            case SBI:
                sb.append("SBC  ");
                return disassembleByte(sb, address);
            case SHLD:
                sb.append("LD   (");
                disassembleWord(sb, address);
                sb.append("),HL");
                return 3;
            case SPHL: sb.append("LD   SP,HL"); return 1;
            case STA:
                sb.append("LD  (");
                disassembleWord(sb, address);
                sb.append("),A");
                return 3;
            case STAX:
                sb.append("LD   (");
                sb.append(REGISTER_PAIR_NAMES3[(opCode & 0x30) >> 4]);
                sb.append("),A");
                return 1;
            case STC:  sb.append("SCF"); return 1;
            case SUB:
                sb.append("SUB  ");
                sb.append(DATA_REGISTER_NAMES2[opCode & 7]);
                return 1;
            case SUI:
                sb.append("SUB  ");
                return disassembleByte(sb, address);
            case XCHG: sb.append("EX   DE,HL"); return 1;
            case XRA:
                sb.append("XOR  ");
                sb.append(DATA_REGISTER_NAMES2[opCode & 7]);
                return 1;
            case XRI:
                sb.append("XOR  ");
                return disassembleByte(sb, address);
            case XTHL: sb.append("XTHL"); return 1;
            }
            return 0;
        } catch (Exception e) {
            errorDescription = e.toString();
            return 0;
        }
    }
}
