package jfrace.device;

import jfrace.*;

import java.awt.AWTEventMulticaster;
import java.awt.event.*;
import java.io.*;
import java.net.URL;
import java.net.URLConnection;
import java.util.Vector;

/**
 *  Class which provides a file as a disk for the CP/M emulator.
 *
 *  @author Franz-Josef Elmer
 */
public class CPMDisk {
    /** Type of action event for reading data. */
    public static final int READ = ActionEvent.ACTION_LAST + 1;

    /** Type of action event for writing data. */
    public static final int WRITE = READ + 1;

    private static final byte EMPTY = (byte) 0xe5;
    private static final byte SPACE = (byte) ' ';
    private static final byte JMP_CMD = (byte) 0xc3;
    private static final int DIR_ENTRY = 0;
    private static final int EMPTY_ENTRY = 1;
    private static final int NO_DIR_ENTRY = 2;
    private static final int CCP_START = 0x35c;
    private static final int CCP_BDOS_SIZE = 44;
    private static final String COPYRIGHT = "COPYRIGHT";
    private static final int CCP_COPYRIGHT = 0x18;


    private String diskName;
    private byte[] data;
    private ActionListener actionListener;
    private boolean dataChanged = false;
    private boolean filesCanBeAdded = false;
    private int freeDirEntry = 0;
    private int freeBlock = 1;

    private Vector allocatedBlocks;
    private int minAllocBlock;
    private int maxAllocBlock;
    private int blockSize;
    private int blockShiftFactor;
    private int sysTracks;
    private int numberOfSectorsPerTrack;
    private int numberOfDirEntries;
    private int numberOfBlocks;
    private int physicalSectorSize; // in units of the logical sector size
    private int[] translationTable;

    /**
     *  Create a disk from the data of the file <code>diskName</code>.
     *  The raw data of the disk will be read from the given file.
     *  Changed data will be written back to the file in the
     *  <a href="#finalize()"><code>finalize</code></a> method.
     *  @exception IOException in the case of a reading error.
     */
    public CPMDisk(String diskName) throws IOException {
        this.diskName = diskName;
        InputStream is = new FileInputStream(diskName);
        data = new byte[is.available()];
        is.read(data);
        is.close();
        if (!calculateDiskParameters()) {
            System.out.println("Invalid CPM disk.");
            throw new IOException("Invalid CPM disk.");
        }
    }

    /**
     *  Create a disk from the data of an URL source.
     *  The raw data of the disk will be read from the source.
     *  Changed data will not be written back.
     *  @param source URL source of the raw data.
     *  @exception IOException in the case of a reading error.
     *  @see finalize
     */
    public CPMDisk(URL source) throws IOException {
        try {
            data = Util.readBinaryData(source);
            if (!calculateDiskParameters()) {
                System.out.println("Invalid CPM disk.");
                throw new IOException("Invalid CPM disk.");
            }
        } catch (IOException ioe) {
            throw ioe;
        } catch (Throwable t) {
            throw new IOException("Caused by " + t);
        }
    }

    /**
     *  Create an empty disk. Empty disks are used to
     *  make files of the real disk system (i.e. not the CP/M
     *  disk system) available for the CP/M emulator.
     *  Changes made by the emulator will not be made persistent.
     *  @param size Number of bytes of the disk.
     */
    public CPMDisk(int size) {
        data = new byte[size];
        for (int i = 0; i < size; i++) {
            data[i] = EMPTY;
        }
        filesCanBeAdded = true;

        numberOfSectorsPerTrack = 40;
        blockShiftFactor = 4;
        blockSize = 2048;
        numberOfBlocks = size / blockSize;
        numberOfDirEntries = blockSize / 32;
        sysTracks = 0;
        physicalSectorSize = 1;
        calculateTranslationTable(1);
    }

    /** Return the number of sectors per track. */
    public int getNumberOfSectorsPerTrack() {
        return numberOfSectorsPerTrack;
    }

    /** Return the number of system tracks. */
    public int getNumberOfSystemTracks() {
        return sysTracks;
    }

    /** Return the number of tracks. */
    public int getNumberOfTracks() {
        return sysTracks + (numberOfBlocks * blockSize)
                               / (numberOfSectorsPerTrack * 128);
    }

    /**
     *  Calculate the disk parameters by analyzing the raw data of
     *  the disk.
     *  @return <code>true</code> if successful
     */
    public boolean calculateDiskParameters() {
        // search for the first directory entry
        int len = Math.min(16384, data.length - 128);
        int dirAddress = 0;
        while (checkTypeOfEntry(dirAddress) != DIR_ENTRY && dirAddress < len) {
            dirAddress += 128;
        }
        if (dirAddress >= len) {
            System.out.println("No directory found");
            return false;
        }

        // estimate number of system tracks by assuming no more than 3
        sysTracks = 0;
        numberOfSectorsPerTrack = 0;
        if (dirAddress > 0) {
            if (dirAddress % 3 == 0 && data.length % (dirAddress / 3) == 0) {
                sysTracks = 3;
            } else if ((dirAddress / 128) % 2 == 0
                       && data.length % (dirAddress / 2) == 0) {
                sysTracks = 2;
            } else {
                sysTracks = 1;
            }
            numberOfSectorsPerTrack = dirAddress / (128 * sysTracks);
        }

        // Collect all directory entries. As a side effect we get the blocksize
        minAllocBlock = 256;
        maxAllocBlock = 0;
        getAllocatedBlocks().setSize(0);
        Vector entries = collectEntriesOfType(dirAddress, len, DIR_ENTRY);

        // Get the maximum number of blocks.
        numberOfBlocks = Math.max(maxAllocBlock + 1,
                                  (data.length - dirAddress) / blockSize);

        // Get the number of directory entries.
        // Uses the minAllocBlock and assumes that all
        // block with lower indices are used for the directory
        numberOfDirEntries = (minAllocBlock * blockSize) / 32;

        physicalSectorSize = 1;
        int skewFactor = 1;
        int[] group = (int[]) entries.elementAt(0);
        if (entries.size() > 1) {
            physicalSectorSize = group[1] / 4;
            skewFactor = group[0];
            group = (int[]) entries.elementAt(1);
            skewFactor = (group[0] - skewFactor) / (128 * physicalSectorSize);
        } else {
            entries = collectEntriesOfType(dirAddress, len, EMPTY_ENTRY);
            if (entries.size() > 1) {
                group = (int[]) entries.elementAt(1);
                physicalSectorSize = group[1] / 4;
                skewFactor = group[0];
                group = (int[]) entries.elementAt(0);
                skewFactor = (skewFactor - group[0] - group[1] * 32)
                             / (128 * physicalSectorSize) + 1;
            }
        }

        // if no system tracks assume 10 physical sectors per track
        if (sysTracks == 0) {
            numberOfSectorsPerTrack = physicalSectorSize * 10;
        }

        // calculate translation table for physical sectors
        calculateTranslationTable(skewFactor);

/*
        System.out.print(sysTracks + " st, "+numberOfSectorsPerTrack+" secs");
        System.out.print(", bs="+blockSize);
        System.out.print(", "+numberOfBlocks);
        System.out.print(", "+numberOfDirEntries);
        System.out.print(", physSize="+physicalSectorSize+", skewFactor="+skewFactor);
        System.out.println();
*/
        return true;
    }

    /**
     *  Set disk parameters.
     *  @param as Address space.
     *  @param address Start address in the address space.
     */
    public void setDiskParameters(AddressSpace as, int address) {
//        System.out.println("setDiskParameters "+Integer.toHexString(address));
        as.setByte(address++, 0, (byte) numberOfSectorsPerTrack);
        as.setByte(address++, 0, (byte) (numberOfSectorsPerTrack >> 8));
        as.setByte(address++, 0, (byte) blockShiftFactor);
        as.setByte(address++, 0, (byte) ((blockSize / 128) - 1));
        as.setByte(address++, 0, (byte) ((blockSize / 1024) - 1));
        as.setByte(address++, 0, (byte) (numberOfBlocks - 1));
        as.setByte(address++, 0, (byte) ((numberOfBlocks - 1) >> 8));
        as.setByte(address++, 0, (byte) (numberOfDirEntries - 1));
        as.setByte(address++, 0, (byte) ((numberOfDirEntries - 1) >> 8));
        as.setByte(address++, 0, (byte) 0x80);
        as.setByte(address++, 0, (byte) 0);
        as.setByte(address++, 0, (byte) numberOfDirEntries);
        as.setByte(address++, 0, (byte) (numberOfDirEntries >> 8));
        as.setByte(address++, 0, (byte) sysTracks);
        as.setByte(address++, 0, (byte) (sysTracks >> 8));
    }

    /**
     *  Translate a logical sector number into a corresponding
     *  logical sector number on the disk. This method is needed
     *  for implementing the BIOS method <code>SECTRAN</code>.
     *  @param sector Sector number from <code>[0,
     *                numberOfSectorsPerTrack)</code>.
     *  @return translated sector number from <code>[1,
     *          numberOfSectorsPerTrack]</code>.
     */
    public int translateSector(int sector) {
        int newSector = translationTable[sector / physicalSectorSize];
        newSector *= physicalSectorSize;
        newSector += sector % physicalSectorSize + 1;
        return newSector;
    }

    private void calculateTranslationTable(int skewFactor) {
        translationTable = new int[numberOfSectorsPerTrack
                                   / physicalSectorSize];
        int index = 0;
        int index0 = 0;
        for (int i = 0; i < translationTable.length; i++) {
            translationTable[i] = index;
//            System.out.print(index+" ");
            index += skewFactor;
            if (index >= translationTable.length) {
                index = ++index0;
            }
        }
    }

    /**
     *  Collect all entries of a given type.
     *  @param address Start address.
     *  @param maxAddress Address of the last entry which will be tested.
     *  @param type Type of the entries.
     *  @return List of two-dimensional integer arrays. The first number is
     *          the address of an entry and the second number counts the
     *          number of consecutive entries of the given type.
     */
    private Vector collectEntriesOfType(int address, int maxAddress,
                                        int type) {
        Vector entries = new Vector();
        int[] group = null;
        boolean lastEntryWasIt = false;
        for (int i = address; i < maxAddress; i += 32) {
            if (checkTypeOfEntry(i) == type) {
                if (lastEntryWasIt) {
                    group[1]++;
                } else {
                    group = new int[2];
                    group[0] = i;
                    group[1] = 1;
                    entries.addElement(group);
                }
                lastEntryWasIt = true;
            } else {
                lastEntryWasIt = false;
            }
        }
        return entries;
    }

    /**
     *  Test wether a directory entry starts at the given address.
     *  The 32 bytes starting at <code>address</code> will be tested.
     *  If they are all empty bytes (i.e. having the value of 0xe5)
     *  <code>EMPTY_ENTRY</code> will be returned. Otherwise
     *  <code>DIR_ENTRY</code> or <code>NO_DIR_ENTRY</code> will be
     *  returned depending on whether the 32 bytes survive all tests
     *  or not. As a side product of the tests <code>blockSize</code>
     *  and <code>blockShiftFactor</code> will be calculated. In
     *  addition all allocated block numbers will be collected in
     *  the vector delivered by the <code>getAllocatedBlocks</code>.
     *  The minimum and maximum of this list are available in
     *  <code>minAllocBlock</code> and <code>maxAllocBlock</code>, resp.
     */
    private int checkTypeOfEntry(int address) {
        int b = data[address++] & 0xff;
        boolean deletedEntry = b == 0xe5;
        if (deletedEntry) {
            // either an empty entry or a deleted directory entry
            // test on empty entry
            deletedEntry = false;
            for (int i = 0; i < 31; i++) {
                if (data[address + i] != EMPTY) {
                    deletedEntry = true;
                    break;
                }
            }
            if (!deletedEntry) {
                return EMPTY_ENTRY;
            }
        }

        // check user number
        if (!deletedEntry && b > 31) {
            return NO_DIR_ENTRY;
        }

        // check file name
        for (int i = 0; i < 11; i++) {
            b = data[address++] & 0x7f;
            if (b < ' ' || b > '~') {
                return NO_DIR_ENTRY;
            }
        }

        int extendNumber = data[address++] & 0xff;

        // next two bytes should be zero
        if (data[address++] != 0 || data[address++] != 0) {
            return NO_DIR_ENTRY;
        }

        // check extend record count
        int extendRecordCount = data[address++] & 0xff;
        if (extendRecordCount > 0x80) {
            return NO_DIR_ENTRY;
        }

        // check that disk blocks are uniqueness
        int numberOfBlocks = 0;
        Integer bn;
        boolean lastBlockNumberWasZero = false;
        Vector ab;
        for (int i = 0; i < 16; i++) {
            b = data[address++] & 0xff;
            if (lastBlockNumberWasZero) {
                if (b != 0) {
                    return NO_DIR_ENTRY;
                }
            } else {
                lastBlockNumberWasZero = b == 0;
                if (!lastBlockNumberWasZero) {
                    if (!deletedEntry) {
                        bn = new Integer(b);
                        ab = getAllocatedBlocks();
                        if (ab.contains(bn)) { // block already in list
                            return NO_DIR_ENTRY;
                        } else {
                            numberOfBlocks++;
                            ab.addElement(bn);
                        }
                    }
                    minAllocBlock = Math.min(minAllocBlock, b);
                    maxAllocBlock = Math.max(maxAllocBlock, b);
                }
            }
        }

        // all tests succeeded, calculate block size
        if (extendNumber == 0 && !deletedEntry) {
            int bsf = 3;
            int bs = 1024;
            while (extendRecordCount * 128 > bs * numberOfBlocks) {
                bsf++;
                bs *= 2;
            }
//            System.out.println(extendRecordCount+" "+numberOfBlocks+" "+bs);
            if (bs > blockSize) {
                blockSize = bs;
                blockShiftFactor = bsf;
            }
        }
        return DIR_ENTRY;
    }

    private Vector getAllocatedBlocks() {
        if (allocatedBlocks == null) {
            allocatedBlocks = new Vector();
        }
        return allocatedBlocks;
    }

    /**
     *  Save changed data to the file given in the constructor.
     *  Data of disks created with constructor for empty disk
     *  will not be saved.
     */
    public void finalize() {
        if (diskName != null && dataChanged) {
            try {
                File bakFile = new File(diskName + ".bak");
                File file = new File(diskName);
                if (bakFile.exists() && !bakFile.delete()) {
                    System.out.println("FATAL ERROR: Couln't delete "
                                       + bakFile.getAbsolutePath());
                } else {
                    file.renameTo(bakFile);
                    FileOutputStream fos = new FileOutputStream(diskName);
                    fos.write(data);
                    fos.close();
                }
            } catch (Exception e) {
                System.out.println("FATAL ERROR: Couldn't save "
                                   + diskName);
            }
        }
    }

    /**
     *  Add an interested listener for reading/writing events.
     */
    public void addActionListener(ActionListener al) {
        actionListener = AWTEventMulticaster.add(actionListener, al);
    }

    /**
     *  Remove a listener for reading/writing events.
     */
    public void removeActionListener(ActionListener al) {
        actionListener = AWTEventMulticaster.remove(actionListener, al);
    }

    /**
     *  Notify interested action listeners.
     *  @param track Track number. Will be the first part of the command
     *               string.
     *  @param sector Sector number. Will be the part of the command
     *                string after the space.
     *  @param type Either <code>WRITE</code> or <code>READ</code>.
     */
    private void notifyActionListeners(int track, int sector, int type) {
        if (actionListener != null) {
            StringBuffer sb = new StringBuffer();
            sb.append(track);
            sb.append(' ');
            sb.append(sector);
            actionListener.actionPerformed(new ActionEvent(this, type,
                                                           new String(sb)));
         }
    }

    /**
     *  Return <code>true</code> if nothing has been written to this
     *  disk.
     */
    public boolean isUnchanged() {
        return filesCanBeAdded;
    }

    /**
     *  Add a file of the real file system to this CP/M disk.
     *  This method only works for instances created with
     *  the constructor for empty disk before the first
     *  invocation of the <code>write</code> method.
     *  @param fileName Name of the file to be added.
     */
    public void add(String fileName) {
        if (filesCanBeAdded) {
            String cpmFileName = fileName;
            int index = fileName.lastIndexOf('\\');
            if (index >= 0) {
                cpmFileName = cpmFileName.substring(index + 1);
            }
            if (cpmFileName.endsWith(".cpm")) {
                int len = cpmFileName.length();
                cpmFileName = cpmFileName.substring(0, len - 3) + "com";
            }
            makeFCB(freeDirEntry, cpmFileName);
            try {
                FileInputStream fis = new FileInputStream(fileName);
                int numberOfBytes = fis.available();
                fis.read(data, freeBlock * blockSize, numberOfBytes);
                int numberOfSectors = (numberOfBytes + 127) / 128;
                int numberOfBlocks = (numberOfBytes + blockSize - 1)
                                     / blockSize;
                System.out.println("numberOfSectors="+numberOfSectors);
                data[freeDirEntry + 12] = (byte) (numberOfSectors / 128);
                data[freeDirEntry + 15] = (byte) (numberOfSectors % 128);
                for (int i = 0; i < numberOfBlocks; i++) {
                    data[freeDirEntry + 16 + i] = (byte) freeBlock++;
                }
                freeDirEntry += 32;
            } catch (Exception e) {
                System.out.println("Error in CPMDisk.add: "+e);
                e.printStackTrace();
            }
        }
    }

    /**
     *  Make for the given fileName an FCB starting at given address.
     *  @param addressFCB Start address on the disk.
     *  @param fileName Name of the file.
     */
    private void makeFCB(int addressFCB, String fileName) {
        data[addressFCB++] = 0;
        int address = addressFCB + 8;
        for (int i = 0; i < 11; i++) {
            data[addressFCB + i] = SPACE;
        }
        for (int i = 11; i < 32; i++) {
            data[addressFCB + i] = 0;
        }
        char c;
        boolean afterDot = false;
        for (int i = 0; i < fileName.length(); i++) {
            c = Character.toUpperCase(fileName.charAt(i));
            if (c == '.') {
                afterDot = true;
                addressFCB = address;
            } else if (afterDot || addressFCB < address) {
                data[addressFCB++] = (byte) c;
            }
        }
    }

    /**
     *  Load CCP and BDOS from the system tracks. The method first
     *  searches for the <code>COPYRIGHT</code> text of the CCP.
     *  From the jump address of the CCP it determines the destination.
     *  @param as Address space of the destination.
     *  @return <code>true</code> if loading was successful.
     */
    public boolean loadCCPBDOS(AddressSpace as) {
        // Search for the COPYRIGHT text
        int sector = 0;
        String str;
        do {
            str = new String(data, CCP_COPYRIGHT + sector * 128,
                             COPYRIGHT.length());
        } while (!str.equals(COPYRIGHT)
                 && ++sector < numberOfSectorsPerTrack);
        int addressCCP = sector * 128;
        if (sector == numberOfSectorsPerTrack
                || data[addressCCP] != JMP_CMD) {
            return false;

        }

        // Obtain destination address
        addressCCP = (data[addressCCP + 1] & 0xff)
                     + ((data[addressCCP + 2] & 0xff) << 8) - CCP_START;
//        System.out.print("sector="+sector+" CCP="+Integer.toHexString(addressCCP));

        // Load CCP and BDOS
        int track = 0;
        for (int i = 0; i < CCP_BDOS_SIZE; i++) {
            readSector(track, ++sector, as, addressCCP);
            if (sector == numberOfSectorsPerTrack) {
                track++;
                sector = 0;
            }
            addressCCP += 128;
        }
//        System.out.println(" BIOS="+Integer.toHexString(addressCCP));
        return true;
    }

    /**
     *  Get a byte.
     *  @param address Absolute address of a byte of the disk.
     *  @return the byte at <code>address</code>.
     */
    public byte getByte(int address) {
        return data[address];
    }

    /**
     *  Read a logical sector from the disk.
     *  @param track Track number.
     *  @param sector Sector number.
     *  @param as Address space.
     *  @param destination Address in the address space.
     *  @return <code>true</code> in the case of successful reading.
     */
    public boolean readSector(int track, int sector, AddressSpace as,
                              int destination) {
        notifyActionListeners(track, sector, READ);
        return read((track * numberOfSectorsPerTrack + sector - 1) * 128,
                    as, destination, 128);
    }

    /**
     *  Read data from the disk.
     *  @param diskAddress Address on the disk of the first byte.
     *  @param as Address space of the CP/M emulator.
     *  @param destination Address in the address space.
     *  @param numberOfBytes Number of bytes to be read.
     *  @return <code>true</code> in the case of successful reading.
     */
    private boolean read(int diskAddress, AddressSpace as, int destination,
                         int numberOfBytes) {
        try {
            for (int i = 0; i < numberOfBytes; i++) {
                as.setByte(destination + i, 0, data[diskAddress + i]);
            }
            return true;
        } catch (Exception e) {
            System.out.println("Reading error from " + diskName
                               + ": diskAddress="
                               + Integer.toHexString(diskAddress)
                               + " destination="
                               + Integer.toHexString(destination));
            e.printStackTrace(System.out);
            return false;
        }
    }

    /**
     *  Write a logical sector to the disk.
     *  @param track Track number.
     *  @param sector Sector number.
     *  @param source Address in the address space.
     *  @param destination Address in the address space.
     *  @return <code>true</code> in the case of successful writing.
     */
    public boolean writeSector(int track, int sector, AddressSpace as,
                               int source) {
        notifyActionListeners(track, sector, WRITE);
        return write((track * numberOfSectorsPerTrack + sector - 1) * 128,
                     as, source, 128);
    }

    /**
     *  Write data to the disk.
     *  @param diskAddress Address on the disk.
     *  @param as Address space of the CP/M emulator.
     *  @param source Address in the address space.
     *  @param numberOfBytes Number of bytes to be read.
     *  @return <code>true</code> in the case of successful writing.
     */
    private boolean write(int diskAddress, AddressSpace as, int source,
                          int numberOfBytes) {
        try {
            dataChanged = true;
            filesCanBeAdded = false;
            for (int i = 0; i < numberOfBytes; i++) {
                data[diskAddress + i] = as.getByte(source + i, 0);
            }
            return true;
        } catch (Exception e) {
            System.out.println("Writing error from " + diskName
                               + ": diskAddress="
                               + Integer.toHexString(diskAddress)
                               + " source="
                               + Integer.toHexString(source));
            e.printStackTrace(System.out);
            return false;
        }
    }
}
