package jfrace.device;

import jfrace.*;
import jfrace.cpu.Intel8080;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;

/**
 *  Device which emulates the BIOS for CP/M V2.2.
 *
 *  @author Franz-Josef Elmer
 */
public class CPMEmulator implements Device, ActionListener,
                                    ItemListener {
    /** BIOS return value in the case of no error. */
    public static final byte OK = (byte) 0;

    /** BIOS return value for an error. */
    public static final byte ERROR = (byte) 0xff;

    /** BIOS return value for a disk error. */
    public static final byte DISK_ERROR = (byte) 1;

    /** BIOS functions */
    public static final int WBOOT = 1;
    public static final int CONST = 2;
    public static final int CONIN = 3;
    public static final int CONOUT = 4;
    public static final int LIST = 5;
    public static final int PUNCH = 6;
    public static final int READER = 7;
    public static final int HOME = 8;
    public static final int SELDSK = 9;
    public static final int SETTRK = 10;
    public static final int SETSEC = 11;
    public static final int SETDMA = 12;
    public static final int READ = 13;
    public static final int WRITE = 14;
    public static final int LISTST = 15;
    public static final int SECTRAN = 16;

    public static final String DRIVE_A = "driveA";
    public static final String DRIVE_B = "driveB";
    public static final String RESET_BUTTON = "reset";

    private static final int PORT_BIAS_ADDRESS = 0x10000;

    private AddressSpace as;
    private int jbiosSystemAdrPointer;
    private int biosOffset;
    private byte[] bios;
    private int port;
    private int addressArg1;
    private int addressArg2;
    private int addressResult;
    private int biosFunction;
    private int biosResult;
    private int biosAddress;
    private int[] diskInfoAddresses;
    private String diskPath;
    private CPMDisk[] disks;
    private FileDialog fileDialog;
    private String directory;
    private int selectedDrive;
    private int selectedTrack;
    private int selectedSector;
    private int dmaAddress;

    private Terminal terminal;
    private Button addButton;

    /**
     *  Create a new instance.
     *  @param as AddressSpace.
     *  @param theDiskPath Directory of the CP/M disks.
     *  @param jbiosSystemAdrPointer Address in <code>AddressSpace</code>
     *                               where JBIOS system address is stored.
     *  @param biosOffset First address of the (J)BIOS.
     *  @param bios JBIOS binary data.
     *  @param port Processor port of the CP/M emulator.
     *  @param terminal Terminal.
     */
    public CPMEmulator(AddressSpace as, String theDiskPath,
                       int jbiosSystemAdrPointer,
                       int biosOffset, byte[] bios,
                       int port, Terminal terminal) {
        diskPath = theDiskPath;
        this.as = as;
        this.jbiosSystemAdrPointer = jbiosSystemAdrPointer;
        this.biosOffset = biosOffset;
        this.bios = bios;
        this.port = port;
        this.terminal = terminal;

        reset();
    }

    /** Set a disk. */
    public void setDisk(int driveIndex, CPMDisk disk) {
        disks[driveIndex] = disk;
        int adr = getWord(diskInfoAddresses[driveIndex] + 1);
        disk.setDiskParameters(as, getWord(adr + 10));
    }

    /** Set 'add' button. */
    public void setAddButton(Button button) {
        addButton = button;
    }

    /**
     *  Return a 16-bit word from the given address.
     *  @apram address the address of the word.
     */
    private int getWord(int address) {
        return (as.getByte(address, 0) & 0xff)
               + ((as.getByte(address + 1, 0) & 0xff) << 8);
    }

    /**
     *  Set a 16-bit word at the given address.
     *  @apram address the address of the word.
     *  @param word the word. Only the lower 16 bits are taken.
     */
    private void setWord(int address, int word) {
        as.setByte(address, 0, (byte) word);
        as.setByte(address + 1, 0, (byte) (word >> 8));
    }

    /**
     *  Process a BIOS call.
     *  @param function Number of BIOS function.
     *  @param arg1 First argument (i.e., BC register).
     *  @param arg2 Second argument (i.e., DE register).
     *  @return byte result (for A) and <code>biosResult</code> (for HL).
     */
    public byte processBIOSFunction(int function, int arg1, int arg2) {
        switch (function) {
        case WBOOT:
            return warmBoot();
        case CONST:
            return terminal.isEmpty() ? OK : ERROR;
        case CONIN:
            return (byte) terminal.getNextChar();
        case CONOUT:
            terminal.writeChar((char) (arg1 & 0xff));
            return OK;
        case LIST:   break;
        case PUNCH:  break;
        case READER: break;
        case HOME:
            return OK;
        case SELDSK:
            return selectDrive(arg1 & 0xf);
        case SETTRK:
            selectedTrack = arg1;
            return OK;
        case SETSEC:
            selectedSector = arg1;
            return OK;
        case SETDMA:
            dmaAddress = arg1;
            return OK;
        case READ:
            return readSector();
        case WRITE:
            return writeSector();
        case LISTST: break;
        case SECTRAN:
            return translateSector(arg1, arg2);
        }
        System.out.println("BIOS: "+function+" "+Integer.toHexString(arg1)+" "+Integer.toHexString(arg2));
        return (byte) ERROR;
    }

    /**
     *  Cold boot. Stops emulation, restores BIOS, reload BDOS and CCP,
     *  and restarts emulation.
     */
    public void coldBoot() {
        System.out.println("Cold boot");
        Emulator emulator = Emulator.getEmulator();

        emulator.stopEmulation();
        System.out.println("emulation stopped");
        try {
            Thread.sleep(300);
        } catch (Exception e) {}

        // reset all devices (including the CPMEmulator)
        int n = as.getNumberOfAddressBanks();
        for (int i = 0; i < n; i++) {
            as.getDevice(i).reset();
        }
        warmBoot();
        as.getCPU(0).setRegisterValue(Intel8080.SP, 0x100);
        as.getCPU(0).setRegisterValue(Intel8080.PC, biosOffset);

        emulator.runEmulation();
        System.out.println("emulation restarted");
    }

    /**
     *  Load CCP and BDOS from the first disk found.
     *  In order to find the CCP the method looks whether there is the
     *  text <code>COPYRIGHT</code> starting at the 24th byte of
     *  an 128 sector.
     *
     *  @return <code>OK</code> if loading was successful.
     */
    public byte warmBoot() {
        System.out.println("Warm boot");
        for (int i = 0; i < disks.length; i++) {
            if (disks[i].loadCCPBDOS(as)) {
                return OK;
            }
        }
        return ERROR;
    }

    private byte selectDrive(int drive) {
        if (drive < diskInfoAddresses.length) {
            selectedDrive = drive;
            biosResult = getWord(diskInfoAddresses[drive] + 1);
            return OK;
        } else {
            biosResult = 0;
            return ERROR;
        }
    }

    private byte translateSector(int sector, int table) {
        biosResult = disks[table].translateSector(sector);
//        System.out.println("translateSector "+sector+" "+table+" "+biosResult);
        return OK;
    }

    private byte readSector() {
        if (disks[selectedDrive].readSector(selectedTrack, selectedSector,
                                            as, dmaAddress)) {
            return OK;
        } else {
            return DISK_ERROR;
        }
    }

    private byte writeSector() {
        if (disks[selectedDrive].writeSector(selectedTrack, selectedSector,
                                             as, dmaAddress)) {
            return OK;
        } else {
            return DISK_ERROR;
        }
    }

//
//  Implementation of the Device interface.
//
    /** Reset. Restores BIOS */
    public void reset() {
        for (int i = 0; i < bios.length; i++) {
            as.setByte(biosOffset + i, 0, bios[i]);
        }
        int jbiosSystemAdr = getWord(jbiosSystemAdrPointer);
        as.setByte(jbiosSystemAdr, 0, (byte) port);
        addressArg1 = jbiosSystemAdr + 1;
        addressArg2 = jbiosSystemAdr + 3;
        addressResult = jbiosSystemAdr + 5;
        biosAddress = getWord(jbiosSystemAdr + 7);
        if (diskInfoAddresses == null) {
            int numberOfDisks = as.getByte(jbiosSystemAdr + 9, 0);
            diskInfoAddresses = new int[numberOfDisks];
            disks = new CPMDisk[numberOfDisks];
            int adr;
            for (int i = 0; i < numberOfDisks; i++) {
                diskInfoAddresses[i] = jbiosSystemAdr + 10 + 3 * i;
                /*
                adr = getWord(diskInfoAddresses[i] + 1);
                adr = getWord(adr + 10);
                System.out.println(i + ":"
                                   + ((char) (as.getByte(diskInfoAddresses[i], 0)
                                              + ((int) 'A')))
                                   + " " + Integer.toHexString(
                                              getWord(diskInfoAddresses[i] + 1)));
                */
            }
        }
    }

    /** Returns 8. */
    public int getNumberOfBits() {
        return 8;
    }

    /** Returns 1. */
    public int getNumberOfUnits() {
        return 1;
    }

    /** Returns 0. */
    public int getValue(int address) {
        return 0;
    }

    /** Does nothing. */
    public void setValue(int address, int value) {}

    /**
     *  Process BIOS function and return register A.
     */
    public byte getByte(int address) {
        byte result = processBIOSFunction(biosFunction,
                                          getWord(addressArg1),
                                          getWord(addressArg2));
        setWord(addressResult, biosResult);
        return result;
    }

    /**
     *  Set the BIOS function code.
     *  @param address Will be ignored.
     *  @param value BIOS function code.
     */
    public void setByte(int address, byte value) {
        biosFunction = value & 0xff;
    }

//
//  Implementation of the ActionListener interface
//

    public void actionPerformed(ActionEvent event) {
        Object source = event.getSource();
        if (source instanceof Button
                && ((Button) source).getName().equals(RESET_BUTTON)) {
            coldBoot();
        } else if (source == disks[2]) {
            if (event.getID() == CPMDisk.WRITE && addButton != null) {
                addButton.disable();
            }
        } else if (disks[2].isUnchanged()) {
            if (fileDialog == null) {
                Component parent = ((Component) terminal).getParent();
                while (!(parent instanceof Frame)) {
                    System.out.println("parent="+parent);
                    parent = parent.getParent();
                }
                fileDialog = new FileDialog((Frame) parent, "Add a file to disk C:");
            }
            if (directory != null) {
                fileDialog.setDirectory(directory);
            }
            fileDialog.show();
            String file = fileDialog.getFile();
            if (file != null) {
                directory = fileDialog.getDirectory();
                System.out.println(directory + file);
                disks[2].add(directory + file);
            }
        }
    }

//
//  Implementation of the ItemListener interface
//

    public void itemStateChanged(ItemEvent event) {
        Choice c = (Choice) event.getSource();
        try {
            CPMDisk disk = new CPMDisk(diskPath + c.getSelectedItem() + ".dsk");
            CPMDisk oldDisk;
            String name = c.getName();
            if (name.equals(DRIVE_A)) {
                oldDisk = disks[0];
                setDisk(0, disk);
                oldDisk.finalize();
            } else if (name.equals(DRIVE_B)) {
                oldDisk = disks[1];
                setDisk(1, disk);
                oldDisk.finalize();
            }
        } catch (Exception e) {
            System.out.println("Couldn't open disk : "
                               + c.getSelectedItem() + ".dsk " + e);
        }

    }
}
