package jfrace.device;

import jfrace.*;
import jfrace.cpu.*;
import jfrace.view.DiskView;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.net.URL;
import java.util.StringTokenizer;

/**
 *  Configurator for the <code>CPMEmulator</code>.
 *  It uses a configuration file.
 *  <a href="../../javadocResources/CPMEmulatorExample.cfg">Here</a>
 *  is an example for such a configuration file.
 *
 *  @author Franz-Josef Elmer
 */
public class CPMEmulatorConfigurator implements Configurator {
    private Emulator emulator = Emulator.getEmulator();
    private AddressSpace as;
    private String packageDevicePrefix;
    private String packageTerminalPrefix;
    private int portBiasAddress;
    private int port;
    private int jbiosSystemAdrPointer = 0xfffe;
    private byte[] bios;
    private int biosOffset;
    private Terminal terminal;
    private String disksPath = "";
    private String bootDisk;

    AddressSpace getAddressSpace() {
        return as;
    }

    String getPackageDevicePrefix() {
        return packageDevicePrefix;
    }

    String getPackageTerminalPrefix() {
        return packageTerminalPrefix;
    }

    int getPort() {
        return port;
    }

    int getJbiosSystemAdrPointer() {
        return jbiosSystemAdrPointer;
    }

    /**
     *  Create an instance of <code>CPMEmulator</code>, configurate
     *  it and add it (as a device) to the address space.
     *  @param as Address space. CPU and RAM are already added.
     *  @param configSource Configuration file either an URL or a local file.
     *  @param packagePrefix Prefix as set in the emulator config file.
     *  @param args The additional command line arguments.
     */
    public boolean configurate(AddressSpace as, String configSource,
                               String packagePrefix, String[] args) {
        this.as = as;
        packageDevicePrefix = packagePrefix + ".device.";
        packageTerminalPrefix = packagePrefix + ".terminal.";

        if (as.getNumberOfCPUs() == 0) {
            System.out.println("CPU missing.");
            return false;
        }
        CPU cpu = as.getCPU(0);
        if (!(cpu instanceof Intel8080)) {
            System.out.println("1. CPU has to be Intel8080 or a subclass.");
            return false;
        }
        portBiasAddress = ((Intel8080) cpu).getPortBiasAddress();

        try {
            URL url = emulator.getAbsoluteURL(configSource);
            BufferedReader br = new BufferedReader(
                                    new InputStreamReader(url.openStream()));
            String line;
            int index;
            int count = 0;
            boolean result = true;
            while (result && (line = br.readLine()) != null) {
                count++;
                if (line.length() > 0 && line.charAt(0) != '#') {
                    index = line.indexOf('=');
                    if (index < 0) {
                        System.out.println("ERROR: '=' missing in");
                        result = false;
                    } else {
                        result = process(line.substring(0, index).trim(),
                                         line.substring(index + 1).trim(),
                                         args);
                    }
                    if (!result) {
                        System.out.println(count + ":" + line);
                    }
                }
            }
            br.close();
            return finishConfiguration(args);
        } catch (Exception e) {
            System.out.println("Error during reading the config file.");
            e.printStackTrace(System.out);
            return false;
        }
    }

    /**
     *  Process a line from the configuration file.
     *  @param type String before '='
     *  @param type String after '='
     *  @param args[] Additional command line arguments.
     */
    private boolean process(String type, String description, String[] args) {
        if (type.equals("port")) {
            port = Util.parseNumber(description);
        } else if (type.equals("JBIOSHexFile")) {
            StringTokenizer st = new StringTokenizer(description);
            int numberOfTokens = st.countTokens();
            if (numberOfTokens < 1) {
                System.out.println("at least one token expected.");
                return false;
            }
            String hexFile = st.nextToken();
            if (numberOfTokens > 1) {
                jbiosSystemAdrPointer = Util.parseNumber(st.nextToken());
            }
            return readJBIOS(hexFile);
        } else if (type.equals("terminal")) {
            return processTerminal(description);
        } else if (type.equals("device")) {
            return processDevice(description);
        } else if (type.equals("disksPath")) {
            disksPath = description + '/';
        } else if (type.equals("bootDisk")) {
            bootDisk = description;
        } else {
            System.out.println("ERROR: unkown type >" + type + "<");
            return false;
        }
        return true;
    }

    /** Process device line. */
    private boolean processDevice(String description) {
        StringTokenizer st = new StringTokenizer(description);
        int numberOfTokens = st.countTokens();
        if (numberOfTokens != 2) {
            System.out.println("class name and port # expected.");
            return false;
        }
        String className = packageDevicePrefix + st.nextToken();
        try {
            Device device = (Device) Class.forName(className).newInstance();
            if (device instanceof AddressSpaceListener) {
                as.addAddressSpaceListener((AddressSpaceListener) device);
            }
            int devicePort = Util.parseNumber(st.nextToken());
            as.addDevice(portBiasAddress + devicePort, 0, device);
            return true;
        } catch (Exception e) {
            e.printStackTrace(System.out);
            return false;
        }
    }

    /** Process terminal line. */
    private boolean processTerminal(String description) {
        StringTokenizer st = new StringTokenizer(description);
        int numberOfTokens = st.countTokens();
        if (numberOfTokens < 1) {
            System.out.println("at least terminal class name expected.");
            return false;
        }
        String className = packageTerminalPrefix + st.nextToken();
        try {
            terminal = (Terminal) Class.forName(className).newInstance();
        } catch (Exception e) {
            e.printStackTrace(System.out);
            return false;
        }

        if (numberOfTokens > 1) {
            int termPort = Util.parseNumber(st.nextToken());
            try {
                as.addDevice(portBiasAddress + termPort, 0,
                             terminal.getDevice());
            } catch (Exception e) {
                e.printStackTrace(System.out);
                return false;
            }
        }
        return true;
    }

    /** Read JBIOS from the given Intel-Hex file. */
    private boolean readJBIOS(String fileName) {
        try {
            URL url = emulator.getAbsoluteURL(fileName);
            BufferedReader br = new BufferedReader(
                                    new InputStreamReader(url.openStream()));
            String line;
            StringBuffer sb = new StringBuffer();
            while ((line = br.readLine()) != null) {
                sb.append(line);
            }
            br.close();
            byte[] data = Util.generateBinaryData(new String(sb));
            for (; biosOffset < data.length && data[biosOffset] == 0;
                 biosOffset++);
            bios = new byte[data.length - biosOffset];
            System.arraycopy(data, biosOffset, bios, 0, bios.length);
            return true;
        } catch (Exception e) {
            e.printStackTrace(System.out);
            return false;
        }
    }

    private boolean finishConfiguration(String[] args) {
        CPMEmulator emu = new CPMEmulator(as, disksPath,
                                          jbiosSystemAdrPointer,
                                          biosOffset, bios, port,
                                          terminal);
        String diskA = disksPath + bootDisk;
        String diskB = disksPath + bootDisk;
        if (args.length > 0) {
            diskB = args[0];
            if (args.length > 1) {
                diskA = diskB;
                diskB = disksPath + args[1];
            }
        }

        try {
            CPMDisk disk0 = null;
            CPMDisk disk1 = null;
            if (emulator.insideAnApplet()) {
                disk0 = new CPMDisk(emulator.getAbsoluteURL(diskA));
                disk1 = new CPMDisk(emulator.getAbsoluteURL(diskB));
            } else {
                disk0 = new CPMDisk(diskA);
                disk1 = new CPMDisk(diskB);
            }
            CPMDisk disk2 = new CPMDisk(194 * 2048);
            emu.setDisk(0, disk0);
            emu.setDisk(1, disk1);
            emu.setDisk(2, disk2);
            if (emu.warmBoot() != CPMEmulator.OK) {
                System.out.println("No bootable disk present.");
                return false;
            }
            as.addDevice(portBiasAddress + port, 0, emu);

            // Show terminal if it is a subclass of java.awt.Component
            if (terminal instanceof Component) {
                showTerminal(emu, disk0, disk1, disk2, diskA, diskB);
                ((Component) terminal).requestFocus();
            }

            // Initialize CPU registers
            as.getCPU(0).setRegisterValue(Intel8080.PC, biosOffset);
            as.getCPU(0).setRegisterValue(Intel8080.SP, 0x100);

            if (!emulator.insideAnApplet()) {
                Runtime.runFinalizersOnExit(true);
            }
            System.out.println("CPMEmulatorConfigurator finished.");
            return true;
        } catch (Exception e) {
            e.printStackTrace(System.out);
            return false;
        }
    }

    private void showTerminal(CPMEmulator emu, CPMDisk disk0,
                              CPMDisk disk1, CPMDisk disk2,
                              String diskA, String diskB) {
        Panel p = new Panel();
        p.setLayout(new BorderLayout());
        p.add("Center", (Component) terminal);
        Button resetButton = new Button("reset");
        resetButton.setName(CPMEmulator.RESET_BUTTON);
        resetButton.addActionListener(emu);
        p.add("South", resetButton);
        if (!emulator.insideAnApplet()) {
            Panel panel = new Panel();
            p.add("North", panel);
            Panel dPanel = new Panel(new GridLayout(2, 1));
            panel.add(dPanel);
            Panel d1Panel = new Panel();
            dPanel.add(d1Panel);
            d1Panel.add(new Label("Drive A:"));
            Choice driveA = makeDiskChoice(disksPath, diskA);
            driveA.setName(CPMEmulator.DRIVE_A);
            driveA.addItemListener(emu);
            d1Panel.add(driveA);
            DiskView dv = new DiskView(disk0.getNumberOfTracks(),
                                       disk0.getNumberOfSectorsPerTrack());
            disk0.addActionListener(dv);
            dPanel.add(dv);

            dPanel = new Panel(new GridLayout(2, 1));
            panel.add(dPanel);
            d1Panel = new Panel();
            dPanel.add(d1Panel);
            d1Panel.add(new Label("Drive B:"));
            Choice driveB = makeDiskChoice(disksPath, diskB);
            driveB.setName(CPMEmulator.DRIVE_B);
            driveB.addItemListener(emu);
            d1Panel.add(driveB);
            dv = new DiskView(disk1.getNumberOfTracks(),
                              disk1.getNumberOfSectorsPerTrack());
            disk1.addActionListener(dv);
            dPanel.add(dv);

            dPanel = new Panel(new GridLayout(2, 1));
            panel.add(dPanel);
            Button addButton = new Button("add file to drive C");
            addButton.addActionListener(emu);
            emu.setAddButton(addButton);
            disk2.addActionListener(emu);
            dPanel.add(addButton);
            dv = new DiskView(disk2.getNumberOfTracks(),
                              disk2.getNumberOfSectorsPerTrack());
            disk2.addActionListener(dv);
            dPanel.add(dv);

            Frame frame = new Frame("CPMEmulator (" + terminal.getName()
                                    + ")");
            frame.setIconImage(Util.loadImage("icon.gif"));
            frame.add(p);
//            frame.setLayout(new BorderLayout());
            frame.addWindowListener(new WindowAdapter() {
                    public void windowClosing(WindowEvent event) {
                        System.exit(0);
                    }
                });
            frame.pack();
            frame.show();
        } else {
            emulator.getApplet().add(p);
        }
    }

    /**
     *  Return a list of all files of type <code>.dsk</code>.
     */
    private static Choice makeDiskChoice(String disksPath,
                                         String selectedDisk) {
        Choice disks = new Choice();
        File f = new File(disksPath);
        if (f.isDirectory()) {
            String file;
            String[] fileList = f.list();
            int index = 0;
            for (int i = 0; i < fileList.length; i++) {
                file = fileList[i];
                if (file.endsWith(".dsk") || file.endsWith(".DSK")) {
                    disks.add(file.substring(0, file.length() - 4));
                    if (file.equals(selectedDisk)) {
                        disks.select(index);
                    }
                    index++;
                }
            }
        }
        return disks;
    }
}
