package jfrace.device;

import jfrace.Device;

/**
 *  This class emulates General Instrument's AY-3-8910
 *  Programmble Sound Generator.
 *  Clock rate is assumed to be 4MHz.
 *  Note: Noise is not implemented.
 *  It uses the undocumented sun.audio package which is a part
 *  of any JDK.
 *
 *  @author Franz-Josef Elmer
 */
public class GI_AY_3_8910 implements Device {
    private static final int DEFAULT_CLOCK_RATE = 3900000;
    private static final int NV = 3; // number of voices

    private int clockRate = DEFAULT_CLOCK_RATE;
    private int regAddress;
    private WaveTablePlayer dummyPlayer;
    private GI_AY_3_8910ToneGenerator[] tgs;

    /**
     *  Creates a new instance.
     */
    public GI_AY_3_8910() {
        tgs = new GI_AY_3_8910ToneGenerator[NV];
        for (int i = 0; i < NV; i++) {
            tgs[i] = new GI_AY_3_8910ToneGenerator(clockRate);
        }
        dummyPlayer = new WaveTablePlayer(100);
        int[] wt = new int[100];
        dummyPlayer.setWaveTable(wt);
        dummyPlayer.play();
    }

//
//  Implementation of the Device interface
//

    /** Reset the sound generator. All voices are stopped. */
    public void reset() {
        for (int i = 0; i < NV; i++) {
            tgs[i].stop();
        }
    }

    /** Return 8. */
    public int getNumberOfBits() {
        return 8;
    }

    /** Return 2. */
    public int getNumberOfUnits() {
        return 2;
    }

    public int getValue(int address) {
        return (int) getByte(address);
    }

    public void setValue(int address, int value) {
        setByte(address, (byte) value);
    }

    public byte getByte(int address) {
        if (address == 0) {
            return -1;
        } else {
            int channel;
            int period;
            if (regAddress < 6) {
                channel = regAddress >> 1;
                period = tgs[channel].getPeriod();
                if ((regAddress % 2) == 0) {
                    return (byte) period;
                } else {
                    return (byte) (period >> 8);
                }
            } else if (regAddress == 6) {       // get noise period
                return -1;
            } else if (regAddress == 7) {       // enable/disable
                int result = 0xff;
                for (int i = 0, mask = 1; i < NV; i++, mask <<= 1) {
                    result = (result << 1) + (tgs[i].isPlaying() ? 0 : 1);
                }
                return (byte) result;
            } else if (regAddress < 11) {       // set amplitude
                channel = regAddress - 8;
                period = tgs[channel].getAmplitude();
                if (tgs[channel].isAmplitudeMode()) {
                    period += 0x10;
                }
                return (byte) period;
            } else {
                return -1;
            }
        }
    }

    public void setByte(int address, byte value) {
        if (address == 0) {
            regAddress = value & 0xf;
        } else {
            int channel;
            int period;
            if (regAddress < 6) {               // set channel period
                channel = regAddress >> 1;
                period = tgs[channel].getPeriod();
                if ((regAddress % 2) == 0) {
                    period = (period & 0xf00) + (value & 0xff);
                } else {
                    period = (period & 0xff) + ((value & 0xf) << 8);
                }
                tgs[channel].setPeriod(period);
            } else if (regAddress == 6) {       // set noise period
                System.out.println("set noise period is not implemented.");
            } else if (regAddress == 7) {       // enable/disable
                for (int i = 0, mask = 1; i < NV; i++, mask <<= 1) {
                    if ((value & mask) == 0) {
                        tgs[i].play();
                    } else {
                        tgs[i].stop();
                    }
                }
            } else if (regAddress < 11) {       // set amplitude
                channel = regAddress - 8;
                tgs[channel].setAmplitude(value);
                tgs[channel].setAmplitudeMode((value & 0x10) != 0);
            } else if (regAddress == 11) {      // set env. period (fine)
                for (int i = 0; i < NV; i++) {
                    period = tgs[i].getEnvelopePeriod();
                    tgs[i].setEnvelopePeriod((period & 0xff00)
                                             + (value & 0xff));
                }
            } else if (regAddress == 12) {      // set env. period (coarse)
                for (int i = 0; i < NV; i++) {
                    period = tgs[i].getEnvelopePeriod();
                    tgs[i].setEnvelopePeriod((period & 0xff)
                                             + ((value & 0xff) << 8));
                }
            } else if (regAddress == 13) {      // set envelope shape
                for (int i = 0; i < NV; i++) {
                    tgs[i].setEnvelopeShape(value);
                }
            }
        }
    }
}
