package jfrace.device;


/**
 *  Simulating class of a voice of the GI's sound generator AY-3-8910.
 *
 *  @author Franz-Josef Elmer
 */
public class GI_AY_3_8910ToneGenerator extends WaveTablePlayer
                                       implements Runnable {
    /** Return the greatest common divisor of two integer numbers. */
    private static final int gcd(int n, int m) {
        int i = Math.min(n, m);
        int n1 = Math.max(n, m);
        int n2;
        while (i != 0) {
            n2 = n1;
            n1 = i;
            i = n2 % n1;
        }
        return n1;
    }

    private static final int MIN_AMPL = 64;
    private static final int HOLD_MASK = 1;
    private static final int ALTERNATE_MASK = 2;
    private static final int ATTACK_MASK = 4;
    private static final int CONTINUE_MASK = 8;
    private static final int REPEAT_MASK = CONTINUE_MASK + HOLD_MASK;

    private int clockRatio;
    private int period = 1;
    private int amplitude;
    private int ampl = MIN_AMPL;
    private boolean amplitudeMode;
    private int envPeriod;
    private int envShape;
    private int envCount;
    private int envStep;
    private int envWaitingTime;
    private Thread envelopeThread;

    /**
     *  Create an instance for a given clock rate.
     *  @param clockRate Clock rate in Hz.
     */
    public GI_AY_3_8910ToneGenerator(int clockRate) {
        super(8000);

        clockRatio = (clockRate + 8 * 8000) / (16 * 8000);
        System.out.println("clockRatio = " + clockRatio
                           + " clockRate = " + (16 * 8000 * clockRatio));
    }


    /** Return the period. */
    public int getPeriod() {
        return period;
    }

    /**
     *  Set the period. Only the 12 lower bits are used.
     */
    public void setPeriod(int value) {
        value = value & 0xfff;
        if (value != period) {
            period = value;
            makeWaveForm(ampl);
        }
    }

    /** Return the amplitude. */
    public int getAmplitude() {
        return amplitude;
    }

    /** Set the amplitude. Only the 4 lower bits are used. */
    public void setAmplitude(int value) {
        value = value & 0xf;
        if (value != amplitude) {
            amplitude = value;
            ampl = calculateAmplitude(amplitude);
            makeWaveForm(ampl);
        }
    }

    private int calculateAmplitude(int a) {
        return (int) (MIN_AMPL * Math.pow(2.0, a * 0.5) + 0.5);
    }

    /** Return <code>true</code> if in the amplitude mode. */
    public boolean isAmplitudeMode() {
        return amplitudeMode;
    }

    /** Set amplitude mode. */
    public void setAmplitudeMode(boolean value) {
        if (amplitudeMode != value) {
            amplitudeMode = value;
        }
    }

    /** Return the envelope period. */
    public int getEnvelopePeriod() {
        return envPeriod;
    }

    /** Set envelope period. */
    public void setEnvelopePeriod(int value) {
        if (envPeriod != value) {
            envPeriod = value;
            envWaitingTime = value / (clockRatio * 16 * 8);
        }
    }

    /** Set envelope shape. Only the lower 4 bits are used. */
    public void setEnvelopeShape(int value) {
        if (envelopeThread == null) {
            envelopeThread = new Thread(this);
            envelopeThread.start();
        }
        envShape = value & 0xf;
        if ((envShape & ATTACK_MASK) == 0) {
            envCount = 15;
            envStep = -1;
        } else {
            envCount = 0;
            envStep = 1;
        }
    }

    /**
     *  Creating a triangular wave form by using the attributs
     *  <code>clockRatio</code> and <code>period</code> and
     *  the parameter <code>ampl</code>.
     */
    private void makeWaveForm(int ampl) {
        int len = Math.min(period / gcd(clockRatio, period), waveTable.length);
        int c4 = 4 * clockRatio;
        int p4 = 4 * period;
        int p3 = 3 * period;
        int p2 = 2 * period;
        int phase, a;
        for (int i = 0; i < len; i++) {
            phase = (i * c4) % p4;
            a = (ampl * (phase - Math.abs(phase - period)
                         + Math.abs(phase - p3) - p2)) / period;
            waveTable[i] = linear2ulaw(a);
        }
        setLength(len);
    }

//
//  Implements the Runnable interface.
//

    /**
     *  Running envelope thread for changing waveform du to the envelope
     *  shape.
     */
    public void run() {
        Thread ct = Thread.currentThread();
        boolean looping = true;
        boolean periodEnd = false;
        while (ct == envelopeThread && amplitudeMode && looping) {
            System.out.println(envWaitingTime+" "+envCount);
            makeWaveForm(calculateAmplitude(envCount));
            try {
                Thread.sleep(envWaitingTime);
            } catch (Exception e) {}
            if (periodEnd) { // end of env. period
                looping = (envShape & REPEAT_MASK) == CONTINUE_MASK;
                if (looping) { // start a new envelope period
                    if ((envShape & ALTERNATE_MASK) == 0) {
                        envCount = 15 - envCount;
                    } else {
                        envStep = - envStep;
                    }
                    periodEnd = false;
                } else {
                    if ((envShape & CONTINUE_MASK) == 0) {
                        envCount = 0;
                        System.out.println(envCount);
                        makeWaveForm(calculateAmplitude(envCount));
                    } else if ((envShape & ALTERNATE_MASK) != 0) {
                        envCount = 15 - envCount;
                        System.out.println(envCount);
                        makeWaveForm(calculateAmplitude(envCount));
                    }
                }
            } else {
                envCount += envStep;
                periodEnd = 15 * (1 + envStep) == 2 * envCount;
            }
        }
        envelopeThread = null;
    }
}
