package jfrace.device;

import jfrace.Device;

import java.awt.Canvas;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Polygon;

/**
 *  A <code>Device</code> simulating an LCD digit.
 *  It is a write-only device which occupies one addressable unit.
 *  Only the five lower bits of data written
 *  to <code>LCDDigit</code> are evaluated where bit 0-3 codes
 *  the index of the segement which will be shown or hide depending
 *  on the value of bit 4. Here is the indexing scheme of the segments:
 *  <p>
 *  <img src="../../javadocResources/LCDDigit.gif" alt="Indexing scheme">
 *
 *  @author Franz-Josef Elmer
 */
public class LCDDigit extends Canvas implements Device {
    private int size = 4;
    private int xSize = 20;
    private int ySize = 20;
    private int tilt = 2;

    private int height = 2 * ySize + 5 * size;
    private int width = xSize + 6 * size + (tilt * height) / 10;


    private boolean[] segments = new boolean[16];
    private Polygon[] segmentPolygons = new Polygon[16];

    public Dimension preferredSize() {
        return new Dimension(width, height);
    }

    /**
     *  Show or hide a segment of the display.
     *  @param index Index of the display. Has to be less than 16.
     *  @param value Segment will be shown/hide if <code>true/false</code>.
     */
    public void setSegment(int index, boolean value) {
        segments[index] = value;
        if (segmentPolygons[index] == null) {
            Polygon p = new Polygon();
            switch (index) {
            case 0: addToPolygon(p, 2 * size, size, xSize, size); break;
            case 1: addToPolygon(p, 2 * size, 2 * size + ySize, xSize, size); break;
            case 2: addToPolygon(p, 2 * size, 3 * size + 2 * ySize, xSize, size); break;
            case 3: addToPolygon(p, size, 2 * size, size, ySize); break;
            case 4: addToPolygon(p, size, 3 * size + ySize, size, ySize); break;
            case 5: addToPolygon(p, 2 * size + xSize, 2 * size, size, ySize); break;
            case 6: addToPolygon(p, 2 * size + xSize, 3 * size + ySize, size, ySize); break;
            case 7: addToPolygon(p, 4 * size + xSize, size + ySize, size, size); break;
            case 8: addToPolygon(p, 4 * size + xSize, 3 * size + ySize, size, size); break;
            case 9: addToPolygon(p, 4 * size + xSize, 3 * size + 2 * ySize, size, size); break;
            }
            segmentPolygons[index] = p;
        }
        repaint();
    }

    private void addToPolygon(Polygon p, int x, int y, int w, int h) {
        int yt = ((height - y) * tilt + 5) / 10;
        p.addPoint(x + yt, y);
        p.addPoint(x + w + yt, y);
        y += h;
        yt =  ((height - y) * tilt + 5) / 10;;
        p.addPoint(x + w + yt, y);
        p.addPoint(x + yt, y);
    }

    /** Paint visible display segments in the foreground color. */
    public void paint(Graphics g) {
        g.setColor(getForeground());
        for (int i = 0; i < segments.length; i++) {
            if (segments[i]) {
                g.fillPolygon(segmentPolygons[i]);
            }
        }
    }

//
//  Implemntation of the Device interface
//

    /** Reset display. Hide all segments. */
    public void reset() {
        for (int i = 0; i < segments.length; i++) {
            setSegment(i, false);
        }
    }

    /** Return 8. */
    public int getNumberOfBits() {
        return 8;
    }

    /** Return 1. */
    public int getNumberOfUnits() {
        return 1;
    }

    /** Nothing to read. Return 0. */
    public int getValue(int address) {
        return 0;
    }

    /**
     *  Show or hide a segment. The four lower bits of
     *  <code>value</code> code the index of the segment
     *  whereas bit 4 codes whether the segment has to be
     *  shown or hide.
     *  @param address Will be ignored.
     *  @param value Segment index and visibility flag.
     */
    public void setValue(int address, int value) {
        setSegment(value & 0xf, (value & 0x10) != 0);
    }

    /** Nothing to read. Return 0. */
    public byte getByte(int address) {
        return 0;
    }

    /**
     *  Show or hide a segment. The four lower bits of
     *  <code>value</code> code the index of the segment
     *  whereas bit 4 codes whether the segment has to be
     *  shown or hide.
     *  @param address Will be ignored.
     *  @param value Segment index and visibility flag.
     */
    public void setByte(int address, byte value) {
        setValue(address, (int) value);
    }
}
