package jfrace.device;

import jfrace.*;

import java.io.*;
import java.net.*;
import java.util.*;

/**
 *  This class emulates the random access memory (RAM) of a
 *  computer. By setting the <code>readOnly</code> flag it can
 *  be turned into a ROM.
 *
 *  @author Franz-Josef Elmer
 */
public class RAM implements Device {
    private static final byte NULL = (byte) 0;
    private static final String MSG_3_ITEMS
        = "At least three items in the description text expected.";
    private static final String MSG_INVALID_URL
        = "The 3. argument of the description text is an invalid URL: ";
    private static final String MSG_IO_ERROR
        = "Couldn't read RAM data: ";

    private int numberOfBits = 8;
    private int numberOfBytes = 1;
    private int mask = 0xff;
    private int numberOfUnits = 0;
    private byte[] memory = new byte[0];
    private boolean readOnly;

    /** Create an empty memory. The addressable unit is one byte. */
    public RAM() {
    }

    /**
     *  Create an empty memory with a given size of the addressable
     *  unit.
     *  @param numberOfBits the size of an addressable unit in number of bits.
     */
    public RAM(int numberOfBits) {
        this.numberOfBits = numberOfBits;
        this.numberOfBytes = (numberOfBits - 1) / 8 + 1;
        mask = 0;
        int bit = 1;
        for (int i = 0; i < numberOfBits; i++) {
            mask = mask | bit;
            bit <<= 1;
        }
    }

    /**
     *  Create an empty memory with a given size of the addressable
     *  unit and a given number of addressable units.
     *  @param numberOfBits The size of an addressable unit in number of bits.
     *  @param numberOfUnits Number of addressable units.
     *  @param readOnlyFlag <code>readOnly</code> flag.
     */
    public RAM(int numberOfBits, int numberOfUnits, boolean readOnlyFlag) {
        this(numberOfBits);
        createMemory(numberOfUnits);
        readOnly = readOnlyFlag;
    }

    /**
     *  Create a memory with a given size of the addressable
     *  unit and filled with raw data from a given source.
     *  The raw data is a sequence of byte where chunks of
     *  <code>numberOfBytes</code> belong to an addressable unit.
     *  <code>numberOfBytes</code> is the number of bytes to cover
     *  an addressable unit with <code>numberOfBits</code> bits.
     *  The data in a chunk is stored in big-endian form (i.e.,
     *  the highest bytes come first).
     *  @param numberOfBits the size of an addressable unit in number of bits.
     *  @param source Source of the raw data.
     *  @param readOnlyFlag <code>readOnly</code> flag.
     *  @exception IOException is thrown when an error occured during
     *                         obtaining the raw data.
     */
    public RAM(int numberOfBits, URL source, boolean readOnlyFlag) throws IOException  {
        this(numberOfBits);
        try {
            memory = Util.readBinaryData(source);
            numberOfUnits = memory.length / numberOfBytes;
        } catch (IOException ioe) {
            throw ioe;
        } catch (Throwable t) {
            throw new IOException("Caused by " + t);
        }
        readOnly = readOnlyFlag;
    }

    /**
     *  Create an instance from a description string and add it to
     *  the given address space.
     *  The description string has the form:<br><br>
     *
     *  <code><i>&lt;number of bits&gt; &lt;first address&gt;</i>
     *  {<i>&lt;size&gt;</i>|<i>&lt;raw data&gt;</i>}
     *  [<i>&lt;mode mask&gt;</i>] [readOnly]</code><br><br>
     *
     *  where
     *  <dl><dt><code><i>&lt;number of bits&gt;</i></code></dt>
     *      <dd>is the number of bits of an addressable unit</dd>
     *      <dt><code><i>&lt;first address&gt;</i></code></dt>
     *      <dd>is the first address in the address space</dd>
     *      <dt><code><i>&lt;size&gt;</i></code></dt>
     *      <dd>the number of addressable units</dd>
     *      <dt><code><i>&lt;raw data&gt;</i></code></dt>
     *      <dd>URL source with initial raw data</dd>
     *      <dt><code><i>&lt;mode mask&gt;</i></code></dt>
     *      <dd>the mode mask</dd>
     *      <dt><code>readOnly</code></dt>
     *      <dd>if the RAM is actually a ROM</dd>
     *  </dl>
     *  The last two terms are optional. If the third parameter
     *  is not a number it will be interpreted as a file name.
     *  The URL source contains the initial content of the RAM.
     *  @param as Address space.
     *  @param description Description string.
     *  @param documentBase URL of the document base if
     *                      <code><i>&lt;raw data&gt;</i></code>
     *                      is a relative URL.
     *  @exception IllegalArgumentException is thrown when the
     *             description string has some error.
     *  @exception AddressBankOverlayException is thrown when
     *             a collision between the RAM and other devices
     *             in the address space occur.
     */
    public static void createRAMAndAddToAS(AddressSpace as, String description,
                                           URL documentBase)
                                        throws AddressBankOverlapException,
                                               IllegalArgumentException {
        StringTokenizer st = new StringTokenizer(description);
        int numberOfTokens = st.countTokens();
        if (numberOfTokens < 3) {
            throw new IllegalArgumentException(MSG_3_ITEMS);
        }
        int numberOfBits = Util.parseNumber(st.nextToken());
        int offset = Util.parseNumber(st.nextToken());
        String thirdToken = st.nextToken();
        int modeMask = 0;
        boolean readOnlyFlag = false;
        if (numberOfTokens > 3) {
            String fourthToken = st.nextToken();
            if (Character.isDigit(fourthToken.charAt(0))) {
                modeMask = Util.parseNumber(fourthToken);
                if (numberOfTokens > 4) {
                    readOnlyFlag = st.nextToken().equalsIgnoreCase("readOnly");
                }
            } else {
                readOnlyFlag = fourthToken.equalsIgnoreCase("readOnly");
            }
        }
        RAM ram = null;
        if (Character.isDigit(thirdToken.charAt(0))) {
            ram = new RAM(numberOfBits, Util.parseNumber(thirdToken),
                          readOnlyFlag);
        } else {
            URL url = null;
            try {
                if (thirdToken.startsWith("http://")) {
                    url = new URL(thirdToken);
                } else {
                    url = new URL(documentBase, thirdToken);
                }
                ram = new RAM(numberOfBits, url, readOnlyFlag);
            } catch (MalformedURLException e) {
                throw new IllegalArgumentException(MSG_INVALID_URL + e);
            } catch (IOException ioe) {
                throw new IllegalArgumentException(MSG_IO_ERROR + ioe);
            }
        }
        as.addDevice(offset, modeMask, ram);
    }

    /**
     *  Return <code>true</code> if read only. That is, all
     *  invocations of <code>setValue</code> and <code>setByte</code>.
     *  will be ignored.
     */
    public boolean isReadOnly() {
        return readOnly;
    }

    /**
     *  Create memory of a given number of addressable units.
     *  @param numberOfUnits number of addressable units.
     */
    public void createMemory(int numberOfUnits) {
        this.numberOfUnits = numberOfUnits;
        memory = new byte[numberOfUnits * numberOfBytes];
    }

    /** Return the memory array. */
    public byte[] getMemory() {
        return memory;
    }

//
//  Implementation of the Device interface
//

    /** Fill the memory with zeros. */
    public void reset() {
        for (int i = 0; i < memory.length; i++) {
            memory[i] = NULL;
        }
    }

    /** Return the size of an addressable unit. */
    public int getNumberOfBits() {
        return numberOfBits;
    }

    /** Return the number of addressable units. */
    public int getNumberOfUnits() {
        return numberOfUnits;
    }

    /**
     *  Get the value of the addressable unit at the given address.
     *  @param address memory address.
     *  @return the value.
     */
    public int getValue(int address) {
        address *= numberOfBytes;
        int result = 0;
        for (int i = 0; i < numberOfBytes; i++) {
            result = (result << 8)
                     + (((int) memory[address + i]) & 0xff);
        }
        return result & mask;
    }

    /**
     *  Set the value of the addressable unit at the given address.
     *  @param address memory address.
     *  @param value the new value.
     */
    public void setValue(int address, int value) {
        if (readOnly) {
            return;
        }

        address *= numberOfBytes;
        int result = 0;
        for (int i = numberOfBytes - 1; i >= 0; i--) {
            memory[address + i] = (byte) (value & 0xff);
            value >>= 8;
        }
    }

    /**
     *  Get the byte at the given address.
     *  This method is similar
     *  to <code>getValue</code> but has a better performance.
     *  <br>WARNING: This method should
     *  be used only when the addressable unit is a byte.
     *  @param address memory address.
     *  @return the value of the byte at <code>address</code>.
     */
    public byte getByte(int address) {
        return memory[address];
    }

    /**
     *  Set the byte at the given address.
     *  This method is similar
     *  to <code>setValue</code> but has a better performance.
     *  <br>WARNING: This method should
     *  be used only when the addressable unit is a byte.
     *  @param address memory address.
     *  @param value new value of the byte at <code>address</code>.
     */
    public void setByte(int address, byte value) {
        if (readOnly) {
            return;
        }

        memory[address] = value;
    }
}
