package jfrace.device;

import jfrace.Device;

import java.awt.Canvas;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.image.IndexColorModel;
import java.awt.image.MemoryImageSource;
import java.awt.event.MouseEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseMotionAdapter;

/**
 *  Simulates a graphic display with the Thomson Graphic Display
 *  Processor (GDP) EF9365. The reserved registers with the addresses
 *  14 and 15 are used to read the screen memory and the light pen
 *  switch, respectively.
 *
 *  @author Franz-Josef
 */
public class ThomsonGDP extends Canvas implements Runnable, Device {
    /** Address of STATUS register. */
    public static final int ADR_STATUS = 0;

    /** Address of CMD register. */
    public static final int ADR_CMD = 0;

    /** Address of CTRL1 register. */
    public static final int ADR_CTRL1 = 1;

    /** Address of CTRL2 register. */
    public static final int ADR_CTRL2 = 2;

    /** Address of CSIZE register. */
    public static final int ADR_CSIZE = 3;

    /** Address of DELTAX register. */
    public static final int ADR_DELTAX = 5;

    /** Address of DELTAY register. */
    public static final int ADR_DELTAY = 7;

    /** Address of X register (MSB). */
    public static final int ADR_X_MSB = 8;

    /** Address of X register (LSB). */
    public static final int ADR_X_LSB = 9;

    /** Address of Y register (MSB). */
    public static final int ADR_Y_MSB = 10;

    /** Address of Y register (LSB). */
    public static final int ADR_Y_LSB = 11;

    /** Address of XLP (light pen) register. */
    public static final int ADR_XLP = 12;

    /** Address of YLP (light pen) register. */
    public static final int ADR_YLP = 13;

    /** Address of MEM (memory) register. */
    public static final int ADR_MEM = 14;

    /** Address of LSWITCH (light pen switch) register. */
    public static final int ADR_LSWITCH = 15;

    /** Command: Select pen. */
    public static final byte PEN = 0;

    /** Command: Select eraser. */
    public static final byte ERASER = 1;

    /** Command: Pen/eraser down. */
    public static final byte DOWN = 2;

    /** Command: Pen/eraser up. */
    public static final byte UP = 3;

    /** Command: Clear screen. */
    public static final byte CLEAR = 4;

    /** Command: Reset everything. */
    public static final byte RESET = 7;

    /** Light pen pressed. */
    public static final byte LPEN_PRESSED = (byte) 0xfb;

    /** Light pen released. */
    public static final byte LPEN_RELEASED = (byte) 0xff;

    /** Command: 5x8 block. */
    public static final byte BLOCK58 = 10;

    /** Command: 4x4 block. */
    public static final byte BLOCK44 = 11;

    private static final String[] CHAR_GENERATOR = {
        "     | MM  | MMM |MMMM | MM  |     ",
        "     |M  M |M   M|M   M| MM  |     ",
        "     |M  M |M MMM|M   M| M   | MMM ",
        "     |M  M |M M M|MMMM |  M  | M  M",
        "     |M  M |M MMM|M    |     | M  M",
        "     |M  M |M    |M    |     | MMM ",
        "     | MM  | MMMM|M    |     | M   ",
        "     |     |     |     |     | M   ",
        "-----+-----+-----+-----+-----+-----",
        "  M  |  M  | MMM | MMM |     |     ",
        "  M  | MM  |M   M|M   M|     |     ",
        "  M  |  M  |M   M|M   M| MM  | MMM ",
        "  M  |  M  |MMMMM|M   M|   M |M  M ",
        "  M  |  M  |M   M|M M M|MMMM |M  M ",
        "     |  M  |M   M|M  M |M  M | MMM ",
        "  M  |  M  |M   M| MM M|MMMMM|   M ",
        "     |     |     |     |     |   MM",
        "-----+-----+-----+-----+-----+-----",
        " M M | MMM |MMMM |MMMM | M   |     ",
        " M M |M   M|M   M|M   M| M   |     ",
        " M M |    M|M   M|M   M| MMMM| M MM",
        "     |  MM |MMMM |MMMM | M  M| MM  ",
        "     | M   |M   M|M M  | M  M| M   ",
        "     |M    |M   M|M  M | M  M| M   ",
        "     |MMMMM|MMMM |M   M|MMMM | M   ",
        "     |     |     |     |     |     ",
        "-----+-----+-----+-----+-----+-----",
        " M M |MMMMM| MMM | MMM |     |     ",
        "MMMMM|    M|M   M|M   M|     |     ",
        " M M |   M |M    |M    |  MM |  MMM",
        "MMMMM|  MM |M    | MMM | M  M| M   ",
        " M M |    M|M    |    M| M   |  MM ",
        "     |    M|M   M|M   M| M   |    M",
        "     |MMMM | MMM | MMM |  MMM| MMM ",
        "     |     |     |     |     |     ",
        "-----+-----+-----+-----+-----+-----",
        "  M  |M  M |MMMM |MMMMM|   M |     ",
        " MMMM|M  M |M   M|  M  |   M |  M  ",
        "M M  |M  M |M   M|  M  | MMM | MMM ",
        " MMM |MMMMM|M   M|  M  |M  M |  M  ",
        "  M M|   M |M   M|  M  |M  M |  M  ",
        "MMMM |   M |M   M|  M  |M  M |  M M",
        "  M  |   M |MMMM |  M  | MMMM|   M ",
        "     |     |     |     |     |     ",
        "-----+-----+-----+-----+-----+-----",
        "MM   |MMMMM|MMMMM|M   M|     |     ",
        "MM  M|M    |M    |M   M|     |     ",
        "   M |MMMM |M    |M   M|  MM |M  M ",
        "  M  |    M|MMMM |M   M| M  M|M  M ",
        " M   |    M|M    |M   M| MMM |M  M ",
        "M  MM|    M|M    |M   M| M   |M  M ",
        "   MM|MMMM |MMMMM| MMM |  MMM| MMMM",
        "     |     |     |     |     |     ",
        "-----+-----+-----+-----+-----+-----",
        " MM  |  MM |MMMMM|M   M|     |     ",
        "M  M | M   |M    |M   M|   MM|     ",
        "M M  |M    |M    |M   M|  M  |M   M",
        " M   |MMMM |MMMM |M   M| MMM |M   M",
        "M M M|M   M|M    |M   M|  M  | M M ",
        "M  M |M   M|M    | M M |  M  | M M ",
        " MM M| MMM |M    |  M  |  M  |  M  ",
        "     |     |     |     |     |     ",
        "-----+-----+-----+-----+-----+-----",
        "  MM |MMMMM| MMM |M   M|     |     ",
        "  MM |    M|M   M|M   M|     |     ",
        "   M |   M |M    |M   M|  MMM|M   M",
        "  M  |  M  |M MMM|M M M| M  M|M   M",
        "     | M   |M   M|M M M| M  M|M M M",
        "     |M    |M   M|MM MM|  MMM|M M M",
        "     |M    | MMMM|M   M|    M| M M ",
        "     |     |     |     | MMM |     ",
        "-----+-----+-----+-----+-----+-----",
        "   M | MMM |M   M|M   M| M   |     ",
        "  M  |M   M|M   M|M   M| M   |     ",
        " M   |M   M|M   M| M M | MMM |MM  M",
        " M   | MMM |MMMMM|  M  | M  M|  MM ",
        " M   |M   M|M   M| M M | M  M|  M  ",
        "  M  |M   M|M   M|M   M| M  M| MM  ",
        "   M | MMM |M   M|M   M| M  M|M  MM",
        "     |     |     |     |     |     ",
        "-----+-----+-----+-----+-----+-----",
        " M   | MMM | MMM |M   M|     |     ",
        "  M  |M   M|  M  |M   M|  M  |     ",
        "   M |M   M|  M  |M   M|     | M  M",
        "   M | MMMM|  M  | M M |  M  | M  M",
        "   M |    M|  M  |  M  |  M  | M  M",
        "  M  |   M |  M  |  M  |  M  |  MMM",
        " M   | MM  | MMM |  M  |  M  |    M",
        "     |     |     |     |     |   MM",
        "-----+-----+-----+-----+-----+-----",
        "  M  |     |    M|MMMMM|     |     |MMMMM",
        "M M M| MM  |    M|    M|     |     |MMMMM",
        " MMM | MM  |    M|   M |   M | MMMM|MMMMM",
        "MMMMM|     |    M|  M  |     |   M |MMMMM",
        " MMM | MM  |    M| M   |   M |  M  |MMMMM",
        "M M M| MM  |M   M|M    |   M | M   |MMMMM",
        "  M  |     | MMM |MMMMM| M M |MMMM |MMMMM",
        "     |     |     |     |  M  |     |MMMMM",
        "-----+-----+-----+-----+-----+-----",
        "     |     |M   M| MMM |     |   MM|     ",
        "  M  | MM  |M  M | M   | M   |  M  |     ",
        "  M  | MM  |M M  | M   | M  M|  M  |     ",
        "MMMMM|     |MM   | M   | M M | M   |     ",
        "  M  | MM  |M M  | M   | MM  |  M  |MMMM ",
        "  M  | MM  |M  M | M   | M M |  M  |MMMM ",
        "     |  M  |M   M| MMM | M  M|   MM|MMMM ",
        "     | M   |     |     |     |     |MMMM ",
        "-----+-----+-----+-----+-----+-----",
        "     |   M |M    |     |     |  M  ",
        "     |  M  |M    |M    | MM  |  M  ",
        "     | M   |M    | M   |  M  |  M  ",
        "     |M    |M    |  M  |  M  |     ",
        "  MM | M   |M    |   M |  M  |  M  ",
        "  MM |  M  |M    |    M|  M  |  M  ",
        "   M |   M |MMMMM|     |  MM |  M  ",
        "  M  |     |     |     |     |     ",
        "-----+-----+-----+-----+-----+-----",
        "     |     |M   M| MMM |     | MM  ",
        "     |     |MM MM|   M |     |   M ",
        "     |MMMMM|M M M|   M |MMMM |   M ",
        "MMMMM|     |M M M|   M |M M M|    M",
        "     |MMMMM|M   M|   M |M M M|   M ",
        "     |     |M   M|   M |M M M|   M ",
        "     |     |M   M| MMM |M M M| MM  ",
        "     |     |     |     |     |     ",
        "-----+-----+-----+-----+-----+-----",
        "     | M   |M   M|  M  |     |     ",
        "     |  M  |MM  M| MMM |     |     ",
        "     |   M |M M M|M M M| MMM |     ",
        "     |    M|M  MM|  M  | M  M|MMMMM",
        "     |   M |M   M|  M  | M  M|    M",
        " MM  |  M  |M   M|  M  | M  M|    M",
        " MM  | M   |M   M|  M  | M  M|     ",
        "     |     |     |     |     |     ",
        "-----+-----+-----+-----+-----+-----",
        "     | MMM | MMM |     |     |M M M",
        "    M|M   M|M   M|  M  |     | M M ",
        "   M |    M|M   M| M   |  MM |M M M",
        "  M  |   M |M   M|MMMMM| M  M| M M ",
        " M   |  M  |M   M| M   | M  M|M M M",
        "M    |     |M   M|  M  | M  M| M M ",
        "     |  M  | MMM |     |  MM |M M M",
        "     |     |     |     |     |     "};

    private static final int WIDTH = 512;
    private static final int HEIGHT = 256;
    private static final int BORDER = 5;

    private static final byte NULL = 0;
    private static final byte ONE = 1;

    private static final int[][] LINE_PATTERNS = {
                {1}, {1, 1, 0, 0}, {1, 1, 1, 1, 0, 0, 0, 0},
                {1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 1, 1, 0, 0}};
    private static final int TIME_STEP = 100;

    // Registers
    private byte regSTATUS = 5;
    private byte regCTRL1;
    private byte regCTRL2;
    private byte regCSIZE;
    private int regDELTAX;
    private int regDELTAY;
    private int regX;
    private int regY;
    private int lightPenX;
    private int lightPenY;
    private byte lightPenSwitch = LPEN_RELEASED;

    private int[] linePattern;

    // Screen
    private byte[] screen = new byte[WIDTH * HEIGHT];
    private IndexColorModel colorModel;
    private MemoryImageSource source;
    private Image image;
    private long lastRepaintTime;
    private boolean repaintRequested;
    private Thread repaintThread;
    private boolean busy;


    /**
     *  Default constructor. Sets the background to black and the
     *  foreground to orange.
     */
    public ThomsonGDP() {
        super();

        addMouseListener(new MouseAdapter() {
                                public void mouseReleased(MouseEvent me) {
                                    lightPenSwitch = LPEN_RELEASED;
                                }
                                public void mousePressed(MouseEvent me) {
                                    requestFocus();
                                    lightPenSwitch = LPEN_PRESSED;
                                }
                            });
        addMouseMotionListener(new MouseMotionAdapter() {
                                public void mouseMoved(MouseEvent me) {
                                    lightPenX = (((me.getX() - BORDER) >> 1)
                                                  & 0xfc) + 1;
                                    lightPenY = HEIGHT - (me.getY() - BORDER);
                                }
                            });

        setBackground(java.awt.Color.black);
        setForeground(java.awt.Color.orange);
        reset();
        repaintThread = new Thread(this);
        repaintThread.start();
    }

    /** Create the color model from the background and forground colors. */
    private void createColorModel() {
        byte[] red = new byte[2];
        byte[] green = new byte[red.length];
        byte[] blue = new byte[red.length];
        int[] color = new int[red.length];
        color[0] = getBackground().getRGB();
        color[1] = getForeground().getRGB();
        for (int i = 0; i < red.length; i++) {
            red[i] = (byte) (color[i] >> 16);
            green[i] = (byte) (color[i] >> 8);
            blue[i] = (byte) color[i];
        }
        colorModel = new IndexColorModel(1, 2, red, green, blue);
    }

    public Dimension preferredSize() {
        return new Dimension(WIDTH + 2 * BORDER, HEIGHT + 2 * BORDER);
    }

    /**
     *  Return <code>true</code> because focus should be traversable with
     *  TAB key.
     */
    public boolean isFocusTraversable() {
        return true;
    }

    /** Directly calls <code>paint</code>. */
    public void update(Graphics g) {
        paint(g);
    }

    /** Paint the screen memory. */
    public void paint(Graphics g) {
        if (source == null) {
            createColorModel();
            source = new MemoryImageSource(WIDTH, HEIGHT, colorModel,
                                           screen, 0, WIDTH);
            source.setAnimated(true);
            image = createImage(source);
        }
        g.drawImage(image, BORDER, BORDER, WIDTH, HEIGHT, this);
        busy = false;
    }

    /**
     *  Set the <code>repaintRequested</code> flag in order to inform
     *  the repainting thread that the screen memory has to be repainted.
     */
    private void repaintScreen() {
        repaintRequested = true;
    }

    /**
     *  Return the 8 horizontal pixels at the current position
     *  given by <code>regX</code> and <code>regY</code>.
     *  Note that the 3 lower bits of <code>regX</code> are ignored.
     */
    private byte getMemoryByte() {
        int adr = ((HEIGHT - 1 - regY) & 0xff) * WIDTH + (regX & 0x1f8) + 7;
        int result = 0;
        for (int i = 0; i < 8; i++, adr--) {
            result <<= 1;
            if (adr >= 0) {
                result += 1 - screen[adr];
            }
        }
        return (byte) result;
    }

    /**
     *  Process ThomsonGDP command.
     *  If busy (i.e. <code>busy == true</code>) due to painting
     *  the screen it waits until ready.
     *  @param cmd Command byte.
     */
    private void processCommand(byte cmd) {
        while (busy) {
            try {
                Thread.sleep(10);
            } catch (Exception e) {}
        }
        int c = cmd & 0xff;
        switch (c) {
        case 0x00: regCTRL1 |= 2; break;
        case 0x01: regCTRL1 &= 0xfd; break;
        case 0x02: regCTRL1 |= 1; break;
        case 0x03: regCTRL1 &= 0xfe; break;
        case 0x04: fillScreen(NULL); break;
        case 0x05: regX = regY = 0; break;
        case 0x06: regX = regY = 0; fillScreen(NULL); break;
        case 0x07: reset(); break;
        case 0x08:
        case 0x09: break;
        case 0x0a: drawCharacter(0x8a); break;
        case 0x0b: drawCharacter(0x8b); break;
        case 0x0c: fillScreen((byte) (regCTRL1 >> 1)); break;
        case 0x0d: regX = 0; break;
        case 0x0e: regY = 0; break;
        case 0x0f: break;
        default:
            if (c < 0x18) {
                drawVector(regDELTAX, regDELTAY, c & 7);
            } else if (c < 0x20) {
                int delta = Math.max(regDELTAX, regDELTAY);
                drawVector(delta, delta, c & 7);
            } else if (c < 0x80) {
                drawCharacter(c);
            } else {
                drawVector((c >> 5) & 3, (c >> 3) & 3, c & 7);
            }
        }
    }

    /**
     *  Fill screen with given color.
     *  @param color index of the color (0 = background color,
     *               1 = foreground color)
     */
    private void fillScreen(byte color) {
        for (int i = 0; i < screen.length; i++) {
            screen[i] = color;
        }
        repaintScreen();
    }

    /**
     *  Draw a character onto the screen.
     *  Uses the character generator <code>CHAR_GENERATOR</code>.
     *  The character scaling is given by <code>regCSIZE</code>.
     *  @param c ASCII character or one of the two block characters.
     */
    private void drawCharacter(int c) {
        int genIndexY = (c & 0xf) * 9;
        int genIndexX = (((c >> 4) & 0xf) - 2) * 6;
        int xScale = (regCSIZE >> 4) & 0xf;
        int yScale = regCSIZE & 0xf;
        byte color = (byte) ((regCTRL1 >> 1) & 1);
        String genLine;
        boolean alter;
        if ((regCTRL2 & 8) == 0) { // characters on horizontal axis
            if ((regCTRL1 & 1) != 0) { // pen down
                int tilt = ((regCTRL2 >> 2) & 1) * xScale;
                int x, gx;
                int y = HEIGHT - regY - 8 * yScale;
                int yw;
                for (int j = 0; j < 8; j++) {
                    genLine = CHAR_GENERATOR[genIndexY++];
                    for (int js = 0; js < yScale; js++, y++) {
                        if (y >= 0 && y < HEIGHT) {
                            yw = y * WIDTH;
                            x = regX + (7 - j) * tilt;
                            gx = genIndexX;
                            for (int i = 0; i < 5; i++) {
                                alter = genLine.charAt(gx++) == 'M';
                                if (alter) {
                                    for (int is = 0; is < xScale; is++, x++) {
                                        if (x >= 0 && x < WIDTH) {
                                            screen[yw + x] = color;
                                        }
                                    }
                                } else {
                                    x += xScale;
                                }
                            }
                        }
                    }
                }
                repaintScreen();
            }
            regX += ((c == 0x8b) ? 4 : 6) * xScale;
        } else {            // character on vertical axis
            if ((regCTRL1 & 1) != 0) { // pen down
                int tilt = ((regCTRL2 >> 2) & 1) * yScale;
                int y, gx;
                int x = regX - 8 * xScale;
                int xw;
                for (int j = 0; j < 8; j++) {
                    genLine = CHAR_GENERATOR[genIndexY++];
                    for (int js = 0; js < xScale; js++, x++) {
                        if (x >= 0 && x < WIDTH) {
                            y = HEIGHT - regY - (7 - j) * tilt;
                            gx = genIndexX;
                            for (int i = 0; i < 5; i++) {
                                alter = genLine.charAt(gx++) == 'M';
                                if (alter) {
                                    for (int is = 0; is < yScale; is++, y--) {
                                        if (y >= 0 && y < HEIGHT) {
                                            screen[y * WIDTH + x] = color;
                                        }
                                    }
                                } else {
                                    y -= yScale;
                                }
                            }
                        }
                    }
                }
                repaintScreen();
            }
            regY += ((c == 0x8b) ? 4 : 6) * yScale;
        }
    }

    /**
     *  Draw a line starting from the current position.
     *  The current position is given by <code>regX</code> and
     *  <code>regY</code>. If the first bit of <code>regCTRL1</code>
     *  is 0 only position will changed and nothing is drawn.
     *  @param deltaX Absolute value of the distance in <i>x</i> for the
     *                new position.
     *  @param deltaY Absolute value of the distance in <i>y</i> for the
     *                new position.
     *  @param direction Direction code (see data sheet of Thomson GDP)
     */
    private void drawVector(int deltaX, int deltaY, int direction) {
        switch (direction) {
        case 0: drawVector(deltaX, 0, 1, 0); break;
        case 1: drawVector(deltaX, deltaY, 1, 1); break;
        case 2: drawVector(0, deltaY, 0, 1); break;
        case 3: drawVector(deltaX, deltaY, -1, 1); break;
        case 4: drawVector(0, deltaY, 0, -1); break;
        case 5: drawVector(deltaX, deltaY, 1, -1); break;
        case 6: drawVector(deltaX, 0, -1, 0); break;
        case 7: drawVector(deltaX, deltaY, -1, -1); break;
        }
    }

    /**
     *  Draw a line starting from current position.
     *  The current position is given by <code>regX</code> and
     *  <code>regY</code>. If the first bit of <code>regCTRL1</code>
     *  is 0 only position will changed and nothing is drawn.
     *  @param deltaX Absolute value of the distance in <i>x</i> for the
     *                new position.
     *  @param deltaY Absolute value of the distance in <i>y</i> for the
     *                new position.
     *  @param dx Sign of the distance in <i>x</i> to the new position.
     *  @param dy Sign of the distance in <i>y</i> to the new position.
     */
    private void drawVector(int deltaX, int deltaY, int dx, int dy) {
        if ((regCTRL1 & 1) == 0) {  // pen up
            regX += dx * deltaX;
            regY += dy * deltaY;
        } else {                    // pen down
            // using Bresenham's algorithm
            byte color = (byte) ((regCTRL1 >> 1) & 1);
            int count = 0;
            int epsilon;
            int x0 = regX;
            int y0 = regY;
            plot(regX, regY, color, count++);
            if (deltaY < deltaX) {
                epsilon = 2 * deltaY - deltaX;
                for (int i = 0; i < deltaX; i++) {
                    if (epsilon > 0) {
                        regY += dy;
                        epsilon += 2 * (deltaY - deltaX);
                    } else {
                        epsilon += 2 * deltaY;
                    }
                    regX += dx;
                    plot(regX, regY, color, count++);
                }
            } else {
                epsilon = 2 * deltaX - deltaY;
                for (int i = 0; i < deltaY; i++) {
                    if (epsilon > 0) {
                        regX += dx;
                        epsilon += 2 * (deltaX - deltaY);
                    } else {
                        epsilon += 2 * deltaX;
                    }
                    regY += dy;
                    plot(regX, regY, color, count++);
                }
            }
            repaintScreen();
        }
    }

    /**
     *  Set the pixel a given position. The pixel is only set if
     *  the current line pattern allows it.
     *  @param x X position.
     *  @param y Y position.
     *  @param color Pixel color.
     *  @param count Pointer into the current line pattern.
     */
    private void plot(int x, int y, byte color, int count) {
        if (0 <= y && y < HEIGHT && 0 <= x && x < WIDTH) {
            if (linePattern[count % linePattern.length] == 1) {
                screen[(HEIGHT - 1 - y) * WIDTH + x] = color;
            }
        }
    }

//
//  Implementation of the Runnable interface.
//

    /**
     *  Run repaint thread. Polling for repaint requests.
     */
    public void run() {
        Thread t = Thread.currentThread();
        while (repaintThread == t) {
            if (repaintRequested && source != null && !busy) {
                busy = true;
                source.newPixels();
                repaintRequested = false;
            }
            try {
                Thread.sleep(TIME_STEP);
            } catch (Exception e) {}
            regSTATUS ^= 2;
        }
    }

//
//  Implementation of the Device interface.
//

    /** Reset. */
    public void reset() {
//        regSTATUS = 4;
//        regCTRL1 = 0;
        regCTRL2 = 0;
        linePattern = LINE_PATTERNS[0];
        regCSIZE = 0x11;
        regDELTAX = 0;
        regDELTAY = 0;
        regX = 0;
        regY = 0;
        fillScreen(NULL);
    }

    /** Return 8. */
    public int getNumberOfBits() {
        return 8;
    }

    /** Return 16. */
    public int getNumberOfUnits() {
        return 16;
    }

    /**
     *  Get the value of a register of ThomsonGDP.
     *  @param address register address.
     *  @return the value.
     *  @see getByte
     */
    public int getValue(int address) {
        return (int) getByte(address);
    }

    /**
     *  Set the value of a register of ThomsonGDP.
     *  @param address register address.
     *  @param value the new value.
     *  @see setByte
     */
    public void setValue(int address, int value) {
        setByte(address, (byte) value);
    }

    /**
     *  Get the byte of a register
     *  @param address register address.
     *  @return the value of the byte at <code>address</code>.
     *  @see getValue
     */
    public byte getByte(int address) {
//        System.out.println("getByte("+address+")");
        switch (address) {
        case ADR_STATUS: return regSTATUS;
        case ADR_CTRL1: return regCTRL1;
        case ADR_CTRL2: return regCTRL2;
        case ADR_CSIZE: return regCSIZE;
        case ADR_DELTAX: return (byte) regDELTAX;
        case ADR_DELTAY: return (byte) regDELTAY;
        case ADR_X_MSB: return (byte) ((regX >> 8) & 0x0f);
        case ADR_X_LSB: return (byte) (regX & 0xff);
        case ADR_Y_MSB: return (byte) ((regY >> 8) & 0x0f);
        case ADR_Y_LSB: return (byte) (regY & 0xff);
        case ADR_XLP: return (byte) lightPenX;
        case ADR_YLP: return (byte) lightPenY;
        case ADR_MEM: return getMemoryByte();
        case ADR_LSWITCH: return lightPenSwitch;
        default: return 0;
        }
    }

    /**
     *  Set the byte of a register.
     *  @param address register address.
     *  @param value new value of the byte at <code>address</code>.
     *  @see setValue
     */
    public void setByte(int address, byte value) {
//        System.out.println("setByte("+address+", "+Integer.toHexString(value & 0xff)+") '"+((char) (Math.max(32, value) & 0x7f)) + "' CTRL1="+regCTRL1);
        switch (address) {
        case ADR_CMD: processCommand(value); break;
        case ADR_CTRL1: regCTRL1 = value; break;
        case ADR_CTRL2:
            regCTRL2 = value;
            linePattern = LINE_PATTERNS[value & 3];
            break;
        case ADR_CSIZE: regCSIZE = value; break;
        case ADR_DELTAX: regDELTAX = value & 0xff; break;
        case ADR_DELTAY: regDELTAY = value & 0xff; break;
        case ADR_X_MSB: regX = (regX & 0xff) + ((value & 0x0f) << 8); break;
        case ADR_X_LSB: regX = (regX & 0xf00) + (value & 0xff); break;
        case ADR_Y_MSB: regY = (regY & 0xff) + ((value & 0x0f) << 8); break;
        case ADR_Y_LSB: regY = (regY & 0xf00) + (value & 0xff); break;
        }
    }
}
