package jfrace.device;

import java.io.ByteArrayInputStream;

import sun.audio.AudioData;
import sun.audio.AudioDataStream;
import sun.audio.AudioPlayer;

/**
 *  Audio player based on a wave table. The player uses the undocumented
 *  audio player <code>sun.audio.AudioPlayer</code>.
 *  The <code>sun.audio</code> package is part of every JDK and of
 *  Java implementation of (presumably) all web browsers supporting
 *  Java. Note that <code>sun.audio.AudioPlayer</code> plays only
 *  8000 Hz sampled u-law encoded audio data. The method
 *  <code>setWaveTable</code> automatically encodes the data.
 *
 *  @author Franz-Josef Elmer
 */
public class WaveTablePlayer {
    private static final int BIAS = 0x84;      // Bias for linear code.
    /*
     * linear2ulaw() - Convert a linear PCM value to u-law.
     * An adaptation from <code>g711.c</code> from Sun Microsystems.
     *
     * In order to simplify the encoding process, the original linear magnitude
     * is biased by adding 33 which shifts the encoding range from (0 - 8158) to
     * (33 - 8191). The result can be seen in the following encoding table:
     *
     *      Biased Linear Input Code        Compressed Code
     *      ------------------------        ---------------
     *      00000001wxyza                   000wxyz
     *      0000001wxyzab                   001wxyz
     *      000001wxyzabc                   010wxyz
     *      00001wxyzabcd                   011wxyz
     *      0001wxyzabcde                   100wxyz
     *      001wxyzabcdef                   101wxyz
     *      01wxyzabcdefg                   110wxyz
     *      1wxyzabcdefgh                   111wxyz
     *
     * Each biased linear code has a leading 1 which identifies the segment
     * number. The value of the segment number is equal to 7 minus the number
     * of leading 0's. The quantization interval is directly available as the
     * four bits wxyz.  * The trailing bits (a - h) are ignored.
     *
     * Ordinarily the complement of the resulting code word is used for
     * transmission, and so the code word is complemented before it is returned.
     *
     * For further information see John C. Bellamy's Digital Telephony, 1982,
     * John Wiley & Sons, pps 98-111 and 472-476.
     */
    /**
     *  Linear 2 byte to 1 byte u-law.
     *  @param pcm_val 16 bit linear value
     *  @return u-law encoded input
     */
    protected static byte linear2ulaw(int pcm_val) {
        int ival = pcm_val;
        int mask, seg, uval;

        /* Get the sign and the magnitude of the value. */
        if (ival < 0) {
            ival = BIAS - pcm_val;
            mask = 0x7F;
        } else {
            ival += BIAS;
            mask = 0xFF;
        }

        /* Convert the scaled magnitude to segment number. */
        seg = search(ival, seg_end);

        /*
         * Combine the sign, segment, quantization bits;
         * and complement the code word.
         */
        if (seg >= 8) {          /* out of range, return maximum value. */
            return (byte) (0x7F ^ mask);
        } else {
            uval = (seg << 4) | ((ival >> (seg + 3)) & 0xF);
            return (byte) (uval ^ mask);
        }
    }

    /** Table needed by <code>linear2ulaw</code>. */
    private static final short[] seg_end =
        {0xFF, 0x1FF, 0x3FF, 0x7FF, 0xFFF, 0x1FFF, 0x3FFF, 0x7FFF};

    /** Method needed by <code>linear2ulaw</code>. */
    private static int search(int val, short[] table) {
        for (int i = 0; i < table.length; i++) {
            if (val <= table[i]) {
                return i;
            }
        }
        return table.length;
    }

/////////////////////////////////////////////////////////////////

    /** Byte array of the wave table. */
    protected byte[] waveTable;

    private int waveFormLength;      // Length of the actual wave form.
    private WaveTableAudioDataStream wtas; // InputStream of the wave table
    private AudioPlayer ap = AudioPlayer.player; // Player singleton
    private boolean playing = false;    // playing flag

    /**
     *  Create an instance with a given size of the wave table.
     *  @param size Wave table size, i.e. the maximum number of bytes
     *              a wave form can have.
     */
    public WaveTablePlayer(int size) {
        waveTable = new byte[size];
        byte zero = (byte) 0xff;
        for (int i = 0; i < size; i++) {
            waveTable[i] = zero;
        }
        wtas = new WaveTableAudioDataStream(waveTable);
        // Applet.newAudioClip(url);
     }

    /**
     *  Return the size of the wave table. This is the maximum number of
     *  bytes a wave form can have.
     */
    public int getSize() {
        return waveTable.length;
    }

    /**
     *  Return the length of the current wave form.
     */
    public int getLength() {
        return waveFormLength;
    }

    /**
     *  Set the length of the actual wave form
     */
    protected void setLength(int value) {
        waveFormLength = Math.min(value, waveTable.length);
        wtas.setSize(waveFormLength);
    }

    /**
     *  Set the wave form of the wave table. In the playing modus this
     *  wave form will be continuously repeated.
     *  The wave form will be stored in the wave table as u-law encoded
     *  bytes.
     *
     *  @param waveForm Array of the wave form.
     */
    public void setWaveTable(int[] waveForm) {
        setLength(Math.min(waveForm.length, waveTable.length));
        for (int i = 0; i < waveFormLength; i++) {
            waveTable[i] = linear2ulaw(waveForm[i]);
        }
    }

    /**
     *  Start playing and set the playing flag. Does nothing if the
     *  playing flag is already set.
     */
    public void play() {
        if (!playing) {
            playing = true;
            ap.start(wtas);
        }
    }

    /**
     *  Return the playing flag.
     */
    public boolean isPlaying() {
        return playing;
    }

    /**
     *  Stop playing and reset the playing flag. Does nothing if the
     *  playing flag is already reset.
     */
    public void stop() {
        if (playing) {
            playing = false;
            ap.stop(wtas);
        }
    }
}

/**
 *  Wrapper class for wave tables. Need for <code>sun.audio.AudioPlayer</code>
 *  to play a wave table in a loop.
 *
 */
class WaveTableAudioDataStream extends ByteArrayInputStream {
    /** Create an instance for the given wave table. */
    public WaveTableAudioDataStream(byte[] waveTable) {
        super(waveTable);
    }

    /** Set the loop size. */
    public void setSize(int size) {
        count = size;
    }

    public int read() {
        int c = super.read();
        if (c == -1) {
            reset();
            c = super.read();
        }
        return c;
    }

    public int read(byte data[], int off, int len) {
        int k, l;
        for (k = 0; k < len;) {
            l = super.read(data, off + k, len - k);
            if (l >= 0) {
                k += l;
            } else {
                reset();
            }
        }
        return k;
    }
}

