package jfrace.terminal;

import jfrace.Device;
import jfrace.Terminal;

import java.awt.*;
import java.awt.event.*;

/**
 *  A simple alphanumerical terminal.
 *
 *  @author Franz-Josef Elmer
 */
public class SimpleTerminal extends Canvas implements Terminal {
    private int numberOfRows = 25;
    private int numberOfColumns = 80;
    private char[][] content;
    private int cursorRow;
    private int cursorCol;
    private Font font = new Font("Monospaced", Font.PLAIN, 12);
    private int lineHeight;
    private int charWidth;
    private int ascent;
    private Image buffer;
    private Graphics bufferG;
    private Dimension oldSize = new Dimension(0, 0);
    private Dimension preferredSize;
    private ActionListener actionListener;
    private ActionEvent newCharEvent;
    private boolean isWaiting;
    private char[] typeAheadBuffer = new char[20];
    private int freePos = 0;
    private int nextPos = 0;

    private FocusListener fl = new FocusAdapter() {
            public void focusGained(FocusEvent e) {
                System.out.println("Focus gained");
            }
        };
    private KeyListener kl = new KeyAdapter() {
            public void keyTyped(KeyEvent event) {
                if (((freePos + 1) % typeAheadBuffer.length) == nextPos) {
                    System.out.print("\007");
                } else {
                    typeAheadBuffer[freePos++] = event.getKeyChar();
                    freePos %= typeAheadBuffer.length;
                    notifyActionListeners(newCharEvent);
                    if (isWaiting) {
                        synchronized (SimpleTerminal.this) {
                            SimpleTerminal.this.notify();
                        }
                    }
                }
            }
        };
    private MouseListener ml = new MouseAdapter() {
            /** Request focus. */
            public void mousePressed(MouseEvent me) {
                requestFocus();
            }
        };

    public SimpleTerminal() {
        super();

        newCharEvent = new ActionEvent(this, 0, "newCharEvent");
        setBackground(Color.black);
        setForeground(Color.green);
        addKeyListener(kl);
        addMouseListener(ml);
        addFocusListener(fl);
        FontMetrics fm = getFontMetrics(font);
        lineHeight = fm.getHeight();
        charWidth = fm.charWidth(' ');
        ascent = fm.getAscent();
        preferredSize = new Dimension(charWidth * numberOfColumns,
                                      lineHeight * numberOfRows);
        content = new char[numberOfRows][];
        for (int i = 0; i < numberOfRows; i++) {
            content[i] = new char[numberOfColumns];
        }
        clearScreen();
    }

    /** Add an action listener. */
    public void addActionListener(ActionListener al) {
        actionListener = AWTEventMulticaster.add(actionListener, al);
    }

    /** Remove an action listener. */
    public void removeActionListener(ActionListener al) {
        actionListener = AWTEventMulticaster.remove(actionListener, al);
    }

    /** Notify action listeners. */
    private void notifyActionListeners(ActionEvent event) {
        if (actionListener != null) {
            actionListener.actionPerformed(event);
        }
    }

    /**
     *  Return <code>true</code> because focus should be traversable with
     *  TAB key.
     */
    public boolean isFocusTraversable() {
        return true;
    }

    public Dimension preferredSize() {
        return preferredSize;
    }

    /** Clear screen and set the cursor into the upper left corner. */
    public void clearScreen() {
        for (int i = 0; i < numberOfRows; i++) {
            for (int j = 0; j < numberOfColumns; j++) {
                content[i][j] = ' ';
            }
        }
        cursorRow = 0;
        cursorCol = 0;
    }

    private void processLineFeed() {
        if (cursorRow == numberOfRows - 1) { // scroll
            scroll();
        } else {
            cursorRow++;
        }
    }

    private void scroll() {
        char[] row = content[0];
        System.arraycopy(content, 1, content, 0, numberOfRows - 1);
        content[cursorRow] = row;
        for (int i = 0; i < numberOfColumns; i++) {
            row[i] = ' ';
        }
    }

    public void update(Graphics g) {
        paint(g);
    }

    public void paint(Graphics g) {
        Dimension size = size();
        if (size.width != oldSize.width || size.height != oldSize.height) {
            if (bufferG != null) {
                bufferG.dispose();
            }
            if (buffer != null) {
                buffer.flush();
            }
            buffer = createImage(size.width, size.height);
            bufferG = buffer.getGraphics();
            oldSize = size;
        }
        bufferG.setColor(getBackground());
        bufferG.fillRect(0, 0, size.width, size.height);
        bufferG.setColor(getForeground());

        bufferG.setFont(font);
        int y = ascent;
        for (int i = 0; i < numberOfRows; i++) {
            bufferG.drawChars(content[i], 0, numberOfColumns, 0, y);
            y += lineHeight;
        }
        int x = cursorCol * charWidth;
        y = cursorRow * lineHeight;
        bufferG.fillRect(x, y, charWidth, lineHeight);
        bufferG.setColor(getBackground());
        bufferG.drawChars(content[cursorRow], cursorCol, 1, x, y + ascent);

        g.drawImage(buffer, 0, 0, null);
    }

//
//  Implementation of the Terminal interface
//

    /** Returns "Simple Terminal". */
    public String getName() {
        return "Simple Terminal";
    }

    /** Returns <code>null</code>. */
    public Device getDevice() {
        return null;
    }

    public boolean isEmpty() {
        return nextPos == freePos;
    }

    public char readChar() {
        if (isEmpty()) {
            return '\0';
        } else {
            char c = typeAheadBuffer[nextPos++];
            nextPos %= typeAheadBuffer.length;
            return c;
        }
    }

    public synchronized char getNextChar() {
        if (isEmpty()) {
            try {
                isWaiting = true;
                wait();
            } catch (Exception e) {}
            isWaiting = false;
        }
        char c = typeAheadBuffer[nextPos++];
        nextPos %= typeAheadBuffer.length;
        return c;
    }

    /**
     *  Write a character onto the screen. The following control
     *  characters will be treated specially:
     *  <table border=1>
     *      <tr><td><code>\r</code></td><td></td></tr>
     *      <tr><td><code>\n</code></td><td></td></tr>
     *      <tr><td><code>\t</code></td><td></td></tr>
     *      <tr><td><code>\f</code></td><td></td></tr>
     *      <tr><td><code>\b</code></td><td></td></tr>
     *  </table>
     */
    public void writeChar(char c) {
        switch (c) {
        case '\r': cursorCol = 0; break;
        case '\n': processLineFeed(); break;
        case '\f': clearScreen(); break;
        case '\b':
            if (--cursorCol < 0) {
                cursorCol = numberOfColumns - 1;
                cursorRow = (cursorRow - 1 + numberOfRows) % numberOfRows;
            }
            break;
        case '\t':
            do {
                writeChar(' ');
            } while ((cursorCol % 8) != 7);
            break;
        default:
            if (c >= ' ') {
//                content[cursorRow][cursorCol++] = Character.toUpperCase(c);
                content[cursorRow][cursorCol++] = c;
            }
        }
        if (cursorCol >= numberOfColumns) { // new line
            cursorCol = 0;
            processLineFeed();
        }
        repaint();
    }
}
