package jfrace.terminal;

import jfrace.Device;
import jfrace.Terminal;
import jfrace.device.ThomsonGDP;

import java.awt.AWTEventMulticaster;
import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Panel;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyAdapter;

/**
 *  A terminal based on Thomson's GDP EF9366 and using Osborne I
 *  command (ESC) sequences.
 *
 *  @author Franz-Josef Elmer
 */
public class ThomsonGDPTerminal extends Panel implements Terminal {
    private static final int NUMBER_OF_ROWS = 25;
    private static final int NUMBER_OF_COLS = 80;
    private static final int X_BIAS = 16;
    private static final int Y_BIAS = 20;
    private static final int LINE_HEIGHT = 9;
    private static final byte ULINE_RIGHT = (byte) 0xe0;
    private static final byte ULINE_LEFT = (byte) 0xe6;

    private ThomsonGDP gdp;

    private ActionListener actionListener;
    private ActionEvent newCharEvent;
    private boolean isWaiting;
    private char[] typeAheadBuffer = new char[20];
    private int freePos = 0;
    private int nextPos = 0;

    private char[][] lines;
    private int cursorRow;
    private int cursorCol;
    private int oldCursorRow;
    private int oldCursorCol;
    private int escapeSequenceIndex = 0;
    private boolean underlining;

    Object thisTerminal; // needed for the KeyAdapter

    /**
     *  Create a new instance.
     */
    public ThomsonGDPTerminal() {
        super();

        gdp = new ThomsonGDP();
        add(gdp);

        thisTerminal = this;

        newCharEvent = new ActionEvent(this, 0, "newCharEvent");
        gdp.addKeyListener(new KeyAdapter() {
                    public void keyTyped(KeyEvent event) {
                        if (((freePos + 1) % typeAheadBuffer.length) == nextPos) {
                            System.out.print("\007");
                        } else {
                            typeAheadBuffer[freePos++] = event.getKeyChar();
                            freePos %= typeAheadBuffer.length;
                            notifyActionListeners(newCharEvent);
                            if (isWaiting) {
                                synchronized (thisTerminal) {
                                    thisTerminal.notify();
                                }
                            }
                        }
                    }
                });

        lines = new char[NUMBER_OF_ROWS][];
        for (int i = 0; i < lines.length; i++) {
            lines[i] = new char[NUMBER_OF_COLS];
            for (int j = 0; j < lines[i].length; j++) {
                lines[i][j] = ' ';
            }
        }
        clearScreen();
    }

    /** Add an action listener. */
    public void addActionListener(ActionListener al) {
        actionListener = AWTEventMulticaster.add(actionListener, al);
    }

    /** Remove an action listener. */
    public void removeActionListener(ActionListener al) {
        actionListener = AWTEventMulticaster.remove(actionListener, al);
    }

    /** Notify action listeners. */
    private void notifyActionListeners(ActionEvent event) {
        if (actionListener != null) {
            actionListener.actionPerformed(event);
        }
    }

    public Dimension preferredSize() {
        return gdp.getPreferredSize();
    }


    /** Clear screen and set the cursor into the upper left corner. */
    public void clearScreen() {
        gdp.setByte(ThomsonGDP.ADR_CMD, ThomsonGDP.RESET);
        gdp.setByte(ThomsonGDP.ADR_CMD, ThomsonGDP.PEN);
        gdp.setByte(ThomsonGDP.ADR_CMD, ThomsonGDP.DOWN);
        for (int i = 0; i < lines.length; i++) {
            clearLine(lines[i]);
        }
        oldCursorRow = -1;
        cursorRow = 0;
        cursorCol = 0;
        drawCursor();
    }

    /** Clear line. That is, fill the array with ' '. */
    private void clearLine(char[] line) {
        for (int i = 0; i < line.length; i++) {
            line[i] = ' ';
        }
    }

    /** Move the cursor down one line. Scroll if necessary. */
    private void processLineFeed() {
        if (cursorRow == lines.length - 1) { // scroll
            deleteLine(0);
        } else {
            cursorRow++;
        }
    }

    /** Insert a character at the current cursor position. */
    private void insertChar(char c) {
        System.out.println("insertChar "+cursorRow+" "+cursorCol);
        for (int i = NUMBER_OF_COLS - 1; i > cursorCol; i--) {
            lines[cursorRow][i] = lines[cursorRow][i - 1];
        }
        lines[cursorRow][cursorCol] = c;
        drawLine(cursorRow, cursorCol, false);
        escapeSequenceIndex = 0;
    }

    /** Delete the character at the current cursor position. */
    private void deleteChar() {
        System.out.println("deleteChar "+cursorRow+" "+cursorCol);
        for (int i = cursorCol; i < NUMBER_OF_COLS - 1; i++) {
            lines[cursorRow][i] = lines[cursorRow][i + 1];
        }
        lines[cursorRow][NUMBER_OF_COLS - 1] = ' ';
        escapeSequenceIndex = 0;
    }

    /** Delete all characters between the cursor and the end of line. */
    private void deleteUntilEndOfLine() {
        for (int i = cursorCol; i < NUMBER_OF_COLS; i++) {
            lines[cursorRow][i] = ' ';
        }
        drawLine(cursorRow, cursorCol, true);
        escapeSequenceIndex = 0;
    }

    /** Insert an empty new line at the cursor position. */
    private void insertLine() {
        synchronized (lines) {
            char[] lastLine = lines[lines.length - 1];
            for (int i = lines.length - 2; i >= cursorRow; i--) {
                lines[i + 1] = lines[i];
            }
            lines[cursorRow] = lastLine;
            clearLine(lastLine);
        }
        redrawScreen();
    }

    /** Delete the line given by the row index. */
    private void deleteLine(int row) {
        synchronized (lines) {
            char[] lastLine = lines[row];
            if (row < lines.length - 1) {
                System.arraycopy(lines, row + 1, lines, row,
                                 lines.length - row - 1);
            }
            lines[lines.length - 1] = lastLine;
            clearLine(lastLine);
        }
        redrawScreen();
    }

    /** Clear the screen and redraw all lines. */
    private void redrawScreen() {
        gdp.setByte(ThomsonGDP.ADR_CMD, ThomsonGDP.CLEAR);
        for (int i = 0; i < lines.length - 1; i++) {
            drawLine(i);
        }
        drawCursor();
        escapeSequenceIndex = 0;
    }

    /** Draw the line with the given index. */
    private void drawLine(int index) {
        int y = getYFromRowIndex(index);
        gdp.setByte(ThomsonGDP.ADR_X_LSB, (byte) X_BIAS);
        gdp.setByte(ThomsonGDP.ADR_X_MSB, (byte) (X_BIAS >> 8));
        char[] line = lines[index];
        int len = getLastIndex(line);
        if (underlining) {
            for (int i = 0; i <= len; i++) {
                gdp.setByte(ThomsonGDP.ADR_Y_LSB, (byte) y);
                gdp.setByte(ThomsonGDP.ADR_CMD, (byte) line[i]);
                gdp.setByte(ThomsonGDP.ADR_Y_LSB, (byte) (y - 1));
                gdp.setByte(ThomsonGDP.ADR_CMD, ULINE_RIGHT);
                gdp.setByte(ThomsonGDP.ADR_CMD, ULINE_RIGHT);
                gdp.setByte(ThomsonGDP.ADR_CMD, ULINE_LEFT);
                gdp.setByte(ThomsonGDP.ADR_CMD, ULINE_LEFT);
            }
        } else {
            gdp.setByte(ThomsonGDP.ADR_Y_LSB, (byte) y);
            for (int i = 0; i <= len; i++) {
                gdp.setByte(ThomsonGDP.ADR_CMD, (byte) line[i]);
            }
        }
    }

    /**
     *  Draw the line with the given row index starting at the given
     *  column. Before drawing erases up to the end of the line if
     *  <code>wholeLine == true</code>.
     */
    private void drawLine(int row, int column, boolean wholeLine) {
        char[] line = lines[row];
        int lastIndex = getLastIndex(line);
        int x = getXFromColumnIndex(column);
        int y = getYFromRowIndex(row);
        gdp.setByte(ThomsonGDP.ADR_Y_LSB, (byte) y);

        // Eraser until rest of line
        gdp.setByte(ThomsonGDP.ADR_X_LSB, (byte) x);
        gdp.setByte(ThomsonGDP.ADR_X_MSB, (byte) (x >> 8));
        gdp.setByte(ThomsonGDP.ADR_CMD, ThomsonGDP.ERASER);
        int len = wholeLine ? NUMBER_OF_COLS : lastIndex;
        for (int i = column; i <= len; i++) {
            gdp.setByte(ThomsonGDP.ADR_CMD, ThomsonGDP.BLOCK58);
        }

        // draw until rest of line
        gdp.setByte(ThomsonGDP.ADR_X_LSB, (byte) x);
        gdp.setByte(ThomsonGDP.ADR_X_MSB, (byte) (x >> 8));
        gdp.setByte(ThomsonGDP.ADR_CMD, ThomsonGDP.PEN);
        if (underlining) {
            for (int i = column; i <= lastIndex; i++) {
                gdp.setByte(ThomsonGDP.ADR_Y_LSB, (byte) y);
                gdp.setByte(ThomsonGDP.ADR_CMD, (byte) line[i]);
                gdp.setByte(ThomsonGDP.ADR_Y_LSB, (byte) (y - 1));
                gdp.setByte(ThomsonGDP.ADR_CMD, ULINE_RIGHT);
                gdp.setByte(ThomsonGDP.ADR_CMD, ULINE_RIGHT);
                gdp.setByte(ThomsonGDP.ADR_CMD, ULINE_LEFT);
                gdp.setByte(ThomsonGDP.ADR_CMD, ULINE_LEFT);
            }
        } else {
            for (int i = column; i <= lastIndex; i++) {
                gdp.setByte(ThomsonGDP.ADR_CMD, (byte) line[i]);
            }
        }
    }

    /**
     *  Return the index of the last non-space character of the given
     *  line. Returns 0 if all characters are spaces.
     */
    private int getLastIndex(char[] line) {
        int len = NUMBER_OF_COLS;
        while (line[--len] == ' ' && len > 0);
        return len;
    }

    /** Draw the cursor at its new position. */
    private void drawCursor() {
        eraseOldCursor();
        drawCursor(cursorCol, cursorRow);
        oldCursorRow = cursorRow;
        oldCursorCol = cursorCol;
    }

    /** Draw a cursor at the given position. */
    private void drawCursor(int cx, int cy) {
        int x = getXFromColumnIndex(cx);
        gdp.setByte(ThomsonGDP.ADR_X_MSB, (byte) (x >> 8));
        gdp.setByte(ThomsonGDP.ADR_X_LSB, (byte) x);
        int y = getYFromRowIndex(cy) - 1;
        gdp.setByte(ThomsonGDP.ADR_Y_LSB, (byte) y);
        gdp.setByte(ThomsonGDP.ADR_CMD, (byte) 0xe0);
        gdp.setByte(ThomsonGDP.ADR_CMD, (byte) 0xc0);
    }

    /** Erase cursor at its old position. */
    private void eraseOldCursor() {
        if (oldCursorRow >= 0) {
            gdp.setByte(ThomsonGDP.ADR_CMD, ThomsonGDP.ERASER);
            drawCursor(oldCursorCol, oldCursorRow);
            gdp.setByte(ThomsonGDP.ADR_CMD, ThomsonGDP.PEN);
        }
    }

    /** Return the x coordinate of the column with given index. */
    private int getXFromColumnIndex(int index) {
        return X_BIAS + 6 * index;
    }

    /** Return the y coordinate of the row with given index. */
    private int getYFromRowIndex(int index) {
        return Y_BIAS + LINE_HEIGHT * (NUMBER_OF_ROWS - 1 - index);
    }

//
//  Implementation of the Terminal interface.
//

    /** Return "Thomson GDP". */
    public String getName() {
        return "Thomson GDP";
    }

    /** Return the instance of <code>ThomsonGDP</code> on which this relies. */
    public Device getDevice() {
        return gdp;
    }

    /** Return <code>true</code> if the type-ahead buffer is empty. */
    public boolean isEmpty() {
        return nextPos == freePos;
    }

    /**
     *  Return a character from the type-ahead buffer. Return
     *  <code>'\0'</code> if the buffer is empty.
     */
    public char readChar() {
        if (isEmpty()) {
            return '\0';
        } else {
            char c = typeAheadBuffer[nextPos++];
            nextPos %= typeAheadBuffer.length;
            return c;
        }
    }

    /** Get the next character typed in. */
    public synchronized char getNextChar() {
        if (isEmpty()) {
            try {
                isWaiting = true;
                wait();
            } catch (Exception e) {}
            isWaiting = false;
        }
        char c = typeAheadBuffer[nextPos++];
        nextPos %= typeAheadBuffer.length;
        return c;
    }

    /**
     *  Write a character onto the screen.
     *  Understands the special commands of the Osborne 1.
     */
    public void writeChar(char c) {
        if (escapeSequenceIndex == 0) {
            switch (c) {
            case '\0':
                gdp.setByte(ThomsonGDP.ADR_CMD, ThomsonGDP.RESET);
                gdp.setByte(ThomsonGDP.ADR_CTRL1, (byte) 9);
                break;
            case '\r': cursorCol = 0; break;
            case '\n': processLineFeed(); break;
            case '\b':
                cursorCol = Math.max(0, cursorCol - 1);
                drawLine(cursorRow, cursorCol, false);
                gdp.setByte(ThomsonGDP.ADR_CMD, ThomsonGDP.ERASER);
                gdp.setByte(ThomsonGDP.ADR_CMD, ThomsonGDP.BLOCK58);
                gdp.setByte(ThomsonGDP.ADR_CMD, ThomsonGDP.PEN);
                break;
            case '\013': cursorRow = Math.max(0, cursorRow - 1); break;
            case '\014':
                cursorCol = Math.min(NUMBER_OF_COLS - 1, cursorCol + 1);
                break;
            case '\032': clearScreen(); break;
            case '\036': cursorCol = cursorRow = 0; break;
            case '\033': escapeSequenceIndex = 1; break;
            default:
                if (c >= ' ' && cursorCol < NUMBER_OF_COLS) {
                    lines[cursorRow][cursorCol] = c;
                    drawLine(cursorRow, cursorCol++, false);
                }
            }
            drawCursor();
        } else {        // Escape sequence
            switch (escapeSequenceIndex) {
            case 1:
                switch (c) {
                case 'l': underlining = true; escapeSequenceIndex = 0; break;
                case 'm': underlining = false; escapeSequenceIndex = 0; break;
                case 'E': insertLine(); break;
                case 'R': deleteLine(cursorRow); break;
                case 'Q': insertChar(c); break;
                case 'W': deleteChar(); break;
                case 'T': deleteUntilEndOfLine(); break;
                case '=': escapeSequenceIndex++; break;
                default: escapeSequenceIndex = 0; break;
                }
                break;
            case 2:
                escapeSequenceIndex++;
                cursorRow = Math.max(0, Math.min(NUMBER_OF_ROWS - 1, c - ' '));
                break;
            case 3:
                escapeSequenceIndex = 0;
                cursorCol = Math.max(0, Math.min(NUMBER_OF_COLS - 1, c - ' '));
                drawCursor();
                break;
            }
        }
    }
}
