package jfrace.view;

import jfrace.AddressSpace;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.util.Vector;

/**
 *  A class visualizing the addressable units (i.e. cells) of an
 *  <a href="../AddressSpace.html"><code>AddressSpace</code></a>.
 *  <p>
 *  The cells are squares arrange in a square (if possible)
 *  with the first cell in the upper left corner and the last
 *  cell in the lower right corner.
 *  <p>
 *  Each cell is shown by small squares (minimum size: one pixel).
 *  Inactive cells are shown in light gray.
 *  If the activity of a cell is set (by the method
 *  <a href="#setActivity(int, int)"><code>setActivity</code></a>)
 *  the color will be changed to a color coding the type
 *  of activity: Red for reading program statements,
 *  green for reading data, and blue for writing data.
 *  This color will decay in a few seconds.
 *  <p>
 *  If one moves the mouse pointer over a cell its address,
 *  its value, and the values of the three following cells are
 *  shown as hex numbers.
 *  If one clicks on a cell all registrated <code>ActionListeners</code>
 *  will be informed. The address of the cell will be stored in the
 *  event ID.
 *
 *  @see jfrace.AddressSpace
 *
 *  @author Franz-Josef ELmer
 */
public class AddressBankViewer extends Canvas implements Runnable {
    /** Type: Reading program */
    public static final int READ_PRG = 0;
    /** Type: Reading data */
    public static final int READ_DATA = 1;
    /** Type: Writing data */
    public static final int WRITE_DATA = 2;

    private static final int MIN_SIZE = 4; // minimum cell size with border
    private static final int TIME_STEP = 200;
    private static final int GRAY = 0xf0;
    private static final int[] COLORS = {0xffff0000, 0xff00ff00, 0xff0000ff};
    private static final Font INFO_FONT = new Font("SansSerif", Font.PLAIN, 12);

    private Thread animationThread;
    private int[] screen;
    private MemoryImageSource source;
    private Image image;
    private int firstAddress;
    private int size;
    private int cellSize;
    private int rowSize; // number of cells per row
    private int rows;
    private int plaquetSize;
    private int bias;
    private int imgWidth;
    private int imgHeight;
    private int rowHeight;
    private int infoYPos;
    private int imgYPos;
    private AddressSpace as;
    private boolean showInfo;
    private int infoAddress;
    private Vector listeners;
    private MouseListener ml = new MouseAdapter() {
            public void mouseExited(MouseEvent me) {
                showInfo = false;
            }
            public void mousePressed(MouseEvent me) {
                int y = me.getY() - imgYPos;
                if (y >= 0) {
                    int address = firstAddress + me.getX() / cellSize
                                  + (y / cellSize) * rowSize;
                    notifyListeners(new ActionEvent(AddressBankViewer.this,
                                                    address, ""));
                }
            }
        };
    private MouseMotionListener mml = new MouseMotionAdapter() {
            public void mouseMoved(MouseEvent me) {
                int y = me.getY() - imgYPos;
                if (y >= 0) {
                    infoAddress = firstAddress + me.getX() / cellSize
                                  + (y / cellSize) * rowSize;
                    showInfo = true;
                    repaint();
                }
            }
        };

    /**
     *  Create an instance for the given <code>AddressSpace</code>
     *  and starts the animation thread.
     *  @param anAddressSpace The address space to be visualized.
     *  @param aFirstAddress The first address to be visualized.
     *  @param aSize The number of cells to be visualized.
     *  @param aCellSize The cell size in pixel
     */
    public AddressBankViewer(AddressSpace anAddressSpace, int aFirstAddress,
                             int aSize, int aCellSize) {
        as = anAddressSpace;
        firstAddress = aFirstAddress;
        size = aSize;
        plaquetSize = cellSize = aCellSize;

        // Choose (i.e. calculate) the size of a row
        if (size <= 32) {
            rowSize = size;
        } else {
            rowSize = (((int) Math.sqrt((double) size) + 8) / 16) * 16;
        }
        rows = (size + rowSize - 1) / rowSize;

        // create image
        imgWidth = cellSize * rowSize;
        imgHeight = cellSize * rows;
        if (cellSize >= MIN_SIZE) {
            imgWidth++;
            imgHeight++;
            plaquetSize--;
            bias = imgWidth + 1;
        }
        rowHeight = cellSize * imgWidth;
        screen = new int[imgWidth * imgHeight];
        for (int i = 0; i < screen.length; i++) {
            screen[i] = 0xff000000;
        }

        // calculate position of info and image
        FontMetrics fm = getFontMetrics(INFO_FONT);
        imgYPos = fm.getHeight();
        infoYPos = fm.getAscent();


        // start animation
        addMouseListener(ml);
        addMouseMotionListener(mml);
        animationThread = new Thread(this);
        animationThread.start();
    }

    public Dimension preferredSize() {
        return new Dimension(imgWidth, imgHeight + imgYPos);
    }

    /**
     *  Subscribe an action listener who is interested in
     *  receiving <code>ActionEvent</code>s. Does nothing if the
     *  given action listener is already in the subscriber list.
     */
    public void addActionListener(ActionListener al) {
        if (listeners == null) {
            listeners = new Vector();
        }
        if (!listeners.contains(al)) {
            listeners.addElement(al);
        }
    }

    /**
     *  Unsubscribe an action listener who is no longer interested in
     *  receiving <code>ActionEvent</code>s. Does nothing if the
     *  given action listener is not in the subscriber list.
     */
    public void removeActionListener(ActionListener al) {
        if (listeners != null) {
            listeners.removeElement(al);
        }
    }

    /**
     *  Send the given action event to all subscribed
     *  <code>ActionListeners</code>.
     */
    private void notifyListeners(ActionEvent anEvent) {
        if (listeners != null) {
            for (int i = 0; i < listeners.size(); i++) {
                ((ActionListener) listeners.elementAt(i))
                                        .actionPerformed(anEvent);
            }
        }
    }

    /**
     *  Set the activity type of at a given address.
     *  @param address Address of the cell.
     *  @param type Type of activity. Possible values are
     *              <code>READ_PRG</code>, <code>READ_DATA</code>,
     *              and <code>WRITE_DATA</code>.
     */
    public synchronized void setActivity(int address, int type) {
        if (type < 0 || type >= COLORS.length) {
            return;
        }
        int col = (address - firstAddress) % rowSize;
        int row = (address - firstAddress) / rowSize;
        if (row < 0 || row >= rows) {
            return;
        }
        int i = bias + (row * imgWidth + col) * cellSize;
        for (int n = 0; n < plaquetSize; n++, i += imgWidth) {
            for (int j = 0; j < plaquetSize; j++) {
                screen[i + j] = COLORS[type];
            }
        }
    }

    public void update(Graphics g) {
        paint(g);
    }

    public void paint(Graphics g) {
        if (source == null) {
            source = new MemoryImageSource(imgWidth, imgHeight,
                                           ColorModel.getRGBdefault(),
                                           screen, 0, imgWidth);
            source.setAnimated(true);
            image = createImage(source);
        }
        g.setColor(getBackground());
        g.fillRect(0, 0, imgWidth, imgYPos);
        g.drawImage(image, 0, imgYPos, imgWidth, imgHeight, this);
        if (showInfo) {
            g.setColor(Color.black);
            g.setFont(INFO_FONT);
            StringBuffer sb = new StringBuffer("(");
            sb.append(Integer.toHexString(infoAddress));
            sb.append(") =");
            for (int i = infoAddress; i < infoAddress + 4; i++) {
                sb.append(' ');
                sb.append(Integer.toHexString(as.getValue(i, 0)));
            }
            g.drawString(new String(sb), 0, infoYPos);
        }
    }

//
//  Implementation of the Runnable interface.
//

    public void run() {
        Thread t = Thread.currentThread();
        while (t == animationThread) {
            if (source != null) {
                boolean changed = false;
                synchronized (screen) {
                    int pixel;
                    int mask;
                    int gray;
                    int newPixel;
                    if (cellSize < MIN_SIZE) {
                        for (int i = 0; i < screen.length; i++) {
                            newPixel = 0xff000000;
                            mask = 0xff;
                            gray = GRAY;
                            for (int n = 0; n < 3; n++) {
                                pixel = screen[i] & mask;
                                if (pixel != gray) {
                                    changed = true;
                                    pixel = ((9 * pixel + gray) / 10) & mask;
                                }
                                newPixel |= pixel;
                                mask <<= 8;
                                gray <<= 8;
                            }
                            screen[i] = newPixel;
                        }
                    } else {
                        int ij;
                        for (int i = 0; i < screen.length; i += imgWidth) {
                            if (i % rowHeight != 0) {
                                for (int j = 0; j < imgWidth; j++) {
                                    if (j % cellSize != 0) {
                                        ij = i + j;
                                        newPixel = 0xff000000;
                                        mask = 0xff;
                                        gray = GRAY;
                                        for (int n = 0; n < 3; n++) {
                                            pixel = screen[ij] & mask;
                                            if (pixel != gray) {
                                                changed = true;
                                                pixel = ((15 * pixel + gray)
                                                        / 16) & mask;
                                            }
                                            newPixel |= pixel;
                                            mask <<= 8;
                                            gray <<= 8;
                                        }
                                        screen[ij] = newPixel;
                                    }
                                }
                            }
                        }
                    }
                    if (changed) {
                        source.newPixels();
                    }
                }
            }
            try {
                Thread.sleep(TIME_STEP);
            } catch (Exception e) {}
        }
    }

}
