package jfrace.view;

import jfrace.CPU;
import jfrace.Register;
import jfrace.Util;

import java.awt.*;
import java.awt.event.*;
import java.util.*;

/**
 *  This class visualizes the state of an CPU.
 *  Each register is shown by its name and its value.
 *  The value is shown bit wise by colored boxes and a hexadecimal
 *  number. If possible the next op-code is shown in its
 *  disassembled form.
 *
 *  @author Franz-Josef Elmer
 */
public class CPUViewer extends Canvas {
    private static final Font TITLE_FONT = new Font("SansSerif", Font.BOLD, 20);
    private static final Font NAME_FONT = new Font("SansSerif", Font.BOLD, 12);
    private static final Font NEXT_INS_FONT = new Font("SansSerif", Font.PLAIN, 12);
    private static final int X0 = 10;
    private static final int DX = 10;
    private static final int Y0 = 30;
    private static final int DY = 15;

    private CPU cpu;
    private int[] regSeq;
    private MemoryView[] regViews;

    private int x1;
    private int yTitle;
    private int totalXSize;
    private int totalYSize;

    private Image buffer;
    private Graphics bufferG;
    private Dimension oldSize = new Dimension(0, 0);
    private Vector listeners;
    private MouseListener ml = new MouseAdapter() {
            /**
             *  Transmits an <code>ActionEvent</code> when the mouse button
             *  is pressed above a bit cell. The bit index is stored in the
             *  type attribute of the <code>ActionEvent</code>.
             */
            public void mousePressed(MouseEvent me) {
                int index;
                int x = me.getX();
                int y = me.getY();
                String regName;
                for (int i = 0; i < regViews.length; i++) {
                    index = regViews[i].getBitIndex(x, y);
                    if (index >= 0) {
                        regName = cpu.getRegister(regSeq[i]).getName();
                        notifyListeners(new ActionEvent(CPUViewer.this, index,
                                                        regName));
                        break;
                    }
                }
            }
        };


    /**
     *  Create a new instance for the given CPU.
     */
    public CPUViewer(CPU aCPU) {
        super();

        cpu = aCPU;
        addMouseListener(ml);

        // build register views:
        // 1. gather registers
        Register[] registers = new Register[cpu.getNumberOfRegisters()];
        regSeq = new int[registers.length];
        for (int i = 0; i < registers.length; i++) {
            registers[i] = cpu.getRegister(i);
            regSeq[i] = i;
        }

        // 2. sort by size
        int index;
        Register r;
        for (int i = 0; i < registers.length; i++) {
            for (int j = i + 1; j < registers.length; j++) {
                if (registers[i].getNumberOfBits()
                        > registers[j].getNumberOfBits()) {
                    r = registers[i];
                    index = regSeq[i];
                    registers[i] = registers[j];
                    regSeq[i] = regSeq[j];
                    registers[j] = r;
                    regSeq[j] = index;
                }
            }
        }

        // 3. Move program counter at the end
        for (int i = 0; i < registers.length - 1; i++) {
            if (registers[i].getType() == Register.PROGRAM_COUNTER) {
                    r = registers[i];
                    index = regSeq[i];
                    registers[i] = registers[registers.length - 1];
                    regSeq[i] = regSeq[registers.length - 1];
                    registers[registers.length - 1] = r;
                    regSeq[registers.length - 1] = index;
            }
        }

        // 4. Compute position of the second column
        FontMetrics fm = getFontMetrics(NAME_FONT);
        for (int i = 0; i < registers.length; i++) {
            x1 = Math.max(x1, fm.stringWidth(registers[i].getName()));
        }
        x1 += X0 + DX;

        // 5. Build views and compute total size
        regViews = new MemoryView[registers.length];
        int maxViewSize = 0;
        int numberOfBits;
        for (int i = 0; i < registers.length; i++) {
            numberOfBits = registers[i].getNumberOfBits();
            regViews[i] = new MemoryView(x1, Y0 + i * DY, numberOfBits);
            maxViewSize = Math.max(maxViewSize, regViews[i].getWidthFor(this));
        }
        totalXSize = x1 + maxViewSize + X0;
        totalYSize = Y0 + (regSeq.length + 2) * DY;

        // 6. Calculate title position
        fm = getFontMetrics(TITLE_FONT);
        yTitle = (Y0 - fm.getHeight()) / 2 + fm.getAscent();
    }

    /**
     *  Subscribe an action listener who is interested in
     *  receiving <code>ActionEvent</code>s. Does nothing if the
     *  given action listener is already in the subscriber list.
     */
    public void addActionListener(ActionListener al) {
        if (listeners == null) {
            listeners = new Vector();
        }
        if (!listeners.contains(al)) {
            listeners.addElement(al);
        }
    }

    /**
     *  Unsubscribe an action listener who is no longer interested in
     *  receiving <code>ActionEvent</code>s. Does nothing if the
     *  given action listener is not in the subscriber list.
     */
    public void removeActionListener(ActionListener al) {
        if (listeners != null) {
            listeners.removeElement(al);
        }
    }

    /**
     *  Send the given action event to all subscribed
     *  <code>ActionListeners</code>.
     */
    private void notifyListeners(ActionEvent anEvent) {
        if (listeners != null) {
            for (int i = 0; i < listeners.size(); i++) {
                ((ActionListener) listeners.elementAt(i))
                                        .actionPerformed(anEvent);
            }
        }
    }

    /**
     *  Show the instruction at the position of the program pointer.
     *  If the CPU supports disassembling the disassemble instruction
     *  is shown otherwise it is shown hexadecimal.
     */
    private String getNextInstruction() {
        StringBuffer sb = new StringBuffer("next: ");
        Register r = cpu.getProgramCounter();
        if (cpu.supportsDisassembling()) {
            cpu.disassemble(sb, r.getValue());
        } else {
            int numberOfDigits = (r.getNumberOfBits() + 3) >> 2;
            int opCode = cpu.getAddressSpace().getValue(r.getValue(), 0);
            Util.appendHex(sb, numberOfDigits, (long) opCode);
        }
        return new String(sb);
    }

    public Dimension preferredSize() {
        return new Dimension(totalXSize, totalYSize);
    }

    /** Update with double-buffering. */
    public void update(Graphics g) {
        Dimension size = size();
        if (size.width != oldSize.width || size.height != oldSize.height) {
            if (bufferG != null) {
                bufferG.dispose();
            }
            if (buffer != null) {
                buffer.flush();
            }
            buffer = createImage(size.width, size.height);
            bufferG = buffer.getGraphics();
            oldSize = size;
        }
        bufferG.setColor(getBackground());
        bufferG.fillRect(0, 0, size.width, size.height);
        bufferG.setColor(getForeground());
        paint(bufferG);

        g.drawImage(buffer, 0, 0, null);
    }

    /** Draw the state of the CPU. */
    public void paint(Graphics g) {
        g.setColor(Color.black);
        g.setFont(TITLE_FONT);
        g.drawString(cpu.getName(), x1, yTitle);
        int y = Y0;
        int dy = regViews[0].getHeight();
        Register r;
        for (int i = 0; i < regSeq.length; i++) {
            g.setColor(Color.black);
            g.setFont(NAME_FONT);
            r = cpu.getRegister(regSeq[i]);
            g.drawString(r.getName(), X0, y + dy);
            regViews[i].setValue(r.getValue());
            regViews[i].draw(g);
            y += DY;
        }
        g.setFont(NEXT_INS_FONT);
        g.drawString(getNextInstruction(), X0, y + dy);
    }
}
