package jfrace.view;

import jfrace.device.CPMDisk;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;

/**
 *  Visialize disk reading (in blue) and writing (in red) operations.
 *
 *  @author Franz-Josef Elmer
 */
public class DiskView extends Canvas implements Runnable, ActionListener {
    private static final int TIME_STEP = 500;
    private Thread animationThread;
    private int numberOfTracks;
    private int numberOfSectors;
    private int[] screen;
    private MemoryImageSource source;
    private Image image;

    public DiskView(int numberOfTracks, int numberOfSectors) {
        this.numberOfTracks = numberOfTracks;
        this.numberOfSectors = numberOfSectors;

        screen = new int[numberOfTracks * numberOfSectors];
        for (int i = 0; i < screen.length; i++) {
            screen[i] = 0xff000000;
        }
        animationThread = new Thread(this);
        animationThread.start();
    }

    public Dimension preferredSize() {
        return new Dimension(numberOfTracks, numberOfSectors);
    }

    private synchronized void setWriting(int track, int sector) {
        screen[(sector - 1) * numberOfTracks + track] |= 0xff0000;
    }

    private synchronized void setReading(int track, int sector) {
        screen[(sector - 1) * numberOfTracks + track] |= 0xff;
    }

    public void update(Graphics g) {
        paint(g);
    }

    public void paint(Graphics g) {
        if (source == null) {
            source = new MemoryImageSource(numberOfTracks, numberOfSectors,
                                           ColorModel.getRGBdefault(),
                                           screen, 0, numberOfTracks);
            source.setAnimated(true);
            image = createImage(source);
        }
        g.drawImage(image, 0, 0, numberOfTracks, numberOfSectors, this);
    }

//
//  Implementation of the Runnable interface.
//

    public void run() {
        Thread t = Thread.currentThread();
        while (t == animationThread) {
            if (source != null) {
                boolean changed = false;
                synchronized (screen) {
                    int pixel;
                    for (int i = 0; i < screen.length; i++) {
                        pixel = screen[i] & 0xffffff;
                        if ((pixel & 0xff0000) != 0) {
                            pixel -= 0x50000;
                            changed = true;
                        }
                        if ((pixel & 0xff) != 0) {
                            pixel -= 5;
                            changed = true;
                        }
                        screen[i] = pixel | 0xff000000;
                    }
                    if (changed) {
                        source.newPixels();
                    }
                }
            }
            try {
                Thread.sleep(TIME_STEP);
            } catch (Exception e) {}
        }
    }

//
//  Implementation of the ActionListener interface
//

    /**
     *  React on events for reading and writing.
     *  The action command contains the track and sector number
     *  (separated by ' '). The event id is either
     *  <code>CPMDisk.WRITE</code> or <code>CPMDisk.READ</code>.
     */
    public void actionPerformed(ActionEvent event) {
        String msg = event.getActionCommand();
        int index = msg.indexOf(' ');
        int track = Integer.parseInt(msg.substring(0, index));
        int sector = Integer.parseInt(msg.substring(index + 1));
        if (event.getID() == CPMDisk.WRITE) {
            setWriting(track, sector);
        } else {
            setReading(track, sector);
        }
    }
}
