package jfrace.view;

import jfrace.*;
import jfrace.device.RAM;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.net.*;
import java.util.*;

/**
 *  An AWT GUI for the emulator. It shows the CPU state
 *  (using <a href="CPUViewer.html"><code>CPUViewer</code></a>),
 *  the memory access (using <a href="AddressBankViewer.html">
 *  <code>AddressBankViewer</code></a>, only op-code reads
 *  are shown), and the value of a selected memory cell (using
 *  <a href="MemoryCellViewer.html"><code>MemoryCellViewer</code></a>).
 *  One can start, stop, and single-step through the emulation.
 *  In addition the memory can be manipulated and loaded from
 *  or stored to a file.
 *
 *  @author Franz-Josef Elmer
 */
public class EmulatorGUI extends Frame implements Runnable {
    private static final int STOPPED = 0;
    private static final int RUNNING = 1;

    private AddressSpace as;
    private CPU cpu;
    private Hashtable cpuRegisterNames;
    private MenuItem openMI;
    private MenuItem saveAsMI;
    private MenuItem exitMI;
    private String lastDirectory;
    private File file;
    private Button stepButton;
    private Button runButton;
    private Button resetButton;
    private Button stopButton;
    private Choice delayChoice;
    private int[] delays = {0, 1, 10, 100, 1000};
    private CPUViewer cpuViewer;
    private MemoryCellViewer mcViewer;
    private AddressBankViewer abViewer;
    private int state = STOPPED;
    private Thread runningThread;
    private ActionListener al = new ActionListener() {
            public void actionPerformed(ActionEvent anEvent) {
                Object source = anEvent.getSource();
                if (source == openMI) {
                    open();
                } else if (source == saveAsMI) {
                    saveAs();
                } else if (source == exitMI) {
                    exit();
                } else if (source == resetButton) {
                    reset();
                } else if (source == stepButton) {
                    step();
                } else if (source == runButton) {
                    runEmulation();
                } else if (source == stopButton) {
                    stop();
                } else if (source == mcViewer) {
                    toggleABitOfMemoryCell(anEvent.getID());
                } else if (source == abViewer) {
                    changeAddressOfMemoryCellViewer(anEvent.getID());
                } else if (source == cpuViewer) {
                    toggleABitOfRegister(anEvent.getActionCommand(),
                                         anEvent.getID());
                }
            }
        };

    /**
     *  Creates a new GUI for the first CPU of the Emulator.
     */
    public EmulatorGUI() {
        super("Computer Emulator");

        as = Emulator.getEmulator().getAddressSpace();
        cpu = as.getCPU(0);

        cpuRegisterNames = new Hashtable();
        for (int i = 0; i < cpu.getNumberOfRegisters(); i++) {
            cpuRegisterNames.put(cpu.getRegister(i).getName(),
                                 new Integer(i));
        }

        addWindowListener(new WindowAdapter() {
                public void windowClosing(WindowEvent we) {
                    we.getWindow().dispose();
                    if (Emulator.getEmulator().insideAnApplet()) {
                        Emulator.getEmulator().getApplet().stop();
                    } else {
                        exit();
                    }
                }
            });
        add("Center", makeGUI());
        if (!Emulator.getEmulator().insideAnApplet()) {
            makeMenuBar();
        }
        setIconImage(Util.loadImage("icon.gif"));
        pack();
    }

    private Panel makeGUI() {
        Panel panel = new Panel();
        panel.setLayout(new BorderLayout());

        Panel northPanel = new Panel();
        panel.add("North", northPanel);

        resetButton = new Button("reset");
        resetButton.addActionListener(al);
        northPanel.add(resetButton);
        stepButton = new Button("step");
        stepButton.addActionListener(al);
        northPanel.add(stepButton);
        delayChoice = new Choice();
        delayChoice.add("no delay");
        delayChoice.add("1 msec");
        delayChoice.add("10 msec");
        delayChoice.add("100 msec");
        delayChoice.add("1 sec");
        delayChoice.select(3);
        northPanel.add(delayChoice);
        runButton = new Button("run");
        runButton.addActionListener(al);
        northPanel.add(runButton);
        stopButton = new Button("stop");
        stopButton.enable(false);
        stopButton.addActionListener(al);
        northPanel.add(stopButton);

        Panel centerPanel = new Panel();
        panel.add("Center", centerPanel);
        cpuViewer = new CPUViewer(cpu);
        cpuViewer.addActionListener(al);
        centerPanel.add(cpuViewer);
        Panel rightPanel = new Panel(new BorderLayout());
        centerPanel.add(rightPanel);
        mcViewer = new MemoryCellViewer(cpu.getAddressSpace(),
                                        cpu.getNumberOfBits(),
                                        new Font("SansSerif", Font.BOLD, 12));
        mcViewer.addActionListener(al);
        rightPanel.add("South", mcViewer);

        int size = 1 << cpu.getNumberOfBits();
        int cellSize = 300 / size;
        abViewer = new AddressBankViewer(cpu.getAddressSpace(), 0,
                                         size * size, cellSize);
        abViewer.addActionListener(al);
        rightPanel.add("Center", abViewer);

        return panel;
    }

    /** Create menu bar. */
    private void makeMenuBar() {
        MenuBar bar = new MenuBar();
        setMenuBar(bar);

        Menu fileMenu = new Menu("File");
        bar.add(fileMenu);

        openMI = new MenuItem("Open...");
        openMI.addActionListener(al);
        fileMenu.add(openMI);
        saveAsMI = new MenuItem("SaveAs...");
        saveAsMI.addActionListener(al);
        fileMenu.add(saveAsMI);
        fileMenu.addSeparator();
        exitMI = new MenuItem("Exit");
        exitMI.addActionListener(al);
        fileMenu.add(exitMI);
    }

    /**
     *  Return a file instance choosen in a <code>FileDialog</code>.
     *  @param type Type of <code>FileDialog</code>. Has to either
     *              <code>FileDialog.LOAD</code> or
     *              <code>FileDialog.SAVE</code>.
     */
    private File getFileFromDialog(int type) {
        FileDialog fd = new FileDialog(this, "Load memory", type);
        if (lastDirectory != null) {
            fd.setDirectory(lastDirectory);
        }
        fd.show();
        lastDirectory = fd.getDirectory();
        return new File(lastDirectory, fd.getFile());
    }

    /**
     *  Open and read binary data into memory from a file
     *  choosen by the user in a file dialog.
     *  Assuming that the first device of <code>AddressSpace</code>
     *  is the <code>RAM</code>.
     */
    private void open() {
        file = getFileFromDialog(FileDialog.LOAD);
        if (file != null) {
            try {
                FileInputStream fis = new FileInputStream(file);
                RAM ram = (RAM) as.getDevice(0);
                fis.read(ram.getMemory(), 0, fis.available());
                fis.close();
            } catch (Exception e) {
                e.printStackTrace(System.out);
            }
            repaint();
        }
    }

    /**
     *  Open a file choosen by the user in a file dialog and save
     *  the binary data from memory into it.
     *  Assuming that the first device of <code>AddressSpace</code>
     *  is the <code>RAM</code>.
     */
    private void saveAs() {
        file = getFileFromDialog(FileDialog.SAVE);
        if (file != null) {
            try {
                FileOutputStream fos = new FileOutputStream(file);
                RAM ram = (RAM) as.getDevice(0);
                fos.write(ram.getMemory());
                fos.close();
            } catch (Exception e) {
                e.printStackTrace(System.out);
            }
        }
    }

    private void exit() {
        if (!Emulator.getEmulator().insideAnApplet()) {
            System.exit(0);
        }
    }

    private void setButtonStates(int newState) {
        switch (newState) {
        case RUNNING:
            state = RUNNING;
            resetButton.disable();
            stepButton.disable();
            delayChoice.disable();
            runButton.disable();
            stopButton.enable();
            if (openMI != null) {
                openMI.disable();
            }
            break;
        case STOPPED:
            state = STOPPED;
            resetButton.enable();
            stepButton.enable();
            delayChoice.enable();
            runButton.enable();
            stopButton.disable();
            if (openMI != null) {
                openMI.enable();
            }
            break;
        }
    }

    public void repaint() {
        super.repaint();
        cpuViewer.repaint();
        abViewer.repaint();
        mcViewer.repaint();
    }

    private void reset() {
        cpu.reset();
        cpuViewer.repaint();
    }

    private void step() {
        cpu.next();
        cpuViewer.repaint();
        mcViewer.repaint();
        abViewer.setActivity(cpu.getProgramCounter().getValue(),
                             AddressBankViewer.READ_PRG);
    }

    private void runEmulation() {
        setButtonStates(RUNNING);
        runningThread = new Thread(this);
        runningThread.start();
    }

    private void stop() {
        runningThread = null;
        setButtonStates(STOPPED);
    }

    private void toggleABitOfRegister(String regName, int bitIndex) {
        Integer i = (Integer) cpuRegisterNames.get(regName);
        if (i != null) {
            int v = cpu.getRegister(i.intValue()).getValue() ^ (1 << bitIndex);
            cpu.setRegisterValue(i.intValue(), v);
            cpuViewer.repaint();
        }
    }

    private void toggleABitOfMemoryCell(int bitIndex) {
        int address = mcViewer.getAddress();
        AddressSpace as = cpu.getAddressSpace();
        as.setValue(address, 0, as.getValue(address, 0) ^ (1 << bitIndex));
        mcViewer.repaint();
        cpuViewer.repaint();
        abViewer.repaint();
    }

    private void changeAddressOfMemoryCellViewer(int address) {
        System.out.println("changeAddressOfMemoryCellViewer="+address);
        mcViewer.setAddress(address);
    }

//
// Implementation of the Runnable interface
//

    /**
     *  Run the emulation in its own thread.
     */
    public void run() {
        int delay = delays[delayChoice.getSelectedIndex()];
        Thread t = Thread.currentThread();
        while (runningThread == t) {
            step();
            if (delay > 0) {
                try {
                    Thread.sleep(delay);
                } catch (Exception e) {}
            }
        }
    }

}
