package jfrace.view;

import jfrace.AddressSpace;

import java.awt.*;
import java.awt.event.*;
import java.util.Vector;

/**
 *  An AWT component which shows the address and the content of an
 *  addressable unit of an <a href="../AddressSpace.html">
 *  <code>AddressSpace</code></a>. The address is shown as a
 *  hexadecimal number. The content is shown as a hex number and
 *  a bit pattern (by using the class <a href="MemoryView.html">
 *  <code>MemoryView</code></a>).
 *  <p>This component is an interactive one. If one clicks onto a cell
 *  of the bit pattern an <code>ActionEvent</code> is sent to all
 *  <code>ActionListener</code>s who have subcribed at this component.
 *
 *  @author Franz-Josef Elmer
 */
public class MemoryCellViewer extends Canvas {
    private static final int X0 = 10;
    private static final int Y0 = 10;
    private static final int DX = 10;

    private MemoryView mv;
    private int numberOfBits;
    private AddressSpace as;
    private Font addressFont;
    private int address;
    private int x1;
    private int y0;
    private int totalWidth;
    private int totalHeight;
    private Vector listeners;
    private MouseListener ml = new MouseAdapter() {
            /**
             *  Transmits an <code>ActionEvent</code> when the mouse button
             *  is pressed above a bit cell. The bit index is stored in the
             *  type attribute of the <code>ActionEvent</code>.
             */
            public void mousePressed(MouseEvent me) {
                int index = mv.getBitIndex(me.getX(), me.getY());
                if (index >= 0) {
                    notifyListeners(new ActionEvent(MemoryCellViewer.this,
                                                    index, ""));
                }
            }
        };

    /**
     *  Constructor.
     *  @param anAddressSpace The address space to which this viewer belongs.
     *  @param aNumber The number of bits shown.
     *  @param anFont The font which is used to paint the address value.
     */
    public MemoryCellViewer(AddressSpace anAddressSpace, int aNumber,
                            Font anFont) {
        as = anAddressSpace;
        numberOfBits = aNumber;
        addressFont = anFont;

        addMouseListener(ml);

        // calculate size
        x1 = X0 + getFontMetrics(addressFont).stringWidth("(AAAA)") + DX;
        mv = new MemoryView(x1, Y0, numberOfBits);
        y0 = Y0 + mv.getHeight();
        totalWidth = x1 + mv.getWidthFor(this);
        totalHeight = mv.getHeight() + 2 * Y0;
    }

    /**
     *  Subscribe an action listener who is interested in
     *  receiving <code>ActionEvent</code>s. Does nothing if the
     *  given action listener is already in the subscriber list.
     */
    public void addActionListener(ActionListener al) {
        if (listeners == null) {
            listeners = new Vector();
        }
        if (!listeners.contains(al)) {
            listeners.addElement(al);
        }
    }

    /**
     *  Unsubscribe an action listener who is no longer interested in
     *  receiving <code>ActionEvent</code>s. Does nothing if the
     *  given action listener is not in the subscriber list.
     */
    public void removeActionListener(ActionListener al) {
        if (listeners != null) {
            listeners.removeElement(al);
        }
    }

    /**
     *  Send the given action event to all subscribed
     *  <code>ActionListeners</code>.
     */
    private void notifyListeners(ActionEvent anEvent) {
        if (listeners != null) {
            for (int i = 0; i < listeners.size(); i++) {
                ((ActionListener) listeners.elementAt(i))
                                        .actionPerformed(anEvent);
            }
        }
    }

    public Dimension preferredSize() {
        return new Dimension(totalWidth, totalHeight);
    }

    /** Return the address of the viewer. */
    public int getAddress() {
        return address;
    }

    /**
     *  Set the address of an addressable unit of the address space
     *  which should be shown by this viewer. Repaint the viewer.
     *  @param anAddress The address.
     */
    public void setAddress(int anAddress) {
        address = anAddress;
        repaint();
    }

    /** Paint address (in hex), bit cells, and value (in hex). */
    public void paint(Graphics g) {
        g.setColor(Color.black);
        g.setFont(addressFont);
        g.drawString('(' + Integer.toHexString(address) + ')', X0, y0);
        mv.setValue(as.getValue(address, 0));
        mv.draw(g);
    }

}
