package jfrace.view;

import java.awt.*;

/**
 *  Class to be used to draw the content of a memory cell
 *  inside an AWT component. The memory cell is shown bitwise
 *  by colored squares (default: green and orange for
 *  bit: 0 and 1, resp.). The value
 *  is also painted as a hexadecimal number right of the squares.
 *
 *  @author Franz-Josef Elmer
 */
public class MemoryView {
    private int x;
    private int y;
    private int xSize = 10;
    private int ySize = 10;
    private int numberOfBits;
    private Color color0 = Color.green.brighter();
    private Color color1 = Color.orange;
    private Font numberFont = new Font("Monospaced", Font.PLAIN, 12);
    private int numberDistance = 5;
    private int numberAscent;
    private int value;

    /**
     *  Create an instance for the given number of bits with
     *  default values for cell size, bit colors, number distance,
     *  and number font.
     *  @param theX The x coordinate of the upper left corner.
     *  @param theY The y coordinate of the upper left corner.
     *  @param aNumberOfBits The number of bits.
     */
    public MemoryView(int theX, int theY, int aNumberOfBits) {
        x = theX;
        y = theY;
        numberOfBits = aNumberOfBits;
        numberDistance += xSize * numberOfBits;
    }

    /**
     *  Create an instance with overwritten all default parameters.
     *  @param theX The x coordinate of the upper left corner.
     *  @param theY The y coordinate of the upper left corner.
     *  @param aNumberOfBits The number of bits.
     *  @param anXSize With of a bit cell in pixels.
     *  @param aYSize Height of a bit cell in pixels.
     *  @param aColor0 Color for bit 0.
     *  @param aColor1 Color for bit 1.
     *  @param aNumberDistance Distance in pixels between bit cells and
     *                         hexadecimal number.
     *  @param aNumberFont The font for the number.
     */
    public MemoryView(int theX, int theY, int aNumberOfBits,
                      int anXSize, int aYSize,
                      Color aColor0, Color aColor1,
                      int aNumberDistance, Font aNumberFont) {
        x = theX;
        y = theY;
        xSize = anXSize;
        ySize = aYSize;
        color0 = aColor0;
        color1 = aColor1;
        numberDistance = xSize * numberOfBits + aNumberDistance;
        numberFont = aNumberFont;
    }

    /**
     *  Return the with of the bit cells plus hex number.
     *  @param c An AWT component needed to calculate the
     *           with of the number.
     */
    public int getWidthFor(Component c) {
        int digitWidth = c.getFontMetrics(numberFont).stringWidth("A");
        return numberDistance + digitWidth * ((numberOfBits + 3) >> 2);
    }

    /** Return the with of a bit cell. */
    public int getCellWidth() {
        return xSize;
    }

    /** Return the height of a bit cell. */
    public int getHeight() {
        return ySize;
    }

    /**
     *  Set the value of the memory cell.
     *  Does not update the view.
     */
    public void setValue(int aValue) {
        value = aValue;
    }

    /**
     *  Return the index of the bit below the given position.
     *  @param theX The x coordinate of the position.
     *  @param theY The y coordinate of the position.
     *  @return -1 if no bit below the point <code>(x, y)</code>.
     */
    public int getBitIndex(int theX, int theY) {
        int index = -1;
        int dy = theY - y;
        if (dy >= 0 && dy <= ySize) {
            index = numberOfBits - 1 - (theX - x) / xSize;
            if (index < 0 || index >= numberOfBits) {
                index = -1;
            }
        }
        return index;
    }

    /**
     *  Draw the bit cells and the hex number.
     *  @param g Graphics context to be drawn.
     */
    public void draw(Graphics g) {
        int mask = 1 << numberOfBits;
        for (int i = 0; i < numberOfBits; i++) {
            mask >>= 1;
            g.setColor((value & mask) == 0 ? color0 : color1);
            g.fillRect(x + xSize * i, y, xSize, ySize);
        }
        g.setColor(Color.black);
        for (int i = 0; i < numberOfBits; i++) {
            g.drawRect(x + xSize * i, y, xSize, ySize);
        }
        g.setFont(numberFont);
        g.drawString(Integer.toHexString(value), x + numberDistance,
                     y + ySize);
    }
}
