package jfrace;

import jfrace.cpu.MockCPU;
import jfrace.device.MockDevice;

import junit.framework.*;

public class AddressSpaceTest extends TestCase {
    public AddressSpaceTest(String name) {
        super(name);
    }

    public static Test suite() {
        return new TestSuite(AddressSpaceTest.class);
    }

    public void testConstructors() {
        AddressSpace as = new AddressSpace();
        assertEquals("getValue(1234, 0)", as.getValue(1234, 0), 0);
        assertEquals("getByte(34, 1)", as.getByte(34, 1) & 0xff, 0);

        int defaultValue = 0x78ab;
        as = new AddressSpace(defaultValue);
        assertEquals("getValue(1234, 0)", as.getValue(1234, 0), defaultValue);
        assertEquals("getByte(34, 1)", as.getByte(34, 1) & 0xff,
                defaultValue & 0xff);
    }

    public void testListenerManagement() {
        AddressSpace as = new AddressSpace();

        MockAddressSpaceListener tasl1 = new MockAddressSpaceListener();
        MockAddressSpaceListener tasl2 = new MockAddressSpaceListener();
        as.addAddressSpaceListener(tasl1);
        as.addAddressSpaceListener(tasl1);
        as.addAddressSpaceListener(tasl2);

        CPU cpu = new MockCPU();
        int offset = 8977;
        as.addCPU(offset, cpu);
        assertEquals("1. AddressSpaceListener.getEventCounter()",
                     tasl1.getEventCounter(), 1);
        assertEquals("2. AddressSpaceListener.getEventCounter()",
                     tasl2.getEventCounter(), 1);
        assertEquals("Both listeners received the same event",
                     tasl1.getLastEvent() == tasl2.getLastEvent(), true);
        AddressSpaceEvent ase = tasl1.getLastEvent();
        assertSame("event.getAddressSpace()", ase.getAddressSpace(), as);
        assertSame("event.getHardware()", ase.getHardware(), cpu);
        assertEquals("event.getType()", ase.getType(), AddressSpaceEvent.ADD);
        assertEquals("event.getAddressOffset()", ase.getAddressOffset(), offset);

        tasl1.resetEventCounter();
        tasl2.resetEventCounter();
        as.removeAddressSpaceListener(tasl1);
        as.removeCPU(cpu);
        assertEquals("1. AddressSpaceListener.getEventCounter()",
                     tasl1.getEventCounter(), 0);
        assertEquals("2. AddressSpaceListener.getEventCounter()",
                     tasl2.getEventCounter(), 1);
        ase = tasl2.getLastEvent();
        assertSame("event.getAddressSpace()", ase.getAddressSpace(), as);
        assertSame("event.getHardware()", ase.getHardware(), cpu);
        assertEquals("event.getType()", ase.getType(), AddressSpaceEvent.REMOVE);
        assertEquals("event.getAddressOffset()", ase.getAddressOffset(), offset);

        Device device = new MockDevice(10);
        offset = 5765;
        int mask = 2468;
        try {
            as.addDevice(offset, mask, device);
        } catch (Exception e) {
            fail(e.toString());
        }
        assertEquals("2. AddressSpaceListener.getEventCounter()",
                     tasl2.getEventCounter(), 2);
        ase = tasl2.getLastEvent();
        assertSame("event.getAddressSpace()", ase.getAddressSpace(), as);
        assertSame("event.getHardware()", ase.getHardware(), device);
        assertEquals("event.getType()", ase.getType(), AddressSpaceEvent.ADD);
        assertEquals("event.getAddressOffset()", ase.getAddressOffset(), offset);
        assertEquals("event.getModeMask()", ase.getModeMask(), mask);
        as.addAddressSpaceListener(tasl1);
        as.removeDevice(device);
        assertEquals("1. AddressSpaceListener.getEventCounter()",
                     tasl1.getEventCounter(), 1);
        assertEquals("2. AddressSpaceListener.getEventCounter()",
                     tasl2.getEventCounter(), 3);
        ase = tasl2.getLastEvent();
        assertSame("event.getAddressSpace()", ase.getAddressSpace(), as);
        assertSame("event.getHardware()", ase.getHardware(), device);
        assertEquals("event.getType()", ase.getType(), AddressSpaceEvent.REMOVE);
        assertEquals("event.getAddressOffset()", ase.getAddressOffset(), offset);
        assertEquals("event.getModeMask()", ase.getModeMask(), mask);
        as.removeAddressSpaceListener(tasl1);
        as.removeAddressSpaceListener(tasl2);
        as.removeAddressSpaceListener(tasl1);
        as.removeAddressSpaceListener(tasl2);
    }

    public void testCPUManagement() {
        AddressSpace as = new AddressSpace();
        assertEquals("getNumberOfCPUs()", as.getNumberOfCPUs(), 0);
        CPU cpu1 = new MockCPU();
        CPU cpu2 = new MockCPU();
        CPU cpu3 = new MockCPU();
        int offset1 = 123;
        int offset2 = 234;
        int offset3 = 345;
        as.addCPU(offset1, cpu1);
        assertEquals("getNumberOfCPUs()", as.getNumberOfCPUs(), 1);
        assertSame("getCPU(0)", as.getCPU(0), cpu1);
        as.addCPU(offset2, cpu2);
        assertEquals("getNumberOfCPUs()", as.getNumberOfCPUs(), 2);
        assertSame("getCPU(0)", as.getCPU(0), cpu1);
        assertSame("getCPU(1)", as.getCPU(1), cpu2);
        as.removeCPU(cpu3);
        assertEquals("getNumberOfCPUs()", as.getNumberOfCPUs(), 2);
        assertSame("getCPU(0)", as.getCPU(0), cpu1);
        assertSame("getCPU(1)", as.getCPU(1), cpu2);
        as.removeCPU(cpu1);
        assertEquals("getNumberOfCPUs()", as.getNumberOfCPUs(), 1);
        assertSame("getCPU(0)", as.getCPU(0), cpu2);
        try {
            as.getCPU(1);
            fail("ArrayIndexOutOfBoundsException expected");
        } catch (ArrayIndexOutOfBoundsException aioobe) {}
        as.addCPU(offset1, cpu1);
        as.addCPU(offset3, cpu3);
        assertEquals("getNumberOfCPUs()", as.getNumberOfCPUs(), 3);
        assertSame("getCPU(0)", as.getCPU(0), cpu2);
        assertSame("getCPU(1)", as.getCPU(1), cpu1);
        assertSame("getCPU(2)", as.getCPU(2), cpu3);
        as.removeCPU(cpu1);
        assertEquals("getNumberOfCPUs()", as.getNumberOfCPUs(), 2);
        assertSame("getCPU(0)", as.getCPU(0), cpu2);
        assertSame("getCPU(2)", as.getCPU(1), cpu3);
        as.removeCPU(cpu3);
        assertEquals("getNumberOfCPUs()", as.getNumberOfCPUs(), 1);
        assertSame("getCPU(0)", as.getCPU(0), cpu2);
        as.removeCPU(cpu2);
        assertEquals("getNumberOfCPUs()", as.getNumberOfCPUs(), 0);
    }

    public void testDeviceManagement() {
        AddressSpace as = new AddressSpace();
        assertEquals("getNumberOfAddressBanks()",
                     as.getNumberOfAddressBanks(), 0);
        try {
            as.addDevice(-1, 0, new MockDevice(1));
            fail("IllegalArgumentException expected");
        } catch (IllegalArgumentException iae) {
        } catch (Exception e) {
            fail("not expected exception:" + e);
        }
        try {
            as.addDevice(0, 0, new MockDevice(0));
        } catch (IllegalArgumentException iae) {
        } catch (Exception e) {
            fail("not expected exception:" + e);
        }
        Device device1 = new MockDevice(10);
        Device device2 = new MockDevice(1);
        Device device3 = new MockDevice(20);
        int offset1 = 100;
        int offset2 = 99;
        int offset3 = 105;
        int offset3b = 120;
        int mask1 = 1024;
        int mask2 = 2048;
        int mask3 = 4096;
        try {
            as.addDevice(offset1, mask1, device1);
        } catch (Exception e) {
            fail(e.toString());
        }
        assertEquals("getNumberOfAddressBanks()", as.getNumberOfAddressBanks(), 1);
        assertSame("getDevice(0)", as.getDevice(0), device1);
        assertEquals("getFirstAddress(0)", as.getFirstAddress(0), offset1);
        assertEquals("getLasstAddress(0)", as.getLastAddress(0),
                     offset1 + device1.getNumberOfUnits());
        assertEquals("getModeMask(0)", as.getModeMask(0), mask1);

        try {
            as.addDevice(88, mask3, device3);
            fail("Exception expected.");
        } catch (Exception e) {}
        assertEquals("getNumberOfAddressBanks()", as.getNumberOfAddressBanks(), 1);

        try {
            as.addDevice(offset2, mask2, device2);
        } catch (Exception e) {
            fail(e.toString());
        }
        assertEquals("getNumberOfAddressBanks()", as.getNumberOfAddressBanks(), 2);
        assertSame("getDevice(0)", as.getDevice(0), device2);
        assertEquals("getFirstAddress(0)", as.getFirstAddress(0), offset2);
        assertEquals("getLasstAddress(0)", as.getLastAddress(0),
                     offset2 + device2.getNumberOfUnits());
        assertEquals("getModeMask(0)", as.getModeMask(0), mask2);
        assertSame("getDevice(1)", as.getDevice(1), device1);
        assertEquals("getFirstAddress(1)", as.getFirstAddress(1), offset1);
        assertEquals("getLasstAddress(1)", as.getLastAddress(1),
                     offset1 + device1.getNumberOfUnits());
        assertEquals("getModeMask(1)", as.getModeMask(1), mask1);

        try {
            as.addDevice(offset3, mask3, device3);
            fail("Exception expected.");
        } catch (Exception e) {}
        assertEquals("getNumberOfAddressBanks()", as.getNumberOfAddressBanks(), 2);

        try {
            as.addDevice(98, mask3, device3);
            fail("Exception expected.");
        } catch (Exception e) {}
        assertEquals("getNumberOfAddressBanks()", as.getNumberOfAddressBanks(), 2);

        try {
            as.addDevice(88, mask3, device3);
            fail("Exception expected.");
        } catch (Exception e) {}
        assertEquals("getNumberOfAddressBanks()", as.getNumberOfAddressBanks(), 2);

        as.removeDevice(device2);
        assertEquals("getNumberOfAddressBanks()", as.getNumberOfAddressBanks(), 1);
        assertSame("getDevice(0)", as.getDevice(0), device1);
        assertEquals("getFirstAddress(0)", as.getFirstAddress(0), offset1);
        assertEquals("getLasstAddress(0)", as.getLastAddress(0),
                     offset1 + device1.getNumberOfUnits());
        assertEquals("getModeMask(0)", as.getModeMask(0), mask1);
        as.removeDevice(device2);
        assertEquals("getNumberOfAddressBanks()", as.getNumberOfAddressBanks(), 1);

        try {
            as.addDevice(offset3b, mask3, device3);
        } catch (Exception e) {
            fail(e.toString());
        }
        assertEquals("getNumberOfAddressBanks()", as.getNumberOfAddressBanks(), 2);
        assertSame("getDevice(0)", as.getDevice(0), device1);
        assertEquals("getFirstAddress(0)", as.getFirstAddress(0), offset1);
        assertEquals("getLasstAddress(0)", as.getLastAddress(0),
                     offset1 + device1.getNumberOfUnits());
        assertEquals("getModeMask(0)", as.getModeMask(0), mask1);
        assertSame("getDevice(1)", as.getDevice(1), device3);
        assertEquals("getFirstAddress(1)", as.getFirstAddress(1), offset3b);
        assertEquals("getLasstAddress(1)", as.getLastAddress(1),
                     offset3b + device3.getNumberOfUnits());
        assertEquals("getModeMask(1)", as.getModeMask(1), mask3);

        try {
            as.addDevice(offset2, mask2, device2);
        } catch (Exception e) {
            fail(e.toString());
        }
        assertEquals("getNumberOfAddressBanks()", as.getNumberOfAddressBanks(), 3);
        assertSame("getDevice(0)", as.getDevice(0), device2);
        assertEquals("getFirstAddress(0)", as.getFirstAddress(0), offset2);
        assertEquals("getLasstAddress(0)", as.getLastAddress(0),
                offset2 + device2.getNumberOfUnits());
        assertEquals("getModeMask(0)", as.getModeMask(0), mask2);
        assertSame("getDevice(1)", as.getDevice(1), device1);
        assertEquals("getFirstAddress(1)", as.getFirstAddress(1), offset1);
        assertEquals("getLasstAddress(1)", as.getLastAddress(1),
                offset1 + device1.getNumberOfUnits());
        assertEquals("getModeMask(1)", as.getModeMask(1), mask1);
        assertSame("getDevice(2)", as.getDevice(2), device3);
        assertEquals("getFirstAddress(2)", as.getFirstAddress(2), offset3b);
        assertEquals("getLasstAddress(2)", as.getLastAddress(2),
                offset3b + device3.getNumberOfUnits());
        assertEquals("getModeMask(2)", as.getModeMask(2), mask3);

        as.removeDevice(device1);
        assertEquals("getNumberOfAddressBanks()", as.getNumberOfAddressBanks(), 2);
        assertSame("getDevice(0)", as.getDevice(0), device2);
        assertEquals("getFirstAddress(0)", as.getFirstAddress(0), offset2);
        assertEquals("getLasstAddress(0)", as.getLastAddress(0),
                offset2 + device2.getNumberOfUnits());
        assertEquals("getModeMask(0)", as.getModeMask(0), mask2);
        assertSame("getDevice(1)", as.getDevice(1), device3);
        assertEquals("getFirstAddress(1)", as.getFirstAddress(1), offset3b);
        assertEquals("getLasstAddress(1)", as.getLastAddress(1),
                offset3b + device3.getNumberOfUnits());
        assertEquals("getModeMask(1)", as.getModeMask(1), mask3);

        as.removeDevice(device3);
        assertEquals("getNumberOfAddressBanks()", as.getNumberOfAddressBanks(), 1);
        assertSame("getDevice(0)", as.getDevice(0), device2);
        assertEquals("getFirstAddress(0)", as.getFirstAddress(0), offset2);
        assertEquals("getLasstAddress(0)", as.getLastAddress(0),
                offset2 + device2.getNumberOfUnits());
        assertEquals("getModeMask(0)", as.getModeMask(0), mask2);
    }

    public void testGetAndSet() {
        int defaultValue = 0xabcd;
        AddressSpace as = new AddressSpace(defaultValue);

        Device device1 = new MockDevice(10);
        Device device2 = new MockDevice(20);
        int offset1 = 100;
        int offset2 = 180;
        int mask1 = 0;
        int mask2 = 3;

        try {
            as.addDevice(offset1, mask1, device1);
        } catch (Exception e) {
            fail(e.toString());
        }
        try {
            as.addDevice(offset2, mask2, device2);
        } catch (Exception e) {
            fail(e.toString());
        }
        try {
            as.setValue(99, 0, 88);
        } catch (Exception e) {
            fail(e.toString());
        }
        assertEquals("getValue(99, 0)", as.getValue(99, 0), defaultValue);
        assertEquals("getByte(99, 0)", as.getByte(99, 0) & 0xff, defaultValue & 0xff);
        try {
            as.setValue(104, 0, 2188);
        } catch (Exception e) {
            fail(e.toString());
        }
        assertEquals("getValue(104, 0)", as.getValue(104, 0), 2188);
        assertEquals("getByte(104, 0)", as.getByte(104, 0) & 0xff, 2188 & 0xff);
        byte b = (byte) 218;
        try {
            as.setValue(105, 0, b);
        } catch (Exception e) {
            fail(e.toString());
        }
        assertEquals("getValue(104, 0)", as.getValue(104, 0), 2188);
        assertEquals("getByte(104, 0)", as.getByte(104, 0) & 0xff, 2188 & 0xff);
        assertEquals("getValue(105, 0)", as.getValue(105, 0), (int) b);
        assertEquals("getByte(105, 0)", (int) as.getByte(105, 0), (int) b);

        try {
            as.setValue(109, 1, 218);
            fail("BusErrorException expected.");
        } catch (BusErrorException bee) {
        } catch (Exception e) {
            fail("BusErrorException expected.");
        }
        try {
            as.getValue(109, 1);
            fail("BusErrorException expected.");
        } catch (BusErrorException bee) {
        } catch (Exception e) {
            fail("BusErrorException expected.");
        }
        try {
            as.setByte(109, 1, b);
            fail("BusErrorException expected.");
        } catch (BusErrorException bee) {
        } catch (Exception e) {
            fail("BusErrorException expected.");
        }
        try {
            as.getByte(109, 1);
            fail("BusErrorException expected.");
        } catch (BusErrorException bee) {
        } catch (Exception e) {
            fail("BusErrorException expected.");
        }
        try {
            as.setValue(199, 8, 218);
            fail("BusErrorException expected.");
        } catch (BusErrorException bee) {
        } catch (Exception e) {
            fail("BusErrorException expected.");
        }
        try {
            as.getValue(199, 8);
            fail("BusErrorException expected.");
        } catch (BusErrorException bee) {
        } catch (Exception e) {
            fail("BusErrorException expected.");
        }
        try {
            as.setByte(199, 8, b);
            fail("BusErrorException expected.");
        } catch (BusErrorException bee) {
        } catch (Exception e) {
            fail("BusErrorException expected.");
        }
        try {
            as.getByte(199, 8);
            fail("BusErrorException expected.");
        } catch (BusErrorException bee) {
        } catch (Exception e) {
            fail("BusErrorException expected.");
        }
        try {
            as.setValue(199, 2, 2188);
            assertEquals("getValue(199, 1)", as.getValue(199, 1), 2188);
            assertEquals("getByte(199, 1)", as.getByte(199, 1) & 0xff, 2188 & 0xff);
            assertEquals("getValue(199, 2)", as.getValue(199, 2), 2188);
            assertEquals("getByte(199, 2)", as.getByte(199, 2) & 0xff, 2188 & 0xff);
        } catch (Exception e) {
            fail(e.toString());
        }

        try {
            as.setValue(190, 1, b);
            assertEquals("getValue(190, 1)", as.getValue(190, 1), (int) b);
            assertEquals("getByte(190, 1)", (int) as.getByte(190, 1), (int) b);
            assertEquals("getValue(190, 2)", as.getValue(190, 2), (int) b);
            assertEquals("getByte(190, 2)", (int) as.getByte(190, 2), (int) b);
        } catch (Exception e) {
            fail(e.toString());
        }
    }
}


