package jfrace;

import jfrace.cpu.MockCPU2;
import jfrace.device.MockDevice2;
import jfrace.device.RAM;

import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintStream;
import java.net.URL;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

public class EmulatorTest extends TestCase {
    private Emulator emu = Emulator.getEmulator();
    private StringBuffer systemOutBuffer = new StringBuffer();
    private OutputStream stringBufferOutputStream = new OutputStream() {
                public void write(int b) throws IOException {
                    systemOutBuffer.append((char) b);
                }
            };
    private PrintStream redirectedSystemOut
            = new PrintStream(stringBufferOutputStream);
    private PrintStream originalSystemOut = System.out;

    public EmulatorTest(String name) {
        super(name);
    }

    public static Test suite() {
        return new TestSuite(EmulatorTest.class);
    }

    protected void setUp() {
        try {
            URL documentBase = new URL("file:///"
                                       + System.getProperty("user.dir")
                                       + '/');
            Emulator.getEmulator().setDocumentBase(documentBase);
        } catch (Exception e) {}
    }

    protected void tearDown() {
         Emulator.getEmulator().stopEmulation();
         try {
            Thread.sleep(200);
         } catch (Exception e) {}
    }

    public void testInvalidConfigFiles() {
        testInvalidConfigFile("nofile.cfg", "Error during reading the config file.");
        testInvalidConfigFile("test1.hex", "ERROR: '=' missing in");
        testInvalidConfigFile("emulatorInvalid1.cfg",
                              "packagePrefix line missing");
        testInvalidConfigFile("emulatorInvalid2.cfg",
                              "ERROR: unkown type >blabla<");
        testInvalidConfigFile("emulatorInvalid3.cfg", "class name missing");
        testInvalidConfigFile("emulatorInvalid4.cfg",
                              "java.lang.ClassNotFoundException");
        testInvalidConfigFile("emulatorInvalid5.cfg",
                              "java.lang.IllegalArgumentException");
        testInvalidConfigFile("emulatorInvalid6.cfg",
                              "'(' missing in a66");
        testInvalidConfigFile("emulatorInvalid7.cfg",
                              "')' missing in a(66");
        testInvalidConfigFile("emulatorInvalid8.cfg", "Error for method bit");
        testInvalidConfigFile("emulatorInvalid9.cfg",
                              "'(' missing in bit66");
        testInvalidConfigFile("emulatorInvalid10.cfg",
                              "')' missing in b(77");
        testInvalidConfigFile("emulatorInvalid11.cfg",
                              "Error for method b(77,88)");
        testInvalidConfigFile("emulatorInvalid12.cfg",
                              "Error for method b(affe)");
        testInvalidConfigFile("emulatorInvalid13.cfg", "three items expected.");
        testInvalidConfigFile("emulatorInvalid14.cfg",
                              "jfrace.AddressBankOverlapException");
        testInvalidConfigFile("emulatorInvalid15.cfg",
                              "java.io.FileNotFoundException");
        testInvalidConfigFile("emulatorInvalid16.cfg", "two items expected.");
        testInvalidConfigFile("emulatorInvalid17.cfg",
                              "java.lang.ClassNotFoundException");
        testInvalidConfigFile("emulatorInvalid18.cfg",
                              "java.lang.IllegalArgumentException: Number");
        testInvalidConfigFile("emulatorInvalid19.cfg",
                              "java.lang.ClassNotFoundException");
    }

    /**
     *  Testing an invalid configuration file.
     *  @param configFile invalid configuration file.
     *  @param expectedSystemOut Expected error message on the console.
     */
    private void testInvalidConfigFile(String configFile,
                                       String expectedSystemOut) {
        configFile = "testingResources/" + configFile;
        systemOutBuffer.setLength(0);
        System.setOut(redirectedSystemOut);
        emu.reset();
        assertEquals("Is " + configFile + " a valid config file?",
                     false, emu.configure(configFile, new String[0]));
        assertSysOutEquals(configFile, expectedSystemOut);
        System.setOut(originalSystemOut);
    }

    private void assertSysOutEquals(String configFile,
                                    String expectedSystemOut) {
        systemOutBuffer.setLength(Math.min(systemOutBuffer.length(),
                                           expectedSystemOut.length()));
        String sysOut = new String(systemOutBuffer);
        assertEquals("System.out for " + configFile, expectedSystemOut, sysOut);
    }

    public void testEmulator0() {
        String configFile = "testingResources/emulator0.cfg";
        emu.reset();
        assertEquals("Is " + configFile + " a valid config file?",
                     true, emu.configure(configFile, new String[0]));

        AddressSpace as = emu.getAddressSpace();
        assertEquals("Address space == null?", false, as == null);
        assertEquals("Default value for outside access", 0xfa,
                     as.getValue(0x1000000, 0));
        assertEquals("jfrace.cpu.", emu.getPackageCPUPrefix());
        assertEquals("jfrace.device.", emu.getPackageDevicePrefix());
        assertEquals("debugging?", false, emu.getDebuging());
        assertEquals("ownThread?", false, emu.getOwnThread());

        int numberOfCPUs = as.getNumberOfCPUs();
        assertEquals("as.getNumberOfCPUs()", 3, numberOfCPUs);
        CPU cpu;
        int[] offsets = {512, 0, 1000};
        for (int i = 0; i < numberOfCPUs; i++) {
            cpu = as.getCPU(i);
            assertEquals("Is " + (i + 1) + ". CPU instanceof MockCPU2?",
                         true, cpu instanceof MockCPU2);
            assertEquals("Is address space for " + (i + 1)
                         + ". CPU the same", as, ((MockCPU2) cpu).as);
            assertEquals("Offset of " + (i + 1) + ". CPU",
                         offsets[i], as.getCPUAddressOffset(i));
            assertEquals("Number of received events of " + (i + 1) + ". CPU",
                         numberOfCPUs - i,
                         ((MockCPU2) cpu).numberOfReceivedEvents);
        }
        assertEquals("1. CPU 1. method", "method a called.",
                     ((MockCPU2) as.getCPU(0)).methodA);
        assertEquals("2. CPU 1. method", "method b called.",
                     ((MockCPU2) as.getCPU(1)).methodB);
        assertEquals("2. CPU 1. method 1. argument", 12,
                     ((MockCPU2) as.getCPU(1)).bNumber);
        assertEquals("3. CPU 1. method", "method c called.",
                     ((MockCPU2) as.getCPU(2)).methodC);
        assertEquals("3. CPU 1. method 1. argument", "affe",
                     ((MockCPU2) as.getCPU(2)).cString);
        assertEquals("3. CPU 2. method", "method d called.",
                     ((MockCPU2) as.getCPU(2)).methodD);
        assertEquals("3. CPU 2. method 1. argument", 23,
                     ((MockCPU2) as.getCPU(2)).dNumber);
        assertEquals("3. CPU 2. method 2. argument", "tiger",
                     ((MockCPU2) as.getCPU(2)).dString);
    }

    public void testEmulator1() {
        String[] args = {"-d", "testingResources/emulator1.cfg"};
        emu.reset();
        emu.main(args);

        AddressSpace as = emu.getAddressSpace();
        assertEquals("Address space == null?", false, as == null);
        assertEquals("Default value for outside access", 0,
                     as.getValue(0x1000000, 0));
        assertEquals("jfrace.cpu.", emu.getPackageCPUPrefix());
        assertEquals("jfrace.device.", emu.getPackageDevicePrefix());
        assertEquals("debugging?", true, emu.getDebuging());
        assertEquals("debugStart", 0, emu.getDebugStart());
        assertEquals("debugEnd", 0, emu.getDebugEnd());
        assertEquals("debugStep", 1, emu.getDebugStep());
        assertEquals("ownThread?", false, emu.getOwnThread());

        assertEquals("as.getNumberOfCPUs()", 0, as.getNumberOfCPUs());

        int[] nBits = {8, 20, 8, 12};
        int[] offsets = {0, 0x20000, 0x30000, 0x40000};
        int[] sizes = {0x10000, 65536, 30990, 15495};
        int[] masks = {0, 0xff, 0, 3};
        int[][] content = {{0, 0, 0},
                           {0, 0, 0},
                           {0x47, 0x49, 0x46, 0x38, 0x39, 0x61},
                           {0x749, 0x638, 0x961}};

        int numberOfDevices = as.getNumberOfAddressBanks();
        assertEquals("as.getNumberOfAddressBanks()", nBits.length,
                     numberOfDevices);
        RAM ram;
        for (int i = 0; i < numberOfDevices; i++) {
            assertEquals("Is " + (i + 1) + ". device RAM?", true,
                         as.getDevice(i) instanceof RAM);
            assertEquals("number of bits of " + (i + 1) + ". RAM", nBits[i],
                         as.getDevice(i).getNumberOfBits());
            assertEquals("offset of " + (i + 1) + ". RAM", offsets[i],
                         as.getFirstAddress(i));
            assertEquals("size of " + (i + 1) + ". RAM", sizes[i],
                         as.getLastAddress(i) - as.getFirstAddress(i));
            assertEquals("mode mask of " + (i + 1) + ". RAM", masks[i],
                         as.getModeMask(i));
            for (int j = 0; j < content[i].length; j++) {
                assertEquals((i + 1) + ". RAM @" + j,
                             content[i][j], as.getValue(offsets[i] + j, 0));
            }
            if (nBits[i] == 8) {
                for (int j = 0; j < content[i].length; j++) {
                    assertEquals((i + 1) + ". RAM @" + j,
                                 (byte) content[i][j], as.getByte(offsets[i] + j, 0));
                }
            }
        }
    }

    public void testEmulator2() {
        String[] args = {"-d33.044.8", "-t", "testingResources/emulator2.cfg",
                         "123", "ape"};
        emu.reset();
        emu.main(args);

        assertEquals("debugging?", true, emu.getDebuging());
        assertEquals("debugStart", 33, emu.getDebugStart());
        assertEquals("debugEnd", 36, emu.getDebugEnd());
        assertEquals("debugStep", 8, emu.getDebugStep());
        assertEquals("ownThread?", true, emu.getOwnThread());

        AddressSpace as = emu.getAddressSpace();
        assertEquals("Address space == null?", false, as == null);
        assertEquals("Default value for outside access", 0,
                     as.getValue(0x1000000, 0));
        assertEquals("jfrace.cpu.", emu.getPackageCPUPrefix());
        assertEquals("jfrace.device.", emu.getPackageDevicePrefix());

        assertEquals("as.getNumberOfCPUs()", 0, as.getNumberOfCPUs());

        int[] nBits = {17, 17, 17};
        int[] offsets = {4, 200, 9999};
        int[] sizes = {7, 7, 7};
        int[] masks = {3, 0, 0};
        int[] nEvents = {2, 3, 0};
        int numberOfDevices = as.getNumberOfAddressBanks();
        assertEquals("as.getNumberOfAddressBanks()", nBits.length,
                     numberOfDevices);
        Device device;
        for (int i = 0; i < numberOfDevices; i++) {
            device = as.getDevice(i);
            assertEquals("Is " + (i + 1) + ". device MockDevice2?", true,
                         device instanceof MockDevice2);
            assertEquals("number of bits of " + (i + 1) + ". device", nBits[i],
                         device.getNumberOfBits());
            assertEquals("offset of " + (i + 1) + ". device", offsets[i],
                         as.getFirstAddress(i));
            assertEquals("size of " + (i + 1) + ". device", sizes[i],
                         as.getLastAddress(i) - as.getFirstAddress(i));
            assertEquals("mode mask of " + (i + 1) + ". device", masks[i],
                         as.getModeMask(i));
            assertEquals("Number of received events of "
                         + (i + 1) + ". device",
                         nEvents[i],
                         ((MockDevice2) device).numberOfReceivedEvents);
        }
        MockDevice2 td = (MockDevice2) as.getDevice(2);
        assertEquals("3. device: Is AddressSpace identical?", true,
                     as == td.as);
        assertEquals("3. device: configResources", "blabla 12 23",
                     td.configResources);
        assertEquals("3. device: packagePrefix", "jfrace",
                     td.packagePrefix);
        assertEquals("3. device: args[0]", args[3], td.args[0]);
        assertEquals("3. device: args[1]", args[4], td.args[1]);
    }

    public void testEmulator3() {
        String[] args = {"-d", "testingResources/emulator3.cfg"};
        systemOutBuffer.setLength(0);
        System.setOut(redirectedSystemOut);
        emu.reset();
        emu.main(args);
        sleep();
        assertSysOutEquals("", "Fibonacci 1\r\nFibonacci 2\r\nFibonacci 3");
        System.setOut(originalSystemOut);
    }

    public void testEmulator3_1() {
        String[] args = {"-d2", "testingResources/emulator3.cfg"};
        systemOutBuffer.setLength(0);
        System.setOut(redirectedSystemOut);
        emu.reset();
        emu.main(args);
        sleep();
        assertSysOutEquals("", "Fibonacci 3\r\nFibonacci 5\r\nFibonacci 8");
        System.setOut(originalSystemOut);
    }

    public void testEmulator3_2() {
        String[] args = {"-d1.2", "testingResources/emulator3.cfg"};
        systemOutBuffer.setLength(0);
        System.setOut(redirectedSystemOut);
        emu.reset();
        emu.main(args);
        sleep();
        assertSysOutEquals("", "Fibonacci 2\r\n1 CPUs");
        System.setOut(originalSystemOut);
    }

    public void testEmulator3_3() {
        String[] args = {"-d1.0.4", "testingResources/emulator3.cfg"};
        systemOutBuffer.setLength(0);
        System.setOut(redirectedSystemOut);
        emu.reset();
        emu.main(args);
        sleep();
        assertSysOutEquals("", "Fibonacci 2\r\nFibonacci 13\r\nFibonacci 89\r\n1 CPUs");
        System.setOut(originalSystemOut);
    }

    private void sleep() {
        try {
            Thread.sleep(200);
        } catch (Exception e) {}
    }

}
