package jfrace.device;

import java.io.*;
import java.net.*;
import jfrace.*;
import junit.framework.*;

public class RAMTest extends TestCase {
    public RAMTest(String name) {
        super(name);
    }

    public static void main(String[] args) {
        junit.textui.TestRunner.run(suite());
    }

    public static Test suite() {
        return new TestSuite(RAMTest.class);
    }

    protected void setUp() {
        try {
            OutputStream os = new FileOutputStream("RAMTest.dat");
            os.write(new byte[] {1, 1, 2, 3, 5, 8, 13, 21, 34, 55});
            os.close();
        } catch (Exception e) {
            fail("Error during creating test file: " + e);
        }
    }

    protected void tearDown() {
        new File("RAMTest.dat").delete();
    }

    public void testByteRam() {
        RAM ram = new RAM();
        assertEquals("getNumberOfBits()", ram.getNumberOfBits(), 8);
        int size = 100;
        ram.createMemory(size);
        assertEquals("getNumberOfUnits()", ram.getNumberOfUnits(), size);
        byte[] memory = ram.getMemory();
        assertEquals("getMemory().length", memory.length, size);

        byte value = (byte) 0xaf;
        ram.setByte(67, value);
        assertEquals("getByte(67)", ram.getByte(67) & 0xff, value & 0xff);
        assertEquals("test memory directly: memory[66]", memory[66] & 0xff, 0);
        assertEquals("test memory directly: memory[67]", memory[67] & 0xff, value & 0xff);
        assertEquals("test memory directly: memory[68]", memory[68] & 0xff, 0);
        ram.setValue(12, 257);
        assertEquals("getValue(12)", ram.getValue(12), 257 & 0xff);
        assertEquals("test memory directly: memory[11]", memory[11] & 0xff, 0);
        assertEquals("test memory directly: memory[12]", memory[12] & 0xff, 257 & 0xff);
        assertEquals("test memory directly: memory[13]", memory[13] & 0xff, 0);
        for (int i = 0; i < memory.length; i++) {
            memory[i] = (byte) (2 * i + 1);
        }

        ram.reset();
        boolean result = true;
        for (int i = 0; i < memory.length; i++) {
            assertEquals("i=" + i, 0, memory[i]);
        }
        for (int i = 0; i < ram.getNumberOfUnits(); i++) {
            assertEquals("i=" + i, 0, ram.getByte(i));
        }
    }

    public void testNBitRam() {
        for (int i = 1; i < 18; i++) {
            testNBitRam(i);
        }
    }

    private void testNBitRam(int n) {
        int mask = (1 << n) - 1;
        RAM ram = new RAM(n);
        assertEquals("getNumberOfBits()", ram.getNumberOfBits(), n);
        int size = 100;
        ram.createMemory(size);
        assertEquals("getNumberOfUnits()", ram.getNumberOfUnits(), size);
        byte[] memory = ram.getMemory();
        assertEquals("getMemory().length", memory.length, ((n + 7) / 8) * size);

        int value1 = 0x987f1234;
        int value2 = 0xabcf5678;
        int value3 = 0x234eacfe;
        ram.setValue(67, value1);
        assertEquals("getByte(67)", ram.getValue(67), value1 & mask);
        ram.setValue(68, value2);
        ram.setValue(69, value3);
        assertEquals("getByte(67)", ram.getValue(67), value1 & mask);
        assertEquals("getByte(68)", ram.getValue(68), value2 & mask);
        assertEquals("getByte(69)", ram.getValue(69), value3 & mask);
    }

    public void testCreateRAMAndAddToAS() {
        AddressSpace as = new AddressSpace();

        RAM ram = null;
        try {
            RAM.createRAMAndAddToAS(as, "17 180 9", null);
            ram = (RAM) as.getDevice(0);
        } catch (Exception e) {
            fail(e.toString());
        }
        assertEquals(17, ram.getNumberOfBits());
        assertEquals(9, ram.getNumberOfUnits());
        assertEquals(false, ram.isReadOnly());
        assertEquals(180, as.getFirstAddress(0));
        assertEquals(189, as.getLastAddress(0));
        assertEquals(0, as.getModeMask(0));
        as.removeDevice(ram);

        try {
            RAM.createRAMAndAddToAS(as, "7 0x180 017 readOnly", null);
            ram = (RAM) as.getDevice(0);
        } catch (Exception e) {
            fail(e.toString());
        }
        assertEquals(7, ram.getNumberOfBits());
        assertEquals(15, ram.getNumberOfUnits());
        assertEquals(true, ram.isReadOnly());
        as.setValue(0x180, 0, 15);
        assertEquals(0, as.getValue(0x180,0));
        assertEquals(0x180, as.getFirstAddress(0));
        assertEquals(0x18f, as.getLastAddress(0));
        assertEquals(0, as.getModeMask(0));
        as.removeDevice(ram);

        try {
            RAM.createRAMAndAddToAS(as, "7 0x180 017 0x1f", null);
            ram = (RAM) as.getDevice(0);
        } catch (Exception e) {
            fail(e.toString());
        }
        assertEquals(7, ram.getNumberOfBits());
        assertEquals(15, ram.getNumberOfUnits());
        assertEquals(false, ram.isReadOnly());
        as.setValue(0x180, 0, 15);
        assertEquals(15, as.getValue(0x180,0));
        assertEquals(0x180, as.getFirstAddress(0));
        assertEquals(0x18f, as.getLastAddress(0));
        assertEquals(31, as.getModeMask(0));
        as.removeDevice(ram);

        try {
            RAM.createRAMAndAddToAS(as, "7 0x180 017 6 readOnly", null);
            ram = (RAM) as.getDevice(0);
        } catch (Exception e) {
            fail(e.toString());
        }
        assertEquals(7, ram.getNumberOfBits());
        assertEquals(15, ram.getNumberOfUnits());
        assertEquals(true, ram.isReadOnly());
        assertEquals(0x180, as.getFirstAddress(0));
        assertEquals(0x18f, as.getLastAddress(0));
        assertEquals(6, as.getModeMask(0));
        as.removeDevice(ram);

        try {
            URL base = new URL("file:///" + System.getProperty("user.dir")
                               + '/');
            RAM.createRAMAndAddToAS(as, "8 0x280 RAMTest.dat readOnly", base);
            ram = (RAM) as.getDevice(0);
        } catch (Exception e) {
            fail(e.toString());
        }
        assertEquals(8, ram.getNumberOfBits());
        assertEquals(10, ram.getNumberOfUnits());
        assertEquals(true, ram.isReadOnly());
        assertEquals(0x280, as.getFirstAddress(0));
        assertEquals(0x28a, as.getLastAddress(0));
        assertEquals(0, as.getModeMask(0));
        assertEquals(1, as.getByte(0x280, 0));
        assertEquals(55, as.getByte(0x289, 0));
        as.setByte(0x289, 0, (byte) 88);
        assertEquals(55, as.getByte(0x289, 0));
        as.removeDevice(ram);
    }
}
